/*

   FILE:
	artis_event_definition.h

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-16

  CHANGELOG:


  DESCRIPTION:

  This file exports the identifiers of the trace event types used in
  ARTIS modules.

  These identifiers have also to be known by the tools which have to
  interpret them ("trace2kiwi", for example).


  IMPORTANT NOTE: All event types defined here MUST be positive. Events
  with negative type are not stored in the events array, since a negative
  type is used to stablish the end of availabel events.

  As a result, if one wants events of a given type NOT to be generated,
  simply redefine the type as -1 in this file.


*/

#ifndef __ARTIS_EVENT_DEFINITION_H
#define __ARTIS_EVENT_DEFINITION_H



#define NTASKS            50         /* Number of tasks in the application */

#define TIMER             100
#define CONTEXT_SW        101
#define OVERHD_START      102
#define OVERHD_STOP       103
#define ACTIV_START       104
#define ACTIV_END         105
#define INITIAL_END       106

#define MUTEX_LOCK        120
#define MUTEX_UNLOCK      121
#define OPT_MUTEX_LOCK    122
#define OPT_MUTEX_UNLOCK  123

#define ADVANCED_INT      -1 /* 200 */
#define INTERRUPT         201
#define INI_START         202
#define INI_END           203
#define FIN_START         204
#define FIN_END           205

#define OPT_START         300
#define OPT_END           301
#define CALL_CONTROL      302
#define EXECUTE_UNTIL     303
#define INSERT_EVENT      304     /*** Insert new event ***/
#define EXTRACT_EVENT     305     /*** Extract event ***/
#define ADVANCE_FINAL     306     /*** Advance the final part ***/
#define META_RULE_ACTION  307     /*** Execution of a Meta-Rule ***/
#define SEND_SCHED_FIFO   308

#define LEVEL_START       400
#define LEVEL_STOP        401
#define LEVEL_SUSPEND     402
#define LEVEL_RESET       403

#define SLS_POLICY        500
#define AA_QUALITY        501
#define EVENT_RATE        502

#define EM_START          600
#define EM_STOP           601
#define SLS_START         602
#define SLS_STOP          603

#define WAKEUP_FIFO       700
#define MCR_BEGIN         701
#define MCR_END           702
#define SLACK_AVAILABLE   703

#define DEBUG_EVENT       800

#endif // #ifndef __ARTIS_EVENT_DEFINITION_H
