/*

   FILE:
	artis_fifo_definition.h

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-16

  CHANGELOG:


  DESCRIPTION:
  
  This file exports the RT-FIFO constants used for synchronization
  between the two scheduler modules in ARTIS: The RSERVER (in
  RT-Linux) and the DSERVER (in the Linux process "linux_side"):

     + SCHED_FIFO -> Primary synchronization FIFO between both
     schedulers. The DSERVER suspends itself in this FIFO, waiting
     until the RSERVER wakes it up in the next slack interval.

     + ADVANCE_FINAL_PART_FIFO -> If the DSERVER wants to advance the
     execution of the RSERVER (and thus not wait until the slack
     interval is exhausted), the DSERVER sends a message through this
     fifo, which wakes the (RT-Linux) FIFO handler of RSERVER.

*/

#ifndef __ARTIS_FIFO_DEFINITION_H
#define __ARTIS_FIFO_DEFINITION_H

#include "artis_conf.h"

#define SCHED_RTFIFO               0


/// Definition of the message type used to communicate through the RT_FIFOs
#define SCHED_MSSG_MAGIC 0x1234

typedef struct sched_mssg_t {
  int           magic;
  systime_t     slack_interval_time;  /* slack interval duration */
  systime_t     cpu_ticks;            /* precise instant when the linux kernel filled this message*/
  systime_t     TimeStamp;           /* unique identifier of each message to let trace */
} sched_mssg_t;





#ifdef ARTIS_CONF_METAREASONING
#define ADVANCE_FINAL_PART_FIFO    1
#endif


#endif // #ifndef__ARTIS_FIFO_DEFINITION_H

