
#include "include/como.h"
#include "parser/red.h"
#include <unistd.h>
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>


/// Maximum number of characters per line
#define TAMLINEA 255
#define TAMMSG	 10240


int read_msg_from_file(char *arg1, char *mensaje) {
	FILE * fd;
	char linea[TAMLINEA];
		
	if((fd=fopen(arg1,"r+"))==NULL){
 		printf("\nError abriendo archivo\n\n");
		return -1;
 	}

	while ((fgets(linea,TAMLINEA,fd))!=NULL){
		strcat(mensaje,linea);
 	}
	
	return 0;
}

void processcommand(char *c) {
	char *arg0 = NULL;
	char *arg1 = NULL;
	char *arg2 = NULL;
	char msg[TAMMSG];
	int i, flag, sent = 0, rec = 0;
	
	for ( i = 0 ; i < TAMMSG ; i++ ) msg[i] = '\0';
	
	if (c)
		arg0 = strtok(c, " ");
	
	if (arg0)
		arg1 = strtok(NULL, " ");
	if (arg1)
		arg2 = strtok(NULL, " ");
	
	// QUIT	
	if ( strncmp("quit", arg0, 4) == 0 ) {
		printf("BYE BYE\n\n");
		exit(0);
	}
	else if ( strcmp("q", arg0) == 0 ) {
		printf("BYE BYE\n\n");
		exit(0);
	}
	
	// SEND
	else if ( strncmp("send", arg0, 4) == 0 ) {  // Send a message
		if ( (arg1) && (arg2) ) {  // If the arguments are a message and a conversation
			flag = read_msg_from_file(arg1, msg);
			if ( flag == 0) {
				como_send(msg, arg2);
				printf("\nMENSSAGE SENT\n\n");
			}
		} else printf("\nParameters missing\n\n");
	}
	
	// VIEW
	else if ( strncmp("view", arg0, 4) == 0 ) {
		print_msg_queue();
		printf("\n\n");
	}
	else if ( strncmp("vd", arg0, 2) == 0 ) {
		print_msg_queue();
		printf("\n\n");
	}
	
	// SO, SEND_OUTBOX
	else if ( strncmp("so", arg0, 2) == 0 ) {
		if (arg1) {	 // Max number of messages to send
			sent = como_send_outbox(atoi(arg1));
			printf("\n%d messages from the outbox sent\n\n", sent);
		} else printf("\nNumber of messages missing\n\n");
	}
	
	// VC
	else if ( strncmp("vc", arg0, 2) == 0 ) {
		if (arg1) {  // conv_id
			if (strcmp("#outbox#", arg1) == 0) { // Asking for the outbox
				printf("\n");
				print_outbox();
				printf("\n");
			} else {   // Asking for a conversation
				printf("\n");
				print_conv(arg1);
				printf("\n");
			}
		}
		else printf("\nMissing conversation-id\n\n");
	}
	
	// VO, VIEW OUTBOX
	else if ( strncmp("vo", arg0, 2) == 0 ) {
		printf("\n");
		print_outbox();
		printf("\n");
	}
	
	// RX, REPLY ERRORS
	else if ( strncmp("rx", arg0, 2) == 0 ) {
		sent = 0;
		if (arg1) {	 // Max number of messages to reply
			sent = como_send_erroneous(atoi(arg1));
			printf("\n%d erroneous messages replied\n\n", sent);
		} else printf("\nNumber of messages missing\n\n");
	}
	
	// LISTEN
	else if ( strncmp("listen", arg0, 6) == 0 ) {
		rec = 0;
		if (arg1) {	 // Max number of messages to reply
			rec = como_receive_inbox(atoi(arg1));
			printf("\n%d messages received\n\n", rec);
		} else printf("\nMissing number of messages\n\n");
	}
	
	// RECEIVE
	else if ( strncmp("receive", arg0, 7) == 0 ) {
		if (arg1) {	 // conv_id
			como_receive(arg1);
			printf("\nMENSSAGES RECEIVED\n\n");
		} else printf("\nMissing conversation-id\n\n");
	}
	
	// CREATE
	else if ( strncmp("create", arg0, 6) == 0 ) {
		if (arg1) {	 // conv_id
			como_create_conversation(arg1);
			printf("\nCONVERSATION CREATED\n\n");
		} else printf("\nMissing conversation-id\n\n");
	}
	
	// CLOSE
	else if ( strncmp("close", arg0, 5) == 0 ) {
		if (arg1) {	 // conv_id
			flag = como_close_conversation(arg1);
			if (flag > -1)
				printf("\nCONVERSATION %s CLOSED\n\n", arg1);
			else
				printf("\nNO SUCH CONVERSATION\n\n");
		} else printf("\nMissing conversation-id\n\n");
	}
	
	// ACCEPT
	else if ( strncmp("accept", arg0, 6) == 0 ) {
		if (arg1) {	 // conv_id
			como_accept_msg(arg1);			
		} else printf("\nMissing conversation-id\n\n");
	}
	
	// DELETE
	else if ( strncmp("del", arg0, 3) == 0 ) {
		if (arg1) {	 // conv_id
			flag = como_delete_conversation(arg1);
			if (flag > -1)
				printf("\nCONVERSATION %s DELETED\n\n", arg1);
			else
				printf("\nNO SUCH CONVERSATION IN TERMINATED CONVERSATIONS LIST\n\n");
		} else printf("\nMissing conversation-id\n\n");
	}
	
	// HELP
	else if ( strncmp("help", arg0, 2) == 0 ) {
		printf("\nCommands:\n");
		printf("\t'view' or 'vd':\t\t\tView message queue\n");
		printf("\t'quit' or 'q':\t\t\tExit the console\n");
		printf("\t'send file':\t\t\tSends the message stored in 'file'\n");
		printf("\t'vc conversation_id':\t\tView conversation named after 'conversation_id'\n");
		printf("\t'receive conversation_id':\tReceive messages from the conversation 'conversation_id'\n");
		printf("\t'create conversation_id':\tCreates a conversation with the given name\n");
		printf("\t'close conversation_id':\tCloses the named conversation. It should be placed in the\n\t\t\t\t\t terminated conversations list for this command to work\n");
		printf("\t'accept conversation_id':\tAccepts the top message from the given conversation\n");
		printf("\t'delete conversation_id':\tPermanently deletes the given conversation\n");
		printf("\t'vo':\t\t\t\tView Outbox\n");
		printf("\t'so N':\t\t\t\tFlush outbox sending as much as 'N' messages\n\n");
	}
	
	else if ( strncmp("\?", arg0, 2) == 0 ) {
printf("\nCommands:\n");
		printf("\t'view' or 'vd':\t\t\tView message queue\n");
		printf("\t'quit' or 'q':\t\t\tExit the console\n");
		printf("\t'send file':\t\t\tSends the message stored in 'file'\n");
		printf("\t'vc conversation_id':\t\tView conversation named after 'conversation_id'\n");
		printf("\t'receive conversation_id':\tReceive messages from the conversation 'conversation_id'\n");
		printf("\t'create conversation_id':\tCreates a conversation with the given name\n");
		printf("\t'close conversation_id':\tCloses the named conversation. It should be placed in the\n\t\t\t\t\t terminated conversations list for this command to work\n");
		printf("\t'accept conversation_id':\tAccepts the top message from the given conversation\n");
		printf("\t'delete conversation_id':\tPermanently deletes the given conversation\n");
		printf("\t'vo':\t\t\t\tView Outbox\n");
		printf("\t'so N':\t\t\t\tFlush outbox sending as much as 'N' messages\n\n");
	}

}

void *mi_listen(void *args) {
	int i = 0;
	Tmensaje m;
	Tmensaje *pm;
	//char direccion_ip[TAMLINEA];
	
	pm = &m;
			
	while(1){
		i = recibir(pm);
  		if (  i > -1  ) 
			put_msg(pm, pm->conversation_id);		
	}
}

int main() {
	char command[TAM_ETIQUETA];
	//pthread_t l;
	//void *nada;
	int i;
	char direccion_ip[TAMLINEA];

	como_init();

	//printf("BEFORE INICIALIZAR_RED\n");		
	//inicializar_red();
	//printf("AFTER DEL INICIALIZAR_RED\n");	
	sleep(2);
  	//obtener_mi_direccion_ip(direccion_ip);
	
	//pthread_create(&l, NULL, mi_listen, nada);

	//sleep(3);
	
	printf("\n\nCOmo COnsole v.01\n=================\nby Gustavo Aranda\n\n");
	while (1) {
		for ( i = 0 ; i < TAM_ETIQUETA ; i++ ) command[i] = '\0';  // Limpia "command"
		printf(">>> ");
		gets(command);
		printf("COMMAND: %s\n", command);
		if ( command[0] != '\0' )
			processcommand(command);	
	}
}
