
#include "include/como.h"
#include "structures/types.h"
#include "parser/debug.h"
#include "parser/global.h"
#include "parser/red.h"
#include "queues/msg_structures.h"
#include "msg_util.h"

extern int get_conv_index(char *c_id);
extern int get_conv_index_no_create(char *c_id);
extern void deltopmsg(int conv);
extern int is_pending(int c);
extern int is_active(int c);
extern int is_terminated(int c);
extern int delete_conv_perm(char *conv_id);


int como_init() {
	char direccion_ip[TAM_DIRECCION_IP];

	initmsglist();

	inicializar_red();
	
        obtener_mi_direccion_ip(direccion_ip);

	fprintf(stderr,"========== COMO_INIT ==========\n");
	fprintf(stderr,"========== COMO_INIT ==========\n");
	fprintf(stderr,"========== COMO_INIT ==========\n");

	return 0;
}

int como_send(char *msg, char* conv_id) {
	Tmensaje fields;
	int c;
	
	// Parsing the message 
	c = extraer_campos_mensaje(msg, &fields);
	
	if (c != -1) {	// The message is processed unless there are parsing errors
		
		// CAUTION: Setting the 'conversation_id' of the message
		// that was just delivered to the 'conv_id' parameter of the function call
		// OVERWRITING the previous value
		strcpy(fields.conversation_id, conv_id);
	
		// Putting it in the outbox. Ready to be shipped
		put_in_outbox(&fields);
	}
	
	return ERROR_NONE;
}

Tmensaje *como_receive(char *conv_id) {
	int c;
	Tmensaje *m = NULL;
	
	c = get_conv_index(conv_id);
	
	// If it's not a terminated one, set it to active 
	if ( is_terminated(c) == 0 )
		set_active_conversation(c);
	
	m = get_top_from_conv(c);

	// We double check things
	if ( m == NULL )
		return NULL;
	else {
		if ( strcmp(m->conversation_id, conv_id) == 0 )
			return m;
		else
			return NULL;
	}
}

Tmensaje *como_receive_service(char *service) {
	// FOR NOW it behaves as 'como_receive' 
	return como_receive(service);
}

int como_create_conversation(char *conv_id) {	
	return get_conv_index(conv_id);
}

int como_close_conversation(char *conv_id) {
	int c;
	
	c = get_conv_index_no_create(conv_id);
	if ( c > -1 ) {
		set_terminated_conversation(c);
		return 0;
	} else
		return c;  // Some error
}

void como_accept_msg(char *conv_id) {
	int c;
	
	c = get_conv_index_no_create(conv_id);
	if ( c > -1)
		deltopmsg(c);
	
	return;
}

int como_delete_conversation(char *conv_id) {
	//int c;
	
	return delete_conv_perm(conv_id);
}

int como_send_erroneous(int n) {
	//int c;
	int i, count = 0;
	msgnode *m;
	
	if (msg_queue[0].top == -1) return 0;  // No menssages
	
	m = &(msg_queue[0].v[msg_queue[0].top]);  // 'top' first
	for ( i = 0 ; i < n ; i++ ) {
		if (m->code != -2)	{    // If the message is usefull, process it
			// Reply the message with a NOT_UNDERSTOOD one
			reply_to_error(&(m->msg));
			
			// Increment the message counter for this round 
			count++;
			
			// Deleting the message from the erroneous queue
			deltopmsg(0);  // "deltopmsg" will update 'top'
			
			// If there are no more messages to send, break the loop 
			if (msg_queue[0].top == -1) return count;
				
			m = &(msg_queue[0].v[msg_queue[0].top]);
		}
	}
	
	return count;
}

int como_send_outbox(int n) {
	int i, c, count = 0;
	msgnode *to_send;
	
	if (outbox.top == -1) return 0;  // Empty mailbox, we leave as we came
	
	to_send = &(outbox.v[outbox.top]);  // 'top' first
	for ( i = 0 ; i < n ; i++ ) {
		if (to_send->code != -2)	{    // If the message is usefull, process it
			// Send the message through the network 
			enviar(to_send->msg);
			
			// Increment the message counter for this round 
			count++;
		
			// 'put_msg' returns the index the conversation has in the 'msg_queue' queue.
			// It is used as the parameter for 'set_active_conversation'
			//c = put_msg(&(to_send->msg), to_send->msg.conversation_id);
			
			// We used to call 'put_msg' in order to store the message in the main queue,
			// but that was annoying the 'como_recive' function, so instead of storing it
			// we just get the conversation index (creating it if it's necessary) and make
			// the conversation active
			c = get_conv_index(to_send->msg.conversation_id);
			set_active_conversation(c);
			
			// Deleting message from the outbox 
			deltopmsg(-1);  // 'deltopmsg' will update 'top'
			
			// If there's nothing else to send, break the loop 
			if (outbox.top == -1) return count;
			
			to_send = &(outbox.v[outbox.top]);
		}
	}
	
	// Return the effective number of messages really sent  
	return count;
}

int como_receive_inbox(int n) {
	int i, c, id, count = 0;
	Tmensaje m;
	
	// Messages should have been stored in the socket's buffer
	// so, a simple call to 'recibir' should be enough to get them
	for ( i = 0 ; i < n ; i++ ) {
		c = recibir(&m);
		if ( c >= 0 ) {
			id = put_msg(&m, m.conversation_id);
			
			// If it's not an error message
			if ( id != 0 ) {			
				// ... and it's not associated to a new or pending conversation,
				// its a brand new conversation and must be signed up in the pending list
				if ( (is_pending(id) == 0) && (is_active(id) == 0) )
					set_pending_conversation(id);
			}
			
			count++;
		}
	}
	
	return count;
}
