
#ifndef COMO_H
#define COMO_H

#include "../parser/mensaje/mensaje.h"
#include "../queues/msg_structures.h"

/// Initializes CoMo
/// @return Error if there is one, zero otherwise 
int como_init();

/// Send a FIPA message
/// @param msg The FIPA message
int como_send(char *msg, char *conv_id);

/// Receive a message of a certain conversation
///
/// The message is NOT deleted from the messages list,
/// it is merely referenced
/// @param conv_id Conversation identifier
/// @return Pointer to the message
Tmensaje *como_receive(char *conv_id);

/// Receive a message relative a service offered by the In-Agent
/// @param service The offered service
/// @return Pointer to the message
Tmensaje *como_receive_service(char *service);

/// Creates a new conversation
/// @param conv_id Desired identifier for the conversation
/// @return Error code if there is one
int como_create_conversation(char *conv_id);

/// Closes a conversation
/// @param conv_id Conversation to close
/// @return Error code if there is one
int como_close_conversation(char *conv_id);

/// Accepts and deletes effectively the first message in the message
/// list of the conversation
/// @param The conversation from which the message is accepted
void como_accept_msg(char *conv_id);

/// Permanently erases a conversation form the 'Terminated Conversations' list
/// @param conv_id  The conversation to delete
/// @return Error code if there is one 
int como_delete_conversation(char *conv_id);

/// Replies erroneous messages
///
/// Read until 'n' messages from the erroneous queue,
/// build their respective replies (NOT_UNDERSTOOD) and
/// send them without passing through the regular outbox
/// TODO: Are the messages erased from the queue?
/// @param n Max number of erroneous messages to process
/// @return Number of erroneous messages processed
int como_send_erroneous(int n);


/// Send (flush) messages from the oubox
///
/// Send until 'n' messages from the outbox
/// @param n Max number of messages to process
int como_send_outbox(int n);

/// Receive a given number of messages
///
/// @param n Max number of messages to receive
/// @return Number of messages received
int como_receive_inbox(int n);

#endif
