
#ifndef MENSAJE_H
#define MENSAJE_H

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include "../utilleria/utilleria.h"

//#define TAM_ETIQUETA 1000
//#define TAM_CAMPO 1000
#define TAM_ETIQUETA 10240
#define TAM_CAMPO 10240


typedef struct {
  char performative [TAM_ETIQUETA];
  char sender [TAM_ETIQUETA];
  char receiver [TAM_ETIQUETA];
  char reply_to [TAM_ETIQUETA];
  char content [TAM_CAMPO];
  char language [TAM_ETIQUETA];
  char encoding [TAM_ETIQUETA];
  char ontology [TAM_ETIQUETA];
  char protocol [TAM_ETIQUETA];
  char conversation_id [TAM_ETIQUETA];
  char reply_with [TAM_ETIQUETA];
  char in_reply_to [TAM_ETIQUETA];
  char reply_by [TAM_ETIQUETA];
  char x_subject [TAM_ETIQUETA];
  char x_priority [TAM_ETIQUETA];
  
  //campo q contendra el valor del tiempo
  struct timeval tiempo_emision;
  struct timeval tiempo_recepcion;

} Tmensaje;


/// Extrae campos de mensaje desde una cadena
///
/// Funcion que dado un mensaje contenido en una cadena (msg) extrae todos
/// sus campos y los devuelve en la estructura campos_mensaje
/// @param msg El mensaje a analizar
/// @param campos_mensaje La estructura a rellenar
/// @return 0 si todo funciono correctamente, -1 en otro caso
int extraer_campos_mensaje(char msg[], Tmensaje * campos_mensaje);



/// Forma un mensaje a partir de una estructura
///
/// Funcion que a partir de los campos del mensaje forma una cadena.
/// Ademas quita la repeticion de todos los espacios en blanco,
/// caracteres de nueva linea y tabuladores.
/// @param campos_mensaje La estructura con el mensaje
/// @param mensaje El mensaje a analizar
/// @return 0 si todo funciono correctamente, -1 en otro caso
int formar_mensaje(Tmensaje campos_mensaje, char mensaje[TAM_CAMPO]);



/// Imprime los campos de un mensaje
///
/// Funcion q presenta los distintos campos q hay rellenados de un mensaje
/// @param campos_mensaje La estructura con el mensaje
void imprimir_campos_mensaje(Tmensaje campos_mensaje);


/// Retira caracteres inutiles
///
/// Funcion que quita todos los caracteres inutiles de una cadena de mensaje,
/// como son los retornos de carro, los tabuladores y los espacios sucesivos
/// @param msg Cadena con el mensaje
void quitar_caracteres_inutiles(char msg[]);



/// Inicializa una estructura de mensaje
///
/// Funcion que dada una estructura de tipo mensaje la inicializa
/// @param campos_mensaje La estructura de mensaje
void inicializar_campos_mensaje(Tmensaje * campos_mensaje);

/// Borra los campos de un mensaje
/// @param m El mensaje
void clear_msg(Tmensaje m);

#endif
