#include "transporte.h"
#include "../utilleria/utilleria.h"
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>

extern long summarize(char *data);
extern unsigned long int inet_addr(const char* cp);
extern char* inet_ntoa(struct in_addr in);
extern int summarize_chunk(unsigned char *partdata);

int connectsock(char * host);
void serialize_chunk(chunk *part, unsigned char *spart);

/// Definicion de un nuevo tipo para guardar todos los sockets con los que esta
/// conectado el cliente
  typedef struct{
  char direccion[TAM_DIRECCION_IP];
  int socket_cliente;
} tipo_cliente;

/// Vector de estructuras que guardara la informacion de los sockets conectados
static tipo_cliente cliente[MAX_ROBOTS];

/// Variable que contiene el socket de escucha del servidor
static int escucha;

/// Direccion Ip sobre la que esta ejecutandose el servidor
static char dir_ip_escucha[TAM_DIRECCION_IP];

static buffer_msg rcv_buffer[MAX_ROBOTS];


int  enviar_udp(char *  msg, char *  destino){
  int i=0;
  int busqueda=TRUE;
  int ncaracteres=0;
  char aux[TEXT_LENGTH];
  long sum;

  // DESACTIVADO
  // Pegamos el resumen del mensaje justo al final del mensaje.
  // Curiosamente, el calculo del resumen no cuenta el propio resumen,
  // o sea, es un resumen de lo que viene antes
  // !DESACTIVADO
  sum = summarize(msg);
  ltoa(sum, aux);
  //strcat(msg, aux);   // DESACTIVADO
  	
  //Comprobamos q el socket de enviar este conectado, en caso contrario lo conectamos

  //Hacemos un recorrido por la tabla
    while ((strlen(cliente[i].direccion)!=0) && (i<MAX_ROBOTS)){
    busqueda=strcmp(cliente[i].direccion,destino);
    //printf("\nValor de las variables %i",busqueda);
    //fflush(stdout);
    if (busqueda==FALSE)
      break;
    i++;
  }

 //Si no lo hemos encontrado lo creamos nuevo
  if (busqueda!=FALSE){
    strcpy(cliente[i].direccion,destino);
    cliente[i].socket_cliente=connectsock(destino);
    //printf("\nSe ha aadido el nuevo host %s",cliente[i].direccion);
	strcpy(aux, "Se ha aadido el nuevo host ");
	strcat(aux, cliente[i].direccion);
	debug_print(aux, 2);
  }

  fflush(stdout);

  //Enviamos los datos
  ncaracteres=send(cliente[i].socket_cliente,msg,strlen(msg),0);

  //Devolvemos el numero de caracteres q se han enviado por la red
  return ncaracteres;
}




/// Conecta a un socket remoto
///
/// Funcion que crea y realiza la conexion con un socket remoto para enviar informacion.
/// @param host El host remoto
/// @return El socket creado y ya conectado
int connectsock(char * host){
  struct hostent * phe; //Contiene la informacion que se obtine a partir de la direccion especificada
  struct sockaddr_in servidor; //Contiene la informacion necesaria para conectarse con el servidor

  //int s,type;
  int s;
  char aux[TEXT_LENGTH];

  //Inicializamos la estructura
  bzero((char *) &servidor, sizeof (servidor));
  servidor.sin_family=AF_INET; //Lo definimos del tipo "direccion de internet", es decir, direccion IP
  servidor.sin_port=htons(SOCKET_RECIBIR);  //Puerto en el que escucha el servidor

  //Hallamos la direccion del host destino, y la copiamos
  if( phe = gethostbyname(host) )
    bcopy(phe->h_addr,(char *) &servidor.sin_addr,phe->h_length);
  else
    if( (servidor.sin_addr.s_addr=inet_addr(host))==INADDR_NONE ){
      perror("Error al hallar el host");
	  debug_print("Error al hallar el host", 2);
      fflush(stdout);
    }

  //Creacion del socket en si
  s=socket(AF_INET,SOCK_DGRAM,0);

  //Si hay un fallo en la creacion
  if (s<0){
    perror("Error en la creacion del socket");
	debug_print("Error en la creacion del socket", 2);
    return -1;
  }

  //Iniciamos la conexion con el servidor
  if ((connect(s,(struct sockaddr * ) & servidor, sizeof(servidor))<0) && 
      (strcmp(host,DIR_IP_GRUPO_MULTICAST)!=0)){
    perror("\nConexion rechazada por el servidor");
    debug_print("Conexion rechazada por el servidor", 2);
    //printf("\n\tEl problema ha estado en %s",host);
	strcpy(aux, "El problema ha estado en ");
    strcat(aux, host);
	debug_print(aux, 2);
  }

  return s; //Devuelve el socket creado y ya conectado
}





int  recibir_udp(char  msg[],char dir_cliente[]){
  //char buffer[TAM_MSG];
  int caracteres;
  int long_cliente;
  struct sockaddr_in cliente;


  //Recibimos los datos
  //printf("\nRecibiendo...\n");
  //debug_print("Recibiendo datos...", 2);
  //Vemos a ver si hay datos disponibles en el socket
  fflush(stdout);

  long_cliente=sizeof(struct sockaddr);
  if((caracteres=recvfrom(escucha,msg,TAM_MSG-1,0, (struct sockaddr *) &cliente, & long_cliente))>=0){
    msg[caracteres]='\0';
    strcpy(dir_cliente,(char *) inet_ntoa(cliente.sin_addr));
  }
  else{
    //    perror("Error recibiendo datos");
	return -1;

  }
  return (caracteres);
}



/// Crea un socket de escucha
///
/// Funcion que crea un socket para la recepcion de mensajes
/// @return el socket creado y conectado, -1 en caso de error
int crear_socket_escucha(void){
  int sock;
  struct sockaddr_in name;

  //printf("\nCreando socket de escucha...");
  debug_print("Creando socket de escucha...", 2);

  //Creamos el socket UDP para recibir datos
  sock=socket(PF_INET,SOCK_DGRAM,0);

  //Si hay un fallo en la creacion del socket lo mostramos
  if (sock<0){
    perror("Error abriendo socket UDP");
	debug_print("Error abriendo socket UDP", 2);
    return -1;
  }

  //Rellenamos la estructura con los datos del servidor
  name.sin_family=AF_INET;
  name.sin_addr.s_addr=htonl(INADDR_ANY);
  name.sin_port=htons(SOCKET_RECIBIR);

  //Le asociamos un nombre al socket
  if (bind(sock, (struct sockaddr *) & name, sizeof name)<0){
    perror("Error asociando nombre al socket");
	debug_print("Error asociando nombre al socket", 2);
    return -1;
  }

  //Le cambiamos las propiedades al socket recien creado para que sea no bloqueante
  if (fcntl(sock,F_SETFL,O_NONBLOCK)!=0){
    perror("Error al hacer el socket no bloqueante");
	debug_print("Error al hacer el socket no bloqueante", 2);
    return -1;
  }

  return sock;
}





/// Presenta los sockets de envio
void presenta_socket_enviar(void){
  int i;
  char aux[TEXT_LENGTH];
  char aux2[255];

  //printf("\n\n");
  for (i=0;i<=MAX_ROBOTS;i++)
    //printf("Direccion: %s \t socket:%i\n",cliente[i].direccion,cliente[i].socket_cliente);
    strcpy(aux, "Direccion: ");
    strcat(aux, cliente[i].direccion);
    strcat(aux, " Socket: ");
    itoa(cliente[i].socket_cliente, aux2);
    strcat(aux, aux2);
    debug_print(aux, 2);
}



char * obtener_dir_ip_servidor(){
  char nom_servidor[TAM_DIRECCION_IP];
  struct hostent * hent_servidor;
//Cogemos la direccion IP de la maquina en la q se esta ejecutando el servidor
  //printf("\nobtener");
  gethostname(nom_servidor,sizeof(nom_servidor));
  //printf("\nservidor %s",nom_servidor);
  if ((hent_servidor=gethostbyname(nom_servidor))==NULL){
    herror("gethostbyname"); //Mostramos el mensaje de error
    return NULL;
  }

  return (char *) inet_ntoa(*((struct in_addr *)hent_servidor->h_addr));
}


/// Pone a un socket a escuchar en modo multicast
/// @param s El socket
/// @return 0 si todo ha ido bien, -1 en caso contrario
int escucha_multicast(int s){
  //Declaramos la estructura necesaria para trabajar con los socket multicast
  struct ip_mreq mreq;

  //printf("\nMe llega %i,%s",s,dir_ip_escucha);
  /* Usamos setsockopt() para indicarle al nucleo q una el socket a un grupo multicast*/
  mreq.imr_multiaddr.s_addr=inet_addr(DIR_IP_GRUPO_MULTICAST); //le decimos la direccion del grupo
  if((mreq.imr_interface.s_addr=inet_addr(obtener_dir_ip_servidor()))==-1){ //le decimos la direccion del servidor comprobando q no ocurre ningun error
    //printf("\nError recogiendo direccion del servidor\n");
	debug_print("Error recogiendo direccion del servidor", 2);
    return -1;
  };

  if (setsockopt(s,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)) < 0) {
    perror("Error uniendo al grupo multicast");
	debug_print("Error uniendo al grupo multicast", 2);
    return -1;
  }
  return 0;
}


 //Devuelve la direccion ip de la maquina sobre la q esta oyendo el servidor
//void obtener_mi_direccion_ip(char * direccion){
//  strcpy(direccion,dir_ip_escucha);
//}


int inicializar_udp(){
  int i=0;
  int j=0, k=0;

  if (escucha!=0) {
    close(escucha);
    //printf("\nEscucha cerrado");
	debug_print("Escucha cerrado", 2);
  }

  //Creamos el socket de escucha
  escucha=crear_socket_escucha();

  strcpy(dir_ip_escucha,obtener_dir_ip_servidor());

  //Modificamos el socket de escucha, aadiendolo al grupo multicast
  if(escucha_multicast(escucha)==-1){
    return -1;
  }

  //INICIALIZACION DE LAS ESTRUCTURAS COMO CLIENTE
  for(i=0;i<MAX_ROBOTS;i++){
    strcpy(cliente[i].direccion,"");
    if (cliente[i].socket_cliente!=0)
      close(cliente[i].socket_cliente);
    cliente[i].socket_cliente=0;
  }

  //Creamos el socket de envio multicast y lo ponemos en la posicion 0
  strcpy(cliente[0].direccion,"multicast");
  cliente[0].socket_cliente=connectsock(DIR_IP_GRUPO_MULTICAST);

  // Limpiar buffers de recepcion
  for ( j = 0 ; j < MAX_ROBOTS ; j++ )
	  for ( k = 0 ; k < MSG_MAX_SIZE ; k++ )
		  rcv_buffer[j][k] = '\0';
  
  return 0;
}


void serialize_chunk(chunk *part, unsigned char *spart) {
	//char aux[12];
	//int i;
	//char debug_aux[TEXT_LENGTH];
	
	return;
}


int send_parted(char *msg, char *dest) {
	chunk part;
	chunk *p_part;
	unsigned char spart[12+CHUNK_SIZE];  // El chunk "serializado"
	int i=0;
	int j=0;
	int k=0;
	int h=0;
  	int busqueda=TRUE;
  	int ncaracteres=0;
  	unsigned char aux[TEXT_LENGTH];
  	//int sum;
	int total_chunks = 0, index_chunks = 0;
	char debug_aux[TEXT_LENGTH];
	int msg_length = 0;
	
	
	p_part = &part;
	
	sprintf(debug_aux, "send_parted: Me llega el mensaje <%s> de longitud %d", msg, strlen(msg));
	debug_print(debug_aux, 2);
		
	//Hacemos un recorrido por la tabla
    while ((strlen(cliente[k].direccion)!=0) && (k<MAX_ROBOTS)){
    	busqueda=strcmp(cliente[k].direccion,dest);
    	//printf("\nValor de las variables %i",busqueda);
    	//fflush(stdout);
    	if (busqueda==FALSE)
      		break;
    	k++;
  	}

 	//Si no lo hemos encontrado lo creamos nuevo
  	if (busqueda!=FALSE){
    	strcpy(cliente[k].direccion,dest);
    	cliente[k].socket_cliente=connectsock(dest);
    	//printf("\nSe ha aadido el nuevo host %s",cliente[i].direccion);
		strcpy(aux, "Se ha aadido el nuevo host ");
		strcat(aux, cliente[k].direccion);
		debug_print(aux, 2);
  	}

 	fflush(stdout);

	// Pre-calculamos algunos campos de los chunks
	total_chunks = ( strlen(msg) / CHUNK_SIZE ) + 1;
	part.total = total_chunks;
  	
	sprintf(debug_aux, "Precalculo de datos de envio hecho. Chunks totales = %d", total_chunks);
	debug_print(debug_aux, 2);
	
	//Enviamos los datos
  	//ncaracteres=send(cliente[i].socket_cliente,msg,strlen(msg),0);
	msg_length = strlen(msg);
	index_chunks = 0;
	for ( i = 0 ; i < msg_length ; i+=CHUNK_SIZE ) {
		
		// Limpieza
		for ( h = 0 ; h < CHUNK_SIZE ; h++ ) part.data[h] = '\0';
		for ( h = 0 ; h < (CHUNK_SIZE+12) ; h++ ) spart[h] = '\0';
				
		sprintf(debug_aux, "Chunk %d Iniciado", index_chunks);
		debug_print(debug_aux, 2);
		for ( j = 0 ; j < CHUNK_SIZE ; j++) part.data[j] = (unsigned char)msg[i+j];
		sprintf(debug_aux, "DATA %s copiado. %d caracteres copiados", part.data, j);
		debug_print(debug_aux, 2);
		if ( j < CHUNK_SIZE-1 ) {  // Ultimo chunk
			sprintf(debug_aux, "Ultimo Chunk");
			debug_print(debug_aux, 2);			
			for ( j = j ; j < CHUNK_SIZE ; j++ ) part.data[j] = '\0';  // Padding
		}
		part.index = index_chunks;
		part.sum = summarize_chunk(part.data);
		sprintf(debug_aux, "Sum del Chunk: %d", part.sum);
		debug_print(debug_aux, 2);
		debug_print("Chunk por serializar",2);
		//serialize_chunk(p_part, spart);
		//##### SERIALIZE CHUNK 
		
		
		sprintf(debug_aux, "serialize_chunk: %d_%d_%d_%s", part.index, part.total, part.sum, part.data);
		debug_print(debug_aux, 3);
	
		sprintf(spart, "%02d#%02d#%06d",part.index,part.total,part.sum);
			
		for ( h = 12 ; h < (CHUNK_SIZE+12) ; h++ ) spart[h] = (unsigned char)part.data[h-12];
	
		sprintf(debug_aux, "serialize_chunk: spart = %s", spart);
		debug_print(debug_aux, 3);
		
		
		
		// ######### END OF SERIALIZE CHUNK
		strcpy(debug_aux, "Chunk serializado:");
		fflush(stdout);
		debug_print(debug_aux, 2);
		strncpy(debug_aux, spart, 12+CHUNK_SIZE);
		debug_print(debug_aux, 2);
		ncaracteres+=send(cliente[k].socket_cliente, (void *)spart, (CHUNK_SIZE+12), 0);		
		debug_print("Chunk enviado",2);
		index_chunks++;
	}

  	//Devolvemos el numero de caracteres q se han enviado por la red
  	return ncaracteres;	
}

//int receive_udp(char msg[], char dir_client[]) {
//	chunk packets[1024];
	
//}


void clean_rcv_buffer(int index_client) {
	int i;
	
	for  ( i = 0 ; i < MSG_MAX_SIZE ; i++ ) {
		rcv_buffer[index_client][i] = '\0';
	}
}

int receive_parted(char  msg[], char dir_cliente[]) {
  //char c[1];
  char buffer[CHUNK_SIZE+12];   // Para no quedarnos cortos
  int caracteres = 1;
  int total_caracteres = 0;
  int long_cliente;
  struct sockaddr_in cliente;
  //int total_chunks;
  //int last_chunk = -1;
  chunk part;
  //int i=0, j=0;
  //char aux[7];
  int sum;
  char debug_aux[TEXT_LENGTH];
  int eop = 0;
  int index_client;
  char nothing[TEXT_LENGTH];

  //Recibimos los datos
  //debug_print("Recibiendo datos...", 2);
  //Vemos a ver si hay datos disponibles en el socket
  fflush(stdout);

  long_cliente=sizeof(struct sockaddr);
  
  
  //while ( (eop == 0) && (caracteres != 0) ) {  // De momento, para hacer pruebas
  if((caracteres=recvfrom(escucha, buffer, CHUNK_SIZE+12, 0, (struct sockaddr *) &cliente, & long_cliente))>=0) {
	total_caracteres += caracteres;
	sprintf(debug_aux, "receive_parted: Me ha llegado el buffer %s", buffer);
	debug_print(debug_aux, 2);
	
	// We use the "nothing" variable just to avoid using NULL and getting a warning.
	// It makes compiling output cleaner
	sscanf(buffer, "%02d#%02d#%06d%s", &part.index, &part.total, &part.sum, nothing);
  	
	// Buscar el resto
  	strncpy(part.data, buffer + 12*sizeof(char), CHUNK_SIZE);
	
  	sprintf(debug_aux, "Recibido el chunk: %d %d %d %s", part.index, part.total, part.sum, part.data);
  	debug_print(debug_aux,2);
  
  	sum = summarize_chunk(part.data);
  	if ( sum != part.sum )  // Falla el checksum ALERTA!!
  		debug_print("receive_parted: Fallo de checksum detectado",2);
  
  	strcpy(dir_cliente,(char *) inet_ntoa(cliente.sin_addr));
  	index_client = summarize(dir_cliente) % MAX_ROBOTS;
	strncat(rcv_buffer[index_client], part.data, CHUNK_SIZE);  // Comprobado el cs, reconstruimos el mensaje	
  	if ( part.index == part.total-1 ) {  // Tengo un mensaje completo
	  eop = 1;
	  strcpy(msg, rcv_buffer[index_client]);
	  clean_rcv_buffer(index_client);
	  return caracteres;
  	} else return -1;
	
  }
  

  
  //sprintf(debug_aux, "Caracteres leidos %d",caracteres);
  //debug_print(debug_aux, 2);
  

  //}
  
  return ERROR_NONE;
}
