
#include "msg_structures.h"

#define DICT_SIZE MQ_SIZE

extern long summarize(char *data);

int get_conv_index(char *c_id);
int get_conv_index_no_create(char *c_id);
int is_terminated(int c);
int del_conv_from_terminated(int c);

/// Instancia efectiva de "msg_queue"
msgvector the_msg_queue[MQ_SIZE];
msgvector *msg_queue;

/// Conversaciones activas
///
/// Vector de enteros que hacen referencia a los indices de la cola de mensajes
int active_convs[MQ_SIZE];

/// Conversaciones pendientes
///
/// Vector de enteros que hacen referencia a los indices de la cola de mensajes
int pending_convs[MQ_SIZE];

/// Conversaciones acabadas
///
/// Vector de enteros que hacen referencia a los indices de la cola de mensajes
int terminated_convs[MQ_SIZE];


typedef struct _dict_pair {
    char key[TEXT_LENGTH];
    int value;
} dict_pair;

typedef dict_pair dict[DICT_SIZE];

/// Diccionario que asocia un nombre de conversacion a un indice
dict names_table;

void init_dict(dict d);

void init_conv_list(int *cl)
{
    int i;

    for (i = 0; i < MQ_SIZE; i++)
	cl[i] = -1;
}

void initmsglist()
{
    int i, j;
    //char aux[TEXT_LENGTH];


    msg_queue = (msgvector *)the_msg_queue;
	

    for (i = 0; i < MQ_SIZE; i++) {
	for (j = 0; j < MQ_SIZE; j++) {
	    //sprintf(aux, "Inicializando elemento %d %d", i, j);
	    //debug_print(aux, 1);
		
		// TODO: Sacar la lineas que no tienen 'v' del bucle
	    msg_queue[i].firstfree = 0;
	    msg_queue[i].last = -1;
	    msg_queue[i].top = -1;
	    msg_queue[i].v[j].code = -2;	// -2 == Vacio , -1 == Ultimo     
	}
    }

    
    init_dict(names_table);
    init_conv_list(pending_convs);
    init_conv_list(active_convs);
    init_conv_list(terminated_convs);

	
	// Para la bandeja de salida
	outbox.firstfree = 0;
	outbox.last = -1;
	outbox.top = -1;
	for (j = 0; j < MQ_SIZE; j++) {
		outbox.v[j].code = -2;
	}
}


/// Actualiza el valor 'firstfree' de una conversacion
/// @param conv La conversacion
void updatefree(int conv)
{
    int i;
    char aux[TEXT_LENGTH];

	if ( conv == -1 ) {  // Para la bandeja de salida
		for (i = 0; i < MQ_SIZE; i++)
			if (outbox.v[i].code == -2) {
	    		outbox.firstfree = i;
	    		sprintf(aux, "Nuevo firstfree en bandeja de salida: %d", i);
	    		debug_print(aux, 1);
	    		return;
			}
	} else {			 // Para una conversacion
		for (i = 0; i < MQ_SIZE; i++)
			if (msg_queue[conv].v[i].code == -2) { 
	    		msg_queue[conv].firstfree = i;
	    		sprintf(aux, "Nuevo firstfree en conv %d: %d", conv, i);
	    		debug_print(aux, 1);
	    		return;
			}
	}
}


/// Copia un mensaje
/// @param dest Mensaje de destino
/// @param orig Mensaje de origen
void copymsgnode(Tmensaje * dest, Tmensaje * orig)
{
    strcpy(dest->performative, orig->performative);
    strcpy(dest->sender, orig->sender);
    strcpy(dest->receiver, orig->receiver);
    strcpy(dest->reply_to, orig->reply_to);
    strcpy(dest->content, orig->content);
    strcpy(dest->language, orig->language);
    strcpy(dest->encoding, orig->encoding);
    strcpy(dest->ontology, orig->ontology);
    strcpy(dest->protocol, orig->protocol);
    strcpy(dest->conversation_id, orig->conversation_id);
    strcpy(dest->reply_with, orig->reply_with);
    strcpy(dest->in_reply_to, orig->in_reply_to);
    strcpy(dest->reply_by, orig->reply_by);
    strcpy(dest->x_subject, orig->x_subject);
    strcpy(dest->x_priority, orig->x_priority);

    dest->tiempo_emision.tv_sec = orig->tiempo_emision.tv_sec;
    dest->tiempo_emision.tv_usec = orig->tiempo_emision.tv_usec;

    return;
}


void deltopmsg(int conv) {
	int top;
	
	if (conv == -1) {  // Es en la bandeja de salida donde hay que operar
		top = outbox.top;
	
		// El nuevo 'top' sera el siguiente del viejo 'top'
		outbox.top = outbox.v[top].code;
	
		// Borramos el contenido del mensaje del viejo 'top'
		clear_msg(outbox.v[top].msg);
	
		// Marcamos el nodo como vacio
		outbox.v[top].code = -2;
		
		// Actualizamos el "firstfree"
		updatefree(-1);
		
		// Actualizamos el "last" a -1 si el array esta vacio
		if (outbox.top == -1) outbox.last = -1;
	} else {
		top = msg_queue[conv].top;
	
		// El nuevo 'top' sera el siguiente del viejo 'top'
		msg_queue[conv].top = msg_queue[conv].v[top].code;
	
		// Borramos el contenido del mensaje del viejo 'top'
		clear_msg(msg_queue[conv].v[top].msg);
	
		// Marcamos el nodo como vacio
		msg_queue[conv].v[top].code = -2;
		
		// Actualizamos el "firstfree"
		updatefree(conv);
		
		// Actualizamos el "last" a -1 si el array esta vacio
		if (msg_queue[conv].top == -1) msg_queue[conv].last = -1;		
	}
}


/// Anyade un mensaje a una conversacion
///
/// @param m El mensaje
/// @param conv La conversacion
void addmsgnode(Tmensaje * m, int conv)
{
    int free, last;
    char aux[TEXT_LENGTH];

    sprintf(aux, "Metiendo mensaje en conversacion %d", conv);
    debug_print(aux, 1);

    free = msg_queue[conv].firstfree;
    last = msg_queue[conv].last;
    msg_queue[conv].v[free].code = -1;	// El nuevo ultimo tiene codigo -1
    copymsgnode(&(msg_queue[conv].v[free].msg), m);
    msg_queue[conv].v[last].code = free;	// El antiguo ultimo ahora apunta al recien llegado
    msg_queue[conv].last = free;	// Actualizamos el nuevo ultimo
    updatefree(conv);		// Actualizamos quien es el primer nodo libre
	
	// Codigo por si el mensaje anyadido es el primero
	if ( msg_queue[conv].top == -1 )  // Si 'top' es -1 aun no se habia metido ningun mensaje
		msg_queue[conv].top = msg_queue[conv].last;
	

    return;
}


/// Hash segun identificador de conversacion
///
/// Funcion de hash que devuelve un indice para la msg_queue basado en el 'conversation_id' del mensaje.
/// Si el 'conversation_id' es '#COMO_ERROR#' devuelve siempre 0
/// @param c_id 'conversation_id' del mensaje
/// @return Indice para msg_queue
int hash_conv(char *c_id)
{
    int index;

    // Si el conversation_id es "#COMO_ERROR#" es que es un mensaje erroneo y por tanto
    // va a la fila 0 de la cola de mensajes
    if (strcmp(c_id, "#COMO_ERROR#") == 0)
		return 0;
    else {
		index = summarize(c_id);
		index = (index % (MQ_SIZE - 1)) + 1;
		return index;
    }
}


int put_msg(Tmensaje * m, char *conv_id)
{
    //int conv;
    int index;
    char aux[TEXT_LENGTH];

    sprintf(aux, "VOY A PONER EL MENSAJE %s", conv_id);
    debug_print(aux, 0);

    // Encontramos el indice adecuado a traves de la conversacion
    // que indica en el mensaje
    //index = hash_conv(m->conversation_id);
    index = get_conv_index(conv_id);

    sprintf(aux, "GET_CONV_INDEX devuelve %d", index);
    debug_print(aux, 0);

	// Si el mensaje pertenece a una conversacion terminada,
	// hay que meterlo en la cola de erroneos
	if ( is_terminated(index) == 1 ) {
		addmsgnode(m, 0);
		return 0;
	}
		
	// Si es un mensaje normal, lo metemos donde toque (si hay sitio)
    if (index != -1) {
		addmsgnode(m, index);
		return index;
    } else {
		sprintf(aux, "GESTOR DE LA COLA DE MENSAJES LLENO. NO SE PUEDE ANYADIR EL MENSAJE");
		debug_print(aux, 1);
		return -1;
    }
}

void print_outbox() {
	//int i = 0;
	int j = 0;
	char aux[TEXT_LENGTH], aux2[TEXT_LENGTH];

	return;  // Disabled by now

	// Imprimir la bandeja de salida
	sprintf(aux, "BANDEJA DE SALIDA: firstfree:%d last:%d top:%d",outbox.firstfree,
		outbox.last,outbox.top);
	debug_print(aux, 1);
	aux[0] = '\0';
	for (j = 0; j < MQ_SIZE; j++) {
	    if (outbox.v[j].code == -2) {	// Mensaje vacio
			sprintf(aux2, "0 ");
			strcat(aux, aux2);
	    } else {		// Mensaje con informacion
		sprintf(aux2, "<conv:%s from:%s to:%s perf:%s code:%d> ",
			outbox.v[j].msg.conversation_id,
			outbox.v[j].msg.sender,
			outbox.v[j].msg.receiver,
			outbox.v[j].msg.performative,
			outbox.v[j].code);
		strcat(aux, aux2);
	    }
	}
	debug_print(aux, 1);
}

void print_conv(char *conv_id) {
	int i = 0, j = 0;
	char aux[TEXT_LENGTH], aux2[TEXT_LENGTH];
	
	return;  // Disabled by now
	
	i = get_conv_index_no_create(conv_id);
		
	if ( i > -1 ) {
		sprintf(aux, "conv:%d firstfree:%d last:%d top:%d", i, msg_queue[i].firstfree, msg_queue[i].last, msg_queue[i].top);
		debug_print(aux, 1);
		//sprintf(aux, "");
		aux[0] = '\0';  // Suggested by Carlos
		for (j = 0; j < MQ_SIZE; j++) {
	    	if (msg_queue[i].v[j].code == -2) {	// Mensaje vacio
				sprintf(aux2, "0 ");
				strcat(aux, aux2);
	    	} else {		// Mensaje con informacion
				sprintf(aux2, "<msg:%s from:%s to:%s perf:%s code:%d> ",
					msg_queue[i].v[j].msg.conversation_id,
					msg_queue[i].v[j].msg.sender,
					msg_queue[i].v[j].msg.receiver,
					msg_queue[i].v[j].msg.performative,
					msg_queue[i].v[j].code);
				strcat(aux, aux2);
	    	}
		}
		debug_print(aux, 1);
	}
}

void print_msg_queue()
{
    int i = 0, j = 0;
    char aux[TEXT_LENGTH], aux2[TEXT_LENGTH];

    return;  // Disbaled for now

	// Imprimir la bandeja de salida
	print_outbox();
		
    // Imprimir la cola de mensajes
    for (i = 0; i < MQ_SIZE; i++) {
	sprintf(aux, "conv:%d firstfree:%d last:%d top:%d", i,
		msg_queue[i].firstfree, msg_queue[i].last, msg_queue[i].top);
	debug_print(aux, 1);
	//sprintf(aux, "");
	aux[0] = '\0';  // Suggested by Carlos
	for (j = 0; j < MQ_SIZE; j++) {
	    if (msg_queue[i].v[j].code == -2) {	// Mensaje vacio
		sprintf(aux2, "0 ");
		strcat(aux, aux2);
	    } else {		// Mensaje con informacion
		sprintf(aux2, "<msg:%s from:%s to:%s perf:%s code:%d> ",
			msg_queue[i].v[j].msg.conversation_id,
			msg_queue[i].v[j].msg.sender,
			msg_queue[i].v[j].msg.receiver,
			msg_queue[i].v[j].msg.performative,
			msg_queue[i].v[j].code);
		strcat(aux, aux2);
	    }
	}
	debug_print(aux, 1);
    }

    // Imprimir los vectores de conversaciones
    for (i = 0; i < TEXT_LENGTH; i++) aux[i] = '\0';		// Limpiar aux
	for (i = 0; i < TEXT_LENGTH; i++) aux2[i] = '\0';		// Limpiar aux2
    strcpy(aux, "CONVERSACIONES PENDIENTES: ");
    for (i = 0; i < MQ_SIZE; i++) {
		sprintf(aux2, "%d ", pending_convs[i]);
		strcat(aux, aux2);
	}
    debug_print(aux, 1);

	for (i = 0; i < TEXT_LENGTH; i++) aux[i] = '\0';		// Limpiar aux
	for (i = 0; i < TEXT_LENGTH; i++) aux2[i] = '\0';		// Limpiar aux2
    strcpy(aux, "CONVERSACIONES ACTIVAS: ");
    for (i = 0; i < MQ_SIZE; i++) {
		sprintf(aux2, "%d ", active_convs[i]);
		strcat(aux, aux2);
	}
    debug_print(aux, 1);
	
	for (i = 0; i < TEXT_LENGTH; i++) aux[i] = '\0';		// Limpiar aux
	for (i = 0; i < TEXT_LENGTH; i++) aux2[i] = '\0';		// Limpiar aux2
    strcpy(aux, "CONVERSACIONES TERMINADAS: ");
    for (i = 0; i < MQ_SIZE; i++) {
		sprintf(aux2, "%d ", terminated_convs[i]);
		strcat(aux, aux2);
	}
    debug_print(aux, 1);

    return;
}

void fprint_msg_queue()
{
    FILE *f;
    int i = 0, j = 0;
    char aux[TEXT_LENGTH], aux2[TEXT_LENGTH];

    return;  // Disabled for now

    f = fopen("receive.log", "w+");
    for (i = 0; i < MQ_SIZE; i++) {
	sprintf(aux, "conv:%d firstfree:%d last:%d ", i,
		msg_queue[i].firstfree, msg_queue[i].last);
	//fprintf(f, "%s\n",aux);
	for (j = 0; j < MQ_SIZE; j++) {
	    if (msg_queue[i].v[j].code == -2) {	// Mensaje vacio
		sprintf(aux2, "0 ");
		strcat(aux, aux2);
	    } else {		// Mensaje con informacion
		sprintf(aux2, "<msg:%s from:%s to:%s perf:%s code:%d> ",
			msg_queue[i].v[j].msg.conversation_id,
			msg_queue[i].v[j].msg.sender,
			msg_queue[i].v[j].msg.receiver,
			msg_queue[i].v[j].msg.performative,
			msg_queue[i].v[j].code);
		strcat(aux, aux2);
	    }
	}
	fprintf(f, "%s\n", aux);
    }

    fclose(f);
    return;

}


/// Inicializa un diccionario
/// @param d El diccionario
void init_dict(dict d)
{
    int i;

    // Para los mensajes erroneos
    strcpy(d[0].key, "#COMO_ERROR#");
    d[0].value = 0;

    for (i = 1; i < DICT_SIZE; i++) {
	strcpy(d[i].key, "\0");
	d[i].value = -1;
    }
}


/// Busca una clave (cadena de caracteres) en un diccionario
/// @param d El diccionario
/// @param c_id La clave
/// @return El valor de esa clave o -1 en caso de que la clave no exista
int has_key(dict d, char *c_id)
{
    int i;

    for (i = 0; i < DICT_SIZE; i++) {
	if (strcmp(d[i].key, c_id) == 0)	// Match
	    return d[i].value;
    }

    return -1;			// No match
}


/// Inserta una clave en un diccionario y le pone de valor inicial su indice interno del diccionario
/// @param d El diccionario
/// @param c_id La clave
/// @return El indice de la clave en el diccionario o -1 si no caben mas claves
int put_key(dict d, char *c_id)
{
    int i;

    for (i = 1; i < DICT_SIZE; i++)	// Buscar una posicin libre
	if (strcmp(d[i].key, "\0") == 0)
	    break;

    if (i == DICT_SIZE)
	return -1;		// No caben mas

    strcpy(d[i].key, c_id);
    d[i].value = i;

    return i;
}

/// Asocia un valor (entero) a una clave en un diccionario
/// @param d El diccionario
/// @param key La clave
/// @param value El valor (entero)
int set_key(dict d, char *key, int value) {
    int i, old_v;

    i = has_key(d, key);

    if (i < 0)
		return i;
    else {
		old_v = d[i].value;
		d[i].value = value;
		return old_v;
    }
}

/// Borra una clave y su valor de un diccionario
/// @param d El diccionario
/// @param key La clave
/// @return -1 en caso de error, 0 en caso contrario
int del_key(dict d, char *key) {
	int i, j;
	
	i = has_key(d, key);
	
	if (i < 0)
		return i;
	else {
		for ( j = 0 ; j < TEXT_LENGTH ; j++ ) d[i].key[j] = '\0';
		d[i].value = -1;
		return 0;
	}
	
}

/// Quita la referencia a una conversacion en un vector de conversaciones
/// @param c El indice de la conversacion en "msg_queue"
/// @param vector El vector de conversacion puede ser "pending_convs", "active_convs" o "terminated_convs"
/// @return El indice que ocupaba la conversacion en el vector o -1 en caso de que no estuviera
int remove_conversation_from_vector(int c, int *vector) {
	int i;
	
	for ( i = 0 ; i < MQ_SIZE ; i++ ) {
		if ( vector[i] == c ) {
			vector[i] = -1;  // Quitamos la conversacion de la lista
			return i;
		}
	}
	
	return -1;  // La conversacion solicitada no estaba en este vector
}

/// Pone una conversacion en la lista de Conversaciones Activas
/// @param c El indice de la conversacion en la cola de mensajes 'msg_queue'
int set_active_conversation(int c)
{
    int i, free = -1;
	
    for (i = 0; i < MQ_SIZE; i++) {
		// Me guardo el indice del primer sitio libre que encuentro en 'free' por 
		// si acaso luego lo necesito y asi no tengo que hacer dos pasadas
		if (active_convs[i] == -1)
			if ( free < 0 )
	    		free = i;
		
		if (active_convs[i] == c) {  // Ya tenemos esa conversacion
	    	return i;
		} 
	}
		
	// No la tenemos y hay que darla de alta
	active_convs[free] = c;
	remove_conversation_from_vector(c, pending_convs);
	
	return ERROR_NONE;
}


/// Pone una conversacion en la lista de Conversaciones Terminadas
/// @param c El indice de la conversacion en la cola de mensajes 'msg_queue'
int set_terminated_conversation(int c)
{
    int i, free = -1;
	
    for (i = 0; i < MQ_SIZE; i++) {
		// Me guardo el indice del primer sitio libre que encuentro en 'free' por 
		// si acaso luego lo necesito y asi no tengo que hacer dos pasadas
		if (terminated_convs[i] == -1)
			if ( free < 0 )
	    		free = i;
		
		if (terminated_convs[i] == c) {  // Ya tenemos esa conversacion
	    	return i;
		} 
	}
		
	// No la tenemos y hay que darla de alta
	terminated_convs[free] = c;
	remove_conversation_from_vector(c, active_convs);
	
	return ERROR_NONE;
}


/// Pone una conversacion en la lista de Conversaciones Pendientes
/// @param c El indice de la conversacion en la cola de mensajes 'msg_queue'
int set_pending_conversation(int c)
{ 
    int i, free = -1;
	
    for (i = 0; i < MQ_SIZE; i++) {
		// Me guardo el indice del primer sitio libre que encuentro en 'free' por 
		// si acaso luego lo necesito y asi no tengo que hacer dos pasadas
		if (pending_convs[i] == -1)
			if ( free < 0 )
	    		free = i;
		
		if (pending_convs[i] == c) {  // Ya tenemos esa conversacion
	    	return i;
		} 
	}
		
	// No la tenemos y hay que darla de alta
	pending_convs[free] = c;

	return ERROR_NONE;	
}

/// Comprueba si una conversacion esta en estado pendiente
/// @param c Identificador de conversacion
/// @return 1 si esta pendiente, 0 en caso contrario
int is_pending(int c) {
	int i;
	
	// Si es un mensaje erroneo, contestar negativamente
	if ( c == 0 ) return 0;
	
	for ( i = 0 ; i < MQ_SIZE ; i++ ) {
		if ( pending_convs[i] == c )
			return 1;
	}
	
	return 0;
}


/// Comprueba si una conversacion esta en estado activo
/// @param c Identificador de conversacion
/// @return 1 si esta activa, 0 en caso contrario
int is_active(int c) {
	int i;
	
	// Si es un mensaje erroneo, contestar negativamente
	if ( c == 0 ) return 0;
	
	for ( i = 0 ; i < MQ_SIZE ; i++ ) {
		if ( active_convs[i] == c )
			return 1;
	}
	
	return 0;
}

/// Comprueba si una conversacion esta en estado terminado
/// @param c Identificador de conversacion
/// @return 1 si esta terminada, o en caso contrario
int is_terminated(int c) {
	int i;
	
	// Si es un mensaje erroneo, contestar negativamente
	if ( c == 0 ) return 0;
	
	for ( i = 0 ; i < MQ_SIZE ; i++ ) {
		if ( terminated_convs[i] == c )
			return 1;
	}
	
	return 0;
}

/// Devuelve un indice a partir de un identificador de conversacion
///
/// Devuelve un indice a partir de un identificador de conversacion.
/// A diferencia de 'hash_conv' lo hace usando el diccionario names_table.
/// Si no encuentra un indice asociado, LO CREA. Si no se desea este comportamiento
/// (por ejemplo, en 'como_close_conversation') hay que usar 'get_conv_index_no_create'.
/// @param c_id Identificador de conversacion
/// @return El indice o -1 si hay un error
int get_conv_index(char *c_id)
{
    int v;

    v = has_key(names_table, c_id);

    if (v > -1)		// Ya existe la conversacion
		return v;		
    else {			// La creamos aqui
		v = put_key(names_table, c_id);
		return v;
    }
}

int get_conv_index_no_create(char *c_id) {
	return has_key(names_table, c_id);    
}

Tmensaje *get_top_from_conv(int c) {
	int top;
	Tmensaje *pm = NULL;
	
	top = msg_queue[c].top;
	if ( top != -1  )  // If we actually have a top
		pm = &( msg_queue[c].v[top].msg );
	
	return pm;
}

Tmensaje *get_last_from_conv(int c) {
	int last;
	Tmensaje *pm;
	
	last = msg_queue[c].last;
	pm = &( msg_queue[c].v[last].msg );
	
	return pm;
}

/// Borra PERMANENTEMENTE una conversacion
/// @param c El indice de la conversacion
/// @return 0 si todo ha ido bien, -1 si ha habido algun error
int delete_conv_perm(char *conv_id) {
	int i, c;
	
	c = get_conv_index_no_create(conv_id);
	
	// Eliminamos su referencia en Conversaciones Terminadas.
	// Si la llamada a "del_conv_from_terminated" falla, es que
	// la conversacion no estaba alli y por tanto no hay que hacer
	// nada mas que devolver un error
	if ( del_conv_from_terminated(c) == 0 ) {
	
		// Borramos cada uno de los mensajes de la conversacion
		for ( i = 0 ; i < MQ_SIZE ; i++ ) {
			clear_msg(msg_queue[c].v[i].msg);
			msg_queue[c].v[i].code = -2;
		}
	
		// Limpiamos la cabecera de la conversacion
		msg_queue[c].last = -1;
		msg_queue[c].top = -1;
		msg_queue[c].firstfree = 0;
		
		// Borramos la referencia en la tabla de nombres
		del_key(names_table, conv_id);
		
	} else return -1;

	return 0;
}

/// Borra una conversacion de Conversaciones Terminadas
/// @param c El indice de la conversacion
/// @return 0 si todo ha ido bien, -1 si la conversacion no estaba en el vector
int del_conv_from_terminated(int c) {
	int i;
	
	for ( i = 0 ; i < MQ_SIZE ; i++ )
		if ( terminated_convs[i] == c ) {
			terminated_convs[i] = -1;  // Marcamos la posicion como vacia
			return 0;
		}
		
	return -1;  // No estaba 
}


int put_in_outbox(Tmensaje *msg){
	int free, last;
    char aux[TEXT_LENGTH];

    sprintf(aux, "Metiendo mensaje en bandeja de salida");
    debug_print(aux, 1);

    free = outbox.firstfree;
    last = outbox.last;
    outbox.v[free].code = -1;	// El nuevo ultimo tiene codigo -1
    copymsgnode(&(outbox.v[free].msg), msg);
    outbox.v[last].code = free;	// El antiguo ultimo ahora apunta al recien llegado
    outbox.last = free;	// Actualizamos el nuevo ultimo
	
	// Si usamos el indice -1 en "updatefree" hace referencia a la bandeja de salida
    updatefree(-1);
	
	// Codigo por si el mensaje anyadido es el primero
	if ( outbox.top == -1 )  // Si 'top' es -1 aun no se habia metido ningun mensaje
		outbox.top = outbox.last;
	

    return 0;
}
