
#ifndef MSG_STRUCTURES_H
#define MSG_STRUCTURES_H

#include "../parser/debug.h"
#include "../structures/types.h"
#include "../parser/mensaje/mensaje.h"

#define MQ_SIZE 20

/// Nodo mensaje
typedef struct msgnode_ {
    int code;
    Tmensaje msg;
} msgnode;

/// Vector de mensajes ya parseados
typedef struct msgvector_ {
    msgnode v[MQ_SIZE];
    int firstfree;  // Primera posicion libre
    int last;		// Ultimo mensaje de la conversacion
	int top;		// Primer mensaje efectivo de la conversacion
} msgvector;

/// Cola de conversaciones
///
/// Cola de conversaciones, compuesta de vectores de mensajes.
/// La posicion 0 hace referencia a los mensajes erroneos
extern msgvector *msg_queue;

/// Bandeja de salida de los mensaje
msgvector outbox;

/// Nodo de los vectores de conversaciones
typedef struct convnode_ {
    int index;
    char name[TEXT_LENGTH];
} convnode;



/// Inicializa la cola de mensajes y las de conversaciones
void initmsglist();


/// Deposita un mensaje recien parseado en la cola
/// @param m El mensaje
/// @param conv_id El identificador de conversacion
/// @return Indice de la conversacion en 'msg_queue' o codigo de error (<0)
int put_msg(Tmensaje * m, char *conv_id);

/// Muestra el contenido de la cola de mensajes
void print_msg_queue();

/// Imprime el contenido de la cola de mensajes en un fichero
void fprint_msg_queue();

/// Pone una conversacion en la lista de Conversaciones Activas
/// @param c El indice de la conversacion en la cola de mensajes 'msg_queue'
/// @return El indice del vector "active_convs" en el que esta la conversacion
int set_active_conversation(int c);

/// Pone una conversacion en la lista de Conversaciones Terminadas
/// @param c El indice de la conversacion en la cola de mensajes 'msg_queue'
/// @return El indice del vector "terminated_convs" en el que esta la conversacion
int set_terminated_conversation(int c);

/// Pone una conversacion en la lista de Conversaciones Pendientes
/// @param c El indice de la conversacion en la cola de mensajes 'msg_queue'
/// @return El indice del vector "pending_convs" en el que esta la conversacion
int set_pending_conversation(int c);

/// Devuelve un puntero al primer mensaje efectivo de una conversacion
/// @param c Indice de la conversacion en "msg_queue"
/// @return El puntero al mensaje
Tmensaje *get_top_from_conv(int c);

/// Devuelve un puntero al ultimo mensaje de una conversacion
/// @param c Indice de la conversacion en "msg_queue"
/// @return El puntero al mensaje
Tmensaje *get_last_from_conv(int c);

//int get_conv_index(char *c_id);

/// Pone un mensaje en la bandeja de salida
/// @param msg El mensaje
/// @return Codigo de error
int put_in_outbox(Tmensaje *msg);

/// Imprime la bandeja de salida
void print_outbox();

/// Imprime una conversacion
/// @param conv_id La conversacion
void print_conv(char *conv_id);

#endif
