/*agenda3.0.h*/
/*** Type of an agent. ***/
#ifndef EST_DATOS_H
#define EST_DATOS_H

#define SPORADIC 0
#define PERIODIC 1

#define NONES 0

#define NEW_AGENT -1

/*** These are the states a KS may take ***/
#define NOT_EXECUTED 0
#define EXECUTING    1 /***It has no sense in this implementation. ***/
#define PAUSED       2 

#include "../../../include/time.h"
///#include "../../include/events.h"

void in();

typedef struct{
 int mks_idn;
 systime_t time;
 void *next;
 void *prev;
}allocat;

typedef struct{
 systime_t LS;
 systime_t LI;
 systime_t AC;
 void *next;
 void *prev;
}hole;

/*** KS structure ***/
typedef struct{
 int       ks_id;    /*** This is the ks's id which is currently in execution.
                       *    We use it to stop and resume a ks.
                       ***/
 char      *name;    /*** ks's identifier.                 ***/
 int       identifier;

 int       kiwi_id;

 int       weight;    /*** Used by rating.                   ***/
 float     coef;
 systime_t execution;
 systime_t duration,  /*** Estimated duration of a ks. Currently this is  provided
                       *   by the user.
                       ***/
           deadline;  /*** No comments.  ***/
 int       result;    /*** Cut if off? ***/
 systime_t last_check;     /*** Last moment Control checked and updated its deadline.***/
 systime_t blk_duration,
	   ks_deliber;    
 float     gain;
 float     ks_coef;

} ks;



/*** KS NODE structure ***/
typedef struct{
 ks  *KS;
 void *next;  /*** We can update the ks of an agent as his current ks has
               *   ended.
               ***/
 void *prev;  /*** So we'll be able to find the reminding deadline of a
               *   periodic agent.      
               ***/     
}n_ks;



/*** MKS structure ***/

typedef struct{
 char  *name;   /*** Agent identifier as it was declared by user. ***/
 int identifier;
 int type;          /*** Sporadic or periodic one?.                   ***/    
 int solution_level;/*** How many solutions have been found.          ***/
 n_ks *p_ks;      /*** Pointer to current ks. This ks remains 
                     *   into a list of ks's. 
                     ***/
 int state;         /*** This is the current ks state. See above.   ***/
                 
 int importance;    /*** No comments.                                ***/
                 
 int number_ks;     /*** Number of ks's into an agent. This field
                     *   is used to find out the quality.
                     ***/
 int type_MKS;
 systime_t mks1_deliber;
 systime_t mks_deliber;
 systime_t mks1_duration;   
 float quality;    /*** No comments.                                ***/
 float real_gain;
 void *mother;
 int empty;
}mks;





/*** MKS NODE structure ***/
typedef struct{
 mks *MKS;
 void  *next, 
       *prev;
}n_mks;



/*** MKS NODE LIST, it'll be instantiated to sub_agendas(tsa, ctsa, rsa) ***/
typedef struct{
 n_mks *first,
         *last;
 int number_mks;
}l_mks;

typedef struct{
 int   ks_n;  /*** This is incremented as a new ks is added to the list. ***/
 n_ks *last,
       *first;
}Lks;



/*** IN-AGENT structure ***/

typedef struct {
	char name[40];
	int in_agent_ident;
	int no_mks;

        int no_critical_mks;
      
	n_mks *mks_list;
}in_agent;


/*** IN_AGENT NODE LIST ***/

typedef struct {
	in_agent *IN_AGENT;
	void *next;
	void *prev;
}n_in_agent;



typedef struct {
	ks *KS;
	systime_t deadline;
	float gain;
	float coef;
	void *sig;
	void *prev;
} KS2;


typedef struct {
	mks *MKS;
	KS2 *KS;
	int importance;
	void *prev;
	void *sig;
}n_mks2;


typedef struct {
	char name[40];
	char type[40];
	n_mks2 *NMKS2;
	int importance;
	char level[40];
	char slot[40];
	void *sig;
}n_regla;


typedef struct{
 int n_allocat;
 allocat *last;
 allocat *first;
}l_allocat;

#endif



