
/******************************agenda.c******************************
*********************************************************************
*** Interface between agendas and Control or 2ond. level planifica-
*** tion.
*********************************************************************
*********************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "../include/dserver/est_datos.h"
#include "../include/dserver/decl.h"
#include "../../include/time.h"

#include "../../include/kdm.h"
#include "../../common/include/artis_conf.h"  /// For the ARTIS_CONF_METAREASONING


#ifdef ARTIS_CONF_METAREASONING
#include "../../include/metaclass.h"
#endif

#ifdef CONTROL_DEBUG
 #include "../rehearsal/exec_ksi_rehearsal.h"
#endif

#define EDF 0
#define BIF 1
#define DBS 2
#define III 3
#define NONE -1

extern Lks l_ks;
extern float system_quality;
extern float system_gain;
extern n_mks *n_mks_actual;
extern l_mks rsa;
extern l_allocat sched;


void init(l_mks *lmks)
{
 lmks->first=NULL;
 lmks->last =NULL;
 lmks->number_mks = 0;
}

void init_l_ks()
{
 l_ks.ks_n =0;
 l_ks.first=NULL;
 l_ks.last =NULL;
}

void init_sched()
{
 sched.n_allocat = 0;
 sched.first = NULL;
 sched.last = NULL;
}

void insert_mks(l_mks *lmks, mks *mks_curr)
{
 n_mks *aux;

 aux=(n_mks *)malloc(sizeof(n_mks));
 aux->MKS=mks_curr;
 aux->next=NULL;
 aux->prev=NULL;
 
 if (lmks->number_mks<=0){
   lmks->first=aux;
   lmks->last =aux;
 }
 else{
   aux->prev=lmks->last;
   lmks->last->next=aux;
   lmks->last=aux;
 }
 lmks->number_mks++;
}

void in_hole(systime_t ls, systime_t t_deliber)
{
  hole *h_aux;
  h_aux = malloc(sizeof(hole));
  h_aux->LS=ls;
  h_aux->LI=ls - t_deliber;
  h_aux->prev = NULL;
  h_aux->next = l_hole;
  if (l_hole!=NULL) l_hole->prev = h_aux;
  l_hole = h_aux;
}
 
void free_hole()
{
  hole *aux;

  while(l_hole!=NULL){
    l_hole->prev=NULL;
    aux = l_hole;
    l_hole = l_hole->next;
    free(aux);
  }
}


mks *out_mks(l_mks *lmks)
{
  mks   *aux;
  n_mks *n_aux;
 
  if (lmks->number_mks<=0) return (mks *)NULL;
  lmks->number_mks--;
  aux  =lmks->last->MKS;
  n_aux=lmks->last;
  if (lmks->first==lmks->last){
    lmks->first=NULL;
    lmks->last =NULL;
  }
  else{
    lmks->last      =n_aux->prev;
    lmks->last->next=NULL;
    n_aux->prev   =NULL;
  }
  free(n_aux);/*** Free from memory this three pointers structure. ***/
  return aux;
}

void in_first(l_mks *lmks, mks *mks_curr)
{
  n_mks *aux;
  
  aux=(n_mks *)malloc(sizeof(n_mks));
  aux->MKS=mks_curr;
  aux->next=NULL;
  aux->prev=NULL;
  if(mks_curr->p_ks->KS->ks_deliber!=0){
    if (lmks->number_mks<=0){
      lmks->first=aux;
      lmks->last =aux;
    }
    else{
      aux->next=lmks->first;
      lmks->first->prev=aux;
      lmks->first=aux;
    }
    lmks->number_mks++;
  }
}

void in_ord_by_importance(l_mks *lmks, mks *mks_curr)
{
  int importance;
  n_mks  *aux, *aux2;
 
  aux=(n_mks *)malloc(sizeof(n_mks));
  aux->MKS  =mks_curr;
  aux->prev=NULL;
  aux->next=NULL;
  importance = aux->MKS->importance;

#ifdef LUIS_DEBUG
  fprintf(stderr,"in_ord_by_importance: mks_name: %s   ag importance: %d\n\n",
	  mks_curr->name, mks_curr->importance);
#endif
  
  if (lmks->last==NULL){
    /*** aux will be first. ***/ 
    lmks->first=aux;
    lmks->last =aux;
  }
  else 
    if (importance >= lmks->first->MKS->importance){
      /*** aux will be first. ***/ 
      aux->next      =lmks->first;
      lmks->first->prev=aux;
      lmks->first      =aux;
    }
    else 
      if (importance <= lmks->last->MKS->importance){
	/*** aux will be last.***/
	aux->prev     =lmks->last;
	lmks->last->next=aux;
	lmks->last      =aux;
      }
      else{
	aux2=lmks->first->next;
	while (importance < aux2->MKS->importance) aux2=aux2->next;
	/*** Within this loop aux2 never will be NULL. ***/    
	aux->next =aux2;
	aux->prev =aux2->prev;
	aux2->prev=aux;
	aux2      =aux->prev;
	aux2->next=aux;
      }
  return;
}

void in_ord_by_deadline(l_mks *lmks, mks *mks_curr)
{
  systime_t   deadline;
  n_mks  *aux, *aux2;
  
  aux=(n_mks *)malloc(sizeof(n_mks));
  aux->MKS  =mks_curr;
  aux->prev=NULL;
  aux->next=NULL;
  deadline= aux->MKS->p_ks->KS->deadline;
  if (lmks->last==NULL){
    /*** aux will be first.***/ 
    lmks->first=aux;
    lmks->last =aux;
  }
  else
    if (deadline <= lmks->first->MKS->p_ks->KS->deadline){
      /*** aux will be first.***/ 
      aux->next       =lmks->first;
      lmks->first->prev =aux;
      lmks->first       =aux;
    }
    else 
      if (deadline >= lmks->last->MKS->p_ks->KS->deadline){
      /*** aux will be last.***/
	aux->prev      =lmks->last;
	lmks->last->next =aux;
	lmks->last       =aux;
      }
      else{
	aux2=lmks->first->next;
	while (deadline > aux2->MKS->p_ks->KS->deadline) 
	  aux2=aux2->next;
	aux->next =aux2;
	aux->prev =aux2->prev;
	aux2->prev=aux;
	aux2      =aux->prev;
	aux2->next=aux;
      }
  return;
}

/*** Depending on heuristic declared by user, we order la by following
 *   EDF, BIF, DBS criteria.
 ***/

void in_ord( l_mks *lmks, mks *mks_curr )
{
 
  lmks->number_mks++;
  switch ( heuristic ){
  case EDF : return ( in_ord_by_deadline( lmks, mks_curr ) );
  case BIF : return ( in_ord_by_importance( lmks, mks_curr ) );
  case DBS : 
  case III :
    return ( in_ord_by_deadline( lmks, mks_curr ) );
  }
}
 
void free_ks(ks *KS)
{
  free(KS->name);
  free(KS);
}

void free_l_ks(n_ks *nks)
{
  n_ks *neig_n_ks;
  n_ks *aux;
 
  aux=nks;
  /*** Free forward n_ksi's. ***/
  nks=nks->next;
  while (nks!=NULL){
    neig_n_ks=nks->next;
    free_ks(nks->KS);
    free(nks);
    nks=neig_n_ks;
  }
  /*** Free backward n_ksi's. ***/
  nks=aux->prev;
  while (nks!=NULL){
    neig_n_ks=nks->prev; 
    free_ks(nks->KS); 
    free(nks);
    nks=neig_n_ks;
  }
  free_ks(aux->KS); 
  free(aux);
}

void free_mks(mks *mks_curr)
{
  free(mks_curr->name);
  free_l_ks(mks_curr->p_ks);
  /*** Don't know what's the a->p_ksi position in 
   *   the list kept by agent, so free_l_ksi().
   ***/
  free(mks_curr);
}
 
void free_l_mks(l_mks *lmks)
{
  n_mks *aux;
 
  while (lmks->first!=lmks->last){
    aux=lmks->last;
    lmks->last=lmks->last->prev;
    free_mks(aux->MKS);
    free(aux);
  }
  free_mks(lmks->last->MKS);
  free(lmks->last);/*=free(lmks->first);*/

  /*** ThreeC 20030910 ***/
  lmks->number_mks = 0;
}

void in_l_ks(ks *ks_curr )
{
  n_ks *aux;
 
  aux=(n_ks *)malloc(sizeof(n_ks));
  aux->KS  = ks_curr;
  aux->next = NULL;
  aux->prev = NULL;
  if (!l_ks.ks_n){
    l_ks.last  = aux;
    l_ks.first = aux;
  }
  else{
    l_ks.last->next  = aux;
    aux->prev         = l_ks.last;
    l_ks.last        = aux;    
  }
  l_ks.ks_n++;
}

float get_ks_gain(ks *ks_curr )
{
  return ( (float) ks_curr->gain );
}

int update_mks(mks *mks_curr )
{
  n_ks *n_ksaux, *n_ksaux2;
  systime_t aux_dead;
  systime_t aux_last;
  int ret;
  float quality_increment, gain_increment;
  mks_curr->solution_level++;
  /*** Another KSI, another solution. ***/

  n_ksaux2 = mks_curr->p_ks;   
  if (mks_curr->p_ks->next != NULL){
    if ( mks_curr->type == PERIODIC ){
      aux_dead = mks_curr->p_ks->KS->deadline;
      /*** Last ksi's remaining deadline.***/
      aux_last = mks_curr->p_ks->KS->last_check;
      n_ksaux = mks_curr->p_ks->next;
      n_ksaux->KS->deadline = aux_dead;
      n_ksaux->KS->last_check = aux_last;
    }
    mks_curr->p_ks = mks_curr->p_ks->next;             
    /*** Go for next ks. ***/
    mks_curr->state = NOT_EXECUTED;   
    /*** Waiting for execution. ***/
    ret = 1;
  }
  else ret = 0;
  
  //#ifndef METAREASONING
  //if ( heuristic == DBS ) {
  //#endif
  gain_increment = get_ks_gain( n_ksaux2->KS ); 
  mks_curr->real_gain += gain_increment;
  system_gain += gain_increment;
  quality_increment = gain_increment * (mks_curr->importance);
  //quality_increment =gain_increment;
  //#ifndef METAREASONING
  //}
  //else quality_increment = ((float)(mks_curr->importance)) / mks_curr->number_ks ;
  //#endif
  mks_curr->quality += quality_increment ;
  /*** Another executed ks->
   *   quality is improved.
   ***/

#ifdef ARTIS_CONF_METAREASONING
  KDM_Set_Value( pMetaClass->systemQuality, 
	( * (int*) KDM_Get_Value( pMetaClass->systemQuality ))
	 + (int) ( quality_increment 
  	           / ( * (int *) KDM_Get_Value( pMetaClass->maxQuality ))) );
#else
  system_quality += quality_increment / MaxQuality; 
#endif
  
  return (ret);
}


mks *element_number(l_mks *lmks, int position )
{
  /*** Returns a pointer to the agent in the position "position"
   *    on the "la" agents list
   ***/
  int i;
 
  if (( lmks->first == lmks->last) && (lmks->last == NULL) )
    return NULL ;  /*** The list is empty ***/
  n_mks_actual = lmks->first ;
  for ( i=1 ; ( i < position ) && ( n_mks_actual != NULL ) ; i++ )
    n_mks_actual = n_mks_actual->next ;
  return ( n_mks_actual != NULL )? n_mks_actual->MKS : NULL ;
}

void out_current(l_mks *lmks )
{
  n_mks *n_aux ;
  n_mks *n_aux2 = NULL ;

  if ( n_mks_actual == NULL )  return ;
  /*** if ( n_mks_actual != NULL ) then the agenda is not empty, and
   *     we had found the agent we were looking for the last
   *     time we tried it.
   ***/

  if((lmks->number_mks==1) && (n_mks_actual==lmks->first) && (n_mks_actual==lmks->last)){
    free(n_mks_actual);
    n_mks_actual=NULL;
    init(lmks);
  }
  else{
    if (n_mks_actual == lmks->first){
      lmks->first = n_mks_actual->next ;
      lmks->first->prev = NULL ;
      free(n_mks_actual);
      n_mks_actual=NULL;
    }
    else if (n_mks_actual == lmks->last){
      lmks->last=n_mks_actual->prev;
      lmks->last->next=NULL;
      free(n_mks_actual);
      n_mks_actual=NULL;
    }
    else{
      n_aux=n_mks_actual->prev;
      n_aux2=n_mks_actual->next;
      n_aux->next=n_aux2;
      n_aux2->prev=n_aux;
      free(n_mks_actual);
      n_mks_actual=n_aux;
    }
    lmks->number_mks--;
  } 
}

/* =================================================================== */
void merge_list(l_mks *lmks, l_mks *lb){
/* =================================================================== */

  if (lb->number_mks > 0){
    if (lmks->number_mks > 0){
      lb->last->next = lmks->first;
      lmks->first->prev = lb->last;
    }
    else	lmks->last=lb->last;
    lmks->first=lb->first;
    lb->first=lb->last= NULL;
    lmks->number_mks += lb->number_mks;
    lb->number_mks = 0;
  }
}

/* ================================================================== */
void in_last(l_mks *lmks, mks *mks_curr){
/* ================================================================== */

  n_mks *aux;
  
  aux = (n_mks *)malloc (sizeof(n_mks));
  aux->MKS = mks_curr;
  aux->next = (n_mks *)NULL;
  aux->prev = (n_mks *)NULL;
  
  if (lmks->number_mks == 0){
    lmks->first = aux;
    lmks->last = aux;
  }
  else {
    lmks->last->next = aux;
    aux->prev = lmks->last;
    lmks->last = aux;
  }
  lmks->number_mks++;
}

mks *first(l_mks *lmks )
{
  if (lmks->number_mks<=0) return(mks *)NULL;
  n_mks_actual = lmks->first ;
  return( n_mks_actual->MKS );
}

mks *last(l_mks *lmks )
{
  if (lmks->number_mks<=0) return(mks *)NULL;
  n_mks_actual = lmks->last;
  return( n_mks_actual->MKS);
}

mks *next(l_mks *lmks )
{
  if ( n_mks_actual==NULL )
    n_mks_actual = lmks->first ;
  else
    n_mks_actual=(n_mks *)n_mks_actual->next;
  if (n_mks_actual==NULL) return (mks *)NULL;
  else                return (n_mks_actual->MKS);
}

mks *prev(l_mks *lmks )
{
  if ( n_mks_actual==NULL )
    n_mks_actual = lmks->last;
  else 
    n_mks_actual=(n_mks *)n_mks_actual->prev;
  if (n_mks_actual==NULL) return(mks *)NULL;
  else 		     return(n_mks_actual->MKS);
}

void in_sched( int mksn, systime_t mks_time )
{
  allocat *auxn;
  
  auxn = (allocat *)malloc( sizeof( allocat ) );
  auxn->next = NULL;
  auxn->prev = NULL;
  auxn->mks_idn = mksn;
  auxn->time = mks_time;
  if ( !sched.n_allocat ) {
    sched.first = auxn;
    sched.last = auxn;
  }
  else {
    sched.first->prev = auxn;
    auxn->next = sched.first;
    sched.first = auxn;
  }
 sched.n_allocat++;
}

void out_sched()
{
  allocat *auxn;
  
  if ( sched.n_allocat ) {
    auxn = sched.first;
    sched.first = sched.first->next;
    free( auxn );
    sched.n_allocat--;
  }
  if ( sched.n_allocat == 0 ) sched.last = NULL;
}

void reset_mks(mks *mks_curr,systime_t t )
{
  
  /*** This function delete t time units from the
   *   deliberation time of ag. 
   ***/
       
  mks_curr->p_ks->KS->ks_deliber = 0;
  mks_curr->mks1_deliber -= t;
}

void free_l_allocat(l_mks *lmks )
{
  allocat *auxn;
  mks *mks_curr;
  
  while ( sched.first != NULL ) {
    auxn = sched.first;
    if ( auxn->mks_idn != NONE ) {
      mks_curr = first ( lmks );
      while ( (mks_curr != NULL) && (auxn->mks_idn != mks_curr->identifier) )
	mks_curr = next( lmks );
      if ( mks_curr != NULL ) reset_mks( mks_curr, auxn->time);
    }
    out_sched();
  }
}

void reset_t_deliber(l_mks *lmks, int mksn, systime_t last, systime_t *add )
{
  mks *mks_curr;
  allocat *auxn;
  systime_t i;
  
  *add = 0;
  /*** We delete from the allocation list (sched) "last" time units and
   *   reset the agent deliberation times associated to these allocations.
   *   But when there are "procedure ks" in sched, sometimes is necessary to 
   *   increase the time units to delete; this is because when we assign 
   *   deliberation time to a procedure ks, we assign it the ks duration and 
   *   we consider that the procedure ks deliberation time can't be < than its
   *   duration.
   ***/  
  if ( sched.n_allocat ) {
    while ( last ) {
      auxn = sched.first;
      
      /*** We go out of the while because there is not allocation units in the 
       *   allocation list (sched).
     ***/
      if ( auxn == NULL ) break;
      i = auxn->time;
      if ( i <= last ) {
        /*** If the agent is the executed (mksn) or a hole (none), we don't search 
         *   in the rsa.
         ***/
	if ( (auxn->mks_idn != mksn) && (auxn->mks_idn != NONE) ) {
	  mks_curr =first( lmks );
	  while ( (mks_curr != (mks *) NULL) &&
		  (auxn->mks_idn != mks_curr->identifier) ) 
	    mks_curr = next( lmks );
	  if (mks_curr != NULL) reset_mks( mks_curr, auxn->time );
	}
	out_sched();
	last-= i;
      }
      else{
	if ( auxn->mks_idn == NONE )  {
	  auxn->time -=last;
	  last = 0;
	}
	else {
          mks_curr = first( lmks );
	  while ( (mks_curr != (mks *) NULL) &&
		  (auxn->mks_idn != mks_curr->identifier) ) 
	    mks_curr = next( lmks );
	  if ( mks_curr != (mks *) NULL ) {
	    reset_mks( mks_curr, auxn->time );
	    *add = auxn->time - last;
	    last = 0;
	    out_sched();
	  } /*** End of if (ag!=NULL) ***/
	  else {
	    /*** The agent has not been found in the rsa ***/
	    *add = auxn->time - last;
	    last = 0;
	    out_sched();
	  }
	}
      }
    } /*** End of while ***/
  }
}


mks *mks_search(l_mks *lmks )
{ 
  int    pos;
  mks *aux_mks;
  
  /*** It returns the first agent in the rsa that has deliberation time. ***/ 
  
  aux_mks = first( lmks );
  if ( aux_mks == NULL ) return (mks *)NULL;
  pos =1;
  while(( aux_mks != NULL ) && ( aux_mks->p_ks->KS->ks_deliber == 0 )) {
    aux_mks = next( lmks );
    pos++;
  }
  if ( aux_mks != NULL ) {
    aux_mks = element_number( lmks, pos );
    return aux_mks;
  }

  return (mks *)NULL ;
}

/***Following functions slow down Control response time, better if this 
 *  function were implemented as macros or inline functions in C++.
 ***/

int paused (mks *mks_curr)
{
  return(mks_curr->state == PAUSED);
}

int not_executed(mks *mks_curr)
{
  return(mks_curr->state == NOT_EXECUTED);
}

int id_ksi(mks *mks_curr)
{
  return (mks_curr->p_ks->KS->ks_id);
}

char * ks_name(mks *mks_curr)
{
  return( mks_curr->p_ks->KS->name);
}


void update_paused(mks *mks_curr, int pid)
{
  mks_curr->state = PAUSED;
  mks_curr->p_ks->KS->ks_id = pid;
}

void deadline(mks *mks_curr, systime_t *t1)
{ 
  if (mks_curr!=NULL) { 
    /*fprintf(stderr,"Soy deadline: %ld%04ld\n",
     *t1 = mks_curr->p_ks->KS->deadline);*/
    *t1 = mks_curr->p_ks->KS->deadline;
    /*fprintf(stderr,"Y ahora %ld%04ld\n", HI(*t1),LO(*t1));*/
  }
  else {
    *t1 = 0;
    /*fprintf(stderr, "deadline vacio\n");*/
  }
}

void deliber(mks *mks_curr, systime_t *t1 )
{
  if (mks_curr!=NULL) *t1 = mks_curr->p_ks->KS->ks_deliber ;
  else *t1 = 0;
}

int ks_execution(mks *mks_curr)
{
  if (mks_curr!=NULL) return( mks_curr->p_ks->KS->execution );
  else return 0;
}

void get_info(mks *mks_curr, systime_t *t1)
{
  systime_t aux1, aux2;
  switch ( heuristic ){
  case EDF : deadline( mks_curr, t1 );break;
  case BIF : deadline( mks_curr, t1 );break;
  case DBS : {
    deliber( mks_curr, &aux1 );
    deadline( mks_curr, &aux2 );
    if (aux1 < aux2) 
      *t1 = aux1;
    else *t1 = aux2;
  }
  }
  /*fprintf(stderr,"en get_info %ld%04ld\n",*t1)*/;
}

int strategy_translation(char strat_orig[]){

int strat_final=-1;

if (strcmp(strat_orig,"EDF")==0) strat_final= 0;
else if (strcmp(strat_orig,"BIF")==0) strat_final= 1;
 else if (strcmp(strat_orig,"DBS")==0) strat_final= 2;
 else if (strcmp(strat_orig,"III")==0) strat_final= 3;

return strat_final;
} 


