/**  
 *  \file dserver.c
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es) after 	
 *       Ana Garcia Fornes
 *	 Andres Terrasa Barrena
 *	 Agustin Espinosa Minguet
 *  \version 3.0
 *
 *  \brief Deliberative Server 3.0
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

#define _POSIX_SOURCE

#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* THREAD STUFF */

#include <unistd.h>

#include "../../include/cpu.h"  ///< To get the cpu frecuency in hz.
#include "../../include/time.h"

#include "../../common/include/artis_conf.h"  ///< For the ARTIS_CONF_METAREASONING

#ifdef ARTIS_CONF_METAREASONING
#include "../../events/include/events/events_internal.h"  ///< For the MAX_EV
#include "../../include/metaclass.h"
#endif

#include "../../common/include/artis_fifo_definition.h"

#include "../include/dserver.h"

#include "../../trace/include/trace.h"
#include "../../linux_core/include/linux_core.h"

#include "../include/dserver/decl.h"

#include "../../include/kdm.h"

///#include "../../meta-rules/meta-rule_store.h"   ---> ARTIS_CONF_METAREASONING


#include "../include/dserver/em.h"

/*  #define TIMING */


#define OPS_PER_SEC 33  /*iteraciones por microsegundo in compute() loop. ADJUST!!*/

#define LEVEL_NOT_READY      0
#define LEVEL_READY          1
#define SCHED_LEVEL          0
#define NOLEVEL             -1

#define SECURITY_INTERVAL  (systime_t)  1000
#define KILL_LEVEL_TIME     (systime_t)  500 /*6000 before ipg*/

#define EVER                 1   /// To make cool for loops

int nlevels;

// Global variables to be shared with other ARTIS modules
int running_level;

/* Control variables */
extern systime_t init_time, origin, end_time, real_time; 
             /* can't use init_time*/
extern systime_t MANDATORY_CT, MANDATORY_I, MANDATORY_R, MANDATORY_S, 
  MANDATORY_T;

#ifdef COMO
extern systime_t MANDATORY_COMO;
extern int NUM_ERROR_MESSAGES, NUM_RECV_MESSAGES, NUM_SEND_MESSAGES;
extern int como_phase;
#endif

/* System-time related variables */
systime_t         auxiliar; //as we can not use many local vbles.
systime_t         current_level_kill_time;
systime_t         now;

/* fifos handlers*/
int                 sched_rtf;

#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_WAKEUP
int               wakeup_rtf;
#endif
extern MetaClass  *pMetaClass;
#endif


static __inline__ unsigned long long rdtsc(void)
{
        unsigned long long ret;
        __asm__ __volatile__("rdtsc"
                            : "=A" (ret)
                            : /* no inputs */);
        return ret;
}


/* Puestas para permitir la compilacion */
void EndControl(void) {}

void DS_stop(int status) {
  /***remote_control_stop();***/
  close(sched_rtf);

#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_WAKEUP
  close(wakeup_rtf);
#endif
#endif

  fprintf(stderr,"About to exit definitively with error: %d\n",status);

  exit(status);
}


#ifdef ARTIS_CONF_METAREASONING
/* ============================================================================ */
int ds_loop() {
/* ============================================================================ */

   /*It's only control*/
  mks *scheduleMKS;
  int dont_inner   =  1; /// To control if it is running the first iteration of the loop (1) or not
  /* interrupt type es para el rating de luis*/
  int interrupt_type = 0;
  int Is_init_deadline = -1,level_end;
  systime_t t1,t2,running_time;

  int bSleep;
  int nTriggerState;
  systime_t lTriggerEndTime;

  systime_t lBeginningTime;

  systime_t lMaxMandatory;

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_DS_DEBUG_2
  trace_event( DEBUG_EVENT, 1, (long long) 0, 0 );
#endif
#endif

  lMaxMandatory = 
    (MANDATORY_T > MANDATORY_CT)? MANDATORY_T + 100: MANDATORY_CT + 100;

  fprintf(stderr,"\t_______________________________________________________\n");
  fprintf(stderr,"\t|                      CONTROL                        |\n");
  fprintf(stderr,"\t-------------------------------------------------------\n");

  //ShowEvents();  /*** What events? ***/ 

  do {

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_DS_DEBUG_2
    trace_event( DEBUG_EVENT, 1, (long long) 1, 0 );
#endif
#endif
    
    bSleep = 0;
    
    if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
       EventRate = * (int *) KDM_Get_Value( pMetaClass->eventRate );
       /*** EventRate may have changed for a Meta-Rule Action ***/
#ifdef ARTIS_TRACE_EVENTS
	mutex_unlock( MUTEX_EM );
#endif
    }

    lBeginningTime=systime_gettime(); 

    get_available(MANDATORY_T + 40, (systime_t) MANDATORY_T * EventRate + 40, 
		   &lTriggerEndTime);
      
    fprintf( stderr, "\t(control.c):: main_control(): lTriggerEndTime: %lld, EventRate: %d, MANDATORY_T: %lld\n",
	     lTriggerEndTime, EventRate, MANDATORY_T );

    if (lTriggerEndTime >= MANDATORY_T){ 

#ifdef ARTIS_TRACE_EVENTS
      trace_event( EM_START, 0, (long long) 0, 0 );
#endif
      
      lTriggerEndTime += lBeginningTime; 

      if ((nTriggerState = trigger(lTriggerEndTime)) == TRIGGER_EXIT) {
	/*** Yankee Go Home !!! ***/
#ifdef ARTIS_TRACE_EVENTS
	trace_event( EM_STOP, 0, (long long) 0, 0 );
#endif
	fprintf(stderr, "\t(is.c)::main_control(): This is the end of the world as we know it\n");

	return TRIGGER_EXIT;
      }
      else if (( EventRate > 1 ) &&
               dont_inner &&
	       ((nTriggerState == TRIGGER_EMPTY)
		///|| (nTriggerState == TRIGGER_NOTIME_ER)
		)) {
#ifdef ARTIS_CONF_AUTO_METAREASONING
                EventRate /= 2.0;
                if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
                     KDM_Set_Value( pMetaClass->eventRate, EventRate );
#ifdef ARTIS_TRACE_EVENTS
                     mutex_unlock( MUTEX_EM );
#endif
                }
#endif
                ///                fprintf(stderr, "(ds_loop): Primera divisin por 2 del EventRate\n");
                ///SetAttentionDegree( ( EventRate / 2.0 ) / MAX_EV )
           }
           else if (( EventRate * 2 ) <= MAX_EV &&
                    dont_inner) {
#ifdef ARTIS_CONF_AUTO_METAREASONING
	          EventRate *= 2;    /*** There is more time for triggering ***/
                  if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
                       KDM_Set_Value( pMetaClass->eventRate, EventRate );
#ifdef ARTIS_TRACE_EVENTS
                       mutex_unlock( MUTEX_EM );
#endif
                  }
#endif

                  ///                  fprintf(stderr, "(ds_loop): Primera multiplicacin por 2 del EventRate\n");
                  ///SetAttentionDegree( ( EventRate * 2.0 ) / MAX_EV );
	   }

    }

#ifdef ARTIS_TRACE_EVENTS
    trace_event( SLS_START, 0, (long long) 0, 0 );
#endif


#ifdef COMO
    if (enough_time(MANDATORY_COMO)) {

	    //como_send_erroneous(NUM_ERROR_MESSAGES);
	    if (como_phase == 1) como_receive_inbox(NUM_RECV_MESSAGES);
	    //como_send_outbox(NUM_SEND_MESSAGES);

	    fprintf(stderr, "=========== COMO WAS HERE ========== \n");

	    if (como_phase == 2) como_phase = 0;
	    else como_phase++;

    }
#endif


    
    if (enough_time(MANDATORY_CT)){
#if defined ARTIS_CONF_DS_DEBUG_1
      fprintf(stderr, "\t(control.c)Let's test conditions...\n");
#endif
      condition_testing();

      if (enough_time(MANDATORY_R) && 
	  (ctsa.number_mks || rsa.number_mks)){
#if defined ARTIS_CONF_DS_DEBUG_1
	fprintf(stderr, "\t(control.c)Let's rate...\n");
#endif
	rating(interrupt_type);

#if defined ARTIS_CONF_DS_DEBUG_1
	fprintf(stderr, "\t(control.c)Let's schedule...\n");
#endif
	scheduleMKS = (mks *)schedule();

	if (scheduleMKS != (mks *)NULL){
	  if (check_deadline(scheduleMKS)) {	  
	    get_info(scheduleMKS,&t1);
	    get_available(MANDATORY_I,t1,&t2);
	    get_real_t_mod(scheduleMKS,t2,&running_time);
	    if ( running_time > MIN_RUNTIME){ /* r3 = running time */
#if defined ARTIS_CONF_DS_DEBUG_2
	      fprintf(stderr,"\t(is.c) ex_until %d -name %s- until %lld, kiwi_id %d\n", 
		      scheduleMKS->p_ks->KS->identifier, 
		      scheduleMKS->name, 
		      running_time,
		      scheduleMKS->p_ks->KS->kiwi_id ); 
#endif

#ifdef ARTIS_TRACE_EVENTS
              // This extern variable is for kiwi purposes
              running_level = scheduleMKS->p_ks->KS->kiwi_id;
#ifdef ARTIS_CONF_DS_DEBUG_1
              fprintf(stderr, "(sls.c) interpret: running_level = %d - until %lld\n", 
                      running_level, running_time);
#endif
#endif

	      level_end = Level_Execute_Until(scheduleMKS->p_ks->KS->identifier,
					running_time, 
					scheduleMKS->p_ks->KS->kiwi_id ); 
	      interpret_end(scheduleMKS, level_end);
	      
#ifdef ARTIS_TRACE_EVENTS
	      trace_event( SLS_STOP, 0, (long long) 0, 0 );
              running_level = 1; ///< DS
	      trace_event( AA_QUALITY, 1, (long long) system_quality, 0 );
#endif
	
	    }
	    else{  /*** There is not enough time to run the scheduled ks ***/
              ///	      fprintf(stderr,"\n No time to run levels. Updating Queues\n");
	      scheduleMKS->p_ks->KS->ks_deliber=0;
	      in_last(&sub_rsa,scheduleMKS);
	      first_call=1;
	      ///dont_inner=1;
	    
	      bSleep = IS_NOTIME;
	    
#if defined ARTIS_CONF_DS_DEBUG_1 
	      fprintf(stderr,"\t Time exhausted\n");
#endif
	    }
	  }
	}
	else{   /*** scheduleMKS == (mks *) NULL ***/
	    first_call=1;
	    ///dont_inner=0;
#if defined ARTIS_CONF_DS_DEBUG_1
	    fprintf(stderr,"SIN AGENTES QUE SCHEDULAR\n");
#endif
	  
	    bSleep = IS_SLEEPING; /// Nothing to execute, better sleep 
	    
#ifdef ARTIS_CONF_DS_DEBUG_1
	    fprintf(stderr,"\nSome spare time avail.. New Event Rate of %d\n",
		    EventRate);
#endif

	  }
      }
      else { /*** There is no time for the rating ***/

	if (ctsa.number_mks || rsa.number_mks || sub_rsa.number_mks) {
	  bSleep = IS_NOTIME; /// There is no time to make the rating

#ifdef ARTIS_CONF_AUTO_METAREASONING
          if (dont_inner) {
            EventRate /= 2;    /// If there is not enough time for the rating
            if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
                 KDM_Set_Value( pMetaClass->eventRate, EventRate );
#ifdef ARTIS_TRACE_EVENTS
                 mutex_unlock( MUTEX_EM );
#endif
            }
            ///            fprintf(stderr, "(ds_loop): Segunda divisin por 2 del EventRate\n");
            ///SetAttentionDegree( ( EventRate / 2.0 ) / MAX_EV );
          }
#endif
        }
	else bSleep = IS_SLEEPING; /// There is nothing to execute, better sleep

	///#ifdef DEBUG
	///fprintf(stderr,"\n No time for the rating. New Event Rate of %d\n",
	///EventRate);
	///#endif
      }
    }
    else if (dont_inner) {   /// There is no time for the condition testing    
#ifdef ARTIS_CONF_AUTO_METAREASONING
      EventRate /= 2; ///If there is not enough time for the condition testing
      if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
           KDM_Set_Value( pMetaClass->eventRate, EventRate );
#ifdef ARTIS_TRACE_EVENTS
           mutex_unlock( MUTEX_EM );
#endif
      }
#endif
      ///      fprintf(stderr, "(ds_loop): Tercera divisin por 2 del EventRate\n");
      ///SetAttentionDegree( ( EventRate / 2.0 ) / MAX_EV );
      ///#ifdef DEBUG
      ///fprintf(stderr,"\nNo time for the cond. testing. New EventRate of %d\n",
      ///EventRate);
      ///#endif
    }

#ifdef ARTIS_CONF_DS_DEBUG_2
    fprintf(stderr, "(main_control) nTriggerState: %d, bSleep: %d, lMaxMandatory: %lld\n",
	    nTriggerState, bSleep, (long) lMaxMandatory);
#endif

    dont_inner = 0; 

  } while ((!((nTriggerState != TRIGGER_OK) && bSleep))
	   && enough_time(lMaxMandatory));  

#ifdef ARTIS_TRACE_EVENTS
  trace_event( SLS_STOP, 0, (long long) 0, 0 );
  trace_event( AA_QUALITY, 1, (long long) system_quality, 0 );
#endif

  if ((bSleep == IS_SLEEPING) &&
      (nTriggerState != TRIGGER_OK) && (nTriggerState != TRIGGER_NOTIME_ER))
    return IS_SLEEPING;
  else return IS_NOTIME;  /// So many things to do, and so little time!!!

}
#else
/* ============================================================================ */
int ds_loop() {
/* ============================================================================ */

   /*It's only control*/
  mks *scheduleMKS;
  int dont_inner   =  1;
  /* interrupt type es para el rating de luis*/
  int interrupt_type = 0;
  int Is_init_deadline = -1,level_end;
  systime_t t1,t2,running_time;
  in_agent *in_agent_curr;

  systime_t lEnd_Time;

  systime_t lBeginningTime;

  lBeginningTime=systime_gettime(); 

  fprintf(stderr,"\t______________________________________________________\n");
  fprintf(stderr,"\t|                     CONTROL                        |\n");
  fprintf(stderr,"\t------------------------------------------------------\n");

  if (enough_time(MANDATORY_T)){
    lEnd_Time = lBeginningTime + MANDATORY_T;
    if (trigger(lEnd_Time) == TRIGGER_EXIT) {/*** Yankee Go Home !!! ***/
      fprintf(stderr, "\t(is.c)::main_control(): This is the end of the world as we know it\n");
      
      return TRIGGER_EXIT;
    }

    if (enough_time(MANDATORY_CT)){
#if defined ARTIS_CONF_DS_DEBUG_1
      fprintf(stderr, "\t(control.c)Let's test conditions...\n");
#endif
      condition_testing();
      if (enough_time(MANDATORY_R)){
#if defined ARTIS_CONF_DS_DEBUG_1
	fprintf(stderr, "\t(control.c)Let's rate...\n");
#endif
	rating(interrupt_type);

	while (enough_time(MANDATORY_S) && dont_inner){

#if defined ARTIS_CONF_DS_DEBUG_1
	  fprintf(stderr, "\t(control.c)Let's schedule...\n");
#endif
	  scheduleMKS = (mks *)schedule();
	  if ((scheduleMKS != (mks *)NULL)){
	    get_info(scheduleMKS,&t1);
	    get_available(MANDATORY_I,t1,&t2);
	    get_real_t(scheduleMKS,t2,&running_time);
	    if ( running_time > MIN_RUNTIME){ ///0){ /* r3 = running time */
	      in_agent_curr=scheduleMKS->mother;
#if defined ARTIS_CONF_DS_DEBUG_1
	      fprintf(stderr,"\t (is.c)Previo execute_until: %d until %lld\n", 
		      scheduleMKS->p_ks->KS->identifier, running_time);
#endif

              // This extern variable is for kiwi purposes
              running_level = scheduleMKS->p_ks->KS->kiwi_id;
#ifdef ARTIS_CONF_DS_DEBUG_1
              fprintf(stderr, "(sls.c) interpret: running_level = %d - UNTIL %lld\n", 
                      running_level, running_time);
#endif
	      level_end = Level_Execute_Until(scheduleMKS->p_ks->KS->identifier,
					running_time, 
					scheduleMKS->p_ks->KS->kiwi_id ); 
	      interpret_end(scheduleMKS, level_end);		
	    }
	    else{  /*** There is not enough time to run the scheduled ks ***/
	      fprintf(stderr,"\n No time to run levels. Updating Queues\n");
	      scheduleMKS->p_ks->KS->ks_deliber=0;
	      in_last(&sub_rsa,scheduleMKS);
	      first_call=1;
	      dont_inner=1;
	      
#if defined ARTIS_CONF_DS_DEBUG_1
	      fprintf(stderr,"\t Time exhausted\n");
#endif
	    }
	  }
	  else{   /*** scheduleMKS == (mks *) NULL ***/
	    first_call=1;
	    dont_inner=0;
#if defined ARTIS_CONF_DS_DEBUG_1
	    fprintf(stderr,"SIN AGENTES QUE SCHEDULAR\n");
#endif
	    
#ifdef ARTIS_TRACE_EVENTS
	    trace_event( AA_QUALITY, 1, (long long) system_quality, 0 );
#endif
	    
#ifdef ARTIS_CONF_DS_DEBUG_1
	    fprintf(stderr,"\n Some spare time. New Event Rate of %d\n", 
		    EventRate);
#endif
	    
	  }	  
	} /*** while enough_time ***/
      }
      else { /*** There is no time for the rating ***/
#ifdef ARTIS_CONF_DS_DEBUG_1
	fprintf(stderr,"\n No time for the rating. New Event Rate of %d\n", 
		EventRate);
#endif
      }
    }
    else {   /*** There is no time for the condition testing ***/

#ifdef ARTIS_TRACE_EVENTS
	    trace_event( AA_QUALITY, 1, (long long) system_quality, 0 );
#endif
#ifdef ARTIS_CONF_DS_DEBUG_1
	    fprintf(stderr,"\n No time for condition testing. Event Rate %d\n",
		    EventRate);
#endif

    }
  }
  else {
#ifdef ARTIS_CONF_DS_DEBUG_1
    fprintf(stderr, "\t(control.c)There is no time like the 1st time...\n");
    fprintf(stderr,"\n No time for triggering. New Event Rate of %d\n",
	    EventRate);
#endif
  }

  return IS_SLEEPING;

}
#endif /// #ifdef ARTIS_CONF_METAREASONING


int DS_start(){ 

	struct  sched_param schd;
        int                 level;
        char                sched_rtf_name[32];

#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_WAKEUP
        char                wakeup_rtf_name[32];
#endif
        int                 bReturnState;
#endif
        
        /*variable spara la segunda parte*/
        systime_t           time_to_die;
        systime_t           slack_interval;
        systime_t           start_time;   //ipg loop activation time
        systime_t           min;          /* ipg min difference time. */
        systime_t           difference;   
                              /* difference between linux and rt_linux*/
        systime_t           last_time;
        /* these two used to measure inteligent server release delay since 
           rt_scheduler activates linux */
        long long           cpu_ticks_now;   /* cpu_ticks since activation */
        systime_t           delay_time;      /* delay time */
        systime_t           kill_start_time; 
                              /* AUXILIAR VARIABLE to measure kill_level time*/
	
        int                 ready_levels_left;
        
        struct sched_mssg_t sched_mssg;
        
        int                 e_m_size;
        int                 nmssg=0;
        int                 i;
        int                 counter;
	int                 the_level;
        systime_t           last;

        /* Control*/
        int dont_inner=0;
        int interrupt_type=0;
        int level_end;

        systime_t running_time,t1,t2,t3;

#ifdef ARTIS_CONF_DS_DEBUG_1 
        fprintf( stderr, "(DS_start) Initializing time\n");
#endif

        MANDATORY_CT = INI_MANDATORY_CT;
        MANDATORY_I  = INI_MANDATORY_I;
        MANDATORY_R  = INI_MANDATORY_R;
        MANDATORY_S  = INI_MANDATORY_S;
	MANDATORY_T  = INI_MANDATORY_T;

#ifdef COMO
	como_phase = 0;
	MANDATORY_COMO = INI_MANDATORY_C;
	NUM_ERROR_MESSAGES = 255;
	NUM_RECV_MESSAGES = 255;
	NUM_SEND_MESSAGES = 255;
#endif
	
        systime_init();

        last_time = systime_gettime();
        init_time=last_time;

        /* Open the RT FIFOs that provides information from the 1st-level
           scheduler about the events produced and the slack time available
        */

#ifdef ARTIS_CONF_DS_DEBUG_1
        fprintf(stderr, "(DS_start) Opening rt_fifos \n");
#endif

        sprintf(sched_rtf_name, "/dev/rtf%d", SCHED_RTFIFO);

        if ((sched_rtf = open(sched_rtf_name, O_RDONLY|O_SYNC)) < 0) {
          fprintf( stderr, "Cannot open rtf %s (%d).\n", sched_rtf_name, 
                   sched_rtf);
          DS_stop(1);
        }

#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_WAKEUP
        sprintf(wakeup_rtf_name, "/dev/rtf%d", ADVANCE_FINAL_PART_FIFO);

        if ((wakeup_rtf = open(wakeup_rtf_name, O_WRONLY|O_NONBLOCK)) < 0) {
          fprintf( stderr, "Cannot open %s (%d).\n", wakeup_rtf_name, 
                   wakeup_rtf);
          DS_stop(1);
        }
#endif
#endif

        init_control();    /// ---> DELIBERATIVE SERVER INITIALIZATION

#ifdef ARTIS_CONF_DS_DEBUG_1 
        fprintf(stderr,"(scheduler.c)Total number of levels %d: \n", nlevels); 
#endif

        /* 
           Now start the endless loop. This loop has 5 stages: 
               (1) Read (and wait) the sched FIFO.
               (2) Process the message -> get the available slack
               (3) Execute the "satisficing cycle" of the DS
        */

        now = systime_gettime();

        fprintf(stderr,"\n:::::::::::::::::::::::::::::::::::::::: START :::::::::::::::::::::::::::::::\n");

        for (;EVER;) {
#if defined ARTIS_CONF_DS_DEBUG_1 || defined TIMING
          now = systime_gettime();

          fprintf(stderr,"\t Read to sleep in %lld [%lld] (self= %d) ----...",
                  now-init_time, rdtsc(), (int) pthread_self());
#endif

          /* 
             (1) Read the sched FIFO. It suspends itself until a message is
             received. Then, update system time.
          */ 

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_DS_DEBUG_2
      trace_event( DEBUG_EVENT, 1, (long long) 25, 0 );
#endif
#endif

          read(sched_rtf, (char *) &sched_mssg, sizeof(sched_mssg_t));

#ifdef ARTIS_TRACE_EVENTS
          trace_event(OPT_START, 0, 0, 0);
#endif

          if (sched_mssg.magic!=SCHED_MSSG_MAGIC) {
            fprintf(stderr,"ERROR READING FIFO, sched_mssg corrupt. Exiting...\n");
            DS_stop(1);
          }

          fprintf(stderr,"AWAKENED [%lld]!!\n", rdtsc());
          cpu_ticks_now = rdtsc() - sched_mssg.cpu_ticks;
          delay_time = (systime_t)((cpu_ticks_now / (long long)CPU_FRECUENCY)*1000);
          
          last_time=now; /* now esta aqui falta de actualizar*/

          now = systime_gettime(); /* actualizo now AQUI */

          start_time=now;
   
#if defined ARTIS_CONF_DS_DEBUG_1
          fprintf(stderr,"\n/--------------------------------------------------------------------------\\\n");
          fprintf(stderr,"|                         A C T I V A T I O N                              |\n");
          fprintf(stderr,"\\--------------------------------------------------------------------------/\n");
          fprintf(stderr,"Waiting in fifo from %lld to [now=%lld]. Total Suspended=%lld , TICKStoMSEC=%lld\n",
                  last_time-init_time, now-init_time,now-last_time,delay_time);
#endif
 
          /* (2) Process the message -> get the available slack
             
                Note: some time must be subtracted from the end of slack,
                   in order for the 2nd-level sched. to return to a known
                   state before being preempted by the rtl scheduler.
          */

          /*** ThreeC 20030812 ***/
          /*** All the time values received from the FLS are in nanoseconds, 
              ***/
          time_left=(systime_t) ( sched_mssg.slack_interval_time );
          end_time= now + (systime_t) time_left;
   
#ifdef ARTIS_CONF_DS_DEBUG_1
          fprintf(stderr,
                  "\t__________________________________________________________________________\n");
          fprintf(stderr,
                  "\t| Mssg No. %d. | RTL Time = %llu | TSC = %llu | slack=%lld |  EndPoint=%lld |\n",
                  ++nmssg, sched_mssg.TimeStamp, rdtsc(), time_left,
                  end_time-init_time);
          fprintf(stderr,
                  "\t--------------------------------------------------------------------------\n");
#endif                        
          
          /***
           *           (3) Execute the "satisficing cycle" of the DS
           ***/
          
          /*It's only control*/
#ifdef ARTIS_CONF_METAREASONING
          if ((bReturnState = ds_loop()) == TRIGGER_EXIT) {   /// ---> DELIBERATIVE SERVER
              /*** Yankee Go Home!!! ***/
#else
              if (ds_loop() == TRIGGER_EXIT)   ///  ---> DELIBERATIVE SERVER
                {
                  /*** Yankee Go Home!!! ***/
#endif
                  
#ifdef ARTIS_TRACE_EVENTS
                  trace_event(OPT_END, 0, 0, 0);
#endif

                  fprintf(stderr,"main_control() decided to finish the cotarrro\n");
                  DS_stop(0);
                  
                }
              
#ifdef ARTIS_TRACE_EVENTS
              else trace_event(OPT_END, 0, 0, 0);
#endif
              
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_DS_DEBUG_2
      trace_event( DEBUG_EVENT, 1, (long long) 68, 0 );
#endif
#endif

#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_WAKEUP
              if (bReturnState == IS_SLEEPING)
                write( wakeup_rtf, "3C", sizeof("3C")); /// Sleeping Beauty
              else write( wakeup_rtf, "NO", sizeof("NO")); /// NO TIME LEFT
#endif 
#endif


#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_DS_DEBUG_2
      trace_event( DEBUG_EVENT, 1, (long long) 69, 0 );
#endif
#endif

  } /* for(;EVER;) loop */
}
       
/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
