/**  
 *  \file em.c
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Event Manager 2.0
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

 /* THREAD STUFF */

#include <pthread.h>
#include <unistd.h>

#include "../../include/time.h"

#include "../../linux_core/include/linux_core.h"
#include "../include/dserver/decl.h"
#include "../../include/kdm.h"

#include "../../events/include/events.h"
#include "../../events/include/events/events_internal.h"

#include "../../common/include/artis_conf.h"  ///< For the ARTIS_CONF_METAREASONING and COMO

#ifdef ARTIS_CONF_METAREASONING
#include "../../include/metaclass.h"

#include "../../include/mutex.h"
#endif


#ifdef ARTIS_TRACE_EVENTS
#include "../../trace/include/trace.h"
#endif

/*#include "../include/est_datos4.0.h"*/
///#include "../include/agenda4.0.h"

/*** #include "../include/time_control4.0.h" ***/

///#include "../include/errs.h"
/****************************************************************/
#include "../include/dserver/em.h"
/****************************************************************/

#define LUIS_DEBUG

/***/
#define DS_DEBUG 

/***/
#define DEBUG

/***/
#define DS_EXEC_DEBUG
/***/
extern n_mks *n_mks_actual;
extern Lks l_ks;
extern l_mks tsa;
extern l_mks ctsa;
extern l_mks rsa;
extern l_mks sub_rsa;
extern n_in_agent *l_in_agent;
extern float system_quality;  
extern float system_gain;
extern systime_t origin;
extern l_allocat sched;
extern systime_t init_deadline;
extern int mks_n;

/// Time points declaration 20040809
/** Mandatory  trigger execution time. This is a estimation based on worst
 *   execution time of this function.
 */

/** Mandatory  condition_testing execution time. This is a estimation based on
 *    worst execution time of this function.
 */
systime_t MANDATORY_CT;///=200;

/** Mandatory  rating execution time. This is a estimation based on worst
 *   execution time of this function.
 */
systime_t MANDATORY_R;///=200;

/** Mandatory  scheduling execution time. This is a estimation based on worst
 *   execution time of this function.
 */
systime_t MANDATORY_S;///=400;

systime_t MANDATORY_I;///=600;
/*** Mandatory interpreting execution time. This is a estimation based on worst
 *   execution time of this function.
 ***/
systime_t MANDATORY_T;///=1000;

systime_t MIN_RUNTIME=61;  /// execute_until gives problems with 50 ms

systime_t control_slack; 
/*** It keeps the available time for Control. ***/

systime_t  real_time;
/*** It keeps the last moment in which control 
 *   called get_real_time().
 ***/

systime_t end_time;
/*** It keeps the moment in which Control has to terminate current invocation.
 ***/

systime_t init_time ;
/*** It keeps the moment in which Control began it's execution or the moment
 *   in which Control begins a new loop.
 ***/
systime_t origin;
/*** It keeps the moment in which Control was activated for first time.
 ***/

/// end timepoints declaration 20040809



#ifdef COMO
/// CoMo parameters

/// For CoMo
systime_t MANDATORY_COMO;///=200;

/// Number of error messages to answer per loop
int NUM_ERROR_MESSAGES = 20;

/// Number of messages to receive per loop
int NUM_RECV_MESSAGES = 20;

/// Number of messages to send per loop
int NUM_SEND_MESSAGES = 20;

/// CoMo stage
int como_phase = 0;

/// end CoMo parameters
#endif



static char      entry_file[256];


///#define CALCULATING_MANDATORY


/*** First call to Control. ***/
/* ============================================================================ */
void init_control(){
/* ============================================================================ */

  //init_periodic_list();

  ///init_rule_array();
  init_l_ks();
  init(&tsa);
  init(&ctsa);
  init(&rsa);
  init(&sub_rsa);
  l_hole=NULL;
  system_quality = 0 ;  /*** Declared in agenda2.0.h ***/

#ifdef ARTIS_CONF_METAREASONING
  mutex_create( MUTEX_EM, sched_get_priority_max(SCHED_FIFO) ); 

  if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
	heuristic = * (int *) KDM_Get_Value( pMetaClass->slsPolicy );
#ifdef ARTIS_TRACE_EVENTS
	mutex_unlock( MUTEX_EM );
#endif
  }
#endif

  if (heuristic ==DBS) {
    init_sched();
    system_gain = 0;
    mks_n = 1;
  }


#ifdef ARTIS_CONF_METAREASONING
/*** ThreeC 20030808 ***/
 LevelsInterrupted = LevelsCompleted = LevelsKilled = 0;

 //EventRate = 1; /*** The events are treated one each control iteration ***/
 /// SetAttentionDegree( 1.0 ); /// MAX_KDM_EV );   /*** The events are treated one each control iteration ***/
 if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
       EventRate = * (int *) KDM_Get_Value( pMetaClass->eventRate );
#ifdef ARTIS_TRACE_EVENTS
	mutex_unlock( MUTEX_EM );
#endif
 }

 /*** EventRate = MAX_KDM_EV; /*** Default value: All the events in the event store should be treated ***/

/// KDM_Set_Value( pMetaClass->maxQuality, MaxQuality );
#endif

 fprintf( stderr, "\t(control.c:END init_control)\n" );
}



/* ============================================================================ */
n_regla *create_regla(char nombre[],char tipo[],int imp,char level[], 
		      char slot[]){
/* ============================================================================ */
  n_regla *aux;
  
  aux=(n_regla *)malloc(sizeof(n_regla));
  strcpy(aux->name, nombre);
  strcpy(aux->type, tipo);
  aux->importance=imp;
  strcpy(aux->level,level);
  strcpy(aux->slot,slot);
  aux->sig=NULL;
  aux->NMKS2=NULL;
  
  return aux;
}


/* ============================================================================ */
void in_regla(n_regla *regla){
/* ============================================================================ */
  int tipo;
  n_regla *lista, *aux;
		
  tipo=num_tipo(regla->type);
  lista=v_reglas[tipo];
  if (lista==NULL)
    v_reglas[tipo]=regla;
  else { 
    while(lista!=NULL){
      aux=lista;
      lista=lista->sig;
    }
    aux->sig=regla;
  }
}

/* ============================================================================ */
n_mks2 *in_mks_regla (char nombre[], int import, n_regla *regla, char mks_v[]){
/* ============================================================================ */
  n_mks2 *aux, *aux2, *aux3;
  n_mks *aux4;
  in_agent *in_ag;
  int enc = 1;
  
  aux=(n_mks2 *)malloc(sizeof(n_mks2));
  aux->importance=import;
  aux->prev=NULL;
  if (regla->NMKS2==NULL)
    regla->NMKS2=aux;
  else {  
    aux2=regla->NMKS2;	
    while (aux2!=NULL){
      aux3=aux2;
      aux2=aux2->sig; 
    }
    aux3->sig=aux;
    aux->prev=aux3;
  }
  aux->KS=NULL;
  aux->sig=NULL;
  aux4=periodic_mks_l.first;/*comprobar*/
  while ((aux4!=NULL) && enc){
    in_ag=aux4->MKS->mother;
    if (strcmp(in_ag->name,mks_v)||
	strcmp(aux4->MKS->name,nombre))
      aux4=aux4->next;
    else enc=0;
  }
  if (aux4!=NULL)
    aux->MKS=aux4->MKS;
  else fprintf(stderr,"ERROR\t(control.c:in_a_regla)Agente original no encontrado\n");

  return aux;
} 

/* ============================================================================ */
void in_ks_mks(char nombre[], systime_t deadline, n_mks2 *n_mks, float gain){
/* ============================================================================ */
  KS2 *aux, *aux2, *lista;
  n_ks *aux3;
  
  aux=(KS2 *)malloc(sizeof(KS2));
  aux->sig=NULL;
  aux->prev=NULL;
  aux->deadline=(systime_t)deadline;
  aux->coef= gain / deadline;
  aux->gain=gain;
  lista=n_mks->KS;
  if (lista==NULL)
    n_mks->KS=aux;
  else {
    while (lista!=NULL){
      aux2=lista;
      lista=lista->sig; 
    }
    aux2->sig=aux;
    aux->prev=aux2;		
  }
  aux3=n_mks->MKS->p_ks;
  while (strcmp(aux3->KS->name,nombre))
    aux3=aux3->next;
  aux->KS=aux3->KS;

#ifdef DS_DEBUG
  fprintf(stderr, "\t(is.c::in_ks_ag) KS: %d, kiwi_id: %d\n", aux->KS->identifier, aux->KS->kiwi_id);
#endif

}			

/* ============================================================= */
void kill_old_mks (l_mks *agenda, int identifier){
/* ============================================================= */

  n_mks *f_mks, *p_mks , *aux2;
  in_agent *aux;

  f_mks = agenda->first;
  p_mks = aux2 = (n_mks *)NULL;

#ifdef DEBUG
  fprintf(stderr, "(is.c::kill_old_mks) Killing %d\n", identifier);
#endif

  while (f_mks != (n_mks *)NULL){
    aux = f_mks->MKS->mother;

#ifdef DEBUG
    fprintf(stderr, "(is.c::kill_old_mks) Is this one %d?\n", aux->in_agent_ident);
#endif

    if (aux->in_agent_ident == identifier){ 
      agenda->number_mks--;
      if (p_mks != (n_mks *)NULL)	/* update pointers */
	p_mks->next = f_mks->next;
      else agenda->first = f_mks->next;
      if (f_mks->next != (n_mks *)NULL){
	aux2 = f_mks->next;
	aux2->prev = p_mks;
      }
      else {
	agenda->last = p_mks;
	aux2 = (n_mks *) NULL;
      }
     
      if (f_mks->MKS->state == PAUSED){/*si tiene niveles en pausa sera el 1*/

#ifdef DEBUG
	fprintf(stderr, "(is.c::kill_old_mks) El pausado se resetea\n");
#endif

	Level_Regenerate(f_mks->MKS->p_ks->KS->identifier, 
		    f_mks->MKS->p_ks->KS->kiwi_id );
      }

#ifdef DEBUG
      fprintf(stderr, "(is.c::kill_old_mks) Free Agent AG\n");
#endif

      free_mks(f_mks->MKS);

#ifdef DEBUG
      fprintf(stderr, "(is.c::kill_old_mks) Free Agent\n");
#endif

      free(f_mks);

#ifdef DEBUG
      fprintf(stderr, "(is.c::kill_old_mks) After Free Agent\n");
#endif

      if (aux2 == (n_mks *)NULL)
	f_mks = (n_mks *)NULL;
      else f_mks = aux2;
    }
    else {
      p_mks=f_mks;
      f_mks=f_mks->next;
    }
  
  } /*** while ***/
 
}


/* ============================================================================ */
in_agent *create_in_agent(char name[10], int ident, int num_in_agent){
/* ============================================================================ */
  in_agent *aux;
	
  aux=(in_agent*)malloc(sizeof(in_agent));
  strcpy(aux->name, name);
  aux->in_agent_ident=ident;
  aux->no_mks=num_in_agent;
  aux->no_critical_mks=0;
  aux->mks_list=NULL;
  return aux;
}

/* ============================================================================ */
n_in_agent *insert_in_agent(in_agent *in_agent_curr, n_in_agent *l_in_agent){
/* ============================================================================ */
  n_in_agent *aux,*aux2,*aux3=NULL;

  aux=(n_in_agent*)malloc(sizeof(n_in_agent));
  aux->IN_AGENT=in_agent_curr;
  aux->next=aux->prev=NULL;
  if (l_in_agent==NULL)
    l_in_agent=aux;
  else {
    aux2=l_in_agent;
    while (aux2!=NULL){
      aux3=aux2;
      aux2=aux2->next;
    }
    aux3->next=aux;
    aux->prev=aux3;
  }
  return l_in_agent;
}

/* ============================================================================ */
void in_mks_at_in_agent(in_agent *in_agent_curr, mks *mks_curr){
/* ============================================================================ */
  n_mks *aux,*aux2,*aux3=NULL;

  aux=(n_mks*)malloc(sizeof(n_mks));
  aux->MKS=mks_curr;
  aux->next=aux->prev=NULL;
  if (in_agent_curr->mks_list==NULL) 
    in_agent_curr->mks_list=aux;
  else {
    aux2=in_agent_curr->mks_list;
    while (aux2!=NULL){
      aux3=aux2;
      aux2=aux2->next;
    }
    aux3->next=aux;
    aux->prev=aux3;
  }
}


/* ============================================================================ */
mks *create_mks(char *name, int ident, char *tipo, char *mode, int importan,
		    int num_ks, in_agent *in_agent_curr){
/* ============================================================================ */
  mks *aux;
  int tipe=0;

  aux=(mks*)malloc(sizeof(mks));
  aux->name=(char *)malloc(strlen(name)+1);
  strcpy(aux->name, name);
  aux->identifier=ident;

  if (strcmp(tipo,"Critical")==0)
  {
	tipe=1;
	in_agent_curr->no_critical_mks++;
  }

  aux->type=tipe;
  aux->importance=importan;
  aux->number_ks=num_ks;
  aux->p_ks=NULL;
  aux->mother=in_agent_curr;
  aux->type_MKS = (strcmp(mode,"RP")==0)?0:1;
  return aux;
}

/* ============================================================================ */
ks *create_ks(char *name, int ident, systime_t duracion, float gain){
/* ============================================================================ */
  ks *aux;

  aux=(ks*)malloc(sizeof(ks));
  aux->name=(char *)malloc(strlen(name)+1);
  strcpy(aux->name, name);
  aux->identifier=ident;
  aux->duration=duracion;
  aux->gain=gain;
  aux->ks_id=0;
  aux->deadline=0;
  return aux;
}

/* ============================================================================ */
void insert_ks(ks *ks_curr, mks *mks_curr){
/* ============================================================================ */
  n_ks *aux,*aux2,*aux3=NULL;

  aux=(n_ks*)malloc(sizeof(n_ks));
  aux->KS=ks_curr;
  aux->next=aux->prev=NULL;
  
  if (mks_curr->p_ks==NULL) 
    mks_curr->p_ks=aux;	
  else {
    aux2=mks_curr->p_ks;
    while (aux2!=NULL) {
      aux3=aux2;
      aux2=aux2->next;
    }
    aux3->next=aux;
    aux->prev=aux3;
  }

#ifdef ARTIS_CONF_METAREASONING
  if (!mutex_system_lock( MUTEX_EM, (systime_t) 2 * WCET_KDM_ACCESS )) {
         KDM_Set_Value( pMetaClass->maxQuality, 
		  (* (int *) KDM_Get_Value( pMetaClass->slsPolicy )) 
		   + (ks_curr->gain * mks_curr->importance) );
#ifdef ARTIS_TRACE_EVENTS
	mutex_unlock( MUTEX_EM );
#endif
  }
#endif

  ks_curr->kiwi_id = ((ks_curr->identifier) - (((in_agent*)(mks_curr->mother))->no_critical_mks)) +1;
}



/* ============================================================================ */
void set_deanboddy_ks(char *name_ks, int kiwi_id, 
		       int identact, systime_t deadline, systime_t duration, float gain, systime_t read_time){
/* ============================================================================ */

  ks *aux_ks;
  
  aux_ks=(ks *)malloc( sizeof(ks) );
  aux_ks->ks_id =      0;
  aux_ks->name=(char *)malloc( strlen(name_ks)+1 );
  strcpy(aux_ks->name, name_ks);

  /*** ThreeC 20030910 ***/
  aux_ks->kiwi_id = kiwi_id;

  aux_ks->weight =      0;
  aux_ks->duration =    duration;
  aux_ks->identifier =  identact;
  aux_ks->deadline =    deadline;
  aux_ks->execution =   0;	
  aux_ks->ks_deliber = 0;		/*** KS Deliber time ***/
  aux_ks->gain =        gain;
  aux_ks->last_check =  read_time; 
 
#ifdef LUIS_DEBUG
  fprintf(stderr,"\t(control.c:set_dean_boddy_ks)set_deanboddy_ksi: name_ks: %s deadline:%lld duration:%lld\n", name_ks,deadline, duration);
#endif

  in_l_ks(aux_ks); 
  /*** We insert this new ks in the agent ks list ***/
}


/* ============================================================================ */
void trigger_periodic_mks(char *mks_name, systime_t deadline, systime_t read_time){
/* ============================================================================ */

  n_mks *n_aux;
  n_ks   *actual_nks;
  systime_t ex_deadline = 0;

#ifdef LUIS_DEBUG
    fprintf(stderr,"\t(control.c:trigger_periodic_agent) first deadline of agent %s: %lld\n",
     mks_name, deadline);
#endif
 
  n_aux=periodic_mks_l.first;
  while((n_aux != NULL) 
	&& (strcmp(n_aux->MKS->name, mks_name)!=0))
    n_aux=n_aux->next;
  if (n_aux == NULL) return;
 
  /*** We search for the periodic agent ( which is in the global and permanent
   *   periodic agent list) we've to activate.
   ***/
     if (n_aux->MKS->type==0){
      set_deanboddy_ks(n_aux->MKS->p_ks->KS->name,
		      n_aux->MKS->p_ks->KS->kiwi_id,
		      n_aux->MKS->p_ks->KS->identifier,
		      deadline,
		      n_aux->MKS->p_ks->KS->duration, 
		      n_aux->MKS->p_ks->KS->gain,
		      read_time);
    }
    else ex_deadline=deadline;
    actual_nks=n_aux->MKS->p_ks->next;
    while(actual_nks!=NULL){
      set_deanboddy_ks(actual_nks->KS->name, 
			actual_nks->KS->kiwi_id,
			actual_nks->KS->identifier, ex_deadline,
			actual_nks->KS->duration, 
			actual_nks->KS->gain,
			read_time);
 
      /*** Next, and next, and next... ks deadline=0. 
       *   When current ks has ended its execution then the
       *    remainding deadline is assigned to the next ks.
       ***/
     actual_nks=actual_nks->next;
    }
  
  trigger_mks(n_aux->MKS->name, PERIODIC, 
		n_aux->MKS->importance, n_aux->MKS->mother); 
  /*** Create and activate a new agent.***/
}


/* ============================================================================ */
///static void CreateStructures( void ) {
/* ============================================================================ */
/*	char structure[10], name[40], type[10];
	int identity, num_MKS, num_crit_MKS, importance, num_levels,
	  nCritMKS = 0 // Number of MKS Crit. in Agent ;
	int bCritMKS;  /// 1 if the MKS being analysed is Critical, 0 if it is NonCritical
	long long int duration;
	float gain;

	in_agent *in_agent_curr, *new_in_agent;
	mks *new_MKS, *nu_mks;
	ks *new_ks, *nu_ks;
	FILE *fd;

	fprintf(stderr,"(control.c-CreateStructures) voy a leer ../app/data/control.txt\n");
	if ((fd=fopen("control.txt", "r"))== NULL){
	   fprintf(stderr, "ERROR \t(control.c:CreateStructures) Control I can't open file\n");
	   exit(1);
	}

	//#ifdef ARTIS_CONF_METAREASONING
	/// ThreeC 20010718
        MaxQuality = 0;

	num_crit_MKS = 0;
	//#endif

	do{
		fscanf(fd,"%s",structure);

		if (!strcmp(structure, "AGENT")){
			fscanf(fd,"%s %d%d", name, &identity, &num_MKS);
			
			num_crit_MKS = num_MKS;

			in_agent_curr=create_in_agent(name, identity, num_MKS, num_crit_MKS);
			l_in_agent=insert_in_agent(in_agent_curr, l_in_agent);
	
#ifdef ARTIS_TRACE_EVENTS
			///			opt_trace_task( "Critical", identity, name ); /// To add names to critical tasks in kiwi
#endif

		}
		else if (!strcmp(structure, "MKS")){
			fscanf(fd,"%s %s %d %d",name, type, &importance, &num_levels);

			if (!( bCritMKS = strcmp(type, "NonCritical") )) {
			  activ->no_critical_agents--;
			}
			else nCritMKS++;		     

			nu_mks=create_mks(name,7,type,importance,num_levels,in_agent_curr);
				 in(&periodic_mks_l, nu_mks);
				 in_mks_at_in_agent(in_agent_curr, nu_mks);
				
			}
		else if (!strcmp(structure, "LEVEL")){
			fscanf(fd,"%s %d %lld %f", 
			       name, &identity, &duration, &gain);
			nu_ks=create_ks(name, identity, duration, gain);

			nu_ks->kiwi_id = identity - nCritMKS + 1;    // Id. for the kiwi (discounting critical levels)

#ifdef DS_DEBUG
			  fprintf( stderr, "(is.c::CreateStructures) Level: %d, Kiwi Id: %d\n", identity, nu_ks->kiwi_id );
#endif

			insert_ks(nu_ks, nu_mks);

			//#ifdef ARTIS_CONF_METAREASONING
			/// ThreeC 20030808
                        MaxQuality += gain * importance;
			//#endif

#ifdef ARTIS_TRACE_EVENTS
			if (!bCritMKS)
			  /// opt_trace_task( "Optional", nu_ksi->kiwi_id, name )
			  ; 
			           /// To add names to critical tasks in kiwi
#endif

			  }
 		else if (!strcmp(structure, "FINAL")){
		  fscanf(fd,"%lld",&duration);
			  //	nu_final=create_final();
			  //	in_fin_at_in_agent(in_agent_curr,nu_final); 
		}
	}while(strcmp(structure, "END"));			
	
#ifdef ARTIS_TRACE_EVENTS
	///	opt_trace_end_names();
#endif

	fclose(fd);	

}
*/


/* ============================================================================ */
void init_rule_array(){
/* ============================================================================ */
    
  char vector[12],cadenaheuristica[6],nombre[20],tipo[40],level[20],slot[20],in_agent_v[20],
    rule_file[256],xb[12];
  int ident,n_mkss,importancia,i,xa;
  systime_t duracion,deadline;
  float gain;
  n_regla *n_reg;
  n_mks2 *n_mks_l;
  FILE *fd;
  void print_rules();
  
  for (i=0;i<8;i++)
    v_reglas[i]=NULL;

  fprintf(stderr, "(control.c:init_rule_array)voy a abrir el fichero ./control/test/control.rules\n");
  if((fd=fopen( "control.rules", "r"))==NULL){ 
    fprintf(stderr, "ERROR  \t(control.c:print_rules): No puedo abrir fichero de reglas\n");
    exit(1);
  }

  fscanf(fd,"%s",vector);
  fprintf(stderr,"(control.c-init_rule_array) START -- --\n");
  while(strcmp(vector,"END") && strcmp(vector,"FIN")){
    if (!strcmp(vector,"HEURISTIC"))
      {
	/** ipg Luis, esto es semi-provisional. En vez de leer un nmero, leemos una cadena. De momento
	    siempre asumir edf, pero habr de hacerse una conversin a enteros o cambiar el formato
	    del fichero 
	    
	if (fscanf(fd,"%d",&heuristic)!= 1) {
	  fprintf(stderr,"ERROR en fichero app/data/control.rules. En HEURISTIC\n");
	  exit(1);
	  }*/
       	if (fscanf(fd,"%s",cadenaheuristica)!= 1) { /* ipg aqui puede cascar como un cerdo */
	  fprintf(stderr,"ERROR en fichero app/data/control.rules. En HEURISTIC\n");
	  exit(1);
	}
	heuristic = strategy_translation(cadenaheuristica);

	fprintf(stderr," Heuristica: %s -> %d \n",cadenaheuristica, heuristic);

      }
    else if (!strcmp(vector,"RULE")){
      if (fscanf(fd,"%s %s %d %s %s",nombre,tipo,&importancia,level,slot)!= 5) 
	{
	  fprintf(stderr,"ERROR en fichero app/data/control.rules. En RULE\n");
	  exit(1);
	}
     fprintf(stderr,"\tRULE: nombre= %s tipo=%s importancia=%d level=%s slot=%s\n",
	     nombre,tipo,importancia,level,slot);
     n_reg=create_regla(nombre,tipo,importancia,level,slot);
     in_regla(n_reg);
    }
    else if (!strcmp(vector,"MKS")){
      if (fscanf(fd,"%s %s %lld %d %s",nombre,in_agent_v,&duracion,&xa,xb)!= 5 )
	{
	  fprintf(stderr,"ERROR en fichero app/data/control.rules. En MKS\n");
	  exit(1);
	} /*<ipgerror>*/
      fprintf(stderr,"\tMKS: nombre= %s activ=%s duracion=%lld xa=%d xb=%s\n",
	      nombre,in_agent_v,duracion,xa,xb);
      
      n_mks_l=in_mks_regla(nombre,(systime_t)duracion,n_reg,in_agent_v);
    }
    else if (!strcmp(vector,"LEVEL:")){
      fscanf(fd,"%s %lld %f",nombre,&deadline,&gain);
      fprintf(stderr,"\t LEVEL: nombre = %s, deadline = %lld, gain = %f\n",
	      nombre,deadline,gain);
      in_ks_mks(nombre,deadline,n_mks_l,gain);
    }
    fscanf(fd,"%s",vector);
  }
  fclose(fd);
 
  print_rules(v_reglas[7]);

}



/* ============================================================================ */
int num_tipo(char type[]){
/* ============================================================================ */
  int tipo;
  if (!strcmp(type, "EXIT_APP"))
    tipo=0;
  else if (!strcmp(type,"CREATION"))
    tipo=1;
  else if (!strcmp(type,"DELETION"))
    tipo=2;
  else if (!strcmp(type,"MODIFICATION"))
    tipo=3;
  else if (!strcmp(type,"NEW_PREDICTION"))
    tipo=4;
  else if (!strcmp(type,"MATCHED_PREDICTION"))
    tipo=5;
  else if (!strcmp(type,"UNMATCHED_PREDICTION"))
    tipo=6;
  else if (!strcmp(type,"CONSIDER_PERIODIC_AGENT"))
    tipo=7;
  else fprintf(stderr,"ERROR\t(control.c:num_tipo) ERROR tipo no correcto");
  
  return tipo;
}


/* ============================================================================ */
void print_in_agent(n_in_agent *in_agent_list){
/* ============================================================================ */
  n_in_agent *aux_in_agent;
  n_mks *aux_mks;
  n_ks   *aux_ks;

  fprintf( stderr, "\t(control.c:print_activ): Imprimiendo lista agentes...\n" );
  aux_in_agent=in_agent_list;
  while (aux_in_agent!=NULL){
    fprintf(stderr, "\t(control.c:print_activ) %s %d %d\n",aux_in_agent->IN_AGENT->name,
	    aux_in_agent->IN_AGENT->in_agent_ident,
	    aux_in_agent->IN_AGENT->no_mks);
    aux_mks=aux_in_agent->IN_AGENT->mks_list;
    if (aux_mks == NULL) fprintf(stderr,"\t(control.c:print_in_agent) NO tengo MKS\n");
    while (aux_mks!=NULL){
      fprintf(stderr, "\t(control.c:print_in_agent) %s %d %d %d %d\n", 
	      aux_mks->MKS->name,
	      aux_mks->MKS->identifier,
	      aux_mks->MKS->type,
	      aux_mks->MKS->importance,
	      aux_mks->MKS->number_ks);
      aux_ks=aux_mks->MKS->p_ks;
      while (aux_ks!=NULL){
	fprintf(stderr,"\t(control.c:print_in_agent) %d %s %ld\n", 
		aux_ks->KS->identifier,
		aux_ks->KS->name,
		aux_ks->KS->duration);
	aux_ks=aux_ks->next;
      }
      aux_mks=aux_mks->next;
    }
    aux_in_agent=aux_in_agent->next;
  }
}


/* ============================================================================ */
void print_rules(n_regla *rule_list){					
/* ============================================================================ */
  n_regla *aux_rule;
  n_mks2   *aux_mks2;
  KS2 	*aux_KS2;

  fprintf(stderr,"\t(control.c:print_rules) imprimo reglas\n");
  aux_rule=rule_list;

  while (aux_rule!=NULL){
     fprintf(stderr,"\t(control.c:print_rules) regla: %s %s %d %s %s\n", aux_rule->name, 
	   aux_rule->type,
	   aux_rule->importance, aux_rule->level,
	   aux_rule->slot);
    aux_mks2=aux_rule->NMKS2;
    while (aux_mks2!=NULL){
      fprintf(stderr,"\t(control.c:print_rules) mks: %s %d\n", aux_mks2->MKS->name,
	     aux_mks2->MKS->importance);
      aux_KS2=aux_mks2->KS;
      while (aux_KS2!=NULL){
	fprintf(stderr,"\t(control.c:print_rules) ks: %s %ld %f\n", 
	       aux_KS2->KS->name,
	       aux_KS2->deadline,
	       aux_KS2->KS->gain);
	aux_KS2=aux_KS2->sig;
      }
      aux_mks2=aux_mks2->sig;
    }
    aux_rule=aux_rule->sig;
  }
  fprintf(stderr,"\t(control.c:print_rules) fin de imprimo reglas\n");
}

/* ========================================================================= */
#ifdef ARTIS_CONF_METAREASONING
/*** ThreeC 20010718 ***/
void ChangeSLS () {
/* ========================================================================= */

  heuristic = (heuristic + 2) % NSLS;

  ///  SetPolicy( heuristic );

  ///if (heuristic == DBS)
  ///free_l_agent(&tsa);

  /*** 
   *    It changes to the next SLS policy. 
   *    It changes from deliberative to reactive SLS policy and from
   *           reactive to deliberative one.
   ***/


     fprintf( stderr, "\t(control.c:ChangeSLS) New SLS Policy: %d \n", 
	      heuristic );
     fprintf( stderr, 
          "\t(control.c:ChangeSLS) Levels: Interrupted: %d, Completed: %d Killed: %d\n", 
             LevelsInterrupted, LevelsCompleted, LevelsKilled);


     LevelsInterrupted = LevelsCompleted = LevelsKilled = 0;

}
#endif


/* ============================================================================ */
void activar(n_mks2 *L_mks){
/* ============================================================================ */
  KS2 *l_ksaux;
  systime_t now;
  void set_deadline();
  void trigger_mks();

  fprintf(stderr, "\t(control.c:activar) Beginning\n");
  
  now=systime_gettime();
  while (L_mks!=NULL){

    fprintf(stderr, "\t(control.c:activar) There are Agents\n");
  
    l_ksaux=L_mks->KS;
    while (l_ksaux!=NULL){
      set_deadline(l_ksaux->KS->name, l_ksaux->KS->kiwi_id, 
		   l_ksaux->KS->identifier,
		   l_ksaux->deadline, l_ksaux->KS->duration,now);
      l_ksaux=l_ksaux->sig;
    }
    trigger_mks(L_mks->KS->KS->identifier,
		  SPORADIC,L_mks->importance,L_mks->MKS->mother);
    L_mks=L_mks->sig;
  }
}	


/* ============================================================================ */
void activar_periodico(n_mks2 *L_mks, int identifier, 
		       systime_t deadline, systime_t read_time){
/* ============================================================================ */
  KS2 *l_ksaux;
  n_ks *real_ks;

  void set_deadline();
  void trigger_mks();
  void trigger_periodic_mks();
  in_agent *aux_in_agent;

  while (L_mks!=NULL){

    aux_in_agent=(in_agent*)(L_mks->MKS->mother);

    if (identifier==aux_in_agent->in_agent_ident){
      l_ksaux=L_mks->KS;
      real_ks=L_mks->MKS->p_ks;
      while (l_ksaux!=NULL){
	real_ks->KS->gain = l_ksaux->gain;
	l_ksaux = l_ksaux->sig;
	real_ks = real_ks->next;
      }
      
      if ((L_mks->MKS->type == 0) || (L_mks->MKS->number_ks > 1))
	trigger_periodic_mks(L_mks->MKS->name,
			     deadline, read_time);
    }
    L_mks=L_mks->sig;      
  }
}					


int TranslateEventIndex (int iEventType){
  switch ( iEventType ){
  case EXIT_APP: return 0;
  case CREATION: return 1;
  case DELETION: return 2;
  case MODIFICATION: return 3;
  case NEW_PREDICTION: return 4;
  case MATCHED_PREDICTION: return 5;
  case UNMATCHED_PREDICTION: return 6;
  case CONSIDER_PERIODIC_AGENT: return 7;
  }
}


/**
 * \fn Trigger_Optional
 *
 * \brief This function activates / deactivates the optional parts of a in-agent.
 */
void Trigger_Optional( float fGain, Event_T regAux ) {
  systime_t EvDeadline = regAux.New_Value.lTime;	
  n_in_agent *pnInAgent = l_in_agent;
  in_agent *pInAgent = pnInAgent->IN_AGENT;
  n_mks *pMks;
  mks *pMKS;

#ifdef DS_EXEC_DEBUG
#ifndef MODULE
  fprintf(stderr, "\t(meta-rule_actions.c::TriggerOptional) The new active Optional %d, with deadline %lld\n",
	  (int) regAux.Old_Value.lTime, EvDeadline ); //KDM_v.i, EvDeadline );
#endif
#endif
  
  /// Looking for the in-agent to activate its optional parts
  if ((int) regAux.Old_Value.lTime == -1) {

#ifdef DS_EXEC_DEBUG
#ifndef MODULE
  fprintf(stderr, "\t(meta-rule_actions.c::TriggerOptional) There is no id to activate, 'cause it is %d, with deadline %d\n",
	  (int) regAux.Old_Value.lTime, (int) EvDeadline ); //KDM_v.i, EvDeadline );
#endif
#endif
  
    return;
  }
  else {
    while (pInAgent->in_agent_ident != (int) regAux.Old_Value.lTime) { //KDM_v.i) {
      pnInAgent = pnInAgent->next;
      pInAgent = pnInAgent->IN_AGENT;
    }
    
#ifdef DS_EXEC_DEBUG
#ifndef MODULE
    fprintf(stderr, "\t(meta-rule_actions.c::TriggerOptional) I have the in-agent to execute\n" );
#endif
#endif
    
    kill_old_mks ( &tsa, (int) regAux.Old_Value.lTime ); //KDM_v.i );
    kill_old_mks ( &rsa, (int) regAux.Old_Value.lTime ); //KDM_v.i );
    
    if (EvDeadline != (systime_t) -1) {
      pMks = pInAgent->mks_list;
      pMKS = pMks->MKS;
      
      ///activar_periodico(lista_regla->NMKS2, regAux.Old_Value.KDM_v.i, EvDeadline, systime_gettime());
      ///activar_periodico(pMKS, regAux.Old_Value.KDM_v.i, EvDeadline, systime_gettime());
      KS2 *l_ksaux;
      n_ks *real_ks;    
      in_agent *aux_in_agent;
      
      aux_in_agent = (in_agent*) (pMKS->mother);
      
      if ( regAux.Old_Value.lTime == aux_in_agent->in_agent_ident ) { //KDM_v.i == aux_in_agent->in_agent_ident ) {
	real_ks = pMKS->p_ks;
	while ( real_ks != NULL ) {
	  real_ks->KS->gain = fGain;
	  real_ks = real_ks->next;
	}
	
	if (( pMKS->type == 0 ) || ( pMKS->number_ks > 1 ))
	trigger_periodic_mks( pMKS->name, EvDeadline, systime_gettime() );
      } 
    }   
  }
}


/*** PANIC!!!: Before we execute a ks, we update its deadline, but not all
 *   deadlines, so if we execute another ks x u.t. later its deadline updating
 *   will be wrong. A problem like this happens with trigger(), condition_
 *   testing(),... . The most correct solution I see is to add a new field
 *   into each ks structure which indicates the moment the ks deadline was
 *   updated.
 ***/

/* ============================================================================ */
int check_deadline(mks *mks_curr){
/* ============================================================================ */
  systime_t time,a1;
  ks *aux;

  time=systime_gettime();
#ifdef LUIS_DEBUG
  fprintf(stderr, "Checking Deadlines of agent %s ...\n", mks_curr->name);
#endif
  /*** How much time since last update?. ***/
  a1 = time - mks_curr->p_ks->KS->last_check;
  mks_curr->p_ks->KS->deadline = (a1 < mks_curr->p_ks->KS->deadline)? mks_curr->p_ks->KS->deadline - a1 : 0; 
                                           
  mks_curr->p_ks->KS->last_check = time;           
  /*** We keep the moment in
   *  which  ksi has been updated.
   ***/
  
#ifdef LUIS_DEBUG
  fprintf(stderr, "Checking Deadlines 2...\n");
#endif
  aux=mks_curr->p_ks->KS;
  if ((mks_curr->state==PAUSED) && (mks_curr->p_ks->KS->deadline <= 0))
    { fprintf(stderr,"\t(control.c:check_deadline) cargarse al %s\n", mks_curr->name);

    Level_Regenerate(mks_curr->p_ks->KS->identifier, mks_curr->p_ks->KS->kiwi_id);

#ifdef ARTIS_CONF_METAREASONING
    /*** ThreeC 20030822 ***/
    LevelsKilled++;
#endif

    return 0;
    }
  else if ((mks_curr->state==NOT_EXECUTED) && 
	   (mks_curr->p_ks->KS->deadline <= mks_curr->p_ks->KS->duration))
    { fprintf(stderr,"\t(control.c:check_deadline) cargarse al %s\n", mks_curr->name);

#ifdef ARTIS_CONF_METAREASONING
    /*** ThreeC 20030822 ***/
    LevelsKilled++;
#endif
    return 0;

    }
  else
    return 1;
  /*** 0=This ks has never been executed before and its estimated duration was
   *   greater than its remainding deadline ( or available execution time ) or
   *   this ks was paused and its deadline has expired.
   *   1=This ks still has enough execution time, by now.
   ***/
}


/* ============================================================================ */
void update_rsa_deadlines(){
/* ============================================================================ */
  n_mks *auxMks,*tmpMks;
  
  auxMks=rsa.last;
  while (auxMks!=(n_mks *)NULL) {
    tmpMks=auxMks->prev;     
    if (check_deadline(auxMks->MKS) ==0){ /*en protov no eliminaba*/
      if (auxMks->next==(n_mks *) NULL)
	rsa.last=tmpMks;
      else ((n_mks *)auxMks->next)->prev = tmpMks;
      if (tmpMks==(n_mks *) NULL)
	rsa.first=auxMks->next;
      else tmpMks->next=auxMks->next;
     rsa.number_mks--;

      free(auxMks);
    }
    auxMks=tmpMks;
  }
}


/* ============================================================================ */
void init_periodic_list(){
/* ============================================================================ */ 
  static void CreateStructures();
  void print_in_agent();
  void print_rules();
  
  /* With the data read from the file, 	*/
  /* create all the structures.		*/
  CreateStructures();
  fprintf(stderr,"\t(control.c:init_periodic_list) i tenjo hestructura\n");
  print_in_agent(l_in_agent);
	
}

/* ============================================================================ */
void calculate_mks_duration(mks *mks_curr, systime_t *t1){
/* ============================================================================ */
  n_ks *ks_aux;
  systime_t duracion = 0;

  ks_aux=mks_curr->p_ks;
  while(ks_aux!=NULL){
    duracion+=ks_aux->KS->duration;
    ks_aux = ks_aux->next;
  }
  *t1= duracion;
}

  
/* ============================================================================ */
void init_sporadic_list(){
/* ============================================================================ */
/*  char vector[10],nombre[20],tipo[20];
  int ident,n_MKS,importancia,n_KS;
  systime_t duracion;
  float gain;
  ks *n_ks;
  mks *nu_mks;
  in_agent *in_agent_curr;
  FILE *fd;

  fprintf(stderr,"(control.c)Voy a analizar el fichero aperiod.txt\n");
  fd=fopen("./app/data/aperiod.txt","r");
			
  fscanf(fd,"%s",vector);
  while(strcmp(vector,"FIN")){
    if (!strcmp(vector,"ACTIVITY")){	  
      fscanf(fd,"%s%d%d",nombre,&ident,&n_agentes);
      in_agent_curr=create_in_agent(nombre, ident, n_MKS, n_MKS);
      l_in_agent=insert_in_agent(in_agent_curr,l_in_agent);
    }
    else 
      if (!strcmp(vector,"AGENTE")){
	fscanf(fd,"%s%d%s%d%d",nombre,&ident,tipo,&importancia,&n_KS);
	nu_agent=create_agent(nombre,ident,tipo,importancia,n_KS,in_agent_curr);

	if (!strcmp(tipo, "NonCritical"))
	  in_agent_curr->no_critical_mks--;

	insert_mks(&non_periodic_mks_l,nu_mks);
      }
      else if (!strcmp(vector,"KS")){
	fscanf(fd,"%s%d%lld%f",nombre,&ident,&duracion,&gain);
	n_ks=create_ks(nombre,ident,duracion,gain);
	insert_ks(n_ks,nu_mks);
      }
    fscanf(fd,"%s",vector);
  }
  fclose(fd);

*/
}


/* ============================================================================ */
void set_deadline(char *ks_name, int kiwi_id, int identInAgent, 
		  systime_t deadlineInAgent,
		  systime_t durationInAgent, systime_t read_time){
/* ============================================================================ */
  ks *aux_ks;
  
  aux_ks=(ks *)malloc(sizeof(ks));
  aux_ks->ks_id =0;    
  /*** It'll be filled inmediately before it's gonna be 
   *   executed.
   ***/
  aux_ks->name=(char *)malloc(strlen(ks_name)+1);
  strcpy(aux_ks->name, ks_name);
  
  /*** ThreeC 20030910 ***/
  aux_ks->kiwi_id = kiwi_id;

  aux_ks->weight  =0;
  aux_ks->identifier =identInAgent;
  aux_ks->duration=durationInAgent;
  aux_ks->deadline=deadlineInAgent;
  aux_ks->last_check=read_time;
  /*** So we can update its deadline 
   *   correctly.
   ***/

#ifdef LUIS_DEBUG
  fprintf(stderr,"\t(control.c:set_dedline) name_ks: %s deadline:%ld duration:%ld\n", ks_name,deadlineInAgent,durationInAgent);
#endif

  in_l_ks(aux_ks);           /*** OK, ks ready to insert. ***/
}


/* ============================================================================ */
void trigger_mks(char *mks_name, int type, int importance, in_agent *mother){
/* ============================================================================ */

  mks *aux_mks;

#ifdef CONTROL_DEBUG 
  int id ;
  systime_t passed_time, a1, a2 ;
#endif
  
  aux_mks=(mks *)malloc(sizeof(mks));
  aux_mks->name=(char *)malloc(strlen(mks_name)+1);
  strcpy(aux_mks->name, mks_name);
  aux_mks->type = type;
  aux_mks->solution_level=0;
  aux_mks->p_ks=l_ks.first;
  /*** Current agent ks is the first of 
   *   newly created 'l_ks'.
   ***/
  aux_mks->state        =NOT_EXECUTED;
  aux_mks->importance   =importance;
  aux_mks->mother       =mother;
  aux_mks->number_ks    =l_ks.ks_n;
  aux_mks->quality      =0;
  aux_mks->empty	  =1;
 
  if (heuristic==DBS){
    aux_mks->mks_deliber = 0;
    calculate_mks_duration(aux_mks, &(aux_mks->mks1_duration));
    aux_mks->mks1_deliber=0;
    aux_mks->real_gain=0;
    aux_mks->identifier=mks_n;
    mks_n++;
  }
  
  init_l_ks();
  /*** We must initialise it so we can use it when 
   *   a new agent appears.
   ***/
  insert_mks(&tsa, aux_mks);

}


/* ============================================================================ */
int trigger( systime_t lEndTriggerTime){
/* ============================================================================ */

  int n_events=0;

  Event_T regAux;

#ifdef ARTIS_CONF_METAREASONING
/*** ThreeC 20030213 ***/
  int nEventTreated = 0;   /*** It counts how many events have 
			    *     been treated in this trigger() 
			    *     execution 
			    ***/
  int nManageEvent;        /*** It gets the output of the 
			    *     ManageEvent function
			    *     --> Debug usage
			    ***/
#endif

  int nExtractOutput;
  systime_t now;

#ifdef CALCULATING_MANDATORY
  /*** It is only used to calculate the MANDATORY times ***/
  systime_t t_beginning, t_end;
#endif

  now=systime_gettime();
#ifdef CALCULATING_MANDATORY
  t_beginning = now;
#endif

#ifdef LUIS_DEBUG
  fprintf(stderr,"\t(control.c:trigger) Triggering...\n");
#endif
  /*** Received events could pass through some filters in order to maximize
   *   quality, reactivity, etc... (following implementations)
   ***/
  
  /*** We must consider a maximum number of events, since Control needs to 
   *   estimate the trigger worst case execution time in order to 2ond. level
   *   planification can find out the remainding execution time ( slack ), in
   *   this way we can give a maximum execution time to a ks.
   ***/

#ifdef ARTIS_CONF_METAREASONING
  if (lEndTriggerTime < now + MANDATORY_T) {
#ifdef DS_DEBUG
    fprintf(stderr, "\t(control.c:trigger) NO TIME: now: %lld, lEndTriggerTime: %lld, MANDATORY_T: %lld\n",
	    now, lEndTriggerTime, MANDATORY_T);
#endif
      return (TRIGGER_NOTIME);
  }
#endif

  /*** Let's see what's coming in. ***/
  
  merge_list(&rsa,&sub_rsa);/* we add the paused and too_time MKS's */

#ifdef DS_DEBUG 
  if ( Events_bIsEmpty() )
    fprintf(stderr, "\t(is.c) No hay na\n");
  else if (lEndTriggerTime <=  now)
    fprintf(stderr, "\t(is.c)::trigger(): Azul y negro, pues el tiempo es: %lld y now es: %lld\n", 
	    lEndTriggerTime, now);
  else 
    fprintf(stderr, "\t(is.c)::trigger() Let's go to the Real Event McCoy\n");

  /***
   *printActiveAttFocus();
  ***/
#endif

  /*** KDM Event Management ***/
  /*** The function treats another event if the event store is not empty, and
   *          there is enough time for the trigger phase, and 
   *          it has not treated the maximum events allowed in one trigger()
   *          execution
   ***/

#ifdef ARTIS_CONF_METAREASONING
/*** ThreeC 20030213 ***/
  nExtractOutput = 0;  
  regAux.iType = 0;
  /// MANDATORY_T should be the wcet of the trigger for one event

  /*** 
   *  While there are events, and there are time enough, and the 
   *    event rate has not been fulfilled and there is no error in 
   *    extracting from the event store and the type of the event is 
   *    not -1
   ***/
  while ( !Events_bIsEmpty() && 
	  ((EventRate == 1) || (lEndTriggerTime >= now + MANDATORY_T)) && 
	  (nEventTreated < EventRate) &&
	  !nExtractOutput && (regAux.iType != -1)) {
#else
  while ( !Events_bIsEmpty() && (lEndTriggerTime > now)) {
#endif

#ifdef DS_DEBUG 
    fprintf(stderr, "\t(is.c) Before Extracting an Event (there are %d events left).\n", *piEvents);
    fprintf(stderr, "\t         EndTriggerTime: %lld, Now: %lld\n", 
	    lEndTriggerTime, now);
#endif

    if (!(nExtractOutput = Events_Extract_Event( &regAux ))) {  
      /*** There is no problem in extracting an event ***/

#ifdef DS_DEBUG 
      fprintf(stderr, "\t(is.c) And now, I have an Event of type %d, TimeStamp:%llu \n", 
	      regAux.iType, regAux.lTimeStamp);
#endif

#ifdef ARTIS_CONF_METAREASONING
      nEventTreated++;
      
      nManageEvent = ManageEvent( regAux, 1 );
    
#ifdef DS_DEBUG
      if (nManageEvent <= 0 )
	fprintf( stderr, "\t(is.c::Trigger()) There was no Meta-Rule to this event? Error was %d \n", nManageEvent );
#endif
#endif    
    
      switch ((int) regAux.iType) {
      case EXIT_APP:  /*** Take care, it is defined in include/rt_common.h, and its value is 1 !!!!! ***/

#ifdef DS_EXEC_DEBUG
	fprintf(stderr, "(is.c) Received EXIT_APP: Exiting application...\n");
	/* Close the RTFIFO and exit */
#endif

#ifdef ARTIS_TRACE_EVENTS
	trace_event( AA_QUALITY, 1, (long long) system_quality, 0 );
#endif

	Events_Free();
      
	return (TRIGGER_EXIT);

#ifdef ARTIS_CONF_METAREASONING
/*** ThreeC 20030213 ***/
      case MODIFICATION:

#ifdef DS_EXEC_DEBUG 
	fprintf(stderr, "\t(is.c) event: MODIFICATION importance: %d timeStamp: %lld ", regAux.iImp, regAux.lTimeStamp );
#endif

	switch (regAux.iDataType) {
	case KDM_INT:

#ifdef DS_EXEC_DEBUG
	  fprintf(stderr, "Old value: %d   New Value: %d INT\n", regAux.Old_Value.KDM_v.i, regAux.New_Value.KDM_v.i );
#endif

	  break;

	  /***
	case KDM_SHORT_INT:

#ifdef DS_EXEC_DEBUG
	  fprintf(stderr, "Old value: %d   New Value: %d SHORT_INT\n", regAux.Old_Value.si, regAux.New_Value.si );
#endif

	  break;
	  

	case KDM_LONG_INT:
	    
#ifdef DS_EXEC_DEBUG
	  fprintf(stderr, "Old value: %ld   New Value: %ld LONG_INT\n", regAux.Old_Value.li, regAux.New_Value.li );
#endif

	  break;
	case KDM_UNSIGNED_INT:

#ifdef DS_EXEC_DEBUG
	  fprintf(stderr, "Old value: %d   New Value: %d UNSIGNED_INT\n", regAux.Old_Value.ui, regAux.New_Value.ui );
#endif

	  break;
	  ***/

	case KDM_FLOAT:

#ifdef DS_EXEC_DEBUG
	  fprintf(stderr, "Old value: %f   New Value: %f FLOAT\n", regAux.Old_Value.KDM_v.f, regAux.New_Value.KDM_v.f );
#endif

	  break;

	  /***	  
	case KDM_DOUBLE:
	    
#ifdef DS_EXEC_DEBUG
	  fprintf(stderr, "Old value: %lf   New Value: %lf DOUBLE\n", regAux.Old_Value.d, regAux.New_Value.d );
#endif

	  break;
	case KDM_UNSIGNED_CHAR:

#ifdef DS_EXEC_DEBUG
	  fprintf(stderr, "Old value: %c   New Value: %c UNSIGNED_CHAR\n", regAux.Old_Value.byte, regAux.New_Value.byte );
#endif

	  break;
	  ***/

	}
	break;
#endif

      case CONSIDER_PERIODIC_AGENT:

#ifdef DS_EXEC_DEBUG
	fprintf(stderr, "\t(is.c) event: CONSIDER_PERIODIC_AGENT importance: %d timeStamp: %lld\n", 
		regAux.iImp, regAux.lTimeStamp );
#endif

	/// Old CONTROL_KS : beginning

#ifdef ARTIS_CONF_AUTO_METAREASONING
	/*** ThreeC 20010718 ***/
	/*** METARULE      ***/
	if (( LevelsInterrupted > LevelsCompleted ) ||
	    ( LevelsKilled > LevelsCompleted ))
	  ChangeSLS ();
	/*** 
	 *   This should be a change between deliberative to reactive SLS policy
	 ***/
#endif   


	/***
	n_mks2 *L_mks;
	n_regla *lista_sin_act;
	KS2 *l_ks;
	n_regla *lista_regla;
	int EvType;
	systime_t EvDeadline;




	/// Beginning of old rules without condition
	lista_sin_act=v_reglas[0];

	while (lista_sin_act !=NULL){
	  L_mks=lista_sin_act->NMKS2;
	  activar(L_mks);
	  lista_sin_act=lista_sin_act->sig;
	}
	/// Ending of old rules without condition




	EvType = TranslateEventIndex(CONSIDER_PERIODIC_AGENT);
	lista_regla=v_reglas[EvType];


	***/


	/*** ThreeC 20030812 ***/
	/*** All the time values received from the FLS are in nanoseconds, 
	 ***/
	///	if (regAux.New_Value.lTime > 0) 
	///  EvDeadline = regAux.New_Value.lTime; /// / 1000.0; 
	  /* Intervalo = deadline relativo */
	  ///else EvDeadline = regAux.New_Value.lTime;


	
	///if (lista_regla!=NULL){
	///  kill_old_agent(&tsa, regAux.Old_Value.KDM_v.i /* Agent identifier */);
	///  kill_old_agent(&rsa, regAux.Old_Value.KDM_v.i /* Agent identifier */);

	///  if (EvDeadline != (long long) -1) 
	///    activar_periodico(lista_regla->NMKS2, regAux.Old_Value.KDM_v.i,
	///			      EvDeadline, systime_gettime());
	///	}


	/***
#ifdef DS_DEBUG
	fprintf(stderr, "\t\t----EVENT---- Type %d, deadline: %lld %lld, identifier: %d\n", 
		EvType, regAux.New_Value.lTime, regAux.Old_Value.KDM_v.i);
#endif
	***/


	break; 
      default: 

#ifdef DS_EXEC_DEBUG
	fprintf(stderr,"\n\n\t\tERROR UNKNOWN EVENT TYPE\n");
#endif

      }
      
    }
    else {

#ifdef DS_DEBUG
      fprintf(stderr, 
	      "\t(em.c) I wasn't able to Extract and Event, and I have an error with number %d, but there are %d events\n", 
	      nExtractOutput, *piEvents);

      Events_ShowAll();  /*** What events? ***/ 
#endif

      return( TRIGGER_ERROR );

    }

#ifdef ARTIS_CONF_METAREASONING
    if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
       EventRate = * (int *) KDM_Get_Value( pMetaClass->eventRate );
       /*** EventRate may have changed for a Meta-Rule Action ***/
#ifdef ARTIS_TRACE_EVENTS
	mutex_unlock( MUTEX_EM );
#endif
    }
#endif

    now=systime_gettime();

  }

#ifdef LUIS_DEBUG
  fprintf(stderr,"\t(control.c:trigger) tsa agents: %d\n", tsa.number_mks);
#endif

#ifdef ARTIS_CONF_METAREASONING
#ifdef CALCULATING_MANDATORY
  t_end=systime_gettime();
  if (EventRate == 1)
    MANDATORY_T = t_end - t_beginning;
  fprintf( stderr, "MANDATORY_T should be: %lld with EventRate %d\n", 
	   MANDATORY_T, EventRate);
#endif
#endif

  if (lEndTriggerTime <= now + MANDATORY_T) {
#ifdef DS_DEBUG
      fprintf(stderr, "\t(control.c:trigger) NO TIME for the Event Rate\n");
#endif
      return (TRIGGER_NOTIME_ER);
  }
  else if (Events_bIsEmpty()) {
#ifdef DS_DEBUG
    fprintf(stderr, "\t(control.c:trigger) NO EVENTS\n");
#endif
    return (TRIGGER_EMPTY);
  }
  else {
#ifdef DS_DEBUG
    fprintf(stderr, "\t(control.c:trigger) OK\n");
#endif
    return (TRIGGER_OK);  /// The Event Rate has been satisfied
  }
  }


/* ============================================================================ */
void condition_testing(){
/* ============================================================================ */

  mks *mks_curr;

#ifdef ARTIS_CONF_METAREASONING
  int old_heuristic = heuristic;
#endif

#ifdef LUIS_DEBUG     
  fprintf(stderr,"\t(control.c) Condition Testing...\n");
#endif

#ifdef ARTIS_CONF_METAREASONING
  if (!mutex_system_lock( MUTEX_EM, (systime_t) WCET_KDM_ACCESS )) {
      heuristic = * (int *) KDM_Get_Value( pMetaClass->slsPolicy );

#ifdef ARTIS_TRACE_EVENTS
      mutex_unlock( MUTEX_EM );
#endif

      if ((heuristic != old_heuristic) && (heuristic == DBS))
         free_l_mks(&tsa);
  }

#endif

#ifdef LUIS_DEBUG
  fprintf(stderr,"\t(control.c:condition_testing) Before checking\n");
#endif
 /*** There are no condition tests but only check_deadline(), perhaps in 
   *   following implementations.
   ***/
  
  while ((mks_curr = (mks *) out_mks(&tsa)) != (mks *) NULL) {
    if (check_deadline(mks_curr)) insert_mks(&ctsa, mks_curr);
  }

#ifdef LUIS_DEBUG
  fprintf(stderr,"\t(control.c:condition_testing) ctsa mks: %d\n", 
	  ctsa.number_mks);
#endif
}


/* ========================================================================= */
void interpret_end (mks *SchedMKS, int TheValue){
/* ========================================================================= */

  if (TheValue == LEVEL_ENDED){
    if (update_mks(SchedMKS))
      in_ord(&rsa,SchedMKS);

#ifdef ARTIS_CONF_METAREASONING
/*** ThreeC 20010718 ***/
    LevelsCompleted ++;
#endif

  }
  else if (TheValue == LEVEL_TIMEOUT){
    SchedMKS->p_ks->KS->ks_deliber = 0;
    update_paused(SchedMKS,SchedMKS->identifier);
    in_last(&sub_rsa, SchedMKS);

#ifdef ARTIS_CONF_METAREASONING
/*** ThreeC 20010718 ***/
    LevelsInterrupted ++;
#endif

  }

  /******************/
  fprintf( stderr, "(control.c:interpret_end: System Quality: %f\n", 
	   system_quality, 0 );

}
