/**************************************************************************
 *
 *                   sls_policies.c
 *                   Second Level Scheduler Policies
 *                   20030906
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>

/**************************************************************************/
#include "../include/dserver/sls_policies.h"
/**************************************************************************/
#include "../include/dserver/agenda.h"
#include "../include/dserver/em.h"
#define DEBUG

mks  *mks_selected;


/* ============================================================================ */
mks *schedule_by_importance(){
/* ============================================================================ */
  int best_position, position ;
  mks *mksactual ;
  mks *mksbest ;
  
  
  best_position = 1 ;
  if (( mksbest = first( &rsa ) ) == NULL )
    /*** There isn't any more agents on the rsa ***/
    return( mksbest );
  
  while ( !check_deadline( mksbest )) {

#ifdef DEBUG
    fprintf(stderr, "(sls_policies.c) Deadline fullfilled\n");
#endif

    out_current( &rsa );

#ifdef DEBUG
    fprintf(stderr, "(sls_policies.c) Out current\n");
#endif

    free_mks( mksbest );

#ifdef DEBUG
    fprintf(stderr, "(sls_policies.c) After free\n");
#endif

    if (( mksbest = next( &rsa )) == NULL ){
      /*** There isn't any more agents on the rsa ***/

#ifdef DEBUG
      fprintf(stderr, "(sls_policies.c) Before return\n");
#endif

      return( mksbest );
    }
  }
  
  /*** If the function arrives till here, in best will be the first agent
   *    of the rsa ( with a non outdated deadline ), and it will have
   *    the biggest importance of the rsa ( because the rsa will be ordered by
   *    by importance ). Now, we'll check if there's another/s agents
   *    whith the same importance in the rsa
   ***/
  for ( position = best_position + 1 ; 
        (mksactual = next(&rsa))!=(mks *)NULL ;
	position++ ){
    if (!check_deadline(mksactual)){
      /*** if it's outdated ---> we pull it out of the agenda ***/
      out_current( &rsa ); 
      free_mks( mksactual );
    }
    else 
      if ( mksbest->importance > mksactual->importance )
	break ; /*** We'll go out of the for ***/
      else /***  mksbest->importance == mksactual->importance ***/
	if ( mksactual->quality < mksbest->quality ){
	  best_position = position ;
	  mksbest = mksactual ;
	}
	else 
	  if ( mksactual->quality == mksbest->quality ){
	    if ( mksactual->p_ks->KS->deadline < mksbest->p_ks->KS->deadline ){
	      best_position = position ;
	      mksbest = mksactual ;
	    }
	    else if ( mksactual->p_ks->KS->deadline ==
		      mksbest->p_ks->KS->deadline )
	      if ( mksactual->p_ks->KS->duration < 
		   mksbest->p_ks->KS->duration ){
		best_position = position ;
		mksbest = mksactual ;
	      }
	      }
  }
  mksbest = element_number( &rsa, best_position );  
#ifdef LUIS_DEBUG
  if (mksbest!=NULL) {
    fprintf(stderr,
	    "\t(control.c:schedule_by_importance) a ejecutar id: %s\n", 
	    mksbest->name);
    fprintf(stderr,"\t(control.c:schedule_by_importance) ejecutar n_mks_actual id: %s, best position: %d\n",
	   n_mks_actual->MKS->name, best_position);
  }
#endif

  out_current( &rsa );
  return( mksbest );
}
 


/* ============================================================================ */
mks *schedule_by_deadline(){
/* ============================================================================ */
  int best_position, position ;
  mks *actual ;
  mks *best ;

  best_position = 1 ;
  if (( best = first( &rsa ) ) == NULL )
    /*** There isn't any more agents on the rsa ***/
    return( best );
  while ( !check_deadline( best )) {

#ifdef DEBUG
    fprintf(stderr, "(sls_policies.c) Deadline fullfilled EDF\n");
#endif

    out_current( &rsa );

#ifdef DEBUG
    fprintf(stderr, "(sls_policies.c) Out current (EDF)\n");
#endif

    free_mks( best );
    if (( best = next( &rsa )) == NULL ){
      /*** There isn't any more agents on the rsa ***/
      return( best );
    }
  }
  
  /*** If the function arrives till here, in best will be the first agent
   *    of the rsa ( with a non outdated deadline ), and it will have
   *    the shortest deadline of the rsa ( because the rsa will be ordered by
   *    by deadline ). Now, we'll check if there's another/s agents
   *    whith the same deadline in the rsa
   ***/
  for ( position = best_position + 1 ; 
	( actual = next( &rsa ) ) != NULL ;
	position++ )
    if ( !check_deadline( actual ) ){
      /*** if it's outdated ---> we pull it out of the agenda ***/
      /*** IS STILL NECESSARY THIS LAST CHECK ? ***/
      out_current( &rsa ); 
      free_mks( actual );
    }
    else 
      if ( best->p_ks->KS->deadline < actual->p_ks->KS->deadline )
	break ; /*** We'll go out of the for ***/
      else 
	/*** best->p_ks->KS->deadline == actual->p_ks->KS->deadline ***/
	if ( best->importance < actual->importance ){
	  best_position = position ;
	  best = actual ;
	}
	else 
	  if ( best->importance == actual->importance ){
	    if ( actual->quality < best->quality ){
	      best_position = position ;
	      best = actual ;
	    }
	    else 
	      if ( actual->quality == best->quality )
		if ( actual->p_ks->KS->duration < 
		     best->p_ks->KS->duration ){
		  best_position = position ;
		  best = actual ;
		}
		}
  best = element_number( &rsa, best_position );  
  /***Only to update the "n_mks_actual" global variable, to do out_current ***/
  
#ifdef LUIS_DEBUG
  if (best!=NULL) {
    fprintf(stderr,"\t(control.c:schedule_by_deadline) a ejecutar id: %s\n", best->name);
    fprintf(stderr,"\t(control.c:schedule_by_deadline) ejecutar n_mks_actual id: %s, best position: %d\n",
	   n_mks_actual->MKS->name, best_position);
  }
#endif
  
  out_current( &rsa );
  return( best );
}


/* ============================================================================ */
void time_to_assign(systime_t *it, systime_t *lt, systime_t *result){
/* ============================================================================ */

  systime_t time_a, time_b, time_c, x;
  n_ks   *aux_ks;
  mks *aux_mks;

  /*** This function returns the minor value between:
   *   time_a=> the difference between "actual" time and the initial time.
   *   time_b=> the difference between "actual" time and the previous agent deadline. 
   *   time_c=> the minor time assignable provided by the agents which have a 
   *   deadline greater than "actual" time.
   ***/
  
  time_b = time_c =(systime_t)10000000;	
  time_a=(*lt > *it)? *lt - *it : 0;
  aux_mks = (mks *)last( &rsa );
  while( (aux_mks!=NULL) && (aux_mks->p_ks->KS->deadline >= *lt) )
    aux_mks = (mks *)prev( &rsa );
  if (aux_mks!=NULL) 
    time_b =(*lt > aux_mks->p_ks->KS->deadline) ? *lt -aux_mks->p_ks->KS->deadline :0;
  aux_mks = next( &rsa );

  while( aux_mks!=NULL ) {
    if (aux_mks->mks1_duration != aux_mks->mks_deliber) {
      aux_ks= aux_mks->p_ks;
      while ((aux_ks!=NULL)
	     &&(aux_ks->KS->duration==aux_ks->KS->ks_deliber))
	aux_ks= aux_ks->next;
      if (aux_ks!=NULL) {
	  x=aux_ks->KS->duration - aux_ks->KS->ks_deliber;
	time_c = time_c < x ? time_c : x;
      }
    }
    aux_mks = next( &rsa );
  } /*** End of while ***/
  x = time_a < time_b ? time_a : time_b;
#ifdef LUIS_DEBUG
  fprintf(stderr,"\t(control.c:time_to_assign) time_a : %lld, time_b : %lld, time_c : %lld\n",
	  time_a,time_b,time_c);
#endif
  if (x < time_c) 
    *result = x; 
  else *result = time_c;
}

/* ============================================================================ */
void algorithm_selection(systime_t t_deliber, systime_t *lt ){
/* ============================================================================ */

  n_ks   *aux_ks, *best_ks;
  mks *aux_mks, *best_mks;
  float best_gain,aux_gain;
  
  /*** This function assigns the time computed by the time_to_assign function
   *   (t_deliber) to the agent actual ks which provides the biggest outfit. 
   *   In case two or more agent provide the same outfit we select the most 
   *   important agent and if all they have the same importance, we select the agent
   *   that has less quality.
   ***/

  best_gain=aux_gain= 0.0;
  best_mks = (mks *) NULL;
  
  /*** We look for the first candidate agent. ***/
  aux_mks = first( &rsa );
  
  while( (aux_mks!=NULL) && (aux_mks->p_ks->KS->deadline < *lt) )
    aux_mks = next( &rsa );
  
  if ( aux_mks!=NULL ){
    while( aux_mks!=NULL ) {
      /*** We see if the actual agent has not the maximum deliberation time. ***/
      if (aux_mks->mks1_duration != aux_mks->mks_deliber) {
	aux_ks = aux_mks->p_ks;
 	while ((aux_ks!=NULL)
	       &&(aux_ks->KS->duration==aux_ks->KS->ks_deliber))
	  aux_ks = aux_ks->next;
	if (aux_ks->KS->duration==t_deliber)
	  aux_gain = (float) aux_ks->KS->gain;
	else aux_gain = -1.0;
	
 	if (aux_gain > best_gain) {
	  best_gain = aux_gain;
	  best_mks = aux_mks;
	  best_ks = aux_ks;
	}
	else 
	  if (aux_gain == best_gain){
	    if (aux_mks->importance > best_mks->importance){
	      best_mks = aux_mks;
	      best_ks = aux_ks;
	    }
	    else 
	      if (aux_mks->importance==best_mks->importance) 
		if (aux_mks->quality < best_mks->quality){
		  best_mks = aux_mks;
		  best_ks = aux_ks;
		}
	  }
      } /*** End of if ***/
      aux_mks = next( &rsa );
    } /*** End of while(aux_mks!=NULL) ***/
    if (best_mks != NULL) {
#ifdef LUIS_DEBUG
      fprintf(stderr,"\t(control.c:algorithm_selection) %s, deliber: %lld\n",best_mks->name,
	      t_deliber);
#endif
      best_mks->mks1_deliber += t_deliber;
      best_mks->mks_deliber+=t_deliber;
      best_ks->KS->ks_deliber+=t_deliber;
      
      /*** We insert a new unit in the allocation list ***/
      in_sched( best_mks->identifier, t_deliber );
    }
  }
  /*** End of if ***/
  /*** Despite not to assign the time, we should insert a hole unit in the 
   *   allocation list so as to reset in a correct way the agent deliberation
   *   time.
   ***/

  if ( best_mks == NULL )  {
    in_sched( NONES, t_deliber );
    in_hole(*lt, t_deliber);
  }
  *lt=*lt - t_deliber;
  
}

/* ============================================================================ */
ks *look_best(l_mks *lmks, systime_t time_rest){ 
/* ============================================================================ */

  n_mks *aux_mks;
  n_ks *aux_ks, *best_ks;
  int best_gain=0;

  aux_mks = lmks->first;
  best_ks = NULL;
  while (aux_mks!=NULL){
    if ((aux_mks->MKS->empty)
	&&(aux_mks->MKS->mks1_deliber!=aux_mks->MKS->mks1_duration)){
      aux_ks = aux_mks->MKS->p_ks; 
      while((aux_ks!=NULL)
	    &&(aux_ks->KS->ks_deliber==aux_ks->KS->duration))
	aux_ks=aux_ks->next;
      if(aux_ks!=NULL){
       	if(aux_ks->KS->duration <= time_rest){
	  if (aux_ks->KS->gain > best_gain){
	    best_ks = aux_ks;
	    best_gain = aux_ks->KS->gain;
	    mks_selected = aux_mks->MKS;
	  }
	}
	else aux_mks->MKS->empty = 0;
	}
    }
    aux_mks = aux_mks->next;
  }
  if (best_ks!=NULL) return best_ks->KS;
  else return (ks *)NULL;
}

/* ============================================================================ */
hole *best_hole(systime_t dead, systime_t time_necessary, int *status){ 
/* ============================================================================ */

  hole *aux,*prev;
  systime_t t1, t2, t3;

  aux = l_hole;
  prev = (hole *) NULL;
  while (aux!=NULL){
    t1= dead - aux->LI;
    t2 = aux->LS - aux->LI;
    t1 += aux->AC;
    t3 = aux->AC + t2;
    if (dead <= aux->LI)  
      if (aux->AC >= time_necessary){
	*status = 1; 
	return prev;
      }
      else return (hole *)NULL;
    else
      if (dead < aux->LS) 
	if (t1 >= time_necessary){
	  *status = 2;
	  return aux;
	}
	else return (hole *)NULL;
      else
	if (dead == aux->LS){
	  if (t2 >= time_necessary){
	    *status = 1;
	    return aux;
	  }
	  else return (hole *)NULL;
	  }
	else 
	  if ((dead >aux->LS)&&(aux->next == NULL)){
	    if (t3 >= time_necessary){
	      *status = 1;
	      return aux; 
	    }
	    else return (hole *)NULL;
	    }
    prev = aux;
    aux = aux->next;
  }
  return (hole *)NULL;
}


/* ============================================================================ */
void AC_update(systime_t t_desc, hole *f_hole){ 
/* ============================================================================ */

  f_hole=f_hole->next;
  while(f_hole != NULL){
    f_hole->AC -= t_desc;
    f_hole = f_hole->next;
  }
}

/* ============================================================================ */
void AC_calculate(systime_t *cont){ 
/* ============================================================================ */

  hole *h_aux;
  systime_t t1;

  *cont = 0;
  h_aux = l_hole;
  while(h_aux != NULL){
    h_aux->AC = *cont;
    h_aux->LS-= h_aux->LI;
    *cont += t1;
    h_aux = h_aux->next;
  }
}


/* ============================================================================ */
void optimize_hole(){ 
/* ============================================================================ */

 
  int status;
  systime_t  temp_hole, time_rest, aux_time, t1;
  ks *b_ks;
  hole *b_hole, *n_hole, *aux, *aux2;
  
  AC_calculate(&time_rest);
  while (time_rest > 0){
    aux = l_hole;
    b_ks = look_best( &rsa , time_rest);
    if (b_ks == NULL) break;
    b_hole = best_hole( mks_selected->p_ks->KS->deadline, 
			b_ks->duration, &status); 
    if (b_hole == NULL) mks_selected->empty = 0;
    else 
      if (status==1){
	b_ks->ks_deliber = temp_hole = b_ks->duration;
	mks_selected->mks_deliber += temp_hole;
	mks_selected->mks1_deliber += temp_hole;
	AC_update(temp_hole, b_hole);
	time_rest = (time_rest > temp_hole)? time_rest-temp_hole :0;
	while( temp_hole > 0){
	  t1 = b_hole->LS - b_hole->LI;
	  if (t1 <= temp_hole){
	    aux = b_hole;
	    aux2=b_hole->prev;
	    aux2->next = b_hole->next;
	    free(aux);
	    temp_hole = (temp_hole > t1)? temp_hole - t1 :0;
	    if (temp_hole > 0) b_hole = b_hole->prev;
	  }
	  else {
	    b_hole->LS-=temp_hole;
	    temp_hole = 0;
	  }
	}
      }
      else {
        b_ks->ks_deliber = temp_hole = b_ks->duration;
	  mks_selected->mks_deliber += temp_hole;
	  mks_selected->mks1_deliber += temp_hole;
	  AC_update(temp_hole, b_hole);
	  aux_time = b_ks->deadline - temp_hole;
	  if (aux_time > b_hole->LI){
	   n_hole = (hole *) malloc(sizeof(hole));
	   n_hole->LI = b_hole->LI;
	   n_hole->LS = aux_time;
	   n_hole->AC = b_hole->AC;
	   b_hole->LI = b_ks->deadline;
	   t1 = n_hole->LS - n_hole->LI;
	   b_hole->AC += t1;
	}
	else 
	  if (aux_time == b_hole->LI)
	    b_hole->LI = b_ks->deadline;
	  else {
	    b_hole->LI = b_ks->deadline;
	    t1 = b_ks->deadline - b_hole->LI;
	    temp_hole=b_ks->duration - t1; 
	    b_hole = b_hole->prev;
	    while( temp_hole > 0){
	      t1 = b_hole->LS - b_hole->LI;
	      if (t1 <= temp_hole){
		aux = b_hole;
		aux2 = b_hole->prev;
		aux2->next = b_hole -> next;
		free(aux);
		temp_hole =(temp_hole >t1)? temp_hole-t1 :0;
		if (temp_hole > 0)
		  b_hole = b_hole -> prev;
	      }
	      else {
			b_hole->LS -=temp_hole;
			temp_hole = 0;
	      }
	    }
	  }
      }
  }
  
}
 
/* ============================================================================ */
void Dean_Boddy_Strategy() {
/* ============================================================================ */

  systime_t last_time, initial_time, t_aux1;
  mks *mks_curr;
  
  fprintf(stderr, "Rating by DBS\n");

  if (( mks_curr = first( &rsa ) ) != NULL ){
    fprintf(stderr,"\t(control.c:Dean_Body....) algo que ratear \n");
    while ( mks_curr != NULL ) {
      if ( !check_deadline( mks_curr ) ) {
    	out_current( &rsa );
    	free_mks( mks_curr );
      }
      mks_curr = next( &rsa );
    }
    if ( rsa.number_mks ) {
      if ( Is_init_deadline == -1 ) {     /*** Control has a new slack. ***/
	free_l_allocat( &rsa );
	init_deadline = rsa.last->MKS->p_ks->KS->deadline;
	Is_init_deadline = 0;
      }       
      last_time = init_deadline < time_left ? init_deadline : time_left;
      /*** We start the D&B algorithm from the 
       *   init_deadline. Remember that this algorithm
       *   works backward.
       ***/
      initial_time = 0;
      free_hole();
      if ( init_deadline != 0 ) 
	do { 
	  time_to_assign(&initial_time, &last_time, &t_aux1);  
	  algorithm_selection( t_aux1, &last_time );
	}while( last_time > initial_time ); 
      optimize_hole();
    }
    else fprintf(stderr,"\t(control.c:Dean_Boddy) No agentes por vencimiento");
  }
}

/* ============================================================================ */
mks *schedule_by_DBS(){
/* ============================================================================ */


  mks *mks_curr;
  n_mks *mks_aux;

  fprintf(stderr, "(schedule_by_DBS)\n");

  if (( mks_curr = first( &rsa ) ) == NULL )
    /*** There isn't any more agents on the rsa ***/
    return( mks_curr );
  while ( mks_curr != NULL ) {
    if ( !check_deadline( mks_curr ) ) {
      out_current( &rsa );
      free_mks( mks_curr );
    }
    mks_curr = next( &rsa );
  }  
  mks_aux=rsa.first;
  if ( rsa.number_mks ) {
	mks_curr = (mks *)mks_search( &rsa ); 
	/*** We get the first agent from the rsa that has
	 *   deliberation time.
	 ***/
	
 	/*if (mks_curr!=NULL) 
	  fprintf(stderr,"Dean_Boddy_S: a ejecutar id: %s\n", mks_curr->name);
	  else fprintf(stderr,"Quiyo, estoy vacio \n");*/
	
 	out_current( &rsa );
 	return( mks_curr );
  }
  else return (mks *) NULL;
}

/* ============================================================================ */
mks *schedule_by_III(){
/* ============================================================================ */


  mks *mks_curr;
  n_mks *mks_aux;

  if (( mks_curr = first( &rsa ) ) == NULL )
    /*** There isn't any more agents on the rsa ***/
    return( mks_curr );

  while ( mks_curr != NULL ) {
   if ( !check_deadline( mks_curr ) ) {
     out_current( &rsa );
     free_mks( mks_curr );
   }
   mks_curr = next( &rsa );
  }
  mks_aux=rsa.first;
  if ( rsa.number_mks ) {
    mks_curr = (mks *)mks_search( &rsa ); 
    /*** We get the first agent from the rsa that has
     *   deliberation time.
     ***/
    
       
    out_current( &rsa );
    return( mks_curr );
  }
  else return (mks *) NULL;
}



/* ============================================================================ */
int ExecuteKSI_dbs(int pid, systime_t time){
/* ============================================================================ */
 int resp;

 /* resp=execute_until(pid,time);*/

 resp=0;
 if (resp==0) return 0;
 else { 
   /* KillKSI(pid);*/
   return 2;
 }
}

