/**  
 *  \file events_internal.h					
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Event Queue Manager 2.0 - Internal Structure
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */


#ifndef _EVENTS_INTERNAL_
#define _EVENTS_INTERNAL_

#include "../events.h"

#define EVENT_STORE "EventStore"         ///< Id for the mbuff


/// The other event type constants are defined in the KDM_events.h

#define MAX_EV                   40 ///< Too many?

typedef union _Event_Value {
  long long lTime;
  KDM_Value KDM_v; /*** KDM_Value is defined in blackboard.h ***/
} Event_Value;

typedef struct _Event{
  int iType;            /**< Event type according the constants defined
                         *     in events.h.
                         *     Bit field.
                         */
  void * pSlot;         ///< Pointer to the slot generating the event
  char iImp;            ///< Event's importance
  int iDataType;        /**< Slot data type. Its value is according the 
                         *      constants defined in blackboard.h
                         */
  long long lTimeStamp;      /** Instant the event record is fulfilled 
                              *     (to be identified in both worlds 
                              *        -> It's the event record unique id.)
                              */
  Event_Value New_Value, 
              Old_Value; /** Values of the slot that generates the event 
                          *     or the slack time
                          */
} Event_T;


typedef struct n_Event{
	Event_T ev;
        int sig;
        int prev;
} n_Event;


typedef struct {
  int pFirst;   ///< -1 if empty
  int pLast;    ///< -1 if empty
} n_Imp;

/** Dynamic array of event queues. Each element of the array has the pointers
 *     to the queue of events which importance is the same as the index of the 
 *     element in the array.
 *  The last element of this array will have the queue of empty nodes. This is to 
 *     mantain a constant number of nodes in the array.
 */
extern n_Imp * aEvImp;   

/** To control if there's anyone accessing the storage 
 *            Values:
 *               - 0 : There's no one accessing
 *               - 1 : There's someone here
 */
extern volatile int *piMutex;  

extern int *piEvents; ///< Current number of events


/*** Function Interface ***/

extern int Events_Insert_Event( Event_T regEvent ); 

extern int Events_Extract_Event( Event_T * regEvent ); 



/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */

#endif
