/*

   FILE:
	mutex.h						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, Universidad Politecnica de Valencia


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    The MUTEX module provides access to CSP-like mutexes to the
    application at both sides (RT-Linux tasks and Linux levels).  

    The CSP protocol sets a priority ceiling to each mutex, being
    computed as the priority of the highest-priority task accessing
    the mutex.  When a task locks a mutex, the task's priority is
    immediately raised to the mutex ceiling. When the task unlocks the
    mutex, it retrieves the priority it had when it locked that
    mutex. The current implementation meets this behaviour only at the
    RT-Linux level.
    
    At the Linux level this functionality is different, since we do
    not have a general-purpose mutual exclusion mechanism between
    RT-Linux task and Linux levels. Instead, the lock operation at the
    Linux level has two parameters: the mutex and the interval that
    the level is going to lock the mutex. Only if the mutex is free
    and the critical section can be completely executed within the
    current slack interval can the level successfully lock the mutex.

    In general, all these functions return 0 if everything went OK and
    some negative number if something went wrong.

*/

#ifndef __MUTEX_H
#define __MUTEX_H

//#ifdef MODULE
//#include "../../rtl_core/include/rtl_core.h"
//#else
//#include "../../linux_core/include/linux_core.h"
//#endif

#include "../common/include/artis_conf.h"
#include "../time/include/time.h"



// Visible definitions:

#define MAX_USER_MUTEXES    256



// Visible functions:

// Mutexes are created in Linux (arbitrary decission!!)

int  mutex_init_module (void);
int  mutex_cleanup_module (void);


#ifndef MODULE
int  mutex_create (int mutex_index, int priority_ceiling); 
#else
#ifdef RTLINUX_ONLY
int  mutex_create (int mutex_index, int priority_ceiling);
#endif
#endif


#ifdef MODULE
int  mutex_lock (int mutex_index);
#else
int  mutex_system_lock (int mutex_index, systime_t blocking_duration);
int  mutex_lock (int mutex_index, systime_t blocking_duration);
#endif

int  mutex_unlock (int mutex_index);

int  mutex_destroy (int mutex_index);

#endif // __MUTEX_H
