/*

   FILE:
	rserver.h						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    This include exports the RSERVER application interface to the
    kernel module containing the RT-Linux part of the ARTIS
    application.

*/


#ifndef __USER_RSERVER_H
#define __USER_RSERVER_H


#include "../common/include/artis_conf.h"
#include "../rtl_core/include/rtl_core.h"
#include "../rserver/include/rserver/rserver.h"


// Visible types:

// Internal task type:
typedef _rt_task_t rt_task_t; 

// Struct for creating tasks:

typedef struct task_create_info {
  int        identifier;
  void      *(*initial_function)(void *);
  void      *initial_parameters;
  void      *(*final_function)(void *);
  void      *final_parameters;
  int        has_optionals;
  int        priority;
  systime_t  offset;
  systime_t  period;
  systime_t  deadline;
  systime_t  initial_wcet;
  systime_t  final_wcet;
  systime_t  blocking_time;
} task_create_info;


#ifdef ARTIS_CONF_METAREASONING
typedef struct rt_behav_task_t {
  rt_task_t *task;
  struct rt_behav_task_t *pNext;
} rt_behav_task_t;

typedef struct rt_behav_t {
  rt_behav_task_t *task_list; /// < Tasks belonging to the behavior
  int nTasks;                 /// < Number of tasks in the behavior
  struct rt_behav_t *pNext;   /// < Ptr to the next behavior
} rt_behav_t;
#endif



// Visible functions :
int  rtl_task_create (rt_task_t *task, struct task_create_info *info);

#ifdef ARTIS_CONF_METAREASONING
int rtl_behav_create( rt_behav_t *pBeh, rt_behav_task_t *aTaskBeh,
                      int nTasks, ... );
#endif

int  rtl_start_application (void);

void rtl_end_application (void);  




#endif // #ifndef __USER_RSERVER_H
