/*

   FILE:
	time.h						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    This include exports the TIME interface to all the parts of ARTIS
    which need time access (including the user).

    In both cases, time are measured as a linear (long long) variable
    in which time is expressed in nanoseconds. This is the natural
    time unit in RT-Linux, but not in Linux. As a result, an explicit
    function for consulting the current time has to be provided to the
    "linux_side" process, while in RT-Linux, the original function is
    used (altough we have renamed it, in order to present a common
    API).

*/

#ifndef __USER_TIME_H
#define __USER_TIME_H



#ifdef MODULE
#include <rtl_time.h>
#else
#include <sys/time.h>
#include <unistd.h>
#endif


// Visible types:

#ifdef MODULE
typedef hrtime_t   systime_t;
#else
typedef long long   systime_t;
#endif


// Visible functions: 

#ifdef MODULE
#define systime_gettime          gethrtime
#define systime_timespec_to_time timespec_to_ns
#define systime_time_to_timespec timespec_from_ns
#else
void           systime_init (void);
systime_t      systime_gettime (void);
systime_t      systime_timeval_to_time (struct timeval t); 
struct timeval systime_time_to_timeval (systime_t t);
#endif





#endif  // #ifndef __USER_TIME_H
