/*!
  \file blackboard.h
  \date Time-stamp: <16-03-2005 00:50>
  \author M. Rebollo <mrebollo@dsic.upv.es>
  \version 0.1
  
  \brief internal structure of the blackboard

  All KSs use a shared memory to mantain their common knowledge
  and to comunicate each other. This structure is based on the
  blackboard model.

  This file provides the user with the data structures needed to
  implement the shared memory. It is supported by an internal module,
  naming storage, 
  
*/

#if !defined _KDMUSER_H_  && !defined _KDMMOD_H_ && !defined KDM_COMPILATION
# error "Never include <include/kdm/blackboard.h> to use KDM."
# error "Only public <kdm.h> files can be included directly."
#endif


#ifndef _BLACKBOARD_H_
#define _BLACKBOARD_H_

#include "globals.h"
#include "static.h"
#include "events.h"
#include "../../../storage/include/storage.h"


/*! 
  \struct tbn_t
  \brief General KDM data structure (storage access) 
*/
struct tbn_t{
  char name[255];              //!< name of the storage space
  Storage_T stdata;            //!< access to the storage data
  unsigned long max_size;      //!< memory reserved for the storage (in bytes)
  unsigned long max_entities;  //!< max. number of entities in the storage
};
 
extern struct tbn_t tbn;



#endif //_BLACKBOARD_H_
