/*!
  \file events.h
  \date Time-stamp: <16-03-2005 00:14>
  \author M. Rebollo <mrebollo@dsic.upv.es>
  \version 0.2

  \brief Internal KDM management for events
*/

#if !defined _KDMUSER_H_  && !defined _KDMMOD_H_ && !defined KDM_COMPILATION
# error "Never include <include/kdm/blackboard.h> to use KDM."
# error "Only public <kdm.h> files can be included directly."
#endif


#ifndef _KDM_EVENTS_H_
#define _KDM_EVENTS_H_

#include "../../../events/include/events.h"


/*!
  \fn int kdm_add_event( void *slot, int event, int imp )
  \brief Adds a new event of interest for the slot 
  \param void *slot identifier for the slot
  \param int event kind of event to add
  \param int imp importance of the event
  \exception slot is NULL
  \return 0: event added; 1: error
*/
int kdm_event_add( void *slot, int event, int imp );

/*!
  \fn int kdm_release_event( void *slot, int event )
  \brief Remove an exising event of interest for the slot 
  \param void *slot identifier for the slot
  \param int event kind of event to delete
  \exception slot is NULL
  \return 0: event deleted; 1: error
*/
int kdm_event_release( void *slot, int event );

/*!
  \fn int kdm_set_event( void *slot, int event, int imp )
  \brief Sets exactly the events of interest for the slot 
  \param void *slot identifier for the slot
  \param int event set of active events
  \param int imp importance of the event
  \exception slot is NULL
  \return 0: event set; 1: error
*/
int kdm_event_set( void *slot, int event, int imp );

#endif //_KDM_EVENTS_H_
