/*! 
  \file include/kdm/kdm.h
  \date Time-stamp: <17-11-2004 23:31>
  \author M. Rebollo <mrebollo@dsic.upv.es>
  \version 0.2

  \brief private access functions to KDM information

  This module provides the implementation of the KDM functions
  that provide access to the set of beliefs of an ARTIS agent.

  You will find the real implementation of the interface functions 
  with the data structures forming the KDM.

  
*/

#if !defined _KDMUSER_H_  && !defined _KDMMOD_H_ && !defined KDM_COMPILATION
# error "Never include <include/kdm/kdm.h> to use KDM."
# error "Only public <kdm.h> files can be included directly."
#endif


#ifndef _KDMPRIV_H_
#define _KDMPRIV_H_


#include "globals.h"
#include "static.h"
#include "events.h"
#include "blackboard.h"


/*!
  \def   kdm_get_slot_value( s )
  \brief Gets the value of slot 's'Replaces this call 
         for a pointer to a function
*/
#define kdm_get_slot_value( s ) static_get[s->type]( s )


/*!
  \def   kdm_set_slot_value( s )
  \brief Inserts a new static value in slot 's'.
         Renames this call to a function with variable arguments

  Note: this function could be replaced for a pointer to a function,
        in the same way that static_get does. 
*/
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
# define kdm_set_slot_value( s,... ) static_set_slot_value( (s), _VA_ARGS__ )
#else
# define kdm_set_slot_value( s,args... ) static_set_slot_value( (s), ##args )
#endif


#endif //_KDMPRIV_H_
