/*!
  \file static.h
  \date Time-stamp: <16-03-2005 00:29>
  \author M. Rebollo <mrebollo@dsic.upv.es>
  \version 0.3

  \brief internal access funtions to static information
*/

#if !defined _KDMUSER_H_  && !defined _KDMMOD_H_ && !defined KDM_COMPILATION
# error "Never include <include/kdm/blackboard.h> to use KDM."
# error "Only public <kdm.h> files can be included directly."
#endif


#ifndef _STATIC_H_
#define _STATIC_H_

#include <stdarg.h>
#include "globals.h"

/*! 
  \struct static_slot 
  \brief  represent a static slot, that is, an slot without any temporal mark
*/
typedef struct _static_slot static_slot;

struct _static_slot{
  kdm_t type;       //!< kdm_t type; data domain (int, doube or char)
  KDM_Value value;  //!< value_t value; slot value
  int event;        //!< int event; triggered events when value changes
  int imp;          //!< int imp;   importance of the events associated
  /* 
     look for offsetof() macro (stddef.h) and kernel.h to
     get the pointer to the struct owner

     void* (*getvalue)();
  */
};


/*!
  \var kdmtype static_get 
  \brief It is an array of function to get access to a KDM_Value
   depending on its datatype
*/
extern kdmtype (*static_get[KDM_PRIMITIVE_TYPES])(const static_slot *slot );


/*!
  \fn    int static_register_access( void );
  \brief Inits the static_get array with the corresponding pointer to get_* function
  \exception static access function don'f found
  \return 0: if all registered; -1: otherwise
*/
int static_register_access( void );


/*!
  \fn    static_set_slot_value( static_slot *slot,... );
  \brief Sets the value of an static slot 'slot'
  \param static_slot *slot Is the slot to be updated
  \param ... is a variable argument with the value to be stored in the slot
  \exception slot is NULL or type undefined
  \return 0: operation realised; -1: error
*/
int static_set_slot_value( static_slot *slot,... );


#endif //_STATIC_H_
