/*!
  \file include/kdm.h
  \date Time-stamp: <16-11-2004 root kdm.h>
  \author M. Rebollo <mrebollo@dsic.upv.es>
  \version 0.2

  \brief ARTIS modules public access functions to KDM information

  This file gives access to the KDM structures to the rest of 
  ARTIS modules, besides the public user functions (to get and set 
  values in the slots).

  The 'init' function creates an space in memory for all KDM structures.
  They have to been called before any other access to the KDM, in
  the ARTIS initialisation phase

  The 'clean' function frees the memory reserved by the 'init' function.
  The user is responsible for previously destroy all the stored data.
*/

#ifndef _KDMMOD_H_
#define _KDMMOD_H_

#include "./kdm/kdm.h"

/*
  Generic functions to init the KDM before use
  and to clean all its resources before it finishes
*/


/*
  \fn int kdm_init( void )
  \brief Prepare the KDM before it can be used
*/
int kdm_init( void );
/*
  \fn int kdm_clean( void )
  \brief Release the KDM resources allocated in the KDM
*/
int kdm_clean( void );



/*
  Functions to mount and unmount the KDM as a module
*/

#ifdef MODULE
/*
  \fn int kdm_init_module( void )
  \brief Called when the module is loaded to init KDM
*/
int kdm_init_module( void );

/*
  \fn void kdm_cleanup_module( void )
  \brief Called when the module is unloaded to release KDM resources
*/
void kdm_cleanup_module( void );

#endif //MODULE

#endif //_KDMMOD_H_
