/*!
  \file static.h
  \date Time-stamp: <16-11-2004 00:34>
  \author M. Rebollo <mrebollo@dsic.upv.es>
  \version 0.2

  \brief internal access funtions to static information
*/

#if !defined _KDMUSER_H_  && !defined _KDMMOD_H_ && !defined KDM_COMPILATION
# error "Never include <include/kdm/blackboard.h> to use KDM."
# error "Only public <kdm.h> files can be included directly."
#endif


#ifndef _STATIC_H_
#define _STATIC_H_

#include <stdarg.h>
#include "globals.h"

#define _kdm_set_
/*! 
  \struct static_slot 
  \brief  represent a static slot, that is, an slot without any temporal mark
*/
typedef struct _static_slot static_slot;

struct _static_slot{
  kdm_t type;       /*!< kdm_t type; data domain (int, doube or char) */ 
  KDM_Value value;  /*!< value_t value; slot value */
  int event;        /*!< int event; triggered events when value changes */
  int imp;          /*!< int imp;   importance of the events associated */
  /* 
     look for offsetof() macro (stddef.h) and kernel.h to
     get the pointer to the struct owner

     void* (*getvalue)();
  */
};

extern kdmtype (*static_get[KDM_PRIMITIVE_TYPES])(const static_slot *slot );

int static_register_access( void );
int static_set_slot_value( static_slot *slot,... );

/*
kdmtype get_char( const static_slot *slot );
kdmtype get_int( const static_slot *slot );
kdmtype get_uint( const static_slot *slot );
kdmtype get_long( const static_slot *slot );
kdmtype get_ulong( const static_slot *slot );
*/

/*
    MRP: Till it works
  
kdmtype get_int64( const static_slot *slot );
kdmtype get_uint64( const static_slot *slot ); 

*/

/*
kdmtype get_float( const static_slot *slot );
kdmtype get_double( const static_slot *slot );
kdmtype get_pointer( const static_slot *slot );


void set_char( static_slot *slot, kdmchar value );
void set_int( static_slot *slot, kdmint value );
void set_uint( static_slot *slot, kdmuint value );
void set_long( static_slot *slot, kdmlong value );
void set_ulong( static_slot *slot, kdmulong value );
void set_float( static_slot *slot, kdmfloat value );
void set_double( static_slot *slot, kdmdouble value );
void set_array( static_slot *slot, kdmpointer pvalue );

*/
#endif //_STATIC_H_
