/**  
 *  \file meta_rule_store.h					
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Meta-Rule Store Manager 2.0 - Internal Structure
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

#ifndef _META_RULE_STORE_H_
#define _META_RULE_STORE_H_

#include "meta-rules_quantities.h"


#include "meta_rule_struct.h"



//#include "../events/KDM_events.h"   
    /*** It is basically for the MAX_KDM_EV_TYPES constant ***/

#define META_RULE_STORE "MetaRuleStore"         ///< Id for the mbuff


/** Max. number of meta-rules: It indicates not only the max. number of 
 *  meta-rules, but also the number of regEventTypes and regpSlot
 */
#define NMAXMETARULES 10   
 

#define METARULE_OK 1


extern aregEventType *paregEventType; ///< To index the M-R Store through the Event Type.

extern int *pn1stEmptyEvType; ///< Ptr to the 1st Empty regEventType
extern regEventType *aregEmptyEventType;  ///< Ptr to the mbuff section of empty regEventType

extern aregpSlot *paregpSlot; ///< To index the M-R Store through the ptr to Slot

extern int *pn1stEmptypSlt; ///< Ptr to the 1st empty regEventType
extern regpSlot *aregEmptypSlot; ///< Ptr to the mbuff section of empty aregpSlot

extern aregActiveAttFocus *paregActiveAttFocus; ///< To index the M-R Store through active Attention Focuses
                                          
/** To implement the Active Attention Focus Array.
 *      Ptr to the mbuff section.
 */
extern regActiveAttFocusNode *pregActiveAttFocusNode; 

extern int *pn1stEmptyAttFocusNode;  ///< Ptr to the 1st empty regActiveAttFocusNode
/** Attention Focus Empty Node Array.
 *      Ptr to the mbuff section of empty regActiveAttFocusNode nodes
 */
extern regActiveAttFocusNode *aregEmptyAttFocusNode;
         
extern int *pn1stEmptyAttentionFocus;  ///< Ptr to the 1st empty regAttentionFocus
/** There is one of this struct for each Attention Focus.
 *    Ptr to the mbuff section of regAttentionFocus.
 */
extern regAttentionFocus *pregAttentionFocus; 

/** Ptr to the mbuff section of regpMetaRuleNodes.
 *      There is one of this structs for each Meta-Rule in an Attention Focus
 */
extern regpMetaRuleNode *pregpMetaRuleNode; 

extern int *pn1stEmptypMetaRuleNode;  ///< Ptr to the 1st empty regpMetaRuleNode
/** Ptr to the mbuff section of empty regpMetaRuleNodes.
 *      This is used to add Meta-Rules to the Attention Focus (learning process)
 */
extern regpMetaRuleNode *aregEmptypMetaRuleNode;

/** Ptr to the mbuff section of regMetaRule.
 *    There is one of this structs for each Meta-Rule
 */
extern regMetaRule *aregMetaRule; 
/** Ptr to the mbuff section of empty regMetaRule
 *    This is useless???
 */
extern regMetaRule *aregEmptyMetaRule; 

extern int *pn1stEmptyCondNode; ///< Ptr to the 1st empty regExprNode
/** Ptr to the mbuff section of empty regExprNode.
 *      They are used to construct the condition part of Meta-Rules.
 */
extern regExprNode *aregEmptyCondNode; 

extern int *pn1stEmptyActNode; ///< Ptr to the 1st empty regActNode
/** Ptr to the mbuff section of empty regActNode.
 *        They are used to construct the action part of Meta-Rules.
 */
extern regActNode *aregEmptyActNode; 

/** To control if there's anyone accessing the storage 
 *           - Values:
 *               - 0 : There's no one accessing
 *               - 1 : There's someone here
 */
extern volatile int *piMRMutex;  


/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */

#endif
