/**  
 *  \file meta_rule_actions.h
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Functions implementing the actions in the Meta-Rules
 *
 *  There are also actions for Meta-Rules:
 *    - To activate an Attention Focus: 
 *        Through the function defined in meta-rule_store.c
 *             void ActivateAttentionFocus( int nAF, float fAttentDegree )
 *    - To deactivate an Attention Focus: 
 *        Through the function defined in meta-rule_store.c
 *             void DeactivateAttentionFocus( int nAF, float fAttentDegree )
 *    - To change the value of an slot (that will be a Meta-Reasoning slot):
 *        Through the function defined in kdm.c & rt_kdm.c
 *             int KDM_Set_Slot_Value( KDM_Handle pHnd, ... ) 
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */


#ifndef _META_RULE_ACTIONS_H_
#define _META_RULE_ACTIONS_H_

#include "../../events/include/events/events_internal.h"

//#include "../control/include/decl4.0.h"  
    /*** For the Meta-Reasoning "slots":
     *     - int heuristic;
     *         This should be called SLS - Second Level Scheduler current policy
     *     - float system_quality ;   
     *         IT WOULDN'T HAVE TO BE A FLOAT 
     *     - int nExecIS;     
     *         Number of times the IS has been executed 
     *     - float MaxQuality; 
     *     - int ReactivityDegree;   
     *         [0..100] 
     *          --> scheduler.c (cuando se puedan adelantar las partes finales)
     *     - int EventRate; 
     *          Max. number of events treated in a control cycle 
     *     - int LevelsCompleted; 
     *           Number of levels completed with the current scheduling policy 
     *     - int LevelsInterrupted;
     *           Number of levels interrupted with the current scheduling policy
     * More Meta - Slots needed: CurrentAttentionFocus 
     ***/


/** This function inserts in the tsa the sporadic level indicated 
 *      by nLevelId
 */
extern void MR_Action_Trigger_Sporadic_Level ( int nLevelId );


/** This function inserts in the tsa the optional levels indicated 
 *      by the event
 */
extern void MR_Action_Trigger_Optional ( float fGain, Event_T regAux );


/// Sets nReactDegree as the Reactivity Degree of the AA
extern void MR_Action_Set_Reactivity_Degree ( int nReactDegree );


/// Sets nSlsPolicy as the current second-level scheduling policy
extern void MR_Action_Set_Policy ( int nSlsPolicy );


/// -> SetAttentionDegree fAttentionDegree --- 
extern void MR_Action_Set_Attention_Degree ( float fAttentionDegree );


/// -> GetAttentionDegree
extern float MR_Action_Get_Attention_Degree ( void );


/// -> SetBehaviour nBehav --- 
extern void MR_Action_Set_Behaviour ( int nBehav );

/// Function for initializing the MR Actions module
extern void MR_Action_Initialize (void);

extern int MR_Action_init_module(void);

extern void MR_Action_cleanup_module(void);
#endif
