/**  
 *  \file meta_rule_interpret.h					
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Meta-Rule Interpreter 2.0 - Public Interface
 *
 *       This module incorporates all the functions to on-line interpret the 
 *     meta-rules. 
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

#ifndef _META_RULE_INTERPRET_H_

#define _META_RULE_INTERPRET_H_

#include "../../events/include/events/events_internal.h"  
                                    ///< Definition of Event_T
/// NOTE 4 ME & MY SHADOW: \todo Reescribir las funciones para eliminar regEvent, en base a Event_Notify, ...


#include "meta_rule_store.h"
#include "meta_rule_actions.h"


/// Evaluate the factor leaving the result in a double valor
extern double Meta_Rule_Eval_Factor ( regTypeFactor regFactor, 
				      Event_T regEvent );


/** Evaluates the expression in aregEmptyCondNode[nCond] 
 *   - It returns:
 *        - 0, if the expr. is evaluated to false.
 *        - <> 0, if the expr. is evaluated to true.
 */
extern int Meta_Rule_Eval_Expr ( int nCond, Event_T regEvent );


/** Executes the action in aregEmptyActNode[nAct]
 *  - Valid data combinations for the regActNode:
 *      - nAct (int -> action code)     regArg1                regArg2
 *      - 0 -> Assign                   pSlot (relative ptr)   pExpr
 *      - 1 -> Function Call            nFunctId               ---
 *      - 2 -> TriggerSporadicLevel     nLevelId               ---
 *      - 3 -> SetReactivityDegree      nReactDegree           ---
 *      - 4 -> SetPolicy                nSlsPolicy             ---
 *      - 5 -> SetFocus                 nAttentionFocus        fFocusAttentDegree
 *      - 6 -> SetAttentionDegree       fAttentionDegree       ---
 *      - 7 -> SetBehaviour             nBehav                 ---
 *      - 8 -> KDM_Add_Event            pSlot                  iType       iImp (nAct from the next regActNode)
 *      - 9 -> KDM_Release_Event        pSlot                  iType
 */
extern void Meta_Rule_Execute_Action ( int nAct, Event_T regEvent );


/** Evaluates the condition and if it is true, executes the actions associated to the Meta Rule.
 *   It should receive also the KDM_Event_T received in the event as a parameter 
 */
extern void Meta_Rule_Interpret ( int nMetaRule, Event_T regEvent );


/** It manages the event in regEvent evaluating as much as nMaxMR meta-rules associated to this event.
 *   It returns the number of MR it has found and executed.
 */
extern int Meta_Rule_Manage_Event ( Event_T regEvent, int nMaxMR );

#endif
