/**
 * \file meta_rule.c
 * \date 2004
 * \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 * \version 2.0
 * 
 * \brief Meta-Rule Store Manager Definition 2.0
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */


#ifdef MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>

#include <rtl.h> /// rtl_printf

#else
#include <stdio.h>
#include <stdarg.h>  ///< For functions' variable arguments 
#endif


///#define NULL (void *)0

///#include "../include/rt_trace.h"

///#include "../include/metareasoning.h"

 
/*** ONLY FOR WINDOWS REHEARSAL ***/
//#include <mbuff.h>
/*** ONLY FOR WINDOWS REHEARSAL ***/



#ifdef MODULE
#include <mbuff.h>
#else

extern int mbuff_alloc(const char *name, int size);
extern int mbuff_free(const char *name, void * mbuf);
extern void * mbuff_attach(const char *name, int size);
extern void mbuff_detach(const char *name, void * mbuf);

#endif

#include "../include/meta_rule/meta_rule_store.h"

#include "../../events/include/events.h"

#include "../../common/include/artis_conf.h"



/** 
 *  To index the M-R Store through the Event Type.
 */
aregEventType *paregEventType; 

int *pn1stEmptyEvType;  ///< Ptr to the 1st Empty regEventType
regEventType *aregEmptyEventType; ///< Ptr to the mbuff section of empty regEventType


aregpSlot *paregpSlot; ///< To index the M-R Store through the ptr to Slot

int *pn1stEmptypSlt; ///< Ptr to the 1st empty regEventType
regpSlot *aregEmptypSlot;  ///< Ptr to the mbuff section of empty aregpSlot


/// To index the M-R Store through active Attention Focuses.
aregActiveAttFocus *paregActiveAttFocus; 
       
                                          
/** 
 *  To implement the Active Attention Focus Array.
 *      Ptr to the mbuff section.
 */
regActiveAttFocusNode *pregActiveAttFocusNode; 
int *pn1stEmptyAttFocusNode; ///< Ptr to the 1st empty regActiveAttFocusNode

/** 
 *   Attention Focus Empty Node Array.
 *      Ptr to the mbuff section of empty regActiveAttFocusNode nodes
 */
regActiveAttFocusNode *aregEmptyAttFocusNode;


int *pn1stEmptyAttentionFocus; ///< Ptr to the 1st empty regAttentionFocus

/** 
 *  There is one of this struct for each Attention Focus.
 *    Ptr to the mbuff section of regAttentionFocus.
 */
regAttentionFocus *pregAttentionFocus; 


/** 
 *  Ptr to the mbuff section of regpMetaRuleNodes.
 *      There is one of this structs for each Meta-Rule in an 
 *      Attention Focus
 */
regpMetaRuleNode *pregpMetaRuleNode; 
int *pn1stEmptypMetaRuleNode; ///< Ptr to the 1st empty regpMetaRuleNode

/** 
 *  Ptr to the mbuff section of empty regpMetaRuleNodes.
 *      This is used to add Meta-Rules to the Attention Focus 
 *      (learning process)
 */
regpMetaRuleNode *aregEmptypMetaRuleNode;

/** 
 *  Ptr to the mbuff section of regMetaRule.
 *    There is one of this structs for each Meta-Rule
 */
regMetaRule *aregMetaRule; 

/** 
 *  Ptr to the mbuff section of empty regMetaRule
 *    This is useless???
 */
regMetaRule *aregEmptyMetaRule; 


int *pn1stEmptyCondNode;  ///< Ptr to the 1st empty regExprNode

/** 
 *  Ptr to the mbuff section of empty regExprNode.
 *      They are used to construct the condition part of Meta-Rules.
 */
regExprNode *aregEmptyCondNode; 


int *pn1stEmptyActNode; ///< Ptr to the 1st empty regActNode

/** 
 *  Ptr to the mbuff section of empty regActNode.
 *        They are used to construct the action part of 
 *        Meta-Rules.
 */
regActNode *aregEmptyActNode; 


/**
 *   To control if there's anyone accessing the storage 
 *            Values:
 *               - 0 : There's no one accessing
 *               - 1 : There's someone here
 */
volatile int *piMRMutex;  


//////////////////////////////////////////////////////////////////////////////
/// Private Functions
//////////////////////////////////////////////////////////////////////////////
/** 
 * \defgroup Private Functions 
 */

//@{

/************************* Meta_Rule_Init_Store *****************************/
/**
 * \fn Meta_Rule_Init_Store
 *
 * Initializes the Meta-Rule Store, shaping all its empty nodes 
 *         This is used in the RTOS level, to contact with the memory allocated
 *             by the RTOS in the initialization phase 
 *                                      ==> in the InitBlackboard()?
 ***/
int Meta_Rule_Init_Store( int nEventTypes, int nExprUsed, int nActUsed,
                          int nMetaRules, int nAttFocus, int nActiveFocus ) {
  int nregExprNode, nregActNode;
  
  nregExprNode = 
          (( 2 * NMAXMETARULES > nExprUsed + 2 * (NMAXMETARULES - nMetaRules))?
          2 * NMAXMETARULES:  nExprUsed + 2 * (NMAXMETARULES - nMetaRules));
                            
  nregActNode = 
          (( 2 * NMAXMETARULES > nActUsed + 2 * (NMAXMETARULES - nMetaRules))?
          2 * NMAXMETARULES:  nActUsed + 2 * (NMAXMETARULES - nMetaRules));

  if (( piMRMutex = ( volatile int * ) mbuff_attach ( META_RULE_STORE, 
                 ( ((int) sizeof( int)) 
                    + ((int) sizeof( aregEventType )* nEventTypes )
                    + ((int) sizeof( regEventType ) * NMAXMETARULES ) 
                    + ((int) sizeof( aregpSlot ) * NMAXMETARULES )  
                    + ((int) sizeof( regpSlot ) * NMAXMETARULES ) 
                    + ((int) sizeof( aregActiveAttFocus ))
                    + ((int) sizeof( regAttentionFocus ) * nAttFocus)
                    + ((int) sizeof( regActiveAttFocusNode ) * nActiveFocus )
                    + ((int) sizeof( regpMetaRuleNode ) * NMAXMETARULES )
                    + ((int) sizeof( regMetaRule ) * NMAXMETARULES )
                    + ((int) sizeof( regExprNode ) * nregExprNode )
                    + ((int) sizeof( regActNode ) * nregActNode)
                    + ((int) sizeof( int ) * 7) 
                    + ((int) sizeof( int ) * 100)
                                 ) )) == NULL )

     /** 
      *   nEventTypes is the number of different event types.
      *   NMAXMETARULES is a constant declared in <meta-rule_store.h>,
      *         and it is the max. number of different meta-rules existing in 
      *         the system at the same time.
      *   nAttFocus represents the number of different attention focuses 
      *         declared.
      *   nActiveFocus represents the number of different attention focuses that 
      *         are initially active.
      *   nExprUsed represents the number of regExprNode used in the initial 
      *         Meta-Rule set.
      *   nActUsed represents the number of regActNode used in the initial 
      *         Meta-Rule set.
      *   nMetaRules represents the initial number of meta-rules.
      *   7 ints for the n1stEmpty... values
      */

    return - 1;  
  else {

    *piMRMutex = 0; ///< Initially there's no one inside

         paregEventType = (aregEventType *) (piMRMutex + 1);
          
         aregEmptyEventType = 
                     (regEventType *) (paregEventType + nEventTypes + 1 );

         paregpSlot = (aregpSlot *) (aregEmptyEventType + NMAXMETARULES + 1);

         aregEmptypSlot = (regpSlot *) (paregpSlot + NMAXMETARULES + 1);
         
         paregActiveAttFocus = 
                     (aregActiveAttFocus *) (aregEmptypSlot + NMAXMETARULES + 1);

         pregActiveAttFocusNode = (regActiveAttFocusNode *) NULL;

         aregEmptyAttFocusNode = 
                             (regActiveAttFocusNode *) (paregActiveAttFocus + 1);

         pregAttentionFocus =  
                (regAttentionFocus *) (aregEmptyAttFocusNode + nActiveFocus + 1);
         
         pregpMetaRuleNode = (regpMetaRuleNode *) NULL;

         /// There is no meta-rule, so all are empty
         aregEmptypMetaRuleNode =  
                      (regpMetaRuleNode *) (pregAttentionFocus + nAttFocus + 1);

         aregMetaRule =  
                   (regMetaRule *) (aregEmptypMetaRuleNode + NMAXMETARULES + 1);
                 
         aregEmptyCondNode = (regExprNode *) (aregMetaRule + NMAXMETARULES + 1 );
            
         aregEmptyActNode =  
                          (regActNode *) (aregEmptyCondNode + nregExprNode + 1);

         pn1stEmptyEvType = (int *) (aregEmptyActNode + nregActNode +1);
         pn1stEmptypSlt = (int *) (pn1stEmptyEvType + 1);
         pn1stEmptyAttFocusNode = (int *) (pn1stEmptypSlt + 1);
         pn1stEmptyAttentionFocus = (int *) (pn1stEmptyAttFocusNode + 1);
         pn1stEmptypMetaRuleNode = (int *) (pn1stEmptyAttentionFocus + 1);
         pn1stEmptyCondNode = (int *) (pn1stEmptypMetaRuleNode + 1);
         pn1stEmptyActNode = (int *) (pn1stEmptyCondNode + 1);

/***
#ifdef MODULE
rtl_printf(" aregEmptyActNode %d\n", (int) aregEmptyActNode );         
#endif
***/
         return( 1 );
  }
}


/*************************** Meta_Rule_Create_Store ******************************/
/**
 * \fn Meta_Rule_Create_Store
 * 
 * Creates the Meta-Rule Store, shaping all its empty nodes 
 *         This is used in the DS level, to allocate the memory that will
 *             be used by the DS to manage meta-rules.
 */
int Meta_Rule_Create_Store( int nEventTypes, int nExprUsed, int nActUsed,
                                  int nMetaRules, int nAttFocus, int nActiveFocus ) {
  int i, nregExprNode, nregActNode;
  
  nregExprNode = 
           (( 2 * NMAXMETARULES > nExprUsed + 2 * (NMAXMETARULES - nMetaRules))?
            2 * NMAXMETARULES:  nExprUsed + 2 * (NMAXMETARULES - nMetaRules));
                            
  nregActNode = 
           (( 2 * NMAXMETARULES > nActUsed + 2 * (NMAXMETARULES - nMetaRules))?
            2 * NMAXMETARULES:  nActUsed + 2 * (NMAXMETARULES - nMetaRules));

  if (( piMRMutex = ( volatile int * ) mbuff_attach ( META_RULE_STORE, 
                  ( ((int) sizeof( int)) 
                    + ((int) sizeof( aregEventType )* nEventTypes )
                    + ((int) sizeof( regEventType ) * NMAXMETARULES ) 
                    + ((int) sizeof( aregpSlot ) * NMAXMETARULES )  
                    + ((int) sizeof( regpSlot ) * NMAXMETARULES ) 
                    + ((int) sizeof( aregActiveAttFocus ))
                    + ((int) sizeof( regAttentionFocus ) * nAttFocus)
                    + ((int) sizeof( regActiveAttFocusNode ) * nActiveFocus )
                    + ((int) sizeof( regpMetaRuleNode ) * NMAXMETARULES )
                    + ((int) sizeof( regMetaRule ) * NMAXMETARULES )
                    + ((int) sizeof( regExprNode ) * nregExprNode )
                    + ((int) sizeof( regActNode ) * nregActNode) 
                    + ((int) sizeof( int ) * 7) 
                    + ((int) sizeof( int ) * 100)
                    ) )) == NULL ) {

     /** 
      *   nEventTypes is the number of different event types.
      *   NMAXMETARULES is a constant declared in <meta-rule_store.h>,
      *         and it is the max. number of different meta-rules existing in 
      *         the system at the same time.
      *   nAttFocus represents the number of different attention focuses 
      *         declared.
      *   nActiveFocus represents the number of different attention focuses that 
      *         are initially active.
      *   nExprUsed represents the number of regExprNode used in the initial 
      *         Meta-Rule set.
      *   nActUsed represents the number of regActNode used in the initial 
      *         Meta-Rule set.
      *   nMetaRules represents the initial number of meta-rules.
      *   7 ints for the n1stEmpty... values
      */

    return - 1;  
  }
  else {

    *piMRMutex = 0; ///< Initially there's no one inside
    
    paregEventType = (aregEventType *) (piMRMutex + 1);
    
    paregEventType[0].nEventType = CREATION;
    paregEventType[1].nEventType = DELETION;
    paregEventType[2].nEventType = MODIFICATION;
    paregEventType[3].nEventType = NEW_PREDICTION;
    paregEventType[4].nEventType = MATCHED_PREDICTION;
    paregEventType[5].nEventType = UNMATCHED_PREDICTION;
    paregEventType[6].nEventType = CONSIDER_PERIODIC_AGENT;
    paregEventType[7].nEventType = TIME_MESSAGE;
    paregEventType[8].nEventType = EXIT_APP;
    for ( i = 0; i < nEventTypes; i++ ) 
      paregEventType[i].pFirst = -1; ////< To indicate it is empty
    
    aregEmptyEventType = (regEventType *) (paregEventType + nEventTypes + 1 );
    for ( i = 0; i < NMAXMETARULES -1; i++ ) {
      aregEmptyEventType[i].nEventType = aregEmptyEventType[i].nEventTypeIndex = -1;
      aregEmptyEventType[i].pNext = i+1;
    }
    aregEmptyEventType[NMAXMETARULES - 1].nEventType = 
      aregEmptyEventType[NMAXMETARULES - 1].nEventTypeIndex = -1;
    aregEmptyEventType[NMAXMETARULES - 1].pNext = 0;
    
    paregpSlot = (aregpSlot *) (aregEmptyEventType + NMAXMETARULES + 1);
    for ( i = 0; i < NMAXMETARULES -1; i++ ) {
      paregpSlot[i].nSlots = 0;
      paregpSlot[i].pFirst = -1;
    }
    
    aregEmptypSlot = (regpSlot *) (paregpSlot + NMAXMETARULES + 1);
    
    for ( i = 0; i < NMAXMETARULES - 1; i++ ) {
      aregEmptypSlot[i].pSlot = (void *) NULL; ///< Initializing all the ptr to slot 
      aregEmptypSlot[i].npSlotIndex = -1;
      aregEmptypSlot[i].pNext = i+1;
    }
    
    aregEmptypSlot[NMAXMETARULES - 1].pSlot = (void *) NULL;
    aregEmptypSlot[NMAXMETARULES - 1].pNext = 0;
    
    paregActiveAttFocus = 
      (aregActiveAttFocus *) (aregEmptypSlot + NMAXMETARULES + 1);
    
    paregActiveAttFocus->nActive = 0;
    paregActiveAttFocus->pFirst = -1;
    
    pregActiveAttFocusNode = (regActiveAttFocusNode *) NULL;

    aregEmptyAttFocusNode = (regActiveAttFocusNode *) (paregActiveAttFocus + 1);
    
    for ( i = 0; i < nActiveFocus - 1; i++ ) {
      aregEmptyAttFocusNode[i].fAttentionDegree = -1.0;
      aregEmptyAttFocusNode[i].pAttFocus = -1;
      aregEmptyAttFocusNode[i].pNext = i+1;
    }
         
    aregEmptyAttFocusNode[nActiveFocus - 1].fAttentionDegree = -1.0;
    aregEmptyAttFocusNode[nActiveFocus - 1].pAttFocus = -1;
    aregEmptyAttFocusNode[nActiveFocus - 1].pNext = -1; ///< The last hasn't got next
    
    pregAttentionFocus = (regAttentionFocus *) (aregEmptyAttFocusNode + nActiveFocus + 1);
         
    ///< Initially, all the attention focuses are empty 
    for (i = 0; i < nAttFocus; i++) {
      pregAttentionFocus[i].nMetaRules = 0;
      pregAttentionFocus[i].pFirst = -1;
    }
         
    pregpMetaRuleNode = (regpMetaRuleNode *) NULL;
    
    ///< There is no meta-rule, so all are empty
    aregEmptypMetaRuleNode = (regpMetaRuleNode *) (pregAttentionFocus + nAttFocus + 1);

    aregMetaRule = (regMetaRule *) (aregEmptypMetaRuleNode + NMAXMETARULES + 1);
               
    /**
     *   A meta-rule, that is, an struct of type regMetaRule, may only
     *     belong to one attention focus at the same time. 
     *   If the user wants that one meta-rule will belong to different 
     *     attention focus at the same time, then several copies will 
     *      be generated.
     *   This defines a 1 on 1 relation between regpMetaRuleNode and 
     *      regMetaRule.
     */
    
    for ( i = 0; i < NMAXMETARULES - 1; i++) {
      aregEmptypMetaRuleNode[i].pMetaRule = i;
            
      aregMetaRule[i].pregEvent = -1;
      aregMetaRule[i].pregEventSlot = -1;
      aregMetaRule[i].pregCondition = -1;
      aregMetaRule[i].pregAction = -1;
      
      aregEmptypMetaRuleNode[i].pNext = i+1;
    }
         
    aregEmptypMetaRuleNode[ NMAXMETARULES - 1 ].pMetaRule = NMAXMETARULES - 1;
            
    aregMetaRule[ NMAXMETARULES - 1 ].pregEvent = -1;
    aregMetaRule[ NMAXMETARULES - 1 ].pregEventSlot = -1;
    aregMetaRule[ NMAXMETARULES - 1 ].pregCondition = -1;
    aregMetaRule[ NMAXMETARULES - 1 ].pregAction = -1;
    
    aregEmptypMetaRuleNode[ NMAXMETARULES - 1 ].pNext = 0;
    
    /** 
     *   A meta-rule, that is, an struct of type regMetaRule, may only 
     *     belong to one attention focus at the same time. 
     *   If the user wants that one meta-rule will belong to different 
     *     attention focus at the same time, then several copies will be 
     *     generated. 
     *   This defines a 1 on 1 relation between regpMetaRuleNode and 
     *     regMetaRule. 
     */
    
    aregEmptyCondNode = (regExprNode *) (aregMetaRule + NMAXMETARULES + 1 );
    
    for ( i = 0; i < nregExprNode - 1; i++ ) {
      aregEmptyCondNode[i].regF1.nType = -1;  ///< It is empty
      aregEmptyCondNode[i].regF2.nType = -1;
      aregEmptyCondNode[i].nOp = 0; ///< No operation
      aregEmptyCondNode[i].pNext = i + 1; 
    }
    
    aregEmptyCondNode[ nregExprNode - 1 ].regF1.nType = -1;  ///< It is empty
    aregEmptyCondNode[ nregExprNode - 1 ].regF2.nType = -1;
    aregEmptyCondNode[ nregExprNode - 1 ].nOp = 0; ///< No operation
    aregEmptyCondNode[ nregExprNode - 1 ].pNext = 0; 
    
    aregEmptyActNode = (regActNode *) (aregEmptyCondNode + nregExprNode + 1);
    
    for ( i = 0; i < nregActNode - 1; i++ ) {
      aregEmptyActNode[i].nAct = -1; ///< It is empty
      aregEmptyActNode[i].pNext = i+1; 
      
      /***aregEmptyActNode[i].regArg1.pSlot = (void *) NULL; 
       *aregEmptyActNode[i].regArg1.nFunctId = -1;
       *aregEmptyActNode[i].regArg1.nLevelId = -1;
       *aregEmptyActNode[i].regArg1.nReactDegree = -1;
       *aregEmptyActNode[i].regArg1.nSlsPolicy = -1;
       *aregEmptyActNode[i].regArg1.nAttentionFocus = -1;
       *aregEmptyActNode[i].regArg1.fAttentionDegree = -1;
       *aregEmptyActNode[i].regArg1.nBehav = -1;
       *aregEmptyActNode[i].regArg2.pExpr = -1;
       *aregEmptyActNode[i].regArg2.fFocusAttentDegree = -1.0;
       ***/
    }
    
    aregEmptyActNode[ nregActNode - 1 ].nAct = -1; ///< It is empty
    aregEmptyActNode[ nregActNode - 1 ].pNext = 0;
    
    pn1stEmptyEvType = (int *) (aregEmptyActNode + nregActNode +1);
    pn1stEmptypSlt = (int *) (pn1stEmptyEvType + 1);
    pn1stEmptyAttFocusNode = (int *) (pn1stEmptypSlt + 1);
    pn1stEmptyAttentionFocus = (int *) (pn1stEmptyAttFocusNode + 1);
    pn1stEmptypMetaRuleNode = (int *) (pn1stEmptyAttentionFocus + 1);
    pn1stEmptyCondNode = (int *) (pn1stEmptypMetaRuleNode + 1);
    pn1stEmptyActNode = (int *) (pn1stEmptyCondNode + 1);
    
    *pn1stEmptyEvType = 0;
    *pn1stEmptypSlt = 0;
    *pn1stEmptyAttFocusNode = 0;
    *pn1stEmptyAttentionFocus = 0;
    *pn1stEmptypMetaRuleNode = 0;
    *pn1stEmptyCondNode = 0;
    *pn1stEmptyActNode = 0;
    
    /***
        #ifndef MODULE         
        fprintf(stderr, "\t(meta-rule_store) aregEmptyActNode %d\n", (int) aregEmptyActNode );
        #endif
    ***/

    return( METARULE_OK );
  }
}


/************************* Meta_Rule_Store_Initialize *************************/
/**
 * \fn Meta_Rule_Store_Initialize
 *
 * Public generic initializing function.
 * It allocates or attaches the memory of the Meta-Rule Store.
 * It must be used both in the RT and non-RT Linux worlds
 */
///int Meta_Rule_Store_Initialize(void) {
  /*** RT-SIDE does an mbuff_attach ***/
///#ifdef MODULE
///  return  InitMetaRuleStore();
  /*while OPTIONAL SIDE does first a mbuff_alloc */
///#else
///  return  CreateMetaRuleStore();
///#endif
///xs}


/******************** Meta_Rule_Get_Empty_Att_Focus_Node *********************/
/**
 *  \fn Meta_Rule_Get_Empty_Att_Focus_Node
 *
 *  Returns two possible values:
 *      -1, if there is no empty regActiveAttFocusNode
 *      n, where aregEmptyAttFocusNode[n] is the empty regActiveAttFocusNode
 */
 int Meta_Rule_Get_Empty_Att_Focus_Node (void) {

  if (*piMRMutex) {  
    /// There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyAttFocusNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine 

  if ((*pn1stEmptyAttFocusNode) == -1) {
    ///< There is no empty regActiveAttFocusNode
        *piMRMutex = 0;
        return -1;  
  }
    else {
           int nAux = *pn1stEmptyAttFocusNode;

           *pn1stEmptyAttFocusNode = aregEmptyAttFocusNode[nAux].pNext;   
           *piMRMutex = 0;
           return nAux;
    }
}



/******************** Meta_Rule_Release_Empty_Att_Focus_Node *********************/
/**
 *  \fn Meta_Rule_Release_Empty_Att_Focus_Node
 *
 *  Returns three possible values:
 *      -2, if there is someone accessing the same memory location.
 *      -1, if there is some problem releasing the regActiveAttFocusNode indicated 
 *          by the arg. nActiveAttFocusNode, where aregEmptyAttFocusNode[nActiveAttFocusNode] 
 *          is the empty regActiveAttFocusNode
 *       0, if there is no problem releasing it.
 */
int Meta_Rule_Release_Empty_Att_Focus_Node ( int nActiveAttFocusNode ) {
  /***
  if (*piMRMutex) {  
    /// There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule.c)::ReleaseEmptyAttFocusNode ERROR ==> LIBERANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine 
  ***/

  int nAux = *pn1stEmptyAttFocusNode;

  *pn1stEmptyAttFocusNode = nActiveAttFocusNode;

  ///  if (nAux != -1) {
    ///< There were empty regActiveAttFocusNodes
    aregEmptyAttFocusNode[*pn1stEmptyAttFocusNode].pNext = nAux;   
    ///  }
    
  ///  *piMRMutex = 0;
  
  return 0;
}






/******************* Meta_Rule_Get_Empty_pMetaRuleNode ********************/
/**
 * \fn Meta_Rule_Get_Empty_pMetaRuleNode
 *
 *  Returns two possible values:
 *      -1, if there is no empty regpMetaRuleNode
 *      n, where aregEmptypMetaRuleNode[n] is the empty regpMetaRuleNode
 *         and aregMetaRule[n] is the empty regMetaRule
 */
int Meta_Rule_Get_Empty_pMetaRuleNode (void) {

  if (*piMRMutex) { ///< There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptypMetaRuleNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

  if ((*pn1stEmptypMetaRuleNode) == -1) { ///< There is no empty regpMetaRuleNode
        *piMRMutex = 0;
        return -1;  
  }
    else {
            int nAux = *pn1stEmptypMetaRuleNode;
            
            *pn1stEmptypMetaRuleNode = aregEmptypMetaRuleNode[nAux].pNext;   
            *piMRMutex = 0;
            return nAux;
    }
}


/******************** Meta_Rule_Get_Empty_Event_Type ***********************/
/**
 * \fn Meta_Rule_Get_Empty_Event_Type
 *
 *  Returns two possible values:
 *      -1, if there is no empty regEventType
 *      n, where aregEmptyEventType[n] is the empty regEventType
 */
int Meta_Rule_Get_Empty_Event_Type (void) {

  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyEventType ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

  if ((*pn1stEmptyEvType) == -1) { ///< There is no empty regEventType
    *piMRMutex = 0;
    return -1;  
  }
  else {
    int nAux = *pn1stEmptyEvType;
    *pn1stEmptyEvType = aregEmptyEventType[nAux].pNext;   
    *piMRMutex = 0;
    return nAux;
  }
}


/******************** Meta_Rule_Get_Empty_pSlot ***********************/
/**
 * \fn Meta_Rule_Get_Empty_pSlot
 *
 *  Returns two possible values:
 *      -1, if there is no empty regpSlot
 *      n, where aregEmptypSlot[n] is the empty regpSlot
 */
int Meta_Rule_Get_Empty_pSlot (void) {

  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptypSlot ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

  if ((*pn1stEmptypSlt) == -1) { ///< There is no empty regpSlot
    *piMRMutex = 0;
    return -1;  
  }
  else {
    int nAux = *pn1stEmptypSlt;
    *pn1stEmptypSlt = aregEmptypSlot[nAux].pNext;   
    *piMRMutex = 0;
    return nAux;
  }
}


/******************** Meta_Rule_Get_Empty_Cond_Node ***********************/
/**
 * \fn Meta_Rule_Get_Empty_Cond_Node
 *
 *  Returns two possible values:
 *      -1, if there is no empty regExprNode
 *      n, where aregEmptyCondNode[n] is the empty regExprNode
 */
int Meta_Rule_Get_Empty_Cond_Node (void) {
  
  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyCondNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

  if ((*pn1stEmptyCondNode) == -1) { ///< There is no empty regExprNode
    *piMRMutex = 0;
    return -1;  
  }
  else {
    int nAux = *pn1stEmptyCondNode;
    *pn1stEmptyCondNode = aregEmptyCondNode[nAux].pNext;   
    *piMRMutex = 0;
    return nAux;
  }
}


/******************** Meta_Rule_Get_Empty_Act_Node ***********************/
/**
 * \fn Meta_Rule_Get_Empty_Act_Node
 *
 *  Returns two possible values:
 *      -1, if there is no empty regActNode
 *      n, where aregEmptyActNode[n] is the empty regActNode
 */
int Meta_Rule_Get_Empty_Act_Node (void) {
  
  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyActNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

  if ((*pn1stEmptyActNode) == -1) { ///< There is no empty regActNode
    *piMRMutex = 0;
    return -1;  
  }
  else {
    int nAux = *pn1stEmptyActNode;
    *pn1stEmptyActNode = aregEmptyActNode[nAux].pNext;   
    *piMRMutex = 0;
    return nAux;
  }
}


/********************** Meta_Rule_Get_Empty_Act_Node *************************/
/***
 *   This is a variation of the previous function that doesn't reorder the
 *      regActNode store, because it has all the regActNode in the same store
 *      empties or not. It is not very efficient.
 ***/
/***
 *int GetEmptyActNode ( int *nRegActIndex ) {
 *   int i;
 *   
 *   if (aregEmptyActNode == NULL)
        *** There is no empty regActNode ***/
/***        return -1;  
 *   else {
 *           for ( i = 0; (i < nregActNode) && (aregEmptyActNode[i].nAct == -1); i++ );
 *           
 *           if ( i == nregActNode )
                *** There is no empty regActNode ***/
/***                return -1;              
 *           else {
 *                   *nRegActIndex = i;
 *                   return 0;
 *           }       
 *   }
 *}        
 ***/


//@}


///////////////////////////////////////////////////////////////////////////////
/// Public Interface Functions 
///////////////////////////////////////////////////////////////////////////////
/** 
 * \defgroup Public Interface Functions
 */

//@{

/******************  Meta_Rule_AF_Store_Creation *****************************/
/**
 * \fn  Meta_Rule_AF_Store_Creation
 *
 * \brief Attention Focus Store (or Meta-Rule Store) initializing function
 *
 *
 *  Creation of the Attention Focus Store (or Meta-Rule Store). 
 *  This function needs as arguments:
 *    - nEventTypes: Number of different event types used in the Meta-Rules 
 *                   defined by the designer.
 *    - nExprUsed: Number of different expression nodes used in the 
 *                   Meta-Rules defined by the designer.
 *    - nActUsed: Number of different action nodes used in the Meta-Rules
 *                   defined by the designer.
 *    - nMetaRules: Number of different Meta-Rules defined by the designer.
 *    - nAttFocus: Number of different Attention Focuses defined by the designer.
 *    - nActiveFocus: Number of different Active Attention Focuses defined by the designer.
 *  
 */
int Meta_Rule_AF_Store_Creation ( int nEventTypes, int nExprUsed, int nActUsed,
                                  int nMetaRules, int nAttFocus, int nActiveFocus ) {
  /*** RT-SIDE does an mbuff_attach ***/
#ifdef MODULE
  return  Meta_Rule_Init_Store( nEventTypes, nExprUsed, nActUsed, nMetaRules, nAttFocus, nActiveFocus );
  /*while OPTIONAL SIDE does first a mbuff_alloc */
#else
  return  Meta_Rule_Create_Store( nEventTypes, nExprUsed, nActUsed, nMetaRules, nAttFocus, nActiveFocus );
#endif
}


/*************************** Meta_Rule_Store_Free ***************************/
/**
 * \fn Meta_Rule_Store_Free
 *
 * It Releases the memory allocated or attached to the Meta-Rule Store.
 * It must be used both in the RT and non-RT Linux worlds
 */
void Meta_Rule_Store_Free(void) {
#ifdef MODULE
  mbuff_detach(META_RULE_STORE,(void *)piMRMutex);
#else
  mbuff_detach(META_RULE_STORE,(void *)piMRMutex);
#endif
}


/************************** Meta_Rule_Create_Att_Focus **********************/
/**
 * \fn Meta_Rule_Create_Att_Focus
 * 
 * \brief Reserves an attention focus node, and return its id or (-1) if error.
 */
int Meta_Rule_Create_Att_Focus(void) {
  return Meta_Rule_Get_Empty_Att_Focus_Node();
}


/************************* Meta_Rule_Create_Expr ***************************/
/**
 * \fn Meta_Rule_Create_Expr
 * 
 * \brief Reserves an expr (cond) node, and returns its id or (-1) if error.
 *
 * It receives 5 arguments:
 *    - nOp: Operation code.
 *    - nType1: Type of the first factor.
 *    - nType2: Type of the second factor.
 *    - First factor.
 *    - Second factor.
 */
int Meta_Rule_Create_Expr(int nOp, int nType1, int nType2, ...) {
  va_list ap;
  int nCond;

  va_start( ap, nType2 );

  if ((nCond = Meta_Rule_Get_Empty_Cond_Node() ) < 0 )
    return nCond;
  else {

    aregEmptyCondNode[nCond].nOp = nOp;
    aregEmptyCondNode[nCond].pNext = -1;

    /// Establish Factor 1 of the expression
    switch( aregEmptyCondNode[nCond].regF1.nType = nType1 ) {
    case 0: ///< pExpr
      aregEmptyCondNode[nCond].regF1.uniType.pExpr = va_arg( ap, int );
      break;

    case 1: ///< pSlot
      aregEmptyCondNode[nCond].regF1.uniType.pSlot = (void *) Storage_Get_Relative( tbn.stdata, va_arg( ap, void * ));
      break;
    case 2: ///< nFunctId
      aregEmptyCondNode[nCond].regF1.uniType.nFunctId = va_arg( ap, int );
      break;
    case 3: ///< nFactor
      aregEmptyCondNode[nCond].regF1.uniType.nFactor = va_arg( ap, int );
      break;
    case 4: /// fFactor
      aregEmptyCondNode[nCond].regF1.uniType.fFactor = (float) va_arg( ap, double );
      break;
    default: ///< error
      return -1;
    }

    /// Establish Factor 2 of the expression
    switch( aregEmptyCondNode[nCond].regF2.nType = nType2 ) {
    case 0: ///< pExpr
      aregEmptyCondNode[nCond].regF2.uniType.pExpr = va_arg( ap, int );
      break;
    case 1: ///< pSlot
      aregEmptyCondNode[nCond].regF2.uniType.pSlot = (void *) Storage_Get_Relative( tbn.stdata, va_arg( ap, void * ));
      break;
    case 2: ///< nFunctId
      aregEmptyCondNode[nCond].regF2.uniType.nFunctId = va_arg( ap, int );
      break;
    case 3: ///< nFactor
      aregEmptyCondNode[nCond].regF2.uniType.nFactor = va_arg( ap, int );
      break;
    case 4: /// fFactor
      aregEmptyCondNode[nCond].regF2.uniType.fFactor = (float) va_arg( ap, double );
      break;
    default: ///< error
      return -1;
    }

    va_end( ap );

    return nCond;
  }
}


/************************* Meta_Rule_Create_Act ***************************/
/**
 * \fn Meta_Rule_Create_Act
 * 
 * \brief Reserves an act node, and returns its id or (-1) if error.
 */
int Meta_Rule_Create_Act(int nAct, ... ) { //int nArg1, float nArg2) {
  va_list ap;
  int nActNode, nActNode2;

  va_start( ap, nAct );

  if (( nActNode = Meta_Rule_Get_Empty_Act_Node () ) < 0 )
    return nActNode;
  else {
    
    aregEmptyActNode[nActNode].pNext = -1;
    
    switch ( aregEmptyActNode[nActNode].nAct = nAct ) {
    case 0: ///< Assign
      aregEmptyActNode[nActNode].regArg1.pSlot = (void *) Storage_Get_Relative( tbn.stdata, va_arg( ap, void * ));
      aregEmptyActNode[nActNode].regArg2.pExpr = va_arg( ap, int );
      break;
    case 1: ///< Function Call
      aregEmptyActNode[nActNode].regArg1.nFunctId = va_arg( ap, int );
      break;
    case 2: ///< TriggerSporadicLevel
      aregEmptyActNode[nActNode].regArg1.nLevelId = va_arg( ap, int );
      break;
    case 3: ///< SetReactivityDegree
      aregEmptyActNode[nActNode].regArg1.nReactDegree = va_arg( ap, int );
      break;
    case 4: ///< SetPolicy
      aregEmptyActNode[nActNode].regArg1.nSlsPolicy = va_arg( ap, int );
      break;
    case 5: ///< SetFocus
      aregEmptyActNode[nActNode].regArg1.nAttentionFocus = va_arg( ap, int );
      aregEmptyActNode[nActNode].regArg2.fFocusAttentDegree = (float) va_arg( ap, double );
      break;
    case 6: ///< SetAttentionDegree
      aregEmptyActNode[nActNode].regArg1.fAttentionDegree = (float) va_arg( ap, double );
      break;
    case 7: ///< SetBehaviour
      aregEmptyActNode[nActNode].regArg1.nBehav = va_arg( ap, int );
      break;
    case 8: ///< KDM_Add_Event
      aregEmptyActNode[nActNode].regArg1.pSlot = (void *) Storage_Get_Relative( tbn.stdata, va_arg( ap, void * ));      
      aregEmptyActNode[nActNode].regArg2.iType = va_arg( ap, int );

      if (( nActNode2 = Meta_Rule_Get_Empty_Act_Node () ) < 0 )
        return nActNode2;
      else {
        aregEmptyActNode[nActNode].nAct = va_arg( ap, int );
      }
      break;
    case 9: ///< KDM_Release_Event
      aregEmptyActNode[nActNode].regArg1.pSlot = (void *) Storage_Get_Relative( tbn.stdata, va_arg( ap, void * ));
      aregEmptyActNode[nActNode].regArg2.iType = va_arg( ap, int );
      break;
    case 10: ///< Trigger_Optional
      aregEmptyActNode[nActNode].regArg1.fGain = (float) va_arg( ap, double );
      break;
    default: ///< Error
      va_end( ap );
      return -1;
    }

    va_end( ap );

    return nActNode;
  }
}


/****************** Meta_Rule_Activate_Attention_Focus ***********************/
/**
 * \fn Meta_Rule_Activate_Attention_Focus
 *
 * Creates a new Active Attention Focus with the given parameters and creates
 *  a new element in the list, placing it at the head of the list 
 *  Returns three possible values:
 *      -2, if the mutex is occupied 
 *      -1, if there is no empty regActiveAttFocusNode
 *      0, The new active attention focus has been inserted
 * The arguments are:
 *      - nAF: Index of the regAttentionFocus
 *      - fAttentDegree: Attention Degree of the Attention Focus to activate 
 */
int Meta_Rule_Activate_Attention_Focus( int nAF, float fAttentDegree ) {
  int nAttFocusNode;

  /// aregEmptyAttFocusNode has an empty node
  if (( nAttFocusNode = Meta_Rule_Get_Empty_Att_Focus_Node()) >= 0) {
    if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::ActivateAttentionFocus ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
      return -2;
    }
    else *piMRMutex = 1; ///< And now, it is mine

    aregEmptyAttFocusNode[nAttFocusNode].fAttentionDegree = fAttentDegree;
    aregEmptyAttFocusNode[nAttFocusNode].pAttFocus = nAF;
    paregActiveAttFocus->nActive++;
    aregEmptyAttFocusNode[nAttFocusNode].pNext = paregActiveAttFocus->pFirst;
    paregActiveAttFocus->pFirst = nAttFocusNode;
    
#ifndef MODULE
    fprintf(stderr, "\t(meta-rule_store.c)::ActivateAttentionFocus %d \n", nAF);
#endif

    *piMRMutex = 0; ///< If U want it, U got it

    return 0;
  }
  else return -1;
}


/***************** Meta_Rule_Deactivate_Attention_Focus ********************/
/**
 * \fn Meta_Rule_Deactivate_Attention_Focus
 *
 * Destroys the Active Attention Focus corresponding to the attention 
 *  focus indicated by nAF 
 *  Returns three possible values:
 *      -2, if the mutex is occupied 
 *      -1, if there is no Active Attention Focus corresponding to nAF
 *      0, The active attention focus has been destroyed
 * The argument is nAF, the Index of the regAttentionFocus
 */
int Meta_Rule_Deactivate_Attention_Focus( int nAF ) {
  int nAttFocusNode, nAttFocusNodePrev;

  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::DeactivateAttentionFocus ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
#endif
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
  rtl_printf("\t(meta_rule_store.c)::DeactivateAttentionFocus \n");      
#else 
  fprintf(stderr, "\t(meta-rule_store.c)::DeactivateAttentionFocus \n");
#endif
#endif

  if (( nAttFocusNodePrev = paregActiveAttFocus->pFirst) == -1) {

#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
          rtl_printf("\t\t\t There is no active attention focus \n");      
#else 
          fprintf(stderr, "\t\t\t There is no active attention focus \n");      
#endif
#endif

          *piMRMutex = 0; ///< If U want it, U got it
          return -1;
  }
  else ///< There are at least one active attention focus
    if ( aregEmptyAttFocusNode[nAttFocusNodePrev].pAttFocus == nAF ) {
      ///< The first one is the one we are looking for

#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
        rtl_printf("\t\t\t The 1st one is the one we are looking for \n");      
#else 
        fprintf(stderr, "\t\t\t The 1st one is the one we are looking for \n");      
#endif
#endif

        paregActiveAttFocus->pFirst = 
                               aregEmptyAttFocusNode[nAttFocusNodePrev].pNext;

        if (Meta_Rule_Release_Empty_Att_Focus_Node( nAttFocusNodePrev ) < 0) {
          *piMRMutex = 0;
          return -1;
        }
        else {
          paregActiveAttFocus->nActive--;
          *piMRMutex = 0;
          return 0;
        }
    }
    else { 
          if (( nAttFocusNode = aregEmptyAttFocusNode[nAttFocusNodePrev].pNext ) 
             == -1 ) {
             /**
              *  There is only one active attention focus, 
              *      and it is not the one we are looking for
              */

#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
            rtl_printf("\t\t\t There is only one active, and it is no the one we are looking for, it is the %d\n",
                       nAttFocusNodePrev);      
#else 
            fprintf(stderr, "\t\t\t There is only one active, and it is no the one we are looking for, it is the %d\n",
                       nAttFocusNodePrev);      
#endif
#endif

             *piMRMutex = 0;
             return -1;
           }
          else { ///< We must look for the one
                  while (( nAttFocusNode != -1 ) &&
                      ( aregEmptyAttFocusNode[nAttFocusNode].pAttFocus != nAF )){
                     nAttFocusNodePrev = nAttFocusNode;
                     nAttFocusNode = aregEmptyAttFocusNode[nAttFocusNode].pNext;
                  }
               
                  if ( nAttFocusNode == -1 ) {  ///< The attention focus was not active

#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
                    rtl_printf("\t\t\t The attention focus was not active \n");      
#else 
                    fprintf(stderr, "\t\t\t The attention focus was not active \n");      
#endif
#endif

                    *piMRMutex = 0;
                    return -1;
                  }
                  else { ///< Deactivate the Attention Focus

#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
                         rtl_printf("\t\t\t Deactivate the attention focus \n");      
#else 
                         fprintf(stderr, "\t\t\t Deactivate the attention focus \n");      
#endif
#endif

                         aregEmptyAttFocusNode[nAttFocusNodePrev].pNext = 
                                    aregEmptyAttFocusNode[nAttFocusNode].pNext;
     

                         if (Meta_Rule_Release_Empty_Att_Focus_Node( nAttFocusNode ) < 0) {
                           *piMRMutex = 0;
                           return -1;
                         }
                         else {
                           paregActiveAttFocus->nActive--;
                           *piMRMutex = 0;
                           return 0;
                         }


                  }
           }
    
          *piMRMutex = 0; ///< If U want it, U got it

    return 0;
    }
}


/*************************** Meta_Rule_Create ********************************/
/**
 * \fn Meta_Rule_Create
 * 
 * \brief Creates a new Meta-Rule node with the given parameters.
 *
 *  Returns three possible values:
 *        -2, if the mutex is occupied
 *        -1, if there is no empty regpMetaRuleNode
 *         n, where aregEmptypMetaRuleNode[n] has the Meta-Rule pointer, and
 *                   aregMetaRule[n] has the corresponding Meta-Rule 
 * The arguments are:
 *      - nET: Event Type (usually MODIFICATION, but who knows...
 *      - pSlot: Pointer to the slot
 *      - nCond: Index of the regExprNode
 *      - nAct: Index of the regActNode 
 */
int Meta_Rule_Create ( int nET, void * pSlot, int nCond, int nAct ) {
  int nMetaRule, nEvType, i, npSlot, nAux, nAuxSlot; 

  if (( nMetaRule = Meta_Rule_Get_Empty_pMetaRuleNode () ) < 0 ) {
        return -1;
  } 
  else if (( nEvType = Meta_Rule_Get_Empty_Event_Type () ) < 0 ) {
          return -1;
	}
	else if (( npSlot = Meta_Rule_Get_Empty_pSlot () ) < 0 ) {
		return -1;
	     }
	     else {   
                   /** 
                    *   aregEmptypMetaRuleNode[nMetaRule] has an empty 
                    *    Meta-Rule ptr node, and aregMetaRule[nMetaRule] 
                    *    has an empty Meta-Rule 
                    *   All the existing event types should be created 
                    *    a priori and indexed through 
                    *    aregEventType *paregEventType;
                    */    
                   if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::CreateMetaRule ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
                     return -2;
                   }
                   else *piMRMutex = 1; ///< And now, it is mine

                   aregEmptyEventType[nEvType].nEventType = nET;
        
                   /**
                    *  Look for the event type in the paregEventType 
                    *    index of Meta-Rules 
                    */
                   for ( i = 0; 
                         (paregEventType[i].nEventType != -1) &&
                           (paregEventType[i].nEventType != nET); 
                         i++ );
                   aregEmptyEventType[nEvType].nEventTypeIndex = i;

                   /**
                    *  The Meta-Rule must be inserted in the index 
                    *    by Event Type 
                    */
                   aregEmptyEventType[nEvType].pNext = 
                                              paregEventType[i].pFirst;
                   paregEventType[i].pFirst = nMetaRule;
    
                   aregEmptypSlot[npSlot].pSlot = (pSlot == NULL)?(void*) NULL:(void *) Storage_Get_Relative( tbn.stdata, pSlot );

                   /**
                    *  Look for the pSlot in the paregpSlot index of 
                    *      Meta-Rules 
                    */      
                   for ( i = 0; 
                           ( i < NMAXMETARULES ) && 
                           (( nAux = paregpSlot[i].pFirst ) != -1 ) &&
                           (( nAuxSlot = aregMetaRule[nAux].pregEventSlot ) 
                                                                      != -1 ) &&
                           ( aregEmptypSlot[nAuxSlot].pSlot != (void *) Storage_Get_Relative( tbn.stdata, pSlot ) ); 
                         i++ );
        
                   /**
                    *  The new Meta-Rule is the new 1st one in the index of
                    *      Meta-Rules with an event associated to that pSlot
                    */   
                   aregEmptypSlot[npSlot].pNext = paregpSlot[i].pFirst;
                   paregpSlot[i].pFirst = nMetaRule;
                   paregpSlot[i].nSlots ++;
                   aregEmptypSlot[npSlot].npSlotIndex = i;
    
                   /// Fulling the empty Meta-Rule
                   aregMetaRule[nMetaRule].pregEvent = nEvType;
                   aregMetaRule[nMetaRule].pregEventSlot = npSlot;
                   aregMetaRule[nMetaRule].pregCondition = nCond;
                   aregMetaRule[nMetaRule].pregAction = nAct;
    
                   *piMRMutex = 0; ///< If U want it, U got it

                   return nMetaRule;
                 }
}
 

/**
 * \fn Meta_Rule_Insert_Cond
 *
 * \brief Inserts the expr. identified by nCond as part of the condition of 
 *        the meta-rule identified by nMetarule (in the last position).
 */
extern int Meta_Rule_Insert_Cond ( int nCond, int nMetaRule ) {
  int nAuxCond, nAuxCond2;

  if ((nAuxCond = aregMetaRule[nMetaRule].pregCondition) == -1) { 
    /// The Meta-Rule does not exist
    return -1;
  }
  else {
    /// Look for the place to insert the expr.
    for ( nAuxCond2 = aregEmptyCondNode[nAuxCond].pNext;
          nAuxCond2 != -1; 
          nAuxCond = nAuxCond2 );

    aregEmptyCondNode[nAuxCond].pNext = nCond;

    return 0;
  }
}


/** 
 * \fn Meta_Rule_Insert_Act
 *
 * \brief Inserts the action identified by nAct as part of the actions of
 *        the meta-rule identified by nMetaRule (in the last position).
 */
extern int Meta_Rule_Insert_Act ( int nAct, int nMetaRule ) {
  int nAuxAct, nAuxAct2;

  if ((nAuxAct = aregMetaRule[nMetaRule].pregAction) == -1) { 
    /// The Meta-Rule does not exist
    return -1;
  }
  else {
    /// Look for the place to insert the action

#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
    rtl_printf("\t\t\t (Meta_Rule_Insert_Act) 1st act. of the meta-rule: %d\n", nAuxAct);      
#else 
    fprintf(stderr , "\t\t\t (Meta_Rule_Insert_Act) 1st act. of the meta-rule: %d\n", nAuxAct);
#endif
#endif    

    for ( nAuxAct2 = aregEmptyActNode[nAuxAct].pNext;
          nAuxAct2 != -1; 
          nAuxAct = nAuxAct2, nAuxAct2 = aregEmptyActNode[nAuxAct].pNext );

#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifdef MODULE
    rtl_printf("\t\t\t (Meta_Rule_Insert_Act) %d Act. of the meta-rule will be inserted after: %d\n", 
               nAct, nAuxAct);      
#else 
    fprintf(stderr , "\t\t\t (Meta_Rule_Insert_Act) %d Act. of the meta-rule will be inserted after: %d\n", 
            nAct, nAuxAct);
#endif
#endif    

    aregEmptyActNode[nAuxAct].pNext = nAct;

    return 0;
  }
}


/***************** Meta_Rule_Insert_In_Att_Focus *****************************/
/**
 * \fn Meta_Rule_Insert_In_Att_Focus
 *
 * Inserts a Meta-Rule node within the Attention Focus indicated.
 *  Returns three possible values:
 *      -2, if the mutex is occupied 
 *      -1, if the mutex is occupied 
 *      0, The Meta-Rule node has been inserted
 * The arguments are:
 *      - nMetarule, where aregEmptypMetaRuleNode[nMetaRule] has the Meta-Rule 
 *                   pointer, and aregMetaRule[nMetaRule] has the 
 *                   corresponding Meta-Rule  
 *      - nAF: Index of the regAttentionFocus
 */
int Meta_Rule_Insert_In_Att_Focus( int nMetaRule, int nAF ) {

  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::InsertMetaRuleInAttFocus ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -1;
  }
  else *piMRMutex = 1; ///<*** And now, it is mine

  /**
   * Adding the Meta-Rule as the head of the Meta-Rule's 
   *      list of the Attention Focus
   */
  aregEmptypMetaRuleNode[nMetaRule].pMetaRule = nMetaRule;
  aregEmptypMetaRuleNode[nMetaRule].pNext = pregAttentionFocus[nAF].pFirst;  
  pregAttentionFocus[nAF].pFirst = nMetaRule;
  pregAttentionFocus[nAF].nMetaRules ++;
  
#ifdef MODULE
  rtl_printf("\t(meta_rule_store.c)::InsertMetaRuleInAttFocus\n");      
#endif
  
  *piMRMutex = 0; ///< If U want it, U got it
  
  return 0;
}


/**
 * \fn Meta_Rule_Print_Active_Att_Focus
 *
 * \brief This function allows to print the contents of the active attention 
 *        focuses in the console (RT-LINUX) or in the error standard output (LINUX).
 */
void Meta_Rule_Print_Active_Att_Focus(void) {
   int i, j, 
     nIndex = paregActiveAttFocus->pFirst, ///< First Active Att. Focus
     nActFocus = paregActiveAttFocus->nActive, ///< Number of Active Attention Focuses
     nAttFocus, ///< Attention Focus
     nMetaRules, ///< Number of Meta-Rules in the Attention Focus
     nMRIndex,  ///< Index of the Meta-Rules in the Attention Focus 
     nMRAux;  ///< Actual Meta-Rule being printed in the Attention Focus
   
#ifdef ARTIS_CONF_META_RULE_DEBUG_1 
#ifndef MODULE
   fprintf( stderr, 
            //printf(
            "\n(meta-rule_store.c)::printActiveAttFocus(): Number of active Focus: %d\n",
            nActFocus );
#else
   rtl_printf("(meta-rule_store.c): Act. Focus: %d", nActFocus);
#endif
#endif
   
   ///< Journey through the Active Attention Focuses
   for ( i = 1; i <= nActFocus; i++ ) {
     /// Journey through the Meta-Rules of the Attention Focus
     nAttFocus = aregEmptyAttFocusNode[nIndex].pAttFocus;
     nMetaRules = pregAttentionFocus[nAttFocus].nMetaRules;
     nMRIndex = pregAttentionFocus[nAttFocus].pFirst;
#ifdef ARTIS_CONF_META_RULE_DEBUG_1 
#ifndef MODULE
     fprintf( stderr, "\tActive Attention Focus %d: \tAttention Degree: %f\n",
              nAttFocus, aregEmptyAttFocusNode[nIndex].fAttentionDegree );
     fprintf( stderr,
              //printf( 
              "\t\tThere are %d Meta-Rules:\n", nMetaRules);
#else
     rtl_printf("        Active Attention Focus %d:        Attention Degree: %d", 
                nAttFocus, aregEmptyAttFocusNode[nIndex].fAttentionDegree);
     rtl_printf("        Meta-Rules: %d", nMetaRules);                         
#endif
#endif
     
     for ( j = 1; j <= nMetaRules; j++ ) {
       nMRAux = aregEmptypMetaRuleNode[nMRIndex].pMetaRule; 
#ifdef ARTIS_CONF_META_RULE_DEBUG_1
#ifndef MODULE
       fprintf( stderr,
                //printf( 
                "\t\t\t%d: Event: %d\n\t\t\t    Action: %d %d \n", j,
                aregEmptyEventType[aregMetaRule[nMRAux].pregEvent].nEventType,
                aregEmptyActNode[aregMetaRule[nMRAux].pregAction].nAct,
                aregEmptyActNode[aregMetaRule[nMRAux].pregAction].regArg1.nSlsPolicy );
#else
       rtl_printf("    %d: Event: %d      Action: %d %d \n", j,
                  aregEmptyEventType[aregMetaRule[nMRAux].pregEvent].nEventType,
                  aregEmptyActNode[aregMetaRule[nMRAux].pregAction].nAct,
                  aregEmptyActNode[aregMetaRule[nMRAux].pregAction].regArg1.nSlsPolicy );
#endif
#endif
       nMRIndex = aregEmptypMetaRuleNode[nMRIndex].pNext;
       
#ifndef MODULE
       fprintf(stderr, "Let's go once again\n");
#endif
     }
#ifndef MODULE
     fprintf(stderr, "Let's go once again to the next AttFocus.\n");
#endif
     nIndex = aregEmptyAttFocusNode[nIndex].pNext;
   }
}


/* ==================================================================== */


#ifdef MODULE
int Meta_Rule_init_module(void) {
  rtl_printf("META-RULE STORE STARTING\n");
  return 0;
}


void Meta_Rule_cleanup_module(void) {
  mbuff_detach(META_RULE_STORE, (void *)piMRMutex);
}
#endif


//@}


/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
