/**
 * \file meta_rule_interpret.c
 * \date 2004
 * \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 * \version 2.0
 * 
 * \brief Meta-Rule Interpreter 2.0
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

#include "../../events/include/events/events_internal.h"  /// Definition of Event_T

#include "../../include/meta_rule.h"
#include "../include/meta_rule/meta_rule_store.h"
#include "../include/meta_rule_actions.h"

#include "../../include/kdm.h"   /// KDM_Add_Event & KDM_Release_Event

///#include "../../kdm/include/kdm.h"      /// KDM_Add_Event & KDM_Release_Event

#ifdef MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>

#include <rtl.h> /// rtl_printf

#else
#include <stdio.h>
#endif

#include "../../common/include/artis_conf.h"  ///< ARTIS_TRACE_EVENTS definition 
#include "../../common/include/artis_event_definition.h"  ///< Trace Event definitions
#include "../../trace/include/trace.h"

#define INTER_DEBUG


/**
 * \fn EvalFactor
 *
 * \brief Evaluate the factor leaving the result in a double valor 
 */
double EvalFactor ( regTypeFactor regFactor, 
		    Event_T regEvent ) {
    double dFactor = 0;
    static_slot *slotAux;
    int nType;

    int EvalExpr ( int nCond, Event_T regEvent );

    switch( regFactor.nType ) {
    case 0: ///< --> pExpr
      dFactor = (double) EvalExpr(regFactor.uniType.pExpr, regEvent);
      break;
      
    case 1: ///< --> pSlot
      if ( regFactor.uniType.pSlot != regEvent.pSlot ) {
	/**
	 *  Reference to an slot that is not the one 
	 *    generating the event
	 */
                   
	/** 
	 *  1st approach: I suppose it won't work with arrays 
	 */
	/**
	 *  It should be checked if this function returns
	 *     KDM_GENERIC_ERROR or KDM_OK
	 */
       	slotAux = ((static_slot *) Storage_Get_Address( tbn.stdata, regFactor.uniType.pSlot ));
      }
      else slotAux = (static_slot *) Storage_Get_Address( tbn.stdata, regEvent.pSlot );

	switch ( slotAux->type ) {
	case KDM_INT:
	  dFactor = (double) slotAux->value.i;
#ifdef INTER_DEBUG
#ifndef MODULE
	  fprintf(stderr, "\t(meta-rules_interpret.c)::EvalFactor) Meta-Rule' nFactor value: %.2lf\n",
		  dFactor );
#endif
#endif
	  break;

	  /***
	      case KDM_SHORT_INT:
	      dFactor = (double) regEvent.New_Value.si;
	      break;
	      case KDM_LONG_INT:
	      dFactor = (double) regEvent.New_Value.li;
	      break;
	      case KDM_UNSIGNED_INT:
	      dFactor = (double) regEvent.New_Value.ui;
	      break;
	      case KDM_DOUBLE:
	      dFactor = (double) regEvent.New_Value.d;
	      break;
	      case KDM_HANDLE:
	      break;
	  ***/
	case KDM_FLOAT:
	  dFactor = (double) slotAux->value.f;
#ifdef INTER_DEBUG
#ifndef MODULE
	  fprintf(stderr, "\t(meta-rules_interpret.c)::EvalFactor) Meta-Rule' nFactor value: %.2lf\n",
		  dFactor );
#endif
#endif
	  break;
	case KDM_CHAR:
	  dFactor = 
	    (double) slotAux->value.c;
#ifdef INTER_DEBUG
#ifndef MODULE
	  fprintf(stderr, "\t(meta-rules_interpret.c)::EvalFactor) Meta-Rule' nFactor value: %.2lf\n",
		  dFactor );
#endif
#endif
	  break;
	default:
#ifdef INTER_DEBUG
#ifndef MODULE
	  fprintf(stderr, "\t(meta-rules_interpret.c)::EvalFactor) KDM_GENERIC_ERROR, con type: %d!!!\n",
		  regEvent.iDataType);
#endif
#endif
	  return -1;  ///< KDM_GENERIC_ERROR;
	} /* switch */
      
#ifdef INTER_DEBUG
#ifndef MODULE
      fprintf(stderr, "\t(meta-rules_interpret.c)::EvalFactor) Meta-Rule' slot value: %.2lf\n",
	      dFactor );
#endif
#endif
      break;
    case 2: ///< --> nFunctId 
      /// \todo ==> What kind of functions there are?
      /***regFactor.uniType.nFunctId;***/  
      break;
    case 3: ///< --> nFactor
      dFactor = (double) (1.0 * regFactor.uniType.nFactor);
#ifdef INTER_DEBUG
#ifndef MODULE
      fprintf(stderr, "\t(meta-rules_interpret.c)::EvalFactor) Meta-Rule' nFactor value: %d, %.2lf\n",
	      regFactor.uniType.nFactor, dFactor );
#endif
#endif
      break;
    case 4: ///< --> fFactor
      dFactor = (double) regFactor.uniType.fFactor;
#ifdef INTER_DEBUG
#ifndef MODULE
      fprintf(stderr, "\t(meta-rules_interpret.c)::EvalFactor) Meta-Rule' nFactor value: %.2lf\n",
	      dFactor );
#endif
#endif
      break;
    }        
    
    return dFactor;
}


/**
 * \fn EvalExpr
 *
 *  Evaluates the expression in aregEmptyCondNode[nCond] 
 *   It returns:
 *        - 0, if the expr. is evaluated to false.
 *        - <> 0, if the expr. is evaluated to true.
 */
int EvalExpr ( int nCond, Event_T regEvent ) {
  int bCond = -1;
  regTypeFactor regFactor1, regFactor2;
  double dFactor1, dFactor2; ///< There are the most generic/big type

    if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef INTER_DEBUG
#ifdef MODULE
    rtl_printf("\t(meta_rules_interpret.c)::EvalExpr ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
#endif
    return -2;
  }
    else *piMRMutex = 1; ///< And now, it is mine

    
    regFactor1 = aregEmptyCondNode[nCond].regF1;
    regFactor2 = aregEmptyCondNode[nCond].regF2;
    
    /// Evaluate the 1st factor of the expr., leaving the result in dFactor1
    *piMRMutex = 0;
    dFactor1 = (double) EvalFactor( regFactor1, regEvent );
    *piMRMutex = 1;

    /// Evaluate the 2nd factor of the expr., leaving the result in dFactor2
    *piMRMutex = 0;
    dFactor2 = (double) EvalFactor( regFactor2, regEvent );
    *piMRMutex = 1;
    
    switch (aregEmptyCondNode[nCond].nOp) {
    case 0: ///< --> no operation
      bCond = dFactor1;                          
      break;
    case 1: ///< --> <>
      bCond = ( dFactor1 != dFactor2 );
      break;                
    case 2: ///< --> <
      bCond = ( dFactor1 < dFactor2 ); 
      break;                
    case 3: ///< --> >
      bCond = ( dFactor1 > dFactor2 ); 
      break;                
    case 4: ///< --> >=
      bCond = ( dFactor1 >= dFactor2 ); 
      break;                
    case 5: ///< --> <=
      bCond = ( dFactor1 <= dFactor2 ); 
      break;                
    case 6: ///< --> = (compare)
      bCond = ( dFactor1 == dFactor2 ); 
      break;                
    case 7: ///< --> +
      bCond = ( dFactor1 + dFactor2 ); 
      break;                
    case 8: ///< --> -
      bCond = ( dFactor1 - dFactor2 ); 
      break;                
    case 9: ///< --> *
      bCond = ( dFactor1 * dFactor2 );
      break;                
    case 10: ///< --> /
      bCond = ( dFactor1 / dFactor2 );
      break;                
    }
    
    *piMRMutex = 0;
    return bCond;
}


/**
 * \fn Meta_Rule_Execute_Action
 *
 * \brief Executes the action in aregEmptyActNode[nAct] 
 *
 * Valid data combinations for the regActNode:
 *      nAct (int -> action code)     regArg1                regArg2
 *      0 -> Assign                   pSlot (relative ptr)   pExpr
 *      1 -> Function Call            nFunctId               ---
 *      2 -> TriggerSporadicLevel     nLevelId               ---
 *      3 -> SetReactivityDegree      nReactDegree           ---
 *      4 -> SetPolicy                nSlsPolicy             ---
 *      5 -> SetFocus                 nAttentionFocus        fFocusAttentDegree
 *      6 -> SetAttentionDegree       fAttentionDegree       ---
 *      7 -> SetBehaviour             nBehav                 ---
 *      8 -> KDM_Add_Event            pSlot                  iType       iImp (nAct from the 
 *                                                                            next regActNode)
 *      9 -> KDM_Release_Event        pSlot                  iType
 *     10 -> Trigger_Optional         fGain                  ---
 */
void Meta_Rule_Execute_Action( int nAct, Event_T regEvent ) {
  int nAuxRes;
  
  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef INTER_DEBUG
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyAttFocusNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
#endif
    //return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

#ifdef ARTIS_TRACE_EVENTS
#ifdef MODULE
  trace_event( META_RULE_ACTION, 0, nAct, 0 ); 
#else
  trace_event( META_RULE_ACTION, 1, nAct, 0 ); 
#endif
#endif
  
  switch ( aregEmptyActNode[nAct].nAct ) {
  case 0: ///< -> Assign pSlot pExpr
    KDM_Set_Value( 
		 *((static_slot *) Storage_Get_Address( tbn.stdata, 
		  	((static_slot *) aregEmptyActNode[nAct].regArg1.pSlot ))), 
		  EvalExpr( aregEmptyActNode[nAct].regArg2.pExpr,
			    regEvent )                           
		  );
    break;
  case 1: /// -> Function Call nFunctId ---
    /// \todo How the ... are we going to manage this?
    break;                
  case 2: ///< -> TriggerSporadicLevel nLevelId ---
#ifdef INTER_DEBUG
#ifndef MODULE
    fprintf(stderr, "\t(meta-rules_interpret.c::ExecuteAction) TriggerSporadicLevel\n");
#endif
#endif
    MR_Action_Trigger_Sporadic_Level( aregEmptyActNode[nAct].regArg1.nLevelId );
    break;                
  case 3: ///< -> SetReactivityDegree nReactDegree ---
#ifdef INTER_DEBUG
#ifndef MODULE
    fprintf(stderr, 
	    "\t(meta-rules_interpret.c::ExecuteAction) SetReactivityDegree %d\n", 
	    (aregEmptyActNode[nAct].regArg1.nReactDegree) );
#endif
#endif
    MR_Action_Set_Reactivity_Degree( aregEmptyActNode[nAct].regArg1.nReactDegree );
    break; 
  case 4: ///< -> SetPolicy nSlsPolicy ---
    MR_Action_Set_Policy( (int) (aregEmptyActNode[nAct].regArg1.nSlsPolicy) );
    break; 
  case 5: ///< -> SetFocus nAttentionFocus fFocusAttentDegree
    if ( aregEmptyActNode[nAct].regArg2.fFocusAttentDegree == 0) {
      
      *piMRMutex = 0; ///< To let myself enter again to change the AF status

      nAuxRes = Meta_Rule_Deactivate_Attention_Focus( aregEmptyActNode[nAct].regArg1.nAttentionFocus );

      *piMRMutex = 1; ///< Let's go on 

#ifdef INTER_DEBUG
#ifndef MODULE
      fprintf(stderr, 
	      "\t(meta-rules_interpret.c::ExecuteAction) DeactivateAttentionFocus %d exit status: %d\n", 
	      (aregEmptyActNode[nAct].regArg1.nAttentionFocus), nAuxRes );
#endif
#endif
    }
    else {
      
      *piMRMutex = 0; ///< To let myself enter again to change the AF status

      Meta_Rule_Activate_Attention_Focus( aregEmptyActNode[nAct].regArg1.nAttentionFocus,
					  aregEmptyActNode[nAct].regArg2.fFocusAttentDegree );
      
      *piMRMutex = 1; ///< Let's go on 
      
#ifdef INTER_DEBUG
#ifndef MODULE
      fprintf(stderr, "\t(meta-rules_interpret.c::ExecuteAction) ActivateAttentionFocus %d\n", 
	      (aregEmptyActNode[nAct].regArg1.nAttentionFocus) );
#endif
#endif
    }
    break; 
  case 6: ///< -> SetAttentionDegree fAttentionDegree ---
    MR_Action_Set_Attention_Degree ( aregEmptyActNode[nAct].regArg1.fAttentionDegree );
#ifdef INTER_DEBUG
#ifndef MODULE
    fprintf(stderr, "\t(meta-rules_interpret.c::ExecuteAction) SetAttentionDegree %f\n", 
	    (float) (aregEmptyActNode[nAct].regArg1.fAttentionDegree) );
#endif
#endif
    break; 
  case 7: ///< -> SetBehaviour nBehav ---
    MR_Action_Set_Behaviour( aregEmptyActNode[nAct].regArg1.nBehav );
#ifdef INTER_DEBUG
#ifndef MODULE
    fprintf(stderr, "\t(meta-rules_interpret.c::ExecuteAction) SetBehaviour %d\n", 
	    (aregEmptyActNode[nAct].regArg1.nBehav) );
#endif
#endif
    break; 
  case 8: ///< -> KDM_Add_Event pSlot iType iImp (nAct from the next regActNode)
    /***/
    KDM_Add_Event( aregEmptyActNode[nAct].regArg1.pSlot, 
		   aregEmptyActNode[nAct].regArg2.iType, 
		   aregEmptyActNode[ aregEmptyActNode[nAct].pNext ].nAct );
    /***/
    break;
  case 9: ///< -> KDM_Release_Event pSlot iType
    /***/
    KDM_Release_Event( aregEmptyActNode[nAct].regArg1.pSlot, 
		       aregEmptyActNode[nAct].regArg2.iType );
    /***/
    break;
  case 10: ///< -> Trigger_Optional fGain
#ifdef INTER_DEBUG
#ifndef MODULE
    fprintf(stderr, "\t(meta-rules_interpret.c::ExecuteAction) Trigger Optional %f\n", 
	    (float) (aregEmptyActNode[nAct].regArg1.fGain) );
#endif
#endif
    MR_Action_Trigger_Optional( aregEmptyActNode[nAct].regArg1.fGain, regEvent );
    break;

    
  }
  
  *piMRMutex = 0;
}


/**
 * \fn InterpretMR
 *
 * Evaluates the condition and if it is true, executes the actions associated
 *      to the Meta Rule.
 *   It should receive also the Event_T received in the event as a parameter 
 ***/
void InterpretMR ( int nMetaRule, Event_T regEvent ) {
  int nCond, bCond, ///< To calculate if the cond is true
    nAct;
  
  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef INTER_DEBUG
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyAttFocusNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
#endif
    //return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

  /// Check if the condition is fullfilled
  if (( nCond = aregMetaRule[nMetaRule].pregCondition ) < 0 )
    /// Empty condition meta-rule
    bCond = 1;
  else
    do {
      *piMRMutex = 0;    
      bCond = EvalExpr( nCond, regEvent );    
      *piMRMutex = 1;
      
      nCond = aregEmptyCondNode[nCond].pNext; ///< Pass to the next sub-condition
    } while ( bCond && ( nCond != -1 ));
  
  if ( bCond ) {
    /// The condition is true, so we must execute the actions related
    nAct = aregMetaRule[nMetaRule].pregAction;
    
    do {
#ifdef INTER_DEBUG
#ifndef MODULE
      fprintf(stderr, "\t(meta-rules_interpret::InterpretMR) Action: nAct %d\n", nAct);
#endif
#endif
      *piMRMutex = 0; 
      Meta_Rule_Execute_Action( nAct, regEvent );
      *piMRMutex = 1;
      nAct = aregEmptyActNode[nAct].pNext;
    } while ( nAct != -1 );
  }

  *piMRMutex = 0;
}


/**
 * \fn ManageEvent
 *
 * It manages the event in regEvent evaluating as much as nMaxMR meta-rules
 *     associated to this event.
 *   It returns the number of MR it has found and executed.
 */
int ManageEvent ( Event_T regEvent, int nMaxMR ) {
  int nEncontrado, i, nAFNodeAux, j, nAFAux, nMR, pMR, nEvType;
  
  if (*piMRMutex) {  ///< There is someone inside!!!!!!!!!!!
#ifdef INTER_DEBUG
#ifdef MODULE
    rtl_printf("\t(meta-rules_interpret.c)::ManageEvent ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
#endif
    return -2;
  }
  else *piMRMutex = 1; ///< And now, it is mine

#ifdef INTER_DEBUG
#ifndef MODULE
  fprintf( stderr, "\t(meta-rules_interpret.c)::ManageEvent No habia nadie dentro, ahora hay que buscar el evento %d\n",
	   regEvent.iType);
#endif
#endif

  nEncontrado = 0; ///< It counts the number of Meta-Rules for the current event have been found 

  /// If there is no meta-rule associated to this event type, go out
  for ( i = 0; 
	((i <= NEVENTTYPES) && 
	 (paregEventType[i].nEventType != regEvent.iType)); 
	i++ );
  if (paregEventType[i].pFirst == -1 ) {
    *piMRMutex = 0; 
    return nEncontrado;
  }
  
#ifdef INTER_DEBUG
#ifndef MODULE
  fprintf( stderr, "\t(meta-rules_interpret.c)::ManageEvent There are meta-rules associated\n");
#endif
#endif
  
  /// Looking through the active Att. Focuses
  for ( i = 1, nAFNodeAux = paregActiveAttFocus->pFirst;
	(( i <= paregActiveAttFocus->nActive) && ( nEncontrado < nMaxMR ));
	i++, nAFNodeAux = aregEmptyAttFocusNode[nAFNodeAux].pNext ) {
    
    /// Looking through the MRs of the AF
    for ( j = 1, nAFAux = aregEmptyAttFocusNode[nAFNodeAux].pAttFocus,
	    nMR = pregAttentionFocus[nAFAux].pFirst;
	  (( j <= pregAttentionFocus[nAFAux].nMetaRules) && (nEncontrado < nMaxMR ));
	  j++, nMR = aregEmptypMetaRuleNode[nMR].pNext ) {
      pMR = aregEmptypMetaRuleNode[nMR].pMetaRule;
      nEvType = aregMetaRule[pMR].pregEvent;
      
      if (( aregEmptyEventType[nEvType].nEventType == regEvent.iType ) &&
	  ( aregEmptypSlot[ aregMetaRule[pMR].pregEventSlot].pSlot  
		== regEvent.pSlot )) {
	/// I' got one more
	nEncontrado ++;
        *piMRMutex = 0;
	
#ifdef INTER_DEBUG
#ifndef MODULE
	fprintf( stderr, "\t(meta-rules_interpret.c)::ManageEvent Interpreting meta-rule number %d of the AF %d\n", pMR, nAFNodeAux);
#endif
#endif
	
	InterpretMR( pMR, regEvent );
	
#ifdef INTER_DEBUG
	Meta_Rule_Print_Active_Att_Focus();
#endif
	
        *piMRMutex = 1;
      }

#ifdef INTER_DEBUG
#ifndef MODULE
      else if ( aregEmptyEventType[nEvType].nEventType != regEvent.iType )
	fprintf( stderr, "\t(meta_rule_interpret.c)::ManageEvent Incorrect Meta-Rule event type: %d\n", 
		 aregEmptyEventType[nEvType].nEventType);
      else if ( aregEmptypSlot[ aregMetaRule[pMR].pregEventSlot].pSlot != regEvent.pSlot )
	fprintf( stderr, "\t(meta_rule_interpret.c)::ManageEvent Incorrect Meta-Rule slot type: %p %p\n", 
		 aregEmptypSlot[ aregMetaRule[pMR].pregEventSlot].pSlot, regEvent.pSlot );
#endif
#endif
    }
  }

  *piMRMutex = 0;

  return nEncontrado;

}
