/*********************** meta_rule_store.c **********************/ 
/*************************** 20040907 ***************************/  


/*  

   FILE:                                
    meta_rule_store.c                                   
                                    
   DESCRIPTION:                             
    Meta-Rule Store Manager 

   Copyright (C) 2004
        Carlos Carrascosa Casamayor             


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

#ifdef MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>

#include <rtl.h> /// rtl_printf

#else
#include <stdio.h>
#endif

///#define NULL (void *)0

///#include "../include/rt_trace.h"

///#include "../include/metareasoning.h"

 
/*** ONLY FOR WINDOWS REHEARSAL ***/
//#include <mbuff.h>
/*** ONLY FOR WINDOWS REHEARSAL ***/



#ifdef MODULE
#include <mbuff.h>
#else

extern int mbuff_alloc(const char *name, int size);
extern int mbuff_free(const char *name, void * mbuf);
extern void * mbuff_attach(const char *name, int size);
extern void mbuff_detach(const char *name, void * mbuf);

#endif

#include "../include/meta_rule/meta_rule_store.h"

#include "../../events/include/events.h"


///#define DEBUG





aregEventType *paregEventType; /*** To index the M-R Store through the
                                *    Event Type.
                                ***/

int *pn1stEmptyEvType;  /*** Ptr to the 1st Empty regEventType ***/
regEventType *aregEmptyEventType; 
                       /*** Ptr to the mbuff section of empty regEventType ***/    


aregpSlot *paregpSlot; /*** To index the M-R Store through the ptr to Slot ***/

int *pn1stEmptypSlt; /*** Ptr to the 1st empty regEventType ***/
regpSlot *aregEmptypSlot;  /*** Ptr to the mbuff section of empty aregpSlot ***/ 


aregActiveAttFocus *paregActiveAttFocus; /*** To index the M-R Store through
                                          *    active Attention Focuses.
                                          ***/
                                          
regActiveAttFocusNode *pregActiveAttFocusNode; 
        /*** To implement the Active Attention Focus Array.
         *      Ptr to the mbuff section.
         ***/
int *pn1stEmptyAttFocusNode; /*** Ptr to the 1st empty regActiveAttFocusNode ***/
regActiveAttFocusNode *aregEmptyAttFocusNode;
        /*** Attention Focus Empty Node Array.
         *      Ptr to the mbuff section of empty regActiveAttFocusNode nodes
         ***/

int *pn1stEmptyAttentionFocus;
        /*** Ptr to the 1st empty regAttentionFocus ***/         
regAttentionFocus *pregAttentionFocus; 
        /*** There is one of this struct for each Attention Focus.
         *    Ptr to the mbuff section of regAttentionFocus.
         ***/


regpMetaRuleNode *pregpMetaRuleNode; 
        /*** Ptr to the mbuff section of regpMetaRuleNodes.
         *      There is one of this structs for each Meta-Rule in an 
         *      Attention Focus
         ***/
int *pn1stEmptypMetaRuleNode; /*** Ptr to the 1st empty regpMetaRuleNode ***/
regpMetaRuleNode *aregEmptypMetaRuleNode;
        /*** Ptr to the mbuff section of empty regpMetaRuleNodes.
         *      This is used to add Meta-Rules to the Attention Focus 
         *      (learning process)
         ***/

regMetaRule *aregMetaRule; /*** Ptr to the mbuff section of regMetaRule.
                            *    There is one of this structs for each Meta-Rule
                            ***/
regMetaRule *aregEmptyMetaRule; 
          /*** Ptr to the mbuff section of empty regMetaRule
           *    This is useless???
           ***/

int *pn1stEmptyCondNode;
        /*** Ptr to the 1st empty regExprNode ***/
regExprNode *aregEmptyCondNode; 
          /*** Ptr to the mbuff section of empty regExprNode.
           *      They are used to construct the condition part of Meta-Rules.
           ***/

int *pn1stEmptyActNode; /*** Ptr to the 1st empty regActNode ***/
regActNode *aregEmptyActNode; /*** Ptr to the mbuff section of empty regActNode.
                               *        They are used to construct the action part of 
                               *        Meta-Rules.
                               ***/

volatile int *piMRMutex;  
                 /*** To control if there's anyone accessing the storage 
                  *            Values:
                  *               - 0 : There's no one accessing
                  *               - 1 : There's someone here
                  ***/


/**************************** InitMetaRuleStore *******************************
 * Initializes the Meta-Rule Store, shaping all its empty nodes 
 *         This is used in the RTOS level, to contact with the memory allocated
 *             by the RTOS in the initialization phase 
 *                                      ==> in the InitBlackboard()?
 ***/
int InitMetaRuleStore(void) {
  int nregExprNode, nregActNode;
  
  nregExprNode = 
          (( 2 * NMAXMETARULES > NEXPRUSED + 2 * (NMAXMETARULES - NMETARULES))?
          2 * NMAXMETARULES:  NEXPRUSED + 2 * (NMAXMETARULES - NMETARULES));
                            
  nregActNode = 
          (( 2 * NMAXMETARULES > NACTUSED + 2 * (NMAXMETARULES - NMETARULES))?
          2 * NMAXMETARULES:  NACTUSED + 2 * (NMAXMETARULES - NMETARULES));

  if (( piMRMutex = ( volatile int * ) mbuff_attach ( META_RULE_STORE, 
                 ( ((int) sizeof( int)) 
                    + ((int) sizeof( aregEventType )* NEVENTTYPES )
                    + ((int) sizeof( regEventType ) * NMAXMETARULES ) 
                    + ((int) sizeof( aregpSlot ) * NMAXMETARULES )  
                    + ((int) sizeof( regpSlot ) * NMAXMETARULES ) 
                    + ((int) sizeof( aregActiveAttFocus ))
                    + ((int) sizeof( regAttentionFocus ) * NATTFOCUS)
                    + ((int) sizeof( regActiveAttFocusNode ) * NACTIVEATTFOCUS )
                    + ((int) sizeof( regpMetaRuleNode ) * NMAXMETARULES )
                    + ((int) sizeof( regMetaRule ) * NMAXMETARULES )
                    + ((int) sizeof( regExprNode ) * nregExprNode )
                    + ((int) sizeof( regActNode ) * nregActNode)
                    + ((int) sizeof( int ) * 7) 
                    + ((int) sizeof( int ) * 100)
                                 ) )) == NULL )

     /*** NEVENTTYPES is a constant declared in <events.h>, 
      *         and it is the number of different event types.
      *   NMAXMETARULES is a constant declared in <meta-rule_store.h>,
      *         and it is the max. number of different meta-rules existing in 
      *         the system at the same time.
      *   NATTFOCUS is a constant declared in ------------------------,
      *         and it represents the number of different attention focuses 
      *         declared.
      *   NACTIVEATTFOCUS is a constant declared in --------------------------,
      *         and it represents the number of different attention focuses that 
      *         are initially active.
      *   NEXPRUSED is a constant declared in ------------------------------,
      *         and it represents the number of regExprNode used in the initial 
      *         Meta-Rule set.
      *   NACTUSED is a constant declared in -------------------------------,
      *         and it represents the number of regActNode used in the initial 
      *         Meta-Rule set.
      *   NMETARULES is a constant declared in -------------------------------,
      *         and it represents the initial number of meta-rules.
      *   7 ints for the n1stEmpty... values
      ***/

    return - 1;  
  else {

         *piMRMutex = 0; /*** Initially there's no one inside ***/

         paregEventType = (aregEventType *) (piMRMutex + 1);
          
         aregEmptyEventType = 
                     (regEventType *) (paregEventType + NEVENTTYPES + 1 );

         paregpSlot = (aregpSlot *) (aregEmptyEventType + NMAXMETARULES + 1);

         aregEmptypSlot = (regpSlot *) (paregpSlot + NMAXMETARULES + 1);
         
         paregActiveAttFocus = 
                     (aregActiveAttFocus *) (aregEmptypSlot + NMAXMETARULES + 1);

         pregActiveAttFocusNode = (regActiveAttFocusNode *) NULL;

         aregEmptyAttFocusNode = 
                             (regActiveAttFocusNode *) (paregActiveAttFocus + 1);

         pregAttentionFocus =  
                (regAttentionFocus *) (aregEmptyAttFocusNode + NACTIVEATTFOCUS + 1);
         
         pregpMetaRuleNode = (regpMetaRuleNode *) NULL;

         aregEmptypMetaRuleNode =  
                      (regpMetaRuleNode *) (pregAttentionFocus + NATTFOCUS + 1);
                /*** There is no meta-rule, so all are empty ***/

         aregMetaRule =  
                   (regMetaRule *) (aregEmptypMetaRuleNode + NMAXMETARULES + 1);
                 
         aregEmptyCondNode = (regExprNode *) (aregMetaRule + NMAXMETARULES + 1 );
            
         aregEmptyActNode =  
                          (regActNode *) (aregEmptyCondNode + nregExprNode + 1);

         pn1stEmptyEvType = (int *) (aregEmptyActNode + nregActNode +1);
         pn1stEmptypSlt = (int *) (pn1stEmptyEvType + 1);
         pn1stEmptyAttFocusNode = (int *) (pn1stEmptypSlt + 1);
         pn1stEmptyAttentionFocus = (int *) (pn1stEmptyAttFocusNode + 1);
         pn1stEmptypMetaRuleNode = (int *) (pn1stEmptyAttentionFocus + 1);
         pn1stEmptyCondNode = (int *) (pn1stEmptypMetaRuleNode + 1);
         pn1stEmptyActNode = (int *) (pn1stEmptyCondNode + 1);

/***
#ifdef MODULE
rtl_printf(" aregEmptyActNode %d\n", (int) aregEmptyActNode );         
#endif
***/
         return( 1 );
  }
}


/*************************** CreateMetaRuleStore ******************************
 * Creates the Meta-Rule Store, shaping all its empty nodes 
 *         This is used in the DS level, to allocate the memory that will
 *             be used by the DS to manage meta-rules.
 ***/
int CreateMetaRuleStore(void) {
  int i, nregExprNode, nregActNode;
  
  nregExprNode = 
           (( 2 * NMAXMETARULES > NEXPRUSED + 2 * (NMAXMETARULES - NMETARULES))?
            2 * NMAXMETARULES:  NEXPRUSED + 2 * (NMAXMETARULES - NMETARULES));
                            
  nregActNode = 
           (( 2 * NMAXMETARULES > NACTUSED + 2 * (NMAXMETARULES - NMETARULES))?
            2 * NMAXMETARULES:  NACTUSED + 2 * (NMAXMETARULES - NMETARULES));

  if (( piMRMutex = ( volatile int * ) mbuff_attach ( META_RULE_STORE, 
                  ( ((int) sizeof( int)) 
                    + ((int) sizeof( aregEventType )* NEVENTTYPES )
                    + ((int) sizeof( regEventType ) * NMAXMETARULES ) 
                    + ((int) sizeof( aregpSlot ) * NMAXMETARULES )  
                    + ((int) sizeof( regpSlot ) * NMAXMETARULES ) 
                    + ((int) sizeof( aregActiveAttFocus ))
                    + ((int) sizeof( regAttentionFocus ) * NATTFOCUS)
                    + ((int) sizeof( regActiveAttFocusNode ) * NACTIVEATTFOCUS )
                    + ((int) sizeof( regpMetaRuleNode ) * NMAXMETARULES )
                    + ((int) sizeof( regMetaRule ) * NMAXMETARULES )
                    + ((int) sizeof( regExprNode ) * nregExprNode )
                    + ((int) sizeof( regActNode ) * nregActNode) 
                    + ((int) sizeof( int ) * 7) 
                    + ((int) sizeof( int ) * 100)
                              ) )) == NULL )

     /*** NEVENTTYPES is a constant declared in <KDM_events.h>, 
      *         and it is the number of different event types.
      *   NMAXMETARULES is a constant declared in <meta-rule_store.h>,
      *         and it is the max. number of different meta-rules existing in 
      *         the system at the same time.
      *   NATTFOCUS is a constant declared in ------------------------,
      *         and it represents the number of different attention focuses 
      *         declared.
      *   NACTIVEATTFOCUS is a constant declared in --------------------------,
      *         and it represents the number of different attention focuses that
      *          are initially active.
      *   NEXPRUSED is a constant declared in ------------------------------,
      *         and it represents the number of regExprNode used in the initial 
      *         Meta-Rule set.
      *   NACTUSED is a constant declared in -------------------------------,
      *         and it represents the number of regActNode used in the initial 
      *         Meta-Rule set.
      *   NMETARULES is a constant declared in -------------------------------,
      *         and it represents the initial number of meta-rules.
      *   7 ints for the n1stEmpty... values
      ***/

    return - 1;  
  else {

         *piMRMutex = 0; /*** Initially there's no one inside ***/

         paregEventType = (aregEventType *) (piMRMutex + 1);
          
         paregEventType[0].nEventType = CREATION;
         paregEventType[1].nEventType = DELETION;
         paregEventType[2].nEventType = MODIFICATION;
         paregEventType[3].nEventType = NEW_PREDICTION;
         paregEventType[4].nEventType = MATCHED_PREDICTION;
         paregEventType[5].nEventType = UNMATCHED_PREDICTION;
         paregEventType[6].nEventType = CONSIDER_PERIODIC_AGENT;
         paregEventType[7].nEventType = TIME_MESSAGE;
         paregEventType[8].nEventType = EXIT_APP;
         for ( i = 0; i < NEVENTTYPES; i++ ) 
           paregEventType[i].pFirst = -1; /*** To indicate it is empty ***/
 
         aregEmptyEventType =  
                           (regEventType *) (paregEventType + NEVENTTYPES + 1 );
         for ( i = 0; i < NMAXMETARULES -1; i++ ) {
            aregEmptyEventType[i].nEventType = 
                                     aregEmptyEventType[i].nEventTypeIndex = -1;
            aregEmptyEventType[i].pNext = i+1;
         }
         aregEmptyEventType[NMAXMETARULES - 1].nEventType = 
                    aregEmptyEventType[NMAXMETARULES - 1].nEventTypeIndex = -1;
         aregEmptyEventType[NMAXMETARULES - 1].pNext = 0;
                  
         paregpSlot = (aregpSlot *) (aregEmptyEventType + NMAXMETARULES + 1);
         for ( i = 0; i < NMAXMETARULES -1; i++ ) {
                  paregpSlot[i].nSlots = 0;
                  paregpSlot[i].pFirst = -1;
         }
         
         aregEmptypSlot = (regpSlot *) (paregpSlot + NMAXMETARULES + 1);

         for ( i = 0; i < NMAXMETARULES - 1; i++ ) {
            aregEmptypSlot[i].pSlot = 
                       (void *) NULL; /*** Initializing all the ptr to slot ***/
            aregEmptypSlot[i].npSlotIndex = -1;
            aregEmptypSlot[i].pNext = i+1;
         }
         
         aregEmptypSlot[NMAXMETARULES - 1].pSlot = (void *) NULL;
         aregEmptypSlot[NMAXMETARULES - 1].pNext = 0;
         
         paregActiveAttFocus = 
                     (aregActiveAttFocus *) (aregEmptypSlot + NMAXMETARULES + 1);

         paregActiveAttFocus->nActive = 0;
         paregActiveAttFocus->pFirst = -1;

         pregActiveAttFocusNode = (regActiveAttFocusNode *) NULL;

         aregEmptyAttFocusNode = 
                       (regActiveAttFocusNode *) (paregActiveAttFocus + 1);
         
         for ( i = 0; i < NACTIVEATTFOCUS - 1; i++ ) {
            aregEmptyAttFocusNode[i].fAttentionDegree = -1.0;
            aregEmptyAttFocusNode[i].pAttFocus = -1;
            aregEmptyAttFocusNode[i].pNext = i+1;
         }
         
         aregEmptyAttFocusNode[NACTIVEATTFOCUS - 1].fAttentionDegree = -1.0;
         aregEmptyAttFocusNode[NACTIVEATTFOCUS - 1].pAttFocus = -1;
         aregEmptyAttFocusNode[NACTIVEATTFOCUS - 1].pNext = -1; 
                                             /*** The last hasn't got next ***/

         pregAttentionFocus =  
            (regAttentionFocus *) (aregEmptyAttFocusNode + NACTIVEATTFOCUS + 1);
         
         /*** Initially, all the attention focuses are empty ***/
         for (i = 0; i < NATTFOCUS; i++) {
            pregAttentionFocus[i].nMetaRules = 0;
            pregAttentionFocus[i].pFirst = -1;
         }
         
         pregpMetaRuleNode = (regpMetaRuleNode *) NULL;

         aregEmptypMetaRuleNode =  
                      (regpMetaRuleNode *) (pregAttentionFocus + NATTFOCUS + 1);
                /*** There is no meta-rule, so all are empty ***/

         aregMetaRule =  
                    (regMetaRule *) (aregEmptypMetaRuleNode + NMAXMETARULES + 1);
               
         /*** A meta-rule, that is, an struct of type regMetaRule, may only
          *     belong to one attention focus at the same time. 
          *   If the user wants that one meta-rule will belong to different 
          *     attention focus at the same time, then several copies will 
          *      be generated.
          *   This defines a 1 on 1 relation between regpMetaRuleNode and 
          *      regMetaRule.
          ***/
         
         for ( i = 0; i < NMAXMETARULES - 1; i++) {
            aregEmptypMetaRuleNode[i].pMetaRule = i;
            
            aregMetaRule[i].pregEvent = -1;
            aregMetaRule[i].pregEventSlot = -1;
            aregMetaRule[i].pregCondition = -1;
            aregMetaRule[i].pregAction = -1;
            
            aregEmptypMetaRuleNode[i].pNext = i+1;
         }
         
         aregEmptypMetaRuleNode[ NMAXMETARULES - 1 ].pMetaRule = 
                                                            NMAXMETARULES - 1;
            
         aregMetaRule[ NMAXMETARULES - 1 ].pregEvent = -1;
         aregMetaRule[ NMAXMETARULES - 1 ].pregEventSlot = -1;
         aregMetaRule[ NMAXMETARULES - 1 ].pregCondition = -1;
         aregMetaRule[ NMAXMETARULES - 1 ].pregAction = -1;
         
         aregEmptypMetaRuleNode[ NMAXMETARULES - 1 ].pNext = 0;

         /*** A meta-rule, that is, an struct of type regMetaRule, may only 
          *     belong to one attention focus at the same time. 
          *   If the user wants that one meta-rule will belong to different 
          *     attention focus at the same time, then several copies will be 
          *     generated. 
          *   This defines a 1 on 1 relation between regpMetaRuleNode and 
          *     regMetaRule. 
          ***/

         aregEmptyCondNode = (regExprNode *) (aregMetaRule + NMAXMETARULES + 1 );
                   
         for ( i = 0; i < nregExprNode - 1; i++ ) {
            aregEmptyCondNode[i].regF1.nType = -1;  /*** It is empty ***/ 
            aregEmptyCondNode[i].regF2.nType = -1;
            aregEmptyCondNode[i].nOp = 0; /*** No operation ***/ 
            aregEmptyCondNode[i].pNext = i + 1; 
         }

         aregEmptyCondNode[ nregExprNode - 1 ].regF1.nType = 
                                                      -1;  /*** It is empty ***/ 
         aregEmptyCondNode[ nregExprNode - 1 ].regF2.nType = -1;
         aregEmptyCondNode[ nregExprNode - 1 ].nOp = 0; /*** No operation ***/ 
         aregEmptyCondNode[ nregExprNode - 1 ].pNext = 0; 
    
         aregEmptyActNode =  
                          (regActNode *) (aregEmptyCondNode + nregExprNode + 1);

         for ( i = 0; i < nregActNode - 1; i++ ) {
            aregEmptyActNode[i].nAct = -1; /*** It is empty ***/
            aregEmptyActNode[i].pNext = i+1; 

            /***aregEmptyActNode[i].regArg1.pSlot = (void *) NULL; 
             *aregEmptyActNode[i].regArg1.nFunctId = -1;
             *aregEmptyActNode[i].regArg1.nLevelId = -1;
             *aregEmptyActNode[i].regArg1.nReactDegree = -1;
             *aregEmptyActNode[i].regArg1.nSlsPolicy = -1;
             *aregEmptyActNode[i].regArg1.nAttentionFocus = -1;
             *aregEmptyActNode[i].regArg1.fAttentionDegree = -1;
             *aregEmptyActNode[i].regArg1.nBehav = -1;
             *aregEmptyActNode[i].regArg2.pExpr = -1;
             *aregEmptyActNode[i].regArg2.fFocusAttentDegree = -1.0;
             ***/
         }

         aregEmptyActNode[ nregActNode - 1 ].nAct = -1; /*** It is empty ***/
         aregEmptyActNode[ nregActNode - 1 ].pNext = 0;

         pn1stEmptyEvType = (int *) (aregEmptyActNode + nregActNode +1);
         pn1stEmptypSlt = (int *) (pn1stEmptyEvType + 1);
         pn1stEmptyAttFocusNode = (int *) (pn1stEmptypSlt + 1);
         pn1stEmptyAttentionFocus = (int *) (pn1stEmptyAttFocusNode + 1);
         pn1stEmptypMetaRuleNode = (int *) (pn1stEmptyAttentionFocus + 1);
         pn1stEmptyCondNode = (int *) (pn1stEmptypMetaRuleNode + 1);
         pn1stEmptyActNode = (int *) (pn1stEmptyCondNode + 1);

         *pn1stEmptyEvType = 0;
         *pn1stEmptypSlt = 0;
         *pn1stEmptyAttFocusNode = 0;
         *pn1stEmptyAttentionFocus = 0;
         *pn1stEmptypMetaRuleNode = 0;
         *pn1stEmptyCondNode = 0;
         *pn1stEmptyActNode = 0;

/***
#ifndef MODULE         
fprintf(stderr, "\t(meta-rule_store) aregEmptyActNode %d\n", (int) aregEmptyActNode );
#endif
***/
         return( METARULE_OK );
  }
}

/************************* Meta_Rule_Store_Initialize *************************
 * Public generic initializing function.
 * It allocates or attaches the memory of the Meta-Rule Store.
 * It must be used both in the RT and non-RT Linux worlds
 ***/
int Meta_Rule_Store_Initialize(void) {
  /*** RT-SIDE does an mbuff_attach ***/
#ifdef MODULE
  return  InitMetaRuleStore();
  /*while OPTIONAL SIDE does first a mbuff_alloc */
#else
  return  CreateMetaRuleStore();
#endif
}

/*************************** Meta_Rule_Store_Free ***************************
 * It Releases the memory allocated or attached to the Meta-Rule Store.
 * It must be used both in the RT and non-RT Linux worlds
 ***/
void Meta_Rule_Store_Free(void) {
#ifdef MODULE
  mbuff_detach(META_RULE_STORE,(void *)piMRMutex);
#else
  mbuff_detach(META_RULE_STORE,(void *)piMRMutex);
#endif
}

/******************** Meta_Rule_Get_Empty_Att_Focus_Node *********************
 *  Returns two possible values:
 *      -1, if there is no empty regActiveAttFocusNode
 *      n, where aregEmptyAttFocusNode[n] is the empty regActiveAttFocusNode
 ***/
 int Meta_Rule_Get_Empty_Att_Focus_Node (void) {

  if (*piMRMutex) {  
    /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyAttFocusNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; /*** And now, it is mine ***/

  if ((*pn1stEmptyAttFocusNode) == -1) {
        /*** There is no empty regActiveAttFocusNode ***/
        *piMRMutex = 0;
        return -1;  
  }
    else {
           int nAux = *pn1stEmptyAttFocusNode;

           *pn1stEmptyAttFocusNode = aregEmptyAttFocusNode[nAux].pNext;   
           *piMRMutex = 0;
           return nAux;
    }
}

/******************* Meta_Rule_Get_Empty_pMetaRuleNode ********************
 *  Returns two possible values:
 *      -1, if there is no empty regpMetaRuleNode
 *      n, where aregEmptypMetaRuleNode[n] is the empty regpMetaRuleNode
 *         and aregMetaRule[n] is the empty regMetaRule
 ***/
 int Meta_Rule_Get_Empty_pMetaRuleNode (void) {

  if (*piMRMutex) {  
    /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptypMetaRuleNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; /*** And now, it is mine ***/

  if ((*pn1stEmptypMetaRuleNode) == -1) {
        /*** There is no empty regpMetaRuleNode ***/
        *piMRMutex = 0;
        return -1;  
  }
    else {
            int nAux = *pn1stEmptypMetaRuleNode;
            
            *pn1stEmptypMetaRuleNode = aregEmptypMetaRuleNode[nAux].pNext;   
            *piMRMutex = 0;
            return nAux;
    }
}

/******************** Meta_Rule_Get_Empty_Event_Type ***********************
 *  Returns two possible values:
 *      -1, if there is no empty regEventType
 *      n, where aregEmptyEventType[n] is the empty regEventType
 ***/
 int Meta_Rule_Get_Empty_Event_Type (void) {

  if (*piMRMutex) {  
    /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyEventType ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; /*** And now, it is mine ***/

  if ((*pn1stEmptyEvType) == -1) {
        /*** There is no empty regEventType ***/
        *piMRMutex = 0;
        return -1;  
  }
    else {
            int nAux = *pn1stEmptyEvType;
            *pn1stEmptyEvType = aregEmptyEventType[nAux].pNext;   
            *piMRMutex = 0;
            return nAux;
    }
}

/******************** Meta_Rule_Get_Empty_pSlot ***********************
 *  Returns two possible values:
 *      -1, if there is no empty regpSlot
 *      n, where aregEmptypSlot[n] is the empty regpSlot
 ***/
 int Meta_Rule_Get_Empty_pSlot (void) {

  if (*piMRMutex) {  
    /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptypSlot ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; /*** And now, it is mine ***/

  if ((*pn1stEmptypSlt) == -1) {
        /*** There is no empty regpSlot ***/
        *piMRMutex = 0;
        return -1;  
  }
    else {
            int nAux = *pn1stEmptypSlt;
            *pn1stEmptypSlt = aregEmptypSlot[nAux].pNext;   
            *piMRMutex = 0;
            return nAux;
    }
}

/******************** Meta_Rule_Get_Empty_Cond_Node ***********************
 *  Returns two possible values:
 *      -1, if there is no empty regExprNode
 *      n, where aregEmptyCondNode[n] is the empty regExprNode
 ***/
 int Meta_Rule_Get_Empty_Cond_Node (void) {

  if (*piMRMutex) {  
    /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyCondNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; /*** And now, it is mine ***/

  if ((*pn1stEmptyCondNode) == -1) {
        /*** There is no empty regExprNode ***/
        *piMRMutex = 0;
        return -1;  
  }
    else {
            int nAux = *pn1stEmptyCondNode;
            *pn1stEmptyCondNode = aregEmptyCondNode[nAux].pNext;   
            *piMRMutex = 0;
            return nAux;
    }
}

/******************** Meta_Rule_Get_Empty_Act_Node ***********************
 *  Returns two possible values:
 *      -1, if there is no empty regActNode
 *      n, where aregEmptyActNode[n] is the empty regActNode
 ***/
 int Meta_Rule_Get_Empty_Act_Node (void) {

  if (*piMRMutex) {  
    /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::GetEmptyActNode ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -2;
  }
  else *piMRMutex = 1; /*** And now, it is mine ***/

  if ((*pn1stEmptyActNode) == -1) {
        /*** There is no empty regActNode ***/
        *piMRMutex = 0;
        return -1;  
  }
    else {
            int nAux = *pn1stEmptyActNode;
            *pn1stEmptyActNode = aregEmptyActNode[nAux].pNext;   
            *piMRMutex = 0;
            return nAux;
    }
}

/********************** Meta_Rule_Get_Empty_Act_Node *************************
 *   This is a variation of the previous function that doesn't reorder the
 *      regActNode store, because it has all the regActNode in the same store
 *      empties or not. It is not very efficient.
 ***/
/***
 *int GetEmptyActNode ( int *nRegActIndex ) {
 *   int i;
 *   
 *   if (aregEmptyActNode == NULL)
        *** There is no empty regActNode ***/
/***        return -1;  
 *   else {
 *           for ( i = 0; (i < nregActNode) && (aregEmptyActNode[i].nAct == -1); i++ );
 *           
 *           if ( i == nregActNode )
                *** There is no empty regActNode ***/
/***                return -1;              
 *           else {
 *                   *nRegActIndex = i;
 *                   return 0;
 *           }       
 *   }
 *}        
 ***/

/****************** Meta_Rule_Activate_Attention_Focus ***********************
 * Creates a new Active Attention Focus with the given parameters and creates
 *  a new element in the list, placing it at the head of the list 
 *  Returns three possible values:
 *      -2, if the mutex is occupied 
 *      -1, if there is no empty regActiveAttFocusNode
 *      0, The new active attention focus has been inserted
 * The arguments are:
 *      - nAF: Index of the regAttentionFocus
 *      - fAttentDegree: Attention Degree of the Attention Focus to activate 
 ***/
int Meta_Rule_Activate_Attention_Focus( int nAF, float fAttentDegree ) {
  int nAttFocusNode;

  if (( nAttFocusNode = Meta_Rule_Get_Empty_Att_Focus_Node()) >= 0) {
    /*** aregEmptyAttFocusNode has an empty node ***/
    if (*piMRMutex) {  
      /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::ActivateAttentionFocus ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
      return -2;
    }
    else *piMRMutex = 1; /*** And now, it is mine ***/

    aregEmptyAttFocusNode[nAttFocusNode].fAttentionDegree = fAttentDegree;
    aregEmptyAttFocusNode[nAttFocusNode].pAttFocus = nAF;
    paregActiveAttFocus->nActive++;
    aregEmptyAttFocusNode[nAttFocusNode].pNext = paregActiveAttFocus->pFirst;
    paregActiveAttFocus->pFirst = nAttFocusNode;
    
#ifndef MODULE
    fprintf(stderr, "\t(meta-rule_store.c)::ActivateAttentionFocus %d \n", nAF);
#endif

    *piMRMutex = 0; /*** If U want it, U got it ***/ 

    return 0;
  }
  else return -1;
}


/***************** Meta_Rule_Deactivate_Attention_Focus ********************
 * Destroys the Active Attention Focus corresponding to the attention 
 *  focus indicated by nAF 
 *  Returns three possible values:
 *      -2, if the mutex is occupied 
 *      -1, if there is no Active Attention Focus corresponding to nAF
 *      0, The active attention focus has been destroyed
 * The argument is nAF, the Index of the regAttentionFocus
 ***/
int Meta_Rule_Deactivate_Attention_Focus( int nAF ) {
  int nAttFocusNode, nAttFocusNodePrev;

  if (*piMRMutex) {  
    /*** There is someone inside!!!!!!!!!!! ***/
#ifdef DEBUG
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::DeactivateAttentionFocus ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
#endif
    return -2;
  }
  else *piMRMutex = 1; /*** And now, it is mine ***/

  ///#ifdef DEBUG
#ifdef MODULE
  rtl_printf("\t(meta_rule_store.c)::DeactivateAttentionFocus \n");      
#else 
  fprintf(stderr, "\t(meta-rule_store.c)::DeactivateAttentionFocus \n");
#endif
  ///#endif

  if (( nAttFocusNodePrev = paregActiveAttFocus->pFirst) == -1) {

#ifdef DEBUG
#ifdef MODULE
          rtl_printf("\t\t\t There is no active attention focus \n");      
#else 
          fprintf(stderr, "\t\t\t There is no active attention focus \n");      
#endif
#endif

          *piMRMutex = 0; /*** If U want it, U got it ***/ 
          return -1;
  }
  else /*** There are at least one active attention focus ***/
    if ( aregEmptyAttFocusNode[nAttFocusNodePrev].pAttFocus == nAF ) {
        /*** The first one is the one we are looking for ***/

#ifdef DEBUG
#ifdef MODULE
        rtl_printf("\t\t\t The 1st one is the one we are looking for \n");      
#else 
        fprintf(stderr, "\t\t\t The 1st one is the one we are looking for \n");      
#endif
#endif

        paregActiveAttFocus->pFirst = 
                               aregEmptyAttFocusNode[nAttFocusNodePrev].pNext;
        paregActiveAttFocus->nActive--;
        *piMRMutex = 0;
        return 0;
    }
    else { 
          if (( nAttFocusNode = aregEmptyAttFocusNode[nAttFocusNodePrev].pNext ) 
             == -1 ) {
             /*** There is only one active attention focus, 
              *      and it is not the one we are looking for
              ***/

#ifdef DEBUG
#ifdef MODULE
            rtl_printf("\t\t\t There is only one active, and it is no the one we are looking for, it is the %d\n",
                       nAttFocusNodePrev);      
#else 
            fprintf(stderr, "\t\t\t There is only one active, and it is no the one we are looking for, it is the %d\n",
                       nAttFocusNodePrev);      
#endif
#endif

             *piMRMutex = 0;
             return -1;
           }
           else { /*** We must look for the one ***/
                  while (( nAttFocusNode != -1 ) &&
                      ( aregEmptyAttFocusNode[nAttFocusNode].pAttFocus != nAF )){
                     nAttFocusNodePrev = nAttFocusNode;
                     nAttFocusNode = aregEmptyAttFocusNode[nAttFocusNode].pNext;
                  }
               
                  if ( nAttFocusNode == -1 ) {
                    /*** The attention focus was not active ***/

#ifdef DEBUG
#ifdef MODULE
                    rtl_printf("\t\t\t The attention focus was not active \n");      
#else 
                    fprintf(stderr, "\t\t\t The attention focus was not active \n");      
#endif
#endif

                    *piMRMutex = 0;
                    return -1;
                  }
                  else { /*** Deactivate the Attention Focus ***/

#ifdef DEBUG
#ifdef MODULE
                         rtl_printf("\t\t\t Deactivate the attention focus \n");      
#else 
                         fprintf(stderr, "\t\t\t Deactivate the attention focus \n");      
#endif
#endif

                         aregEmptyAttFocusNode[nAttFocusNodePrev].pNext = 
                                    aregEmptyAttFocusNode[nAttFocusNode].pNext;
                         paregActiveAttFocus->nActive--;
                         *piMRMutex = 0;
                         return 0;
                  }
           }
    
    *piMRMutex = 0; /*** If U want it, U got it ***/ 

    return 0;
    }
}


/*************************** Meta_Rule_Create ********************************
 * Creates a new Meta-Rule node with the given parameters.
 *  Returns three possible values:
 *        -2, if the mutex is occupied
 *        -1, if there is no empty regpMetaRuleNode
 *         n, where aregEmptypMetaRuleNode[n] has the Meta-Rule pointer, and
 *                   aregMetaRule[n] has the corresponding Meta-Rule 
 * The arguments are:
 *      - nET: Event Type (usually MODIFICATION, but who knows...
 *      - pSlot: Pointer to the slot
 *      - nCond: Index of the regExprNode
 *      - nAct: Index of the regActNode 
 ***/
 int Meta_Rule_Create ( int nET, void * pSlot, int nCond, int nAct ) {
   int nMetaRule, nEvType, i, npSlot, nAux, nAuxSlot; 

  if (( nMetaRule = Meta_Rule_Get_Empty_pMetaRuleNode () ) < 0 ) {
        return -1;
  } 
  else if (( nEvType = Meta_Rule_Get_Empty_Event_Type () ) < 0 ) {
		return -1;
	}
	else if (( npSlot = Meta_Rule_Get_Empty_pSlot () ) < 0 ) {
		return -1;
	     }
	     else {   
                   /*** aregEmptypMetaRuleNode[nMetaRule] has an empty 
                    *    Meta-Rule ptr node, and aregMetaRule[nMetaRule] 
                    *    has an empty Meta-Rule 
                    *   All the existing event types should be created 
                    *    a priori and indexed through 
                    *    aregEventType *paregEventType;
                    ***/    
                   if (*piMRMutex) {  
                     /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::CreateMetaRule ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
                     return -2;
                   }
                   else *piMRMutex = 1; /*** And now, it is mine ***/

                   aregEmptyEventType[nEvType].nEventType = nET;
        
                   /*** Look for the event type in the paregEventType 
                    *    index of Meta-Rules 
                    ***/
                   for ( i = 0; 
                         (i <= NEVENTTYPES) 
                         && (paregEventType[i].nEventType != nET); 
                         i++ );
                   aregEmptyEventType[nEvType].nEventTypeIndex = i;

                   /*** The Meta-Rule must be inserted in the index 
                    *    by Event Type 
                    ***/
                   aregEmptyEventType[nEvType].pNext = 
                                              paregEventType[i].pFirst;
                   paregEventType[i].pFirst = nMetaRule;
    
                   aregEmptypSlot[npSlot].pSlot = pSlot;
                   /*** Look for the pSlot in the paregpSlot index of 
                    *      Meta-Rules 
                    ***/
      
                   for ( i = 0; 
                           ( i < NMAXMETARULES ) && 
                           (( nAux = paregpSlot[i].pFirst ) != -1 ) &&
                           (( nAuxSlot = aregMetaRule[nAux].pregEventSlot ) 
                                                                      != -1 ) &&
                           ( aregEmptypSlot[nAuxSlot].pSlot != pSlot ) ; i++ );
        
                   /*** The new Meta-Rule is the new 1st one in the index of
                    *      Meta-Rules with an event associated to that pSlot
                    ***/   
                   aregEmptypSlot[npSlot].pNext = paregpSlot[i].pFirst;
                   paregpSlot[i].pFirst = nMetaRule;
                   paregpSlot[i].nSlots ++;
                   aregEmptypSlot[npSlot].npSlotIndex = i;
    
                   /*** Fulling the empty Meta-Rule ***/ 
                   aregMetaRule[nMetaRule].pregEvent = nEvType;
                   aregMetaRule[nMetaRule].pregEventSlot = npSlot;
                   aregMetaRule[nMetaRule].pregCondition = nCond;
                   aregMetaRule[nMetaRule].pregAction = nAct;
    
                   *piMRMutex = 0; /*** If U want it, U got it ***/ 

                   return nMetaRule;
                 }
}
 
/***************** Meta_Rule_Insert_In_Att_Focus *****************************
 * Inserts a Meta-Rule node within the Attention Focus indicated 
 *  Returns three possible values:
 *      -1, if the mutex is occupied 
 *      0, The Meta-Rule node has been inserted
 * The arguments are:
 *      - nMetarule, where aregEmptypMetaRuleNode[nMetaRule] has the Meta-Rule 
 *                   pointer, and aregMetaRule[nMetaRule] has the 
 *                   corresponding Meta-Rule  
 *      - nAF: Index of the regAttentionFocus
 ***/
int Meta_Rule_Insert_In_Att_Focus( int nMetaRule, int nAF ) {

  if (*piMRMutex) {  
    /*** There is someone inside!!!!!!!!!!! ***/
#ifdef MODULE
    rtl_printf("\t(meta_rule_store.c)::InsertMetaRuleInAttFocus ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
    return -1;
  }
  else *piMRMutex = 1; /*** And now, it is mine ***/

    /*** Adding the Meta-Rule as the head of the Meta-Rule's 
     *      list of the Attention Focus
     ***/
    aregEmptypMetaRuleNode[nMetaRule].pMetaRule = nMetaRule;
    aregEmptypMetaRuleNode[nMetaRule].pNext = pregAttentionFocus[nAF].pFirst;  
    pregAttentionFocus[nAF].pFirst = nMetaRule;
    pregAttentionFocus[nAF].nMetaRules ++;

#ifdef MODULE
  rtl_printf("\t(meta_rule_store.c)::InsertMetaRuleInAttFocus\n");      
#endif
    
    *piMRMutex = 0; /*** If U want it, U got it ***/ 

    return 0;
}





#ifdef MODULE
int init_module(void) {
  rtl_printf("META-RULE STORE STARTING\n");
  return 0;
}

void cleanup_module(void)
/* ==================================================================== */
{
  mbuff_detach(META_RULE_STORE, (void *)piMRMutex);
}

#endif

void Meta_Rule_Print_Active_Att_Focus(void) {
   int i, j, 
          nIndex = paregActiveAttFocus->pFirst, /*** First Active Att. Focus ***/
          nActFocus = paregActiveAttFocus->nActive, 
                 /*** Number of Active Attention Focuses ***/ 
          nAttFocus, /*** Attention Focus ***/
          nMetaRules, /*** Number of Meta-Rules in the Attention Focus ***/
          nMRIndex,  /*** Index of the Meta-Rules in the Attention Focus ***/ 
          nMRAux;/*** Actual Meta-Rule being printed in the Attention Focus ***/

#ifdef DEBUG   
#ifndef MODULE
   fprintf( stderr, 
            //printf(
     "\n(meta-rule_store.c)::printActiveAttFocus(): Number of active Focus: %d\n",
     nActFocus );
#else
  rtl_printf("(meta-rule_store.c): Act. Focus: %d", nActFocus);
#endif
#endif

   /*** Journey through the Active Attention Focuses ***/  
   for ( i = 1; i <= nActFocus; i++ ) {
      /*** Journey through the Meta-Rules of the Attention Focus ***/
      nAttFocus = aregEmptyAttFocusNode[nIndex].pAttFocus;
      nMetaRules = pregAttentionFocus[nAttFocus].nMetaRules;
      nMRIndex = pregAttentionFocus[nAttFocus].pFirst;
#ifdef DEBUG    
#ifndef MODULE
      fprintf( stderr, "\tActive Attention Focus %d: \tAttention Degree: %f\n",
                          nAttFocus, aregEmptyAttFocusNode[nIndex].fAttentionDegree );
      fprintf( stderr,
               //printf( 
            "\t\tThere are %d Meta-Rules:\n", nMetaRules);
#else
      rtl_printf("        Active Attention Focus %d:        Attention Degree: %d", 
                          nAttFocus, aregEmptyAttFocusNode[nIndex].fAttentionDegree);
      rtl_printf("        Meta-Rules: %d", nMetaRules);                         
#endif
#endif
      
      for ( j = 1; j <= nMetaRules; j++ ) {
        nMRAux = aregEmptypMetaRuleNode[nMRIndex].pMetaRule; 
#ifdef DEBUG
#ifndef MODULE
        fprintf( stderr,
                 //printf( 
                 "\t\t\t%d: Event: %d\n\t\t\t    Action: %d %d \n", j,
          aregEmptyEventType[aregMetaRule[nMRAux].pregEvent].nEventType,
          aregEmptyActNode[aregMetaRule[nMRAux].pregAction].nAct,
          aregEmptyActNode[aregMetaRule[nMRAux].pregAction].regArg1.nSlsPolicy );
#else
       rtl_printf("    %d: Event: %d      Action: %d %d \n", j,
       aregEmptyEventType[aregMetaRule[nMRAux].pregEvent].nEventType,
       aregEmptyActNode[aregMetaRule[nMRAux].pregAction].nAct,
       aregEmptyActNode[aregMetaRule[nMRAux].pregAction].regArg1.nSlsPolicy );
#endif
#endif
        nMRIndex = aregEmptypMetaRuleNode[nMRIndex].pNext;

#ifndef MODULE
        fprintf(stderr, "Let's go once again\n");
#endif
      }
#ifndef MODULE
      fprintf(stderr, "Let's go once again to the next AttFocus.\n");
#endif
      nIndex = aregEmptyAttFocusNode[nIndex].pNext;
   }
}

/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
