/*

   FILE:
	rserver.h						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    The RSERVER module implements the Reactive Server (also known as
    the "first-level scheduler") of the ARTIS architecture. It is
    basically a scheduler of RT-Linux real-time threads, adapted to 
    match the requirements of ARTIS.

    This module offers two interfaces: 

       a) To the application real-time module. It exports the task
       type and functions to create and configure the application
       tasks and to start and stop the application.
       (this is specified in ../../include/rserver.h)
       
       b) To other fixed parts of the ARTIS architecture, such as the
       RT-Linux MUTEX module, the following:
          + a pointer to the current running task 
	  + functions to udpate and schedule (required in each syscall)

*/

#ifndef __RSERVER_H
#define __RSERVER_H

#include "./rserver/rserver.h"
#include "../../include/rserver.h"
// Visible types:

//#typedef _rt_task_t rt_task_t;



// Visible functions:

#define TASK_PRIO(t) (t->priority)

void rtl_update (void);
void rtl_schedule (void);

int  rserver_init_module (void);
void rserver_cleanup_module (void);




// Visible global variables:

extern rt_task_t  *rtl_running;



#endif // #ifndef __RSERVER_H
