
/*

   FILE:
	rserver.h						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    This private header file defines some constants and types which are
    internal to the RSERVER module.

*/


#ifndef __PRIVATE_RSERVER_H
#define __PRIVATE_RSERVER_H


#include "../../../common/include/artis_conf.h"
#include "../../../time/include/time.h"
#include "../../../rtl_core/include/rtl_core.h"


#define RT_TASK_MAGIC 0x754d2774

#define RTL_LOWEST_PRIORITY 1000000


enum rt_task_state_t {RT_TASK_INITIAL, RT_TASK_OPTIONAL, RT_TASK_FINAL,
		      RT_TASK_DELAYED, RT_TASK_DORMANT, RT_TASK_LINUX};


typedef struct _rt_task_struct {
  rt_context_t    *context;
  int             free;
  int             magic;
  int             state;
  int             priority;
  int             basepriority;            
  int             identifier;
  int             optional_func_id;

#ifdef ARTIS_CONF_METAREASONING
  // ThreeC: Behavior
  int nBehav;  
     /// To mark the behavior a task belongs to. It will be useful in the transition between behs.
#endif

  systime_t       deadline;
  systime_t       absolute_deadline;
  systime_t       period;
  systime_t       resume_time;
  systime_t       offset;
  systime_t       execution_left;
  systime_t       initial_wcet;
  systime_t       final_wcet;
  systime_t       blocking_time;
  systime_t       slack;
  void            *(*initial_function)(void *initial_data);
  void           *initial_data;
  void            *(*final_function)(void *final_data);
  void           *final_data;
  int             num_optionals;
  struct _rt_task_struct *next;
} _rt_task_t;

#endif // #ifndef __PRIVATE_RSERVER_H
