
/*

   FILE:
								
	rtl_hal.h						
	
   AUTHOR:
        * Originally written by Michael Barabanov, Victor Yodaiken
        * Copyright (C) Finite State Machine Labs Inc., 1998-1999
        * Released under the terms of the GPL Version 2

        Copyright (C) 1999-2003
	GTI-IA, UPV

        This is free software; you can redistribute it and/or modify
	it under terms of the GNU General Public License as published
	by the Free Soft- ware Foundation; either version 2, or (at
	your option) any later ver- sion.  It is distributed in the
	hope that it will be useful, but WITH- OUT ANY WARRANTY;
	without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
	License for more details.  You should have received a copy of
	the GNU General Public License; see file COPYING.  If not,
	write to the Free Software Foundation, 59 Temple Place - Suite
	330, Boston, MA 02111-1307, USA.

   DESCRIPTION:	

       Hardware-dependent definitions for the x86 architecture.

   CREATION DATE:

        26-Aug-1999

   REVISED:

        Oct-2003  Revised compatibility with RT-Linux 3.2-pre1
*/


#ifndef __RTL_HAL_H__
#define __RTL_HAL_H__ 


/* 
   Interrupt-related definitions: save interrupts.
     Enable/Disable interrupt definitions are taken from RTL.
*/

#define rtl_save_interrupts(s) \
__asm__ __volatile__("pushfl ; popl %0":"=g" (s): /* no input */ :"memory")


/* 
   Definition of a task context.
   Must be defined here because it contains hw-dependent types,
   like the fpu registers.
*/

#include <arch/rtl_fpu.h>


typedef struct __context_struct {
	int               *stack;	/* hardcoded */
	int               fpu_initialized;
	RTL_FPU_CONTEXT   fpu_regs;
	int               uses_fp;
	int               *stack_bottom;
        unsigned long     interrupt_state;
} rt_context_t;


extern rt_context_t rtl_linux_context;	
#define rtl_is_linux_context(context) (context==&rtl_linux_context)



/* 
   Floating-Point Unit definitions
*/

extern void rtl_fpu_save (rt_context_t *context);
extern void rtl_fpu_restore (rt_context_t *context);
extern void rtl_context_init_fpu (rt_context_t *c, rt_context_t *fpu_owner);



/* 
    Context-switch and context stack initialization
*/

#define __STR(x) #x
#define STR(x) __STR(x)

#define rtl_switch_to(current_task_ptr, new_task) \
	__asm__ __volatile__( \
	"pushl %%eax\n\t" \
	"pushl %%ebp\n\t" \
	"pushl %%edi\n\t" \
	"pushl %%esi\n\t" \
	"pushl %%edx\n\t" \
	"pushl %%ecx\n\t" \
	"pushl %%ebx\n\t" \
	"movl (%%ebx), %%edx\n\t" /* get current */ \
	"pushl $1f\n\t" \
	"movl %%esp, (%%edx)\n\t" \
	"movl (%%ecx), %%esp\n\t" \
	"movl %%ecx, (%%ebx)\n\t" /* store current */\
	"ret\n\t" \
"1:	popl %%ebx\n\t" \
	"popl %%ecx\n\t" \
	"popl %%edx\n\t" \
	"popl %%esi\n\t" \
	"popl %%edi\n\t" \
	"popl %%ebp\n\t" \
	"popl %%eax\n\t" \
	: /* no output */ \
	: "c" (new_task), "b" (current_task_ptr) \
	);

#define rtl_init_stack(task,fn,data,rt_startup)\
	{\
	*--(task->stack) = (int) data;\
	*--(task->stack) = (int) fn;\
	*--(task->stack) = 0;	/* dummy return addr*/\
	*--(task->stack) = (int) rt_startup;\
	}





#endif
