/*  

   FILE:

	rtl_hal.c

   AUTHOR:
        * Originally written by Michael Barabanov, Victor Yodaiken
        * Copyright (C) Finite State Machine Labs Inc., 1998-1999
        * Released under the terms of the GPL Version 2

        Copyright (C) 1999-2003
	GTI-IA, UPV

        This is free software; you can redistribute it and/or modify
	it under terms of the GNU General Public License as published
	by the Free Soft- ware Foundation; either version 2, or (at
	your option) any later ver- sion.  It is distributed in the
	hope that it will be useful, but WITH- OUT ANY WARRANTY;
	without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
	License for more details.  You should have received a copy of
	the GNU General Public License; see file COPYING.  If not,
	write to the Free Software Foundation, 59 Temple Place - Suite
	330, Boston, MA 02111-1307, USA.


   DESCRIPTION:

        Hardware-dependent functions for the x86 architecture.

   CREATION DATE:
        26-Aug-1999

   REVISED:

	14-Nov-2003  Incorporated to ARTIS v2 source tree

        11-Oct-2003 
	    Newly adapted for version 3.2-pre1 in order to match 
	    <rtlinux>/schedulers/i386/switch.c
*/


#include <asm/system.h>
#include "../include/rtl_core/rtl_hal.h"
#include "../include/rtl_core.h"


static int rtl_linux_saved_ts;


int 
is_set_TS (void) {
  int current_TS;
  __asm__ ("movl %%cr0,%%eax;andl $8,%%eax;movl %%eax,%0":"=m" (current_TS)
	   /*output */
	   :				/* no inputs */
	   :	   "ax");
  return current_TS;
}






void 
rtl_fpu_save (rt_context_t * current_c) {
  if (rtl_is_linux_context (current_c)) {
    rtl_linux_saved_ts = is_set_TS ();
    clts ();
  }
  __asm__ ("fsave %0":"=m" (current_c->fpu_regs));
}






void 
rtl_fpu_restore (rt_context_t * current_c) {
  clts ();			/* just in case */
  __asm__ ("frstor %0":"=m" (current_c->fpu_regs));
  if (rtl_is_linux_context (current_c) && rtl_linux_saved_ts) {
    stts ();
  }
}





void 
rtl_context_init_fpu (rt_context_t * c, rt_context_t * fpu_owner) {
/*
   (ATB) Note: in RTL 3.2-pre1, this function is void, but this seems
   to work, so I don't remove it (yet)
*/
  int was_set_TS;
  
  if (fpu_owner) {
    if ((was_set_TS = is_set_TS ()))
      clts ();
    __asm__ ("fsave %0":"=m" (fpu_owner->fpu_regs));
    __asm__ ("finit");
    __asm__ ("fsave %0":"=m" (c->fpu_regs));
    __asm__ ("frstor %0":"=m" (fpu_owner->fpu_regs));
    if (was_set_TS) {
      stts ();
    }
  }
  else {
    clts ();
    __asm__ ("finit");
  }
  return;
}
