/*  
   FILE:
		
       rtl_timer.c						       
	
   AUTHOR:

        Copyright (C) 1999-2001
	GTI-IA, UPV

        This is free software; you can redistribute it and/or modify
	it under terms of the GNU General Public License as published
	by the Free Soft- ware Foundation; either version 2, or (at
	your option) any later ver- sion.  It is distributed in the
	hope that it will be useful, but WITH- OUT ANY WARRANTY;
	without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
	License for more details.  You should have received a copy of
	the GNU General Public License; see file COPYING.  If not,
	write to the Free Software Foundation, 59 Temple Place - Suite
	330, Boston, MA 02111-1307, USA.


   DESCRIPTION: 

        Simple interface to access the HW timer in order to set interrupts.
	(The API resembles the RT-Linux v1 timer functions a lot :)

   CREATION DATE:

        26-Aug-1999

   REVISED:

       12-Oct-2003  Revised compatibility with RTL 3.2-pre1

       8-Apr-2001   Added compatibility for RT-Linux V3

*/



#include <linux/errno.h>
#include <rtl_time.h>

#include "../../time/include/time.h"






int 
rtl_set_timer (systime_t time)
{

  clockid_t local_clock = rtl_getbestclock (cpu_logical_map (0));

  if (!local_clock) {
    return -EINVAL;
  }

  //  return local_clock->settimer (local_clock, time - gethrtime ());
  return local_clock->settimer (local_clock, time);

}





int 
rtl_no_timer (void)
{
  clockid_t local_clock = rtl_getbestclock (cpu_logical_map (0));

  if (!local_clock) {
    return -EINVAL;
  }

  return local_clock->settimer (local_clock, HRTIME_INFINITY);
}







int 
rtl_request_timer (void (*handler) (struct pt_regs * regs))
{
  int ret;
  clockid_t local_clock = rtl_getbestclock (cpu_logical_map (0));

  if (!local_clock) {
    return -EINVAL;
  }

  ret = rtl_setclockhandler (local_clock, (clock_irq_handler_t) handler);

  if (ret == 0) {
    local_clock->init (local_clock);
  }

  return ret;
}






int 
rtl_free_timer (void)
{
  int ret;
  clockid_t local_clock = rtl_getbestclock (cpu_logical_map (0));

  if (!local_clock) {
    return -EINVAL;
  }

  ret = rtl_unsetclockhandler (local_clock);
  local_clock->uninit (local_clock);

  return ret;
}







int 
rtl_set_oneshot_mode (void)
{
  int ret;
  clockid_t local_clock = rtl_getbestclock (cpu_logical_map (0));

  if (!local_clock) {
    return -EINVAL;
  }

  ret = local_clock->settimermode (local_clock, RTL_CLOCK_MODE_ONESHOT);
  if (ret == 0) {
    ret = local_clock->settimer (local_clock, HRTIME_INFINITY);
  }

  return ret;
}







systime_t 
rtl_set_periodic_mode (systime_t period)
{
  int ret;
  clockid_t local_clock = rtl_getbestclock (cpu_logical_map (0));

  if (!local_clock) {
    return -EINVAL;
  }

  ret = local_clock->settimermode (local_clock, RTL_CLOCK_MODE_PERIODIC);

  if (ret == 0) {
    ret = local_clock->settimer (local_clock, period);
  }

  if (ret == 0) {
    return clock_gethrtime (local_clock);
  }
  else {
    return HRTIME_INFINITY;
  }
}
