/*

  STORAGE.C

  Copyright (C) 1998 - 2001
  Ana Garcia Fornes
  Andres Terrasa Barrena
  Agustin Espinosa Minguet
  Ignacio Pachs Giner

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write
  to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston,
  MA 02111-1307, USA.                                                      
*/

#include "../include/storage.h"

#ifdef ARTIS
#ifdef MODULE
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#endif
#include <mbuff.h>
#endif


#ifndef MODULE
#include <stdio.h>
#endif


#define STORAGE_MAGIC 0xFDD00000

///#define DEBUG  // For debugging



//////////////////////////////////////////
//  TYPES
//////////////////////////////////////////

typedef unsigned int internal_address ;

typedef struct {        
  internal_address First_Frame;   
  internal_address Next_Frame_Free;
  unsigned int Item_Size;
  unsigned int How_Many_Items;
} _Pool_Header;



typedef struct {
  int                     Magic;
  internal_address        Next_Free;
  internal_address        Last_Address;
  Storage_Entity_Size_T   Storage_Size;
  unsigned long           Max_Entities;
  internal_address        Entity_Index_Table;
} _Storage_Header;




//////////////////////////////////////////
//  EXTERNAL
//////////////////////////////////////////

static inline void * 
EXTERNAL (internal_address  Relative_Addr,
	  void             *Origin
	  ) 
{
  return (void*) (Relative_Addr + (unsigned int) Origin);
}


//////////////////////////////////////////
//  INTERNAL
//////////////////////////////////////////

static inline internal_address 
INTERNAL (void *External_Address ,
	  void *Origin
	  ) 
{
  return (unsigned int) External_Address - (unsigned int) Origin;
}


#ifdef DEBUG
//////////////////////////////////////////
//  Show_Storage
//////////////////////////////////////////

static void
Show_Storage (Storage_T Stoid
      )
{
  _Storage_Header  *EXT_Header_Ptr;
  void            **EXT_Entity_Index_Table;
  unsigned long     i;
  
  EXT_Header_Ptr = (_Storage_Header *) Stoid ;

#ifdef MODULE
  rtl_printf ("(rtl_storage) Cabecera del storage %u\n", 
	      (unsigned int) EXT_Header_Ptr);
  rtl_printf ("  Magic          -> %x\n",  EXT_Header_Ptr->Magic);
  rtl_printf ("  Next_Free      -> %u\n",  EXT_Header_Ptr->Next_Free);
  rtl_printf ("  Last_Address   -> %u\n",  EXT_Header_Ptr->Last_Address);
  rtl_printf ("  Storage_Size   -> %lu\n", EXT_Header_Ptr->Storage_Size);
  rtl_printf ("  Entities_Table ->\n");
#else  
  printf ("(storage) Cabecera del storage %u\n", 
	  (unsigned int) EXT_Header_Ptr);
  printf ("  Magic          -> %x\n",  EXT_Header_Ptr->Magic);
  printf ("  Next_Free      -> %u\n",  EXT_Header_Ptr->Next_Free);
  printf ("  Last_Address   -> %u\n",  EXT_Header_Ptr->Last_Address);
  printf ("  Storage_Size   -> %lu\n", EXT_Header_Ptr->Storage_Size);
  printf ("  Entities_Table ->\n");
#endif

  EXT_Entity_Index_Table = EXTERNAL (EXT_Header_Ptr->Entity_Index_Table, Stoid);

  for (i = 0; i < EXT_Header_Ptr->Max_Entities; i++ ) {
#ifdef MODULE
    rtl_printf ("    %lu -> %u\n", i, (unsigned int) EXT_Entity_Index_Table[i]);
#else
    printf ("    %lu -> %u\n", i, (unsigned int) EXT_Entity_Index_Table[i]);
#endif
  }
#ifdef MODULE
  rtl_printf ("\n");
#else
  printf ("\n");
#endif
} 


//////////////////////////////////////////
//  Show_Pool
//////////////////////////////////////////

static void
Show_Pool (Storage_T        Stoid,
	   internal_address Pool
      )
{
  
  _Pool_Header     *EXT_Pool;
  internal_address *EXT_Ptr;
  int               i;

  EXT_Pool = EXTERNAL (Pool, Stoid);

#ifdef MODULE
  rtl_printf ("(rtl_storage) Cabecera del pool %u\n", (unsigned int) Pool);
  rtl_printf ("  First_Frame     -> %u\n",  EXT_Pool->First_Frame);
  rtl_printf ("  Item_Size       -> %u\n",  EXT_Pool->Item_Size);
  rtl_printf ("  How_Many_Items  -> %u\n",  EXT_Pool->How_Many_Items);

  
  rtl_printf ("  Chain           ->\n");
  rtl_printf ("    Next_Frame_Free-> %u ",  EXT_Pool->Next_Frame_Free);
#else  
  printf ("(storage) Cabecera del pool %u\n", (unsigned int) Pool);
  printf ("  First_Frame     -> %u\n",  EXT_Pool->First_Frame);
  printf ("  Item_Size       -> %u\n",  EXT_Pool->Item_Size);
  printf ("  How_Many_Items  -> %u\n",  EXT_Pool->How_Many_Items);

  
  printf ("  Chain           ->\n");
  printf ("    Next_Frame_Free-> %u ",  EXT_Pool->Next_Frame_Free);
#endif

  if (EXT_Pool->Next_Frame_Free != 0) {
    i = 0;
    EXT_Ptr = EXTERNAL (EXT_Pool->Next_Frame_Free, Stoid);
    
    while (*EXT_Ptr != 0) {
      if ( i >= EXT_Pool->How_Many_Items) {
#ifdef MODULE
	rtl_printf (" -> NO 0 pointer\n");
#else
	printf (" -> NO 0 pointer\n");
#endif
	break;
      }
#ifdef MODULE
      rtl_printf ("-> %u ", *EXT_Ptr);
#else
      printf ("-> %u ", *EXT_Ptr);
#endif
      EXT_Ptr = EXTERNAL (*EXT_Ptr, Stoid);
      i = i + 1;
    }
#ifdef MODULE
    rtl_printf ("-> %u ", *EXT_Ptr);
#else
    printf ("-> %u ", *EXT_Ptr);
#endif
  }
#ifdef MODULE
  rtl_printf ("\n");
#else
  printf ("\n");
#endif
} 
#endif   ///DEBUG


//////////////////////////////////////////
//  _Storage_Create
//////////////////////////////////////////

int
_Storage_Create (void           *Initial_Address,
		 unsigned long   Storage_Size,
		 unsigned long   Max_Entities,
		 Storage_T      *Stoid
		)
{
  
  _Storage_Header  *EXT_Header_Ptr;
  internal_address *EXT_Entity_Index_Table;
  unsigned long     i;


  EXT_Header_Ptr = (_Storage_Header *) Initial_Address;


  if (EXT_Header_Ptr->Magic == STORAGE_MAGIC) {
    // Storage was previously created, just bind it
#ifdef DEBUG
#ifdef MODULE
    rtl_printf("(rtl_storage.c) Storage already exists\n");
#else
    printf("(storage.c) Storage already exists\n");
#endif
#endif
    *Stoid = Initial_Address;
    return STORAGE_SUCCESS;
  }

#ifdef DEBUG
#ifdef MODULE
    rtl_printf("(rtl_storage.c) NEW Storage\n");
#else
    printf("(storage.c) NEW Storage\n");
#endif
#endif
  
  EXT_Header_Ptr->Next_Free = sizeof(_Storage_Header) +         
    Max_Entities * sizeof (void *);    

  EXT_Header_Ptr->Last_Address = Storage_Size - 1;  

  if (EXT_Header_Ptr->Next_Free >= EXT_Header_Ptr->Last_Address) {
    *Stoid = NULL;
    return STORAGE_FAILURE;
  } 

  EXT_Header_Ptr->Magic              = STORAGE_MAGIC;
  EXT_Header_Ptr->Storage_Size       = Storage_Size;
  EXT_Header_Ptr->Max_Entities       = Max_Entities;
  EXT_Header_Ptr->Entity_Index_Table = sizeof(_Storage_Header); 

  EXT_Entity_Index_Table = EXTERNAL (EXT_Header_Ptr->Entity_Index_Table, Initial_Address);

  for (i = 0; i < Max_Entities; i++ ) {
    EXT_Entity_Index_Table[i] = 0;
  }

  *Stoid = Initial_Address;  


#ifdef DEBUG
  Show_Storage (*Stoid);
#endif


  return STORAGE_SUCCESS;    
}





//////////////////////////////////////////
// Artis Storage Management
//////////////////////////////////////////

#ifdef ARTIS
int
Storage_Create (char           *Name,
		unsigned long   Storage_Size,
		unsigned long   Max_Entities,
		Storage_T      *Stoid
		)
{
  volatile void *Initial_Address;
  unsigned long Real_Size;

  Real_Size = Storage_Size;

  Initial_Address = (volatile void *) mbuff_attach(Name, (size_t) Real_Size);

#ifdef DEBUG
#ifdef MODULE
  rtl_printf("(rtl_storage) Name: %s Address: %p\n",
	     Name, (void *) Initial_Address);
#else
  printf("(storage) Name: %s Address: %p\n",
	     Name, (void *) Initial_Address);
#endif
#endif

  if (Initial_Address == NULL) {
    return STORAGE_FAILURE;
  }

  return _Storage_Create ((void *) Initial_Address, 
			  Real_Size, Max_Entities, Stoid);
}


//////////////////////////////////////////
//  Storage_Destroy
//////////////////////////////////////////

int Storage_Destroy (char      *Name,
		     Storage_T  Stoid) 
{

  mbuff_detach(Name, (void *) Stoid);
  return STORAGE_SUCCESS;
}



#endif //ARTIS








//////////////////////////////////////////
//  Storage_Create_Object
//////////////////////////////////////////

void *
Storage_Create_Object (Storage_T             Stoid,
		       Storage_Entity_Size_T Size_Of_Object,
		       unsigned long         Entity_Id
		       )
{
  internal_address   The_Result;
  internal_address   Next_Free;
  _Storage_Header  *EXT_Header_Ptr;
  internal_address  *EXT_Entity_Index_Table;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return NULL;
  }

  The_Result  = EXT_Header_Ptr->Next_Free;  
  Next_Free   = EXT_Header_Ptr->Next_Free + Size_Of_Object;
 
  if (Next_Free > EXT_Header_Ptr->Last_Address) {
    return NULL;
  }

  EXT_Header_Ptr->Next_Free = Next_Free;

  EXT_Entity_Index_Table = EXTERNAL (EXT_Header_Ptr->Entity_Index_Table, Stoid);
  EXT_Entity_Index_Table[Entity_Id] = The_Result;

#ifdef DEBUG
#ifdef MODULE
    rtl_printf("(rtl_storage.c) New Storage object\n");
#else
    printf("(storage.c) New Storage object\n");
#endif
#endif
  


#ifdef DEBUG
  Show_Storage (Stoid);
#endif


  return (void *) The_Result; 
}




//////////////////////////////////////////
//  Storage_Get_Object
//////////////////////////////////////////

void *
Storage_Get_Object (Storage_T     Stoid,
		    unsigned long Entity_Id
		    )
{
  _Storage_Header *EXT_Header_Ptr;
  internal_address *EXT_Entity_Index_Table;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return NULL;
  }

  if (Entity_Id >= EXT_Header_Ptr->Max_Entities) {
    return NULL;
  }

  EXT_Entity_Index_Table = (internal_address *) EXTERNAL (EXT_Header_Ptr->Entity_Index_Table, Stoid);

  return (void *) EXT_Entity_Index_Table[Entity_Id];
}




//////////////////////////////////////////
//  Storage_Create_Pool
//////////////////////////////////////////

Storage_Pool_Handler_T
Storage_Create_Pool (Storage_T             Stoid,
		     Storage_Entity_Size_T Item_Size, 
		     unsigned int          How_Many_Items,
		     unsigned long         Entity_Id
		     )

{
  int                    Pool_Size;
  int                    i;
  _Pool_Header      *EXT_Pool;
  internal_address   Pool_Start_Address;
  internal_address  *P;


  if (Item_Size < sizeof (void *) ) {
    Item_Size = sizeof (void *);
  }

  Pool_Size = sizeof (_Pool_Header) + How_Many_Items * Item_Size;

  Pool_Start_Address = (internal_address ) Storage_Create_Object (Stoid, Pool_Size, Entity_Id);

  if ( Pool_Start_Address == (internal_address) NULL) {
    return NULL;
  }
  
  EXT_Pool = EXTERNAL (Pool_Start_Address, Stoid);

  EXT_Pool->Item_Size       = Item_Size;
  EXT_Pool->How_Many_Items  = How_Many_Items;
  EXT_Pool->Next_Frame_Free = Pool_Start_Address + sizeof (_Pool_Header); 
  EXT_Pool->First_Frame     = Pool_Start_Address + sizeof (_Pool_Header); 


  P = (internal_address *) EXTERNAL (EXT_Pool->Next_Frame_Free, Stoid);

  for (i = 1; i < How_Many_Items ; i++) {
    P[i - 1] = EXT_Pool->First_Frame + Item_Size * i;
 }

  P[How_Many_Items - 1] = 0;


#ifdef DEBUG
  Show_Storage (Stoid);
  Show_Pool (Stoid, Pool_Start_Address);
#endif


  return (void *) Pool_Start_Address;
}




//////////////////////////////////////////
//  Storage_Get_Pool
//////////////////////////////////////////

Storage_Pool_Handler_T
Storage_Get_Pool (Storage_T     Stoid,
		  unsigned long Entity_Id
		  )
{
  return (Storage_Pool_Handler_T) Storage_Get_Object (Stoid, Entity_Id);
}
  



//////////////////////////////////////////
//  Storage_Delete_Pool
//////////////////////////////////////////

int
Storage_Delete_Pool(Storage_T              Stoid,
		    Storage_Pool_Handler_T The_Pool
		    )
{
  _Pool_Header     *EXT_Pool;
  _Storage_Header  *EXT_Header_Ptr;
  internal_address  *P;
  int                i;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return STORAGE_FAILURE;
  }

  EXT_Pool = (_Pool_Header *) EXTERNAL ((internal_address) The_Pool, Stoid);

  EXT_Pool->Next_Frame_Free = EXT_Pool->First_Frame;

  P = (internal_address *) EXTERNAL (EXT_Pool->Next_Frame_Free, Stoid);

  for (i = 1; i < EXT_Pool->How_Many_Items ; i++) {
    P[i - 1] = EXT_Pool->First_Frame + EXT_Pool->Item_Size * i;
 }

  P[EXT_Pool->How_Many_Items - 1] = 0;


#ifdef DEBUG
  Show_Pool (Stoid, (internal_address) The_Pool);
#endif

  
  return STORAGE_SUCCESS;
}




//////////////////////////////////////////
//  Storage_Destroy_Pool
//////////////////////////////////////////
//  DANGEROUS FUNCTION
//  
//  This function destroys a pool and
//  all the pools and objects which are
//  after this pool in the storage !!!!!
//////////////////////////////////////////

int
Storage_Destroy_Pool (Storage_T              Stoid,
		      Storage_Pool_Handler_T The_Pool
		      )
{
  _Storage_Header  *EXT_Header_Ptr;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return STORAGE_FAILURE;
  }

  EXT_Header_Ptr->Next_Free = (internal_address) The_Pool;

#ifdef DEBUG
  Show_Storage (Stoid);
#endif

  return STORAGE_SUCCESS;
}






//////////////////////////////////////////
//  Storage_Create_Item
//////////////////////////////////////////

void *
Storage_Create_Item (Storage_T              Stoid,
		     Storage_Pool_Handler_T In_Pool
		     )
{
  _Pool_Header     *EXT_Pool;
  internal_address  The_Item;
  internal_address *EXT_Ptr; 
  _Storage_Header  *EXT_Header_Ptr;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return NULL;
  }

  EXT_Pool = (_Pool_Header *) EXTERNAL ((internal_address) In_Pool, Stoid);

  The_Item = EXT_Pool->Next_Frame_Free;

  if (The_Item == 0) {
    return NULL;
  }
     
  EXT_Ptr = (internal_address *) EXTERNAL (The_Item, Stoid);

  EXT_Pool->Next_Frame_Free = *EXT_Ptr;


#ifdef DEBUG
  Show_Pool (Stoid, (internal_address) In_Pool);
#endif


  return (void *) The_Item;
}




//////////////////////////////////////////
//  Storage_Destroy_Item
//////////////////////////////////////////

int
Storage_Destroy_Item (Storage_T               Stoid,
		      Storage_Pool_Handler_T  In_Pool,
		      void                    *One_Item
		      )
{

  internal_address  The_Item;
  internal_address *EXT_Ptr; 
  _Pool_Header     *EXT_Pool;
  _Storage_Header  *EXT_Header_Ptr;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return STORAGE_FAILURE;
  }

  EXT_Pool = (_Pool_Header *) EXTERNAL ((internal_address) In_Pool, Stoid);

  The_Item = (internal_address) One_Item;

  EXT_Ptr  = (internal_address *) EXTERNAL (The_Item, Stoid);

  *EXT_Ptr = EXT_Pool->Next_Frame_Free;

  EXT_Pool->Next_Frame_Free = The_Item;


#ifdef DEBUG
  Show_Pool (Stoid, (internal_address) In_Pool);
#endif


  return STORAGE_SUCCESS;
}




//////////////////////////////////////////
//  Storage_Get_Item
//////////////////////////////////////////

void *
Storage_Get_Item (Storage_T              Stoid,
		  Storage_Pool_Handler_T In_Pool, 
		  unsigned int           iItem_Position
		  )
{
  _Storage_Header  *EXT_Header_Ptr;
  _Pool_Header     *EXT_Pool;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return NULL;
  }

  EXT_Pool = (_Pool_Header *) EXTERNAL ((internal_address) In_Pool, Stoid);

  if (iItem_Position >= EXT_Pool->How_Many_Items) {
    return NULL;
  }

  return (void *) EXT_Pool->First_Frame + iItem_Position * EXT_Pool->Item_Size;
}




//////////////////////////////////////////
//  Storage_Get_Address
//////////////////////////////////////////

inline void * 
Storage_Get_Address (Storage_T  Stoid,
		     void      *internal_addr
		     )
{
  _Storage_Header        *EXT_Header_Ptr;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return NULL;
  }

  return EXTERNAL ((unsigned int) internal_addr, Stoid);
}


//////////////////////////////////////////
//  Storage_Get_Relative
//////////////////////////////////////////

inline void * 
Storage_Get_Relative (Storage_T  Stoid,
		     void      *external_addr
		     )
{
  _Storage_Header        *EXT_Header_Ptr;


  EXT_Header_Ptr = (_Storage_Header *) Stoid;

  if (EXT_Header_Ptr->Magic != STORAGE_MAGIC) {
    return NULL;
  }

  return INTERNAL ((unsigned int) external_addr, Stoid);
}


#ifdef MODULE
int storage_init_module (void) {
  return 0;
}

void storage_cleanup_module (void) {
}

#endif
