/*

   FILE:
	time.c						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-11-05

  CHANGELOG:


  DESCRIPTION:

    This file implements the TIME interface to all the parts of ARTIS
    which need time access (including the user).

    In both cases, time are measured as a linear (long long) variable
    in which time is expressed in nanoseconds. This is the natural
    time unit in RT-Linux, but not in Linux. As a result, an explicit
    function for consulting the current time has to be provided to the
    "linux_side" process, while in RT-Linux, the original function is
    used (altough we have renamed it, in order to present a common
    API).

    In Linux, the interface function for consulting the system time is
    "gettimeofday", which returns the current time in a timeval
    structure. This structure is as follows:

      struct timeval {
               long    tv_sec; 
               long    tv_usec;
       };

     where "tv_sec" are seconds and "tv_usec" are microseconds.

     In this implementation, the current time in Linux is returned as
     a systime_t (long long int) value containing the current time
     relative to the moment "linux_side" started to run, and converted
     to nanoseconds.



*/

#include <sys/time.h>
#include <unistd.h>

#include "../include/time.h"




#ifndef MODULE
static long init_seconds;
#endif




#ifdef MODULE



#else // ! MODULE


#define NORMALIZE(t) \
     (systime_t) ((((systime_t) t.tv_sec) * 1000000000) + \
                  (((systime_t) t.tv_usec) * 1000))





void 
systime_init (void) {
  struct timeval ini_time;
  
  gettimeofday(&ini_time, NULL);
  init_seconds = ini_time.tv_sec;
}




systime_t  
systime_gettime (void) {

  struct timeval time_now;

  gettimeofday(&time_now, NULL);

  time_now.tv_sec -= init_seconds;

  return  NORMALIZE(time_now);
}





systime_t
systime_timeval_to_time (struct timeval st) {

  return NORMALIZE(st);
}



struct timeval  
systime_time_to_timeval (systime_t t) {
  struct timeval st;

  t /= 1000;
  st.tv_usec = (long) (t % 1000000);
  st.tv_sec  = (long) (t / 1000000);
  return st;
}




#endif // #ifdef MODULE



