/*
  file: parser.c
  date: dec 04
  author: J. Fabregat

  purpose: translate xml file with entities definition into C structures

  based on kdm parser by M. Rebollo
*/

#include <stdio.h>
#include <string.h>
#include <expat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define MARK(f, msg) //printf( "(%s) %s(%d)\n", f, msg, userdata->depth )
//#define IMPRIME_AGENTE


#define AGENT 0

#define BEHAVIOR 1 
#define IN_AGENT 2
#define MKS 3
#define KS 4

#define BEHAVIOR_NAME 11
#define BEHAVIOR_IN_AGENTS 12

#define IN_AGENT_NAME 21
#define IN_AGENT_IMPORTANCE 22
#define IN_AGENT_PERIOD 23
#define IN_AGENT_DEADLINE 24
#define IN_AGENT_PERCEPTION 25
#define IN_AGENT_COGNITION 26
#define IN_AGENT_ACTION 27

#define MKS_IMPORTANCE 31
#define MKS_METHOD 32
#define MKS_TYPE 33
#define MKS_LEVEL 34
#define MKS_NAME 35

#define MKS_LEVEL_QUALITY 342

#define KS_CODEFILE 41
#define KS_NAME 42
#define KS_WCET 43

#define TYPE_IN_AGENT 100
#define TYPE_MKS 101
#define TYPE_KS 102

#define METHOD_REFINEMENT 103
#define METHOD_MULTIPLE 104
#define TYPE_MANDATORY 105
#define TYPE_OPTIONAL 106

#define MAXBUFF 256

typedef struct
{
  int type;
  int id;
}element;

typedef struct
{
  int id;
  char *name;
  element *in_agents;
  int n_inagents;
}behavior_data;

typedef struct
{
  int id;
  char *name;
  int importance;
  unsigned long deadline;
  unsigned long period;
  element *perception;
  int n_perception;
  element *cognition;
  int n_cognition;
  element action;
}in_agent_data;

typedef struct
{
  int id;
  element *ks;
  int n_ks;
  float quality;
}level;

typedef struct
{
  int id;
  char *name;
  int importance;
  int method;
  int type;
  level *levels;
  int n_levels;
}mks_data;

typedef struct
{
int id;
char *codefile;
char *name;
unsigned long wcet;
}ks_data;

typedef struct userdata{
  XML_Parser parser;
  int wrong_version;
  int depth;
  int entity;
  behavior_data *behaviors;
  in_agent_data *in_agents;
  mks_data *mks;
  ks_data *ks;
  int n_behaviors;
  int n_in_agents;
  int n_mks;
  int n_ks;
  char cname[MAXBUFF];
  char sname[MAXBUFF];
  char stype[MAXBUFF];
  char slength[MAXBUFF];
  char elemtype[MAXBUFF];
  char iname[MAXBUFF];
  char svalue[MAXBUFF];
}XML_State;



enum deeplevel{ TBNLVL, CLASSLVL, SLOTLVL, ARRAYLVL, CONSTLVL };

void analyse( const char *filename );


static void
//------------------------------------------------------------
xmldecl_handler(void            *userData,
                const XML_Char  *version,
                const XML_Char  *encoding,
                int              standalone)
//------------------------------------------------------------
{
  static const XML_Char Version_1_0[] = {'1', '.', '0', 0};

  int i, len;  
  XML_State *userdata = userData;

  len = strlen( Version_1_0 );
  for (i = 0; i < len; i++) {
    if (version[i] != Version_1_0[i]) {
      userdata->wrong_version = 1;
      //clear all handlers
      XML_SetCharacterDataHandler(userdata->parser, NULL);
      XML_SetElementHandler( userdata->parser, NULL, NULL );
      XML_SetCharacterDataHandler( userdata->parser, NULL );
      XML_SetCommentHandler( userdata->parser, NULL );
      XML_SetProcessingInstructionHandler( userdata->parser, NULL );
      fprintf( stderr, "Wrong version of XML document\n" ); 
      exit( -1 );
    } //if wrong version detected
  }//for i
} //xmldecl_handler()


static void
//------------------------------------------------------------
xmlstartelem_handler( void        *userData,
		      const char  *el, 
		      const char  **attr )
//------------------------------------------------------------
{
  int i;
  XML_State *userdata = userData;
  int schema_file = 0;

  MARK( "xmlstartelem_handler", el );

  if ((strcmp(el,"agent")==0)&&((userdata->depth) == 0))
  {
	(userdata->entity) = AGENT; 
	userdata->n_behaviors=0;
	userdata->n_in_agents=0;
	userdata->n_mks=0;
	userdata->n_ks=0;
	userdata->behaviors=NULL;
	userdata->in_agents=NULL;
	userdata->mks=NULL;
	userdata->ks=NULL;
   }
  else if (userdata->depth == 1)
  {
    if (strcmp(el,"behavior")==0)
    {
	(userdata->entity) = BEHAVIOR;
 	userdata->behaviors=realloc(userdata->behaviors,(1+userdata->n_behaviors)*sizeof(behavior_data));
	userdata->behaviors[userdata->n_behaviors].id=atoi(attr[1]);
    }
    else if (strcmp(el,"in-agent")==0)
    {
 	userdata->in_agents=realloc(userdata->in_agents,(1+userdata->n_in_agents)*sizeof(in_agent_data));
	(userdata->in_agents[userdata->n_in_agents].id)=atoi(attr[1]);
	(userdata->entity) = IN_AGENT; 
    }
    else if (strcmp(el,"ks")==0)
    {
 	userdata->ks=realloc(userdata->ks,(1+userdata->n_ks)*sizeof(ks_data));
	(userdata->ks[userdata->n_ks].id)=atoi(attr[1]);
	(userdata->entity) = KS; 
    }
    else if (strcmp(el,"mks")==0)
    {
	(userdata->entity) = MKS; 
 	userdata->mks=realloc(userdata->mks,(1+userdata->n_mks)*sizeof(mks_data));
	(userdata->mks[userdata->n_mks].id)=atoi(attr[1]);
	userdata->mks[userdata->n_mks].n_levels=0;
    }
  }
  else if (userdata->depth > 1)
  {
	switch (userdata->entity)
	{
	  case BEHAVIOR: 
		if(strcmp(el,"name")==0) userdata->entity=BEHAVIOR_NAME;
		if(strcmp(el,"in-agents")==0) 
		{
			userdata->entity=BEHAVIOR_IN_AGENTS;
			userdata->behaviors[userdata->n_behaviors].n_inagents=0;
		}
		break;
  	  case IN_AGENT: 
		if(strcmp(el,"name")==0) userdata->entity=IN_AGENT_NAME;
		if(strcmp(el,"importance")==0) userdata->entity=IN_AGENT_IMPORTANCE;
		if(strcmp(el,"period")==0) userdata->entity=IN_AGENT_PERIOD;
		if(strcmp(el,"deadline")==0) userdata->entity=IN_AGENT_DEADLINE;
		if(strcmp(el,"perception")==0) 
		{
			userdata->entity=IN_AGENT_PERCEPTION;
			userdata->in_agents[userdata->n_in_agents].n_perception=0;
		}
		if(strcmp(el,"cognition")==0) 
		{
			userdata->entity=IN_AGENT_COGNITION;
			userdata->in_agents[userdata->n_in_agents].n_cognition=0;
		}
		if(strcmp(el,"action")==0) 
			userdata->entity=IN_AGENT_ACTION;
		break;
	  case MKS: 
		if(strcmp(el,"name")==0) userdata->entity=MKS_NAME;
		if(strcmp(el,"importance")==0) userdata->entity=MKS_IMPORTANCE;
		if(strcmp(el,"method")==0) userdata->entity=MKS_METHOD;
		if(strcmp(el,"type")==0) userdata->entity=MKS_TYPE;
		if(strcmp(el,"level")==0) 
		{	
		userdata->entity=MKS_LEVEL;
		userdata->mks[userdata->n_mks].levels=realloc( (userdata->mks[userdata->n_mks].levels) , (1+userdata->mks[userdata->n_mks].n_levels) * (sizeof (level) ) );
		   for (i = 0; attr[i]; i += 2) 
		   {
 		     if( strcmp( attr[i], "id" ) == 0 ) userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].id=atoi(attr[i+1]);
		   }
		userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].ks=NULL;
		userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].n_ks=0;
		}
		break;
	  case KS: 
		if(strcmp(el,"codefile")==0) userdata->entity=KS_CODEFILE;
		if(strcmp(el,"name")==0) userdata->entity=KS_NAME;
		if(strcmp(el,"wcet")==0) userdata->entity=KS_WCET;
		break;
	  case BEHAVIOR_IN_AGENTS:
		if(strcmp(el,"element")==0)
		{
		userdata->behaviors[userdata->n_behaviors].in_agents=realloc(userdata->behaviors[userdata->n_behaviors].in_agents,(1+userdata->behaviors[userdata->n_behaviors].n_inagents)*sizeof(element));
		for (i = 0; attr[i]; i += 2) 
		{
 		     if( strcmp( attr[i], "id" ) == 0 )
		     userdata->behaviors[userdata->n_behaviors].in_agents[userdata->behaviors[userdata->n_behaviors].n_inagents].id=atoi(attr[i+1]);
    		} //for i
		userdata->behaviors[userdata->n_behaviors].in_agents[userdata->behaviors[userdata->n_behaviors].n_inagents].type=TYPE_IN_AGENT;
		userdata->behaviors[userdata->n_behaviors].n_inagents++;
		}
		break;
	  case IN_AGENT_PERCEPTION:
		if(strcmp(el,"element")==0)
		{
		userdata->in_agents[userdata->n_in_agents].perception=realloc(userdata->in_agents[userdata->n_in_agents].perception,(1+userdata->in_agents[userdata->n_in_agents].n_perception)*sizeof(element) ) ;
		for (i = 0; attr[i]; i += 2) 
		{
 		     if( strcmp( attr[i], "id" ) == 0 )
		     {
		     userdata->in_agents[userdata->n_in_agents].perception[userdata->in_agents[userdata->n_in_agents].n_perception].id=atoi(attr[i+1]);
		     }
    		} //for i
		userdata->in_agents[userdata->n_in_agents].perception[userdata->in_agents[userdata->n_in_agents].n_perception].type=TYPE_MKS;
		userdata->in_agents[userdata->n_in_agents].n_perception++;
		}
		break;
	  case IN_AGENT_COGNITION:
		if(strcmp(el,"element")==0)
		{
		userdata->in_agents[userdata->n_in_agents].cognition=realloc(userdata->in_agents[userdata->n_in_agents].cognition,(1+userdata->in_agents[userdata->n_in_agents].n_cognition)*sizeof(element));
		for (i = 0; attr[i]; i += 2) 
		{
 		     if( strcmp( attr[i], "id" ) == 0 )
		     userdata->in_agents[userdata->n_in_agents].cognition[userdata->in_agents[userdata->n_in_agents].n_cognition].id=atoi(attr[i+1]);
		}
		userdata->in_agents[userdata->n_in_agents].cognition[userdata->in_agents[userdata->n_in_agents].n_cognition].type=TYPE_MKS;
		userdata->in_agents[userdata->n_in_agents].n_cognition++;
		}
		break;
	  case IN_AGENT_ACTION:
		if(strcmp(el,"element")==0)
		{
		for (i = 0; attr[i]; i += 2) 
		{
 		     if( strcmp( attr[i], "id" ) == 0 )
  		     userdata->in_agents[userdata->n_in_agents].action.id=atoi(attr[i+1]);
		}
		userdata->in_agents[userdata->n_in_agents].action.type=TYPE_KS;
		}
		break;
	  case MKS_LEVEL:
		if(strcmp(el,"element")==0)
		{
		userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].ks=realloc(userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].ks,((1+userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].n_ks)*sizeof(element)));
		  for (i = 0; attr[i]; i += 2) 
		  {
 		     if( strcmp( attr[i], "id" ) == 0 )
		     userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].ks[userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].n_ks].id=atoi(attr[i+1]);
		  }
		userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].ks[userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].n_ks].type=TYPE_KS;
		userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].n_ks++;	
		}


		if(strcmp(el,"quality")==0) userdata->entity=MKS_LEVEL_QUALITY; 
		break;
	}
}

  (userdata->depth)++;
} //xmlstartelem_handler()


static void
//------------------------------------------------------------
xmlendelem_handler( void        *userData, 
		    const char  *el )
//------------------------------------------------------------
{
  XML_State *userdata = userData;

  MARK( "xmlendelem_handler", el );
  (userdata->depth)--;
  if (userdata->depth == 1)
  {
    if (strcmp(el,"behavior")==0) userdata->n_behaviors++;
    else if (strcmp(el,"in-agent")==0) userdata->n_in_agents++;
    else if (strcmp(el,"ks")==0) userdata->n_ks++;
    else if (strcmp(el,"mks")==0) userdata->n_mks++;
  }
  else if (userdata->depth > 1)
  {
   switch(userdata->entity)
    {
	case BEHAVIOR_NAME:
		if(strcmp(el,"name")==0) userdata->entity=BEHAVIOR;
		break;
	case BEHAVIOR_IN_AGENTS:
		if(strcmp(el,"in-agents")==0) userdata->entity=BEHAVIOR;
	case IN_AGENT_NAME:
		if(strcmp(el,"name")==0) userdata->entity=IN_AGENT;
		break;
	case IN_AGENT_IMPORTANCE:
		if(strcmp(el,"importance")==0) userdata->entity=IN_AGENT;
		break;
	case IN_AGENT_PERIOD:
		if(strcmp(el,"period")==0) userdata->entity=IN_AGENT;
		break;
	case IN_AGENT_DEADLINE:
		if(strcmp(el,"deadline")==0) userdata->entity=IN_AGENT;
		break;
	case IN_AGENT_PERCEPTION:
		if(strcmp(el,"perception")==0) userdata->entity=IN_AGENT;
		break;
	case IN_AGENT_COGNITION:
	        if(strcmp(el,"cognition")==0) userdata->entity=IN_AGENT;
		break;
	case IN_AGENT_ACTION:
	        if(strcmp(el,"action")==0) userdata->entity=IN_AGENT;
		break;
	case MKS_NAME:
		if(strcmp(el,"name")==0) userdata->entity=MKS;
		break;
	case MKS_IMPORTANCE:
	        if(strcmp(el,"importance")==0) userdata->entity=MKS;
		break;
	case MKS_METHOD:
	        if(strcmp(el,"method")==0) userdata->entity=MKS;
		break;
	case MKS_TYPE:
	        if(strcmp(el,"type")==0) userdata->entity=MKS;
		break;
	case MKS_LEVEL:
		if(strcmp(el,"level")==0) 
    		{
	 	  userdata->mks[userdata->n_mks].n_levels++;
	 	  userdata->entity=MKS;
    		}
		break;
	case MKS_LEVEL_QUALITY:
		if(strcmp(el,"quality")==0) userdata->entity=MKS_LEVEL;
		break;
	case KS_CODEFILE:
		if(strcmp(el,"codefile")==0) userdata->entity=KS;
		break;
	case KS_NAME:
		if(strcmp(el,"name")==0) userdata->entity=KS;
		break;
	case KS_WCET:
		if(strcmp(el,"wcet")==0) userdata->entity=KS;
		break;
	}//end switch*/
    }//end if
} //xmlendelem_handler() 



static void
//------------------------------------------------------------
xmltext_handler( void       *userData, 
		 const char *el, 
		 int        len )
//------------------------------------------------------------
{
  int i, j;
  char *buff;
  XML_State *userdata = userData;

  switch(userdata->entity)
  {
	case BEHAVIOR_NAME:
		userdata->behaviors[userdata->n_behaviors].name=malloc((1+len)*sizeof(char));
		snprintf(userdata->behaviors[userdata->n_behaviors].name,len+1,el);
		break;
	case IN_AGENT_NAME:
		userdata->in_agents[userdata->n_in_agents].name=malloc((1+len)*sizeof(char));
		snprintf(userdata->in_agents[userdata->n_in_agents].name,len+1,el);
break;
	case IN_AGENT_IMPORTANCE:
  		if((buff=malloc((1+len)*sizeof(char)))!=NULL)
		{
		snprintf(buff,len+1,el);
		userdata->in_agents[userdata->n_in_agents].importance=atoi(buff);
		}
		break;
	case IN_AGENT_PERIOD:
  		if((buff=malloc((1+len)*sizeof(char)))!=NULL)
		{
		snprintf(buff,len+1,el);
		userdata->in_agents[userdata->n_in_agents].period=atoi(buff);
		}
		break;
	case IN_AGENT_DEADLINE:
  		if((buff=malloc((1+len)*sizeof(char)))==NULL) exit(-1);
		snprintf(buff,len+1,el);
		userdata->in_agents[userdata->n_in_agents].deadline=atoi(buff);
		break;
	case MKS_NAME:
		userdata->mks[userdata->n_mks].name=malloc((1+len)*sizeof(char));
		snprintf(userdata->mks[userdata->n_mks].name,len+1,el);
		break;
	case MKS_IMPORTANCE:
  		if((buff=malloc((1+len)*sizeof(char)))==NULL) exit(-1);
		snprintf(buff,len+1,el);
		userdata->mks[userdata->n_mks].importance=atoi(buff);
		break;
	case MKS_METHOD:
  		if((buff=malloc((1+len)*sizeof(char)))==NULL) exit(-1);
		snprintf(buff,len+1,el);
		if(strcmp(buff,"Refinement")==0)userdata->mks[userdata->n_mks].method=METHOD_REFINEMENT;
		else if (strcmp(buff,"Multiple")==0)userdata->mks[userdata->n_mks].method=METHOD_MULTIPLE;
		break;
	case MKS_TYPE:
  		if((buff=malloc((1+len)*sizeof(char)))==NULL) exit(-1);
		snprintf(buff,len+1,el);
		if(strcmp(buff,"Mandatory")==0)userdata->mks[userdata->n_mks].type=TYPE_MANDATORY;
		else if (strcmp(buff,"Optional")==0)userdata->mks[userdata->n_mks].type=TYPE_OPTIONAL;
		break;
	case MKS_LEVEL_QUALITY:
  		if((buff=malloc((1+len)*sizeof(char)))==NULL) exit(-1);
		snprintf(buff,len+1,el);
		userdata->mks[userdata->n_mks].levels[userdata->mks[userdata->n_mks].n_levels].quality=atof(buff);
		break;
	case KS_CODEFILE:
		userdata->ks[userdata->n_ks].codefile=malloc((1+len)*sizeof(char));
		snprintf(userdata->ks[userdata->n_ks].codefile,len+1,el);
		break;
	case KS_NAME:
		userdata->ks[userdata->n_ks].name=malloc((1+len)*sizeof(char));
		snprintf(userdata->ks[userdata->n_ks].name,len+1,el);
		break;
	case KS_WCET:
  		if((buff=malloc((1+len)*sizeof(char)))==NULL) exit(-1);
		snprintf(buff,len+1,el);
		userdata->ks[userdata->n_ks].wcet=atoi(buff);
		break;
     }//End switch 
} //xmltext_handler()



int print_agent(XML_State *userdata)
{
int i,a,m;

printf("---AGENTE ARTIS---\n\n");
printf("\nCOMPORTAMIENTOS %d\n",userdata->n_behaviors);
for(i=0; i<userdata->n_behaviors; i++) //IMPRIMIMOS LOS COMPORTAMIENTOS
	{
	printf("Comportamiento %d\n\tId:%d\n\tName:%s\n\tNº In-Agents:%d\n",i,userdata->behaviors[i].id,userdata->behaviors[i].name,userdata->behaviors[i].n_inagents);
	for(a=0; a<userdata->behaviors[i].n_inagents; a++)
		{
		printf("\t\tin-agent %d\n",a);
		printf("\t\t Id:%d Type:%d\n",userdata->behaviors[i].in_agents[a].id,userdata->behaviors[i].in_agents[a].type);
		}
	}
printf("\nIN-AGENTS %d\n",userdata->n_in_agents);
for(i=0; i<userdata->n_in_agents; i++) //IMPRIMIMOS LOS IN-AGENTS
	{
	printf("In-Agent %d\n",i);
	printf("\tId:%d\n\tName:%s\n\tDeadline:%d\n\tPeriod:%d\n",userdata->in_agents[i].id,userdata->in_agents[i].name,userdata->in_agents[i].deadline,userdata->in_agents[i].period);
	printf("\tPercepción (%d mks)\n",userdata->in_agents[i].n_perception);
		for(a=0; a<userdata->in_agents[i].n_perception; a++)
		{
		printf("\t\tmks %d\n",a);
		printf("\t\t Id:%d Type:%d\n",userdata->in_agents[i].perception[a].id,userdata->in_agents[i].perception[a].type);
		}
	printf("\tCognición (%d mks)\n",userdata->in_agents[i].n_cognition);
		for(a=0; a<userdata->in_agents[i].n_cognition; a++)
		{
		printf("\t\tmks %d\n",a);
		printf("\t\t Id:%d Type:%d\n",userdata->in_agents[i].cognition[a].id,userdata->in_agents[i].cognition[a].type);
		}
printf("\tAcción (1 ks)\n");
	printf("\t\t Id:%d\n\t\tType:%d\n",userdata->in_agents[i].action.id,userdata->in_agents[i].action.type);
	}
printf("\nMKS %d\n",userdata->n_mks);
for(i=0; i<userdata->n_mks; i++) //IMPRIMIMOS LOS MKS
	{
	printf(" Mks %d\n",i);
	printf("\tId: %d\n\tName: %s\n\tImportance: %d\n\tMethod: %d\n\tType: %d\n",userdata->mks[i].id,userdata->mks[i].name,userdata->mks[i].importance,userdata->mks[i].method,userdata->mks[i].type);
	printf("\tLevels %d\n",userdata->mks[i].n_levels);
	for(a=0; a<userdata->mks[i].n_levels; a++)
		{
			printf("\t\tLevel %d Id:%d Quality:%d\n",a,userdata->mks[i].levels[a].id,userdata->mks[i].levels[a].quality);
			for(m=0;m<userdata->mks[i].levels[a].n_ks;m++)
			printf("\t\t -Ks%d(Id:%d)\n",m,userdata->mks[i].levels[a].ks[m].id);
		}
	}
printf("\nKS %d\n",userdata->n_ks);
for(i=0; i<userdata->n_ks; i++) //IMPRIMIMOS LOS KS
	{
	printf(" Ks %d\n",i);
	printf("\tId: %d\n\tName: %s\n\tCodefile: %s\n\tWCET: %d\n",userdata->ks[i].id,userdata->ks[i].name,userdata->ks[i].codefile,userdata->ks[i].wcet);
	}


return 0;
}


int test_agent(XML_State *userdata)
{
int b,bi,i,imp,imc,m,ml,mlk,k;

for(b=0; b<userdata->n_behaviors; b++) //IMPRIMIMOS LOS COMPORTAMIENTOS
  {
  printf("behavior %d\n",b);
  for(bi=0; bi<userdata->behaviors[b].n_inagents; bi++)
    for(i=0; i<userdata->n_in_agents; i++)
      if(userdata->behaviors[b].in_agents[bi].id==userdata->in_agents[i].id) 
        {
	  printf("  El In-Agent %d posición %d pertence al comportamiento\n",userdata->in_agents[i].id,i);
	  printf("\tMks de percepción:\n");
	  for(imp=0;imp<userdata->in_agents[i].n_perception; imp++) for (m=0; m<userdata->n_mks; m++) if(userdata->in_agents[i].perception[imp].id==userdata->mks[m].id) 
	    {
	    printf("\t\tEl mks con id %d y posición %d se usa en la percepción del inagent\n",userdata->mks[m].id,m);
	    for(ml=0;ml<userdata->mks[m].n_levels;ml++)
	      for(mlk=0;mlk<userdata->mks[m].levels[ml].n_ks;mlk++) 
	        for(k=0;k<userdata->n_ks;k++)
		  if(userdata->mks[m].levels[ml].ks[mlk].id==userdata->ks[k].id)
		    printf("\t\t  -El ks %d id %d se usa en percepción en el nivel %d\n",mlk,userdata->ks[k].id,ml);
	    }
	  printf("\tMks de cognición:\n");
	  for(imc=0;imc<userdata->in_agents[i].n_cognition; imc++) for (m=0; m<userdata->n_mks; m++) if(userdata->in_agents[i].cognition[imc].id==userdata->mks[m].id) 
	    {
	    printf("\t\tEl mks con id %d y posición %d se usa en la percepción del inagent\n",userdata->mks[m].id,m);
	    for(ml=0;ml<userdata->mks[m].n_levels;ml++) 
	      for(mlk=0;mlk<userdata->mks[m].levels[ml].n_ks;mlk++) 
	        for(k=0;k<userdata->n_ks;k++)
		  if(userdata->mks[m].levels[ml].ks[mlk].id==userdata->ks[k].id)
		    printf("\t\t  -El ks %d id %d se usa en cognición en el nivel %d\n",mlk,userdata->ks[k].id,ml);
	    }
	  printf("\tKs de acción:\n");
	  for(k=0;k<userdata->n_ks;k++) if(userdata->in_agents[i].action.id==userdata->ks[k].id) printf("\t\tEl ks con id %d y posición %d es la parte de accion del agente\n",userdata->ks[k].id,k);
        }
  }
}

void generate_control_data(XML_State *userdata) // TO DO fi_x está mal generado, los ks que aparecen no son los correctos.
{
int i,m,ml,mlk,im,k;
FILE *f;
char mks_type[9];
char mks_mode[3];

printf("Generate control_data.c\n");
if(f=fopen("./control_data.c","w+"))
 {

fprintf(f,"#include <artis/dserver.h>\n");
fprintf(f,"#include <artis/em.h>\n");
fprintf(f,"#include <artis/agenda.h>\n");

fprintf(f,"void CreateStructures()\n{\n");
fprintf(f,"  inagent *nu_inagent;\n");
fprintf(f,"  mks *nu_mks;\n");
fprintf(f,"  ks *nu_ks;\n\n");

for(i=0;i<userdata->n_in_agents;i++)
  {
  fprintf(f,"  nu_inagent=create_in_agent(\"%s\", %d, %d);\n",userdata->in_agents[i].name,userdata->in_agents[i].id,userdata->in_agents[i].n_perception+userdata->in_agents[i].n_cognition);
  fprintf(f,"  l_in_agent=insert_in_agent(nu_inagent, l_in_agent);\n\n");

	// MKS de percepcion del In-agent
	for(im=0;im<userdata->in_agents[i].n_perception; im++) 
	  for(m=0;m<userdata->n_mks;m++) 
	    if(userdata->in_agents[i].perception[im].id==userdata->mks[m].id) 
	     {
	     if(userdata->mks[m].type==TYPE_MANDATORY) strcpy(mks_type,"Critical");
	     else strcpy(mks_type,"Optional");
	     if(userdata->mks[m].method==METHOD_REFINEMENT) strcpy(mks_mode,"RP");
	     else strcpy(mks_mode,"MP");
	     
	     fprintf(f,"    nu_mks=create_mks(\"%s\", %d, \"%s\", \"%s\", %d, %d, in_agent);\n",userdata->mks[m].name,userdata->mks[m].id,mks_type,mks_mode,userdata->mks[m].importance,userdata->mks[m].n_levels);
	     fprintf(f,"    insert_mks( &periodic_mks_l, nu_mks);\n");
	     fprintf(f,"    in_mks_at_in_agent(in_agent, nu_mks);\n\n");
	 
	     for(ml=0;ml<userdata->mks[m].n_levels;ml++) 
		for(mlk=0;mlk<userdata->mks[m].levels[ml].n_ks;mlk++)
		  for(k=0;k<userdata->n_ks;k++) 
		    if(userdata->mks[m].levels[ml].ks[mlk].id==userdata->ks[k].id)
		     {
		     fprintf(f,"\tnu_ks=create_ks(\"%s\", %d, %d, %f);\n",userdata->ks[k].name,k,userdata->ks[k].wcet*1000,userdata->mks[m].levels[ml].quality);
		     fprintf(f,"\tinsert_ks(nu_ks, nu_mks);\n\n");
		     }

	     }
	// MKS de cognicion del In-agent
	for(im=0;im<userdata->in_agents[i].n_cognition; im++) 
	  for(m=0;m<userdata->n_mks;m++) 
	    if(userdata->in_agents[i].cognition[im].id==userdata->mks[m].id) 
	     {
	     if(userdata->mks[m].type==TYPE_MANDATORY) strcpy(mks_type,"Critical");
	     else strcpy(mks_type,"Optional");
	     if(userdata->mks[m].method==METHOD_REFINEMENT) strcpy(mks_mode,"RP");
	     else strcpy(mks_mode,"MP");
	     fprintf(f,"    nu_mks=create_mks(\"%s\", %d, \"%s\", \"%s\", %d, %d:, in_agent);\n",userdata->mks[m].name,userdata->mks[m].id,mks_type,mks_mode,userdata->mks[m].importance,userdata->mks[m].n_levels);
	     fprintf(f,"    insert_mks( &periodic_mks_l, nu_mks);\n");
	     fprintf(f,"    in_mks_at_in_agent(in_agent, nu_mks);\n\n");
	 
	     for(ml=0;ml<userdata->mks[m].n_levels;ml++) 
		for(mlk=0;mlk<userdata->mks[m].levels[ml].n_ks;mlk++)
		  for(k=0;k<userdata->n_ks;k++) 
		    if(userdata->mks[m].levels[ml].ks[mlk].id==userdata->ks[k].id)
		     {
		     fprintf(f,"\tnu_ks=create_ks(\"%s\", %d, %d, %f);\n",userdata->ks[k].name,k,userdata->ks[k].wcet*1000,userdata->mks[m].levels[ml].quality);
		     fprintf(f,"\tinsert_ks(nu_ks, nu_mks);\n\n");
		     }

	     }
  }
 fprintf(f,"}\n");
 }
else printf("File control_data.c could not be opened\n");
fclose(f);
}



void generate_rtl_tasks(XML_State *userdata)
{
FILE *f;
int b,bi,i,im,a,m,mlk,k;
int rtks;
int *initial_wcet;
int *final_wcet;

initial_wcet=malloc(userdata->n_in_agents*sizeof(int));
final_wcet=malloc(userdata->n_in_agents*sizeof(int));;

printf("Generate rtl_tasks.c\n");
if(f=fopen("./rtl_tasks.c","w+"))
 {
 fprintf(f,"#include <linux/module.h>\n");
 fprintf(f,"#include <linux/kernel.h>\n");
 fprintf(f,"#include <linux/version.h>\n\n");

 fprintf(f,"#include <artis/artis_conf.h>\n");
 fprintf(f,"#include <artis/artis_user.h>\n");
 fprintf(f,"#include <artis/artis_app.h>\n");
 fprintf(f,"#include <artis/kdm.h>\n");

 fprintf(f,"#include \"userclass.h\"\n\n");

 fprintf(f,"MODULE_LICENSE(\"GPL\");\n\n");

 fprintf(f,"#define MUTEX1     1\n");

 fprintf(f,"#define N_TASKS    %d\n",userdata->n_in_agents);

 fprintf(f,"rt_task_t task[N_TASKS];\n\n");

 fprintf(f,"#ifdef ARTIS_CONF_METAREASONING\n");
 fprintf(f,"#define N_BEHAVS  %d\n",userdata->n_behaviors);

 fprintf(f,"rt_behav_task_t aTaskBeh[N_BEHAVS][N_TASKS];\n");

 fprintf(f,"rt_behav_t aBeh[N_BEHAVS];\n");
 fprintf(f,"#endif\n\n");

// Funciones de los KS
for(k=0;k<userdata->n_ks;k++)
{
   rtks=0;
   for(i=0;i<userdata->n_in_agents;i++) 
   {
     for(im=0;im<userdata->in_agents[i].n_perception;im++)
     { 
       for(m=0;m<userdata->n_mks;m++) 
         if((userdata->in_agents[i].perception[im].id==userdata->mks[m].id)&&(userdata->mks[m].type==TYPE_MANDATORY))
	   for(mlk=0;mlk<userdata->mks[m].levels[0].n_ks;mlk++)
	     if(userdata->mks[m].levels[0].ks[mlk].id==userdata->ks[k].id) rtks=1;
   }
   for(im=0;im<userdata->in_agents[i].n_cognition;im++) 
   {
     for(m=0;m<userdata->n_mks;m++) 
       if((userdata->in_agents[i].cognition[im].id==userdata->mks[m].id)&&(userdata->mks[m].type==TYPE_MANDATORY))
	 for(mlk=0;mlk<userdata->mks[m].levels[0].n_ks;mlk++)
	     if(userdata->mks[m].levels[0].ks[mlk].id==userdata->ks[k].id) rtks=1;
   }
   if(userdata->in_agents[i].action.id==userdata->ks[k].id) rtks=1;
  }

  if(rtks==1) fprintf(f,"int %s(){\n//Codigo de %s\n}\n",userdata->ks[k].name,userdata->ks[k].codefile);
} // Fin de for de funciones de los KS



 // Partes Iniciales y finales
 for(i=0;i<userdata->n_in_agents;i++) 
 {
   initial_wcet[i]=0;
   final_wcet[i]=0;
   fprintf(f,"void *fi_%d (void *dummy) \n{\n",i);
   for(im=0;im<userdata->in_agents[i].n_perception;im++)
   { 
     for(m=0;m<userdata->n_mks;m++) 
       if((userdata->in_agents[i].perception[im].id==userdata->mks[m].id)&&(userdata->mks[m].type==TYPE_MANDATORY))
	 for(mlk=0;mlk<userdata->mks[m].levels[0].n_ks;mlk++)
	   for(k=0;k<userdata->n_ks;k++)
	     if(userdata->mks[m].levels[0].ks[mlk].id==userdata->ks[k].id)
		{
		  fprintf(f,"  %s();\n",userdata->ks[k].name);
	    	  initial_wcet[i]+=userdata->ks[k].wcet;
		}

   }

   for(im=0;im<userdata->in_agents[i].n_cognition;im++) 
   {
     for(m=0;m<userdata->n_mks;m++) 
       if((userdata->in_agents[i].cognition[im].id==userdata->mks[m].id)&&(userdata->mks[m].type==TYPE_MANDATORY))
	 for(mlk=0;mlk<userdata->mks[m].levels[0].n_ks;mlk++)
	   for(k=0;k<userdata->n_ks;k++)
	     if(userdata->mks[m].levels[0].ks[mlk].id==userdata->ks[k].id)
		{
		  fprintf(f,"  %s();\n",userdata->ks[k].name);
	    	  initial_wcet[i]+=userdata->ks[k].wcet;
		}
   }
   fprintf(f,"  return (void *) 0;\n}\n");

   for(k=0;k<userdata->n_ks;k++) 
     if(userdata->in_agents[i].action.id==userdata->ks[k].id) 
	{
        fprintf(f,"void *ff_%d (void *dummy) \n{\n  %s();\n  return (void *) 0;\n}\n\n\n",i,userdata->ks[k].name);
	final_wcet[i]+=userdata->ks[k].wcet;
	}
 }

 fprintf(f,"int init_module(void) \n{\n\n");

 fprintf(f," struct task_create_info info[N_TASKS];\n");
 fprintf(f," int i;\n");

// Bucle de creación de los In-Agents
 fprintf(f,"//In-Agent definition\n\n");
 for(i=0;i<userdata->n_in_agents;i++)
 {
  fprintf(f,"  info[%d].identifier    = %d;\n",i,userdata->in_agents[i].id);
  fprintf(f,"  info[%d].priority      = %d;\n",i,userdata->in_agents[i].importance);
  fprintf(f,"  info[%d].initial_function   = fi_%d;\n",i,i);
  fprintf(f,"  info[%d].initial_parameters = NULL;\n",i);
  fprintf(f,"  info[%d].final_function     = ff_%d;\n",i,i);
  fprintf(f,"  info[%d].final_parameters   = NULL;\n",i);
  fprintf(f,"  info[%d].has_optionals      = 1;\n",i);
  fprintf(f,"  info[%d].offset        = (systime_t)  0;\n",i);
  fprintf(f,"  info[%d].period        = (systime_t) %d;\n",i,userdata->in_agents[i].period*1000);
  fprintf(f,"  info[%d].deadline      = (systime_t) %d;\n",i,userdata->in_agents[i].deadline*1000);
  fprintf(f,"  info[%d].initial_wcet  = (systime_t) %d;\n",i,initial_wcet[i]);
  fprintf(f,"  info[%d].final_wcet    = (systime_t) %d;\n",i,final_wcet[i]);
  fprintf(f,"  info[%d].blocking_time = (systime_t)  0;\n\n",i);
 }

 fprintf(f," mutex_create(MUTEX1, 1);\n\n");
 fprintf(f," for ( i = 0; i < N_TASKS; i++ )\n");
 fprintf(f,"   rtl_task_create (&task[i], &info[i]);\n\n");

 fprintf(f,"#ifdef ARTIS_CONF_METAREASONING\n");

//Bucle de creación de los comportamientos TODO: las tareas no corresponden
 fprintf(f,"//behavior definition\n");
 for(b=0;b<userdata->n_behaviors;b++)
 {
 fprintf(f," rtl_behav_create ( &(aBeh[%d]), aTaskBeh[%d], %d",i,i,userdata->behaviors[i].n_inagents);
 for(bi=0;bi<userdata->behaviors[b].n_inagents;bi++)
   for(i=0;i<userdata->n_in_agents;i++)
     if(userdata->behaviors[b].in_agents[bi].id==userdata->in_agents[i].id)
	fprintf(f,", &(task[%d])",i);
 fprintf(f,");\n");
 }


 fprintf(f,"#endif\n\n");
 fprintf(f," rtl_start_application();\n\n");
 fprintf(f," return 0;\n}\n\n\n");

 fprintf(f,"void cleanup_module(void) \n{\nrtl_end_application();\n}\n");
 }


 else printf("File rtl_tasks.c could not be opened\n");
}

void generate_optional_list(XML_State *userdata)
{
FILE *f;
int m,ml,mlk,k;
int optks;

printf("Generate optional_list.c\n");
if(f=fopen("./optional_list.c","w+"))
 {

 fprintf(f,"#include <stdio.h>\n");
 fprintf(f,"#include <artis/linux_core.h>\n");
 fprintf(f,"#include <artis/dserver.h>\n");

 fprintf(f,"#include <artis/kdm.h>\n");
 fprintf(f,"#include \"userclass.h\"\n\n");

 fprintf(f,"#include <artis/mutex.h>\n");

 fprintf(f,"#define MUTEX2 2\n\n");



// Funciones de los KS


for(k=0;k<userdata->n_ks;k++)
{
 optks=0;
 for (m=0;m<userdata->n_mks;m++)
   for (ml=0;ml<userdata->mks[m].n_levels;ml++)
     {
     if((userdata->mks[m].type==TYPE_MANDATORY)&&(ml==0))
	if(ml+1<userdata->mks[m].n_levels) ml++;
	else break;
     for(mlk=0;mlk<userdata->mks[m].levels[ml].n_ks;mlk++)
	  if(userdata->ks[k].id==userdata->mks[m].levels[ml].ks[mlk].id)
	    optks=1;
     }
 if(optks==1) fprintf(f,"int %s(){\n//Codigo de %s\n}\n",userdata->ks[k].name,userdata->ks[k].codefile);
}




 fprintf(f,"\nvoid rtfunc(void *dummy){}\n");

 for (m=0;m<userdata->n_mks;m++)
   for (ml=0;ml<userdata->mks[m].n_levels;ml++)
     {
     if((userdata->mks[m].type==TYPE_MANDATORY)&&(ml==0))
	if(ml+1<userdata->mks[m].n_levels) ml++;
	else break;
     fprintf(f,"void optional_level_%d(void *dummy)\n{\n",userdata->mks[m].levels[ml].id);
     for(mlk=0;mlk<userdata->mks[m].levels[ml].n_ks;mlk++)
	for(k=0;k<userdata->n_ks;k++)
          {
	  if(userdata->ks[k].id==userdata->mks[m].levels[ml].ks[mlk].id)
	    fprintf(f,"  %s();\n",userdata->ks[k].name);
          }
     fprintf(f,"}\n");
     }
 

fprintf(f,"\n\nvoid fill_optional_levels(void) \n{\n");
 for (m=0;m<userdata->n_mks;m++)
   for (ml=0;ml<userdata->mks[m].n_levels;ml++)
     {
     if((userdata->mks[m].type==TYPE_MANDATORY)&&(ml==0))
	fprintf(f,"  if (Level_Create(%d, rtfunc, NULL) != LINUX_CORE_SUCCESS)\n\tDS_stop(1);\n",userdata->mks[m].levels[ml].id);
     else fprintf(f,"  if (Level_Create(%d, optional_level_%d, NULL) != LINUX_CORE_SUCCESS)\n\tDS_stop(1);\n",userdata->mks[m].levels[ml].id,userdata->mks[m].levels[ml].id);
     }

 fprintf(f,"}\n");


fprintf(f,"\n\nvoid empty_optional_levels(void) \n{\n");
 for (m=0;m<userdata->n_mks;m++)
   for (ml=0;ml<userdata->mks[m].n_levels;ml++)
     {
	fprintf(f,"  if (Level_Destroy(%d) != LINUX_CORE_SUCCESS)\n\tfprintf(stderr, \"(empty_optional_levels) Error destroying level\\n\");\n",userdata->mks[m].levels[ml].id);
     }

 fprintf(f,"}\n");






 }
else printf("File optional_list.c could not be opened\n");


}
void 
//-----------------------------------------------------------
analyse( const char *filename )
//-----------------------------------------------------------
{
  FILE *fin;               //xml source file
  XML_Parser parser;       //parser for the elements
  XML_State userdata;      //global user data
  char buff[MAXBUFF];      //string with element to parse
  int len, done, line;

  // open xml file 
  if( (fin = fopen( filename, "r" )) == NULL ){
    fprintf( stderr, "Error opening file %s\n", filename );
    exit( -1 );
  } // if error opening file

  // create parser and analyse document
  parser = XML_ParserCreate( NULL );
  if( !parser ){
    fprintf(stderr, "Couldn't allocate memory for parser\n");
    exit( -1 );
  } // if not enough memory
  
  // define handlers for analysing xml code
  XML_SetUserData( parser, &userdata );
  XML_SetXmlDeclHandler( parser, xmldecl_handler );
  if( strstr( filename, ".xml" ) != NULL )
    XML_SetElementHandler( parser, xmlstartelem_handler, xmlendelem_handler );
  else{
    fprintf( stderr, "Error: invalid filetype\n" );
    XML_ParserFree( parser );
    exit( -1 );
  } //else (not an xml or xsd file)
  XML_SetCharacterDataHandler( parser, xmltext_handler );

  // analising xml file
  userdata.parser = parser;
  userdata.wrong_version = 0;
  userdata.depth = 0;
  //userdata.is_a_class = 0;
  for( line = 1, done = 0; fgets( buff, MAXBUFF, fin ) != NULL; line++ ){
#ifdef DEBUG    
    printf( "line %d(%d): %s\n", line, strlen(buff), buff );
#endif    
    if( XML_Parse( parser, buff, strlen( buff ), done) == XML_STATUS_ERROR) {
	fprintf( stderr, "Parse error at line %d:\n%s\n",
              XML_GetCurrentLineNumber( parser ),
              XML_ErrorString( XML_GetErrorCode( parser )));
	exit( -1 );
    } //if error in xml file 
  
  } //for !EOF
 #ifdef IMPRIME_AGENTE
  print_agent(&userdata);
 #endif
//  test_agent(&userdata);
 generate_control_data(&userdata);
 generate_rtl_tasks(&userdata);
 generate_optional_list(&userdata);

  // free memory and close files
  XML_ParserFree( parser );
  fclose( fin );

} //analyse()


int 
//------------------------------------------------------------
main( int  argc, 
      char *argv[] )
//------------------------------------------------------------
{
  // check sintax
  if( argc < 2 ){
    fprintf( stderr, "Usage: parser <file.xml>\n" );
    return -1;
  } //if bad sintax

  // analyse xml file
//  kdm_beginparsing( argv[1] );
  analyse( argv[1] );
//  kdm_endparsing(argv[1] );
  
  printf( "** XML file processed **\n" );
  return 0;

} //main()

