/*!
  \file kdmparser.h
  \date Time-stamp: <21-03-2005 12:54>
  \author M. Rebollo
  \version 0.1
  
  \brief kdm specific parsing functions
*/

/*!
  \def MAXBUFF
  \brief max length for a token
*/
#define MAXBUFF     256



/*!
  \fn int kdm_beginparsing( const char *xmlfile )
  \brief Inits the needed files before begin parsing
*/
int kdm_beginparsing( const char *xmlfile );

/*!
  \fn void kdm_endparsing( const char *xmlfile )
  \brief Closes the used files after finishing parsing
*/
void kdm_endparsing( const char *xmlfile );




/*!
  \fn void kdm_startclass( const char *classname )
  \brief Called when an opening label for the class is readed.
*/
void kdm_startclass( const char *classname );

/*!
  \fn void kdm_endclass( const char *classname )
  \brief Called when closing label for the class is readed.
*/
void kdm_endclass( const char *classname );

/*!
  \fn void kdm_addslot()
  \brief Called to add a new slot to he user files.
*/
void kdm_addslot( const char *classname, 
		  const char *slotname, 
		  const char *slottype );

/*!
  \fn void kdm_addslotvalue()
  \brief Called to add a new value for an atomic slot.
*/
void kdm_addslotvalue( const char *instname, 
		       const char *slotname, 
		       const char *slotvalue,
		       const char *eventtype,
		       const char *eventimp );
/*!
  \fn void kdm_addarray()
  \brief Called to add an array-typed slot.
*/
void kdm_addarray( const char *classname,
		   const char *slotname, 
		   const char *slottype,
		   const char *slotsubtype,
		   const char *slotsize );
/*!
  \fn void kdm_addinstance( const char *classname, const char *instname )
  \brief Called to add a new instance to the user files.
*/
void kdm_addinstance( const char *classname, 
		      const char *instname );
