/*!
  \file symtab.c
  \date Time-stamp: <20-03-2005 16:01>
  \author M. Rebollo
  \version 0.1

  \brief implements a hash table for a dictionary 
           about XML and ARTIS expressions
*/

#include <stdlib.h>
#include <assert.h>
#include "symtab.h"

char sdatatype[7][10] = { "error", "void", "int", "float", "char", 
			  "array", "class" };  
  
int 
//-------------------------------------------------
h( const char *item )
//-------------------------------------------------
{
  int key = 0;

  while( (*item) != '\0' )
    key += (*(item++));

  return key % NKEYS;
} //h()


symbol_table* 
//-------------------------------------------------
st_newtable()
//-------------------------------------------------
{
  int i;
  symbol_table *st;

  st = (symbol_table*)malloc( NKEYS * sizeof(st_item*) );

  for( i = 0; i < NKEYS; st[i++] = NULL )
    ; //empty
  
  return st;
} //st_newtable()


void
//-------------------------------------------------
st_deltable( symbol_table *st )
//-------------------------------------------------
{
  int i;
  st_item *tray;

  assert( st != NULL );

  //delete all the trays
  for( i = 0; i < NKEYS; i++ )
    if( st[i] != NULL )
      while( st[i] != NULL ){
	tray = st[i];
	st[i] = st[i]->next;
	free( tray );
      } //while there are elements left
 
  //frees the hash table memory
  free( st );
  st = NULL;
} //hdeltable()


void
//-------------------------------------------------
st_printtable( symbol_table *st )
//-------------------------------------------------
{
  int i;
  st_item *tray;

  assert( st != NULL );

  //print the heading
  printf( "KEY\t%-10.4s\tCATEG\tTYPE\n","NAME" );
  for( i = 0; i < 72; i++ )
  	printf( "-" );
  printf( "\n" );
  //search all the keys
  for( i = 0; i < NKEYS; i++ )
    if( st[i] != NULL ){
      //search all the trays in each key
      tray = st[i];
      while( tray != NULL ){
	printf( "%d\t", i );
	printf( "%10s\t%d\t%d\n", 
		tray->name, tray->categ, tray->type );
	tray = tray->next;
      } //while there are elements
    } //if the key has items 
} //st_printtable()


st_item*  
//-------------------------------------------------
st_search( const char *elem, 
	   symbol_table *st )
//-------------------------------------------------
{
  st_item *pitem;

  assert( st != NULL );
  assert( elem != NULL && strlen( elem ) > 0 );

  //obtains the element's key
  pitem = st[h( elem )];

  //search the element in the list
  while( pitem != NULL ){
    if( !strcmp( elem, pitem->name ) )
      return pitem;
    else
      pitem = pitem->next;
  }
  return NULL; //not found
} //st_search()


void 
//-------------------------------------------------
st_add_symbol( const char *name,
	       int cat,
	       int type,
	       symbol_table *st )
//-------------------------------------------------
{
  int key;
  st_item *tray;

  assert( st != NULL );

  if( st_search( name, st ) == NULL ){
    //insert new element in the head of the list
    key = h( name );
    tray = st[key];
    st[key] = (st_item*)malloc( sizeof(st_item) );
    strcpy( st[key]->name, name );
    st[key]->categ = cat;
    st[key]->type = type;
    st[key]->ref = NULL;
    st[key]->next = tray;
  } //if is a new element

} //st_add_symbol()


void  
//-------------------------------------------------
st_add_array( const char *name,
	      int cat,
	      int type,
	      int subtype,
	      int size,
	      symbol_table *st )
//-------------------------------------------------
{
  int key;
  st_item *tray;

  assert( st != NULL );
  
  if( st_search( name, st ) == NULL ){
    //insert new element in the head of the list
    key = h( name );
    tray = st[key];
    st[key] = (st_item*)malloc( sizeof( st_item ) );
    strcpy( st[key]->name, name );
    st[key]->categ = cat;
    st[key]->type = type;
    st[key]->next = tray;

    //creates the array description
    st[key]->ref = (void*) malloc( sizeof( at_item ) );
    ((at_item*)st[key]->ref)->type = subtype;
    ((at_item*)st[key]->ref)->size = size;
    ((at_item*)st[key]->ref)->ref = NULL;
  } //if is a new element

} //st_add_array()


void  
//-------------------------------------------------
st_del( const char *elem, 
	symbol_table *st )
//-------------------------------------------------
{
  st_item *tray, *prevtray;
  int key;

  assert( st != NULL );
  
  if( (tray = st_search( elem, st )) != NULL ){
    key = h( elem );
    //extracts the item from the list
    if( tray == st[key] ){
      //it is the first item in the list
      st[key] = tray->next;
    }
    else{
      //(it is the 2nd or more)
      prevtray = st[key];
      //gets the previous item (pointer maintenance)
      while( prevtray->next != tray && prevtray->next != NULL )
	prevtray = prevtray->next;
      //jump over the item to be deleted
      prevtray->next = tray->next;
    } //else (several items in the key)

    //frees the memory
    free( tray );
  } //if found
  
} //st_del()


