/*
  file: parser.c
  date: dec 04
  author: J. Fabregat

  purpose: translate xml file with MetaClass definition into C structures

  based on kdm parser by M. Rebollo
*/

#include <stdio.h>
#include <string.h>
#include <expat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define MAXBUFF 256

#define MC 100
#define NM 0
#define MQ 1
#define SP 2
#define RD 3
#define ER 4
#define CB 5

#define MARK(f, msg) //printf( "(%s) %s(%d)\n", f, msg, userdata->depth )

typedef struct userdata{
  XML_Parser parser;
  int wrong_version;

  int element;
  char *name;
  int maxQuality;
  int slsPolicy;
  int reactivityDegree;
  int eventRate;
  int nCurrentBehavior;

  int depth;
  char cname[MAXBUFF];
  char sname[MAXBUFF];
  char stype[MAXBUFF];
  char slength[MAXBUFF];
  char elemtype[MAXBUFF];
  char iname[MAXBUFF];
  char svalue[MAXBUFF];
}XML_State;



enum deeplevel{ TBNLVL, CLASSLVL, SLOTLVL, ARRAYLVL, CONSTLVL };

void analyse( const char *filename );


static void
//------------------------------------------------------------
xmldecl_handler(void            *userData,
                const XML_Char  *version,
                const XML_Char  *encoding,
                int              standalone)
//------------------------------------------------------------
{
  static const XML_Char Version_1_0[] = {'1', '.', '0', 0};

  int i, len;  
  XML_State *userdata = userData;

  len = strlen( Version_1_0 );
  for (i = 0; i < len; i++) {
    if (version[i] != Version_1_0[i]) {
      userdata->wrong_version = 1;
      //clear all handlers
      XML_SetCharacterDataHandler(userdata->parser, NULL);
      XML_SetElementHandler( userdata->parser, NULL, NULL );
      XML_SetCharacterDataHandler( userdata->parser, NULL );
      XML_SetCommentHandler( userdata->parser, NULL );
      XML_SetProcessingInstructionHandler( userdata->parser, NULL );
      fprintf( stderr, "Wrong version of XML document\n" ); 
      exit( -1 );
    } //if wrong version detected
  }//for i
} //xmldecl_handler()


static void
//------------------------------------------------------------
xmlstartelem_handler( void        *userData,
		      const char  *el, 
		      const char  **attr )
//------------------------------------------------------------
{
  XML_State *userdata = userData;

  MARK("xmlstartelem_handler", el);

  switch(userdata->depth)
  {

  case 0:
	if (strcmp(el,"MetaClass")==0) { userdata->depth++; userdata->element=MC; }
	break;

  case 1:
	if (strcmp(el,"name")==0) userdata->element=NM;
	if (strcmp(el,"maxQuality")==0) userdata->element=MQ;
	if (strcmp(el,"slsPolicy")==0) userdata->element=SP;
	if (strcmp(el,"reactivityDegree")==0) userdata->element=RD;
	if (strcmp(el,"eventRate")==0) userdata->element=ER;
	if (strcmp(el,"nCurrentBehavior")==0) userdata->element=CB;
	break;
  }


 /* Aqui código a ejecutar al leer la <etiqueta> */
	// En '*el' tienes la cadena 'etiqueta' y en '**attrib' un vector de elementos rollo argc

} //xmlstartelem_handler()


static void
//------------------------------------------------------------
xmlendelem_handler( void        *userData, 
		    const char  *el )
//------------------------------------------------------------
{
  XML_State *userdata = userData;

  MARK("xmlstartelem_handler", el);

  switch(userdata->depth)
  {

  case 0:
	if (strcmp(el,"MetaClass")==0) userdata->depth--;
	break;

  case 1:
	if (strcmp(el,"name")==0) userdata->element=MC;
	if (strcmp(el,"maxQuality")==0) userdata->element=MC;
	if (strcmp(el,"slsPolicy")==0) userdata->element=MC;
	if (strcmp(el,"reactivityDegree")==0) userdata->element=MC;
	if (strcmp(el,"eventRate")==0) userdata->element=MC;
	if (strcmp(el,"nCurrentBehavior")==0) userdata->element=MC;
	break;
  }

} //xmlendelem_handler() 



static void
//------------------------------------------------------------
xmltext_handler( void       *userData, 
		 const char *el, 
		 int        len )
//------------------------------------------------------------
{
  XML_State *userdata = userData;

  char *buff;



  switch(userdata->element){
	case NM:
		userdata->name=malloc((1+len)*sizeof(char));
		snprintf(userdata->name,len+1,el);
		MARK("text handler name", userdata->name);
		break;

	case MQ:
		buff=malloc((1+len)*sizeof(char));
		snprintf(buff,len+1,el);
		userdata->maxQuality=atoi(buff);
		MARK("text handler MQ", buff);
		break;

	case SP:
		buff=malloc((1+len)*sizeof(char));
		snprintf(buff,len+1,el);
		userdata->slsPolicy=atoi(buff);
		MARK("text handler SP", buff);
		break;

	case RD:
		buff=malloc((1+len)*sizeof(char));
		snprintf(buff,len+1,el);
		userdata->reactivityDegree=atoi(buff);
		MARK("text handler RD", buff);
		break;

	case ER:
		buff=malloc((1+len)*sizeof(char));
		snprintf(buff,len+1,el);
		userdata->eventRate=atoi(buff);
		MARK("text handler ER", buff);
		break;

	case CB:
		buff=malloc((1+len)*sizeof(char));
		snprintf(buff,len+1,el);
		userdata->nCurrentBehavior=atoi(buff);
		MARK("text handler CB", buff);
		break;
	}

  MARK("xmlstartelem_handler", el);
} //xmltext_handler()


int print_mclass(XML_State *userdata)
{

printf(" ************* METACLASS ************* \n");
printf(" Name: %s\n",userdata->name);
printf(" maxQuality: %d\n",userdata->maxQuality);
printf(" slsPolicy: %d\n",userdata->slsPolicy);
printf(" reactivityDegree: %d\n",userdata->reactivityDegree);
printf(" eventRate: %d\n",userdata->eventRate);
printf(" nCurrentBehavior: %d\n",userdata->nCurrentBehavior);
printf(" ************* METACLASS ************* \n");


return 0;
}

int generate_metaclass(XML_State *userdata)
{
FILE *f;

printf("Generate metaclass.c\n");

if(f=fopen("./metaclass.c","w+"))
 {
	fprintf(f,"/* file: metaclass.c\n generated from metaclass.xml\n\n PLEASE, DO NOT MODIFIY THIS FILE\n*/\n\n\n");
	fprintf(f,"#ifdef MODULE\n#include <linux/module.h>\n#include <linux/kernel.h>\n#include <linux/version.h>\n");
	fprintf(f,"#include <rtl.h>\n#else\n#include <stdio.h>\n#include <stdlib.h>\n#endif\n\n");

	fprintf(f,"#include <artis/include/metaclass.h>\n\nMetaClass *pMetaClass;\n\n");

	fprintf(f,"int meta_kdm_load(void)\n{\n  // Creates the objects for the Meta-KDM\n");

	fprintf(f,"#ifdef MODULE\n  if ((pMetaClass = (MetaClass *) Storage_Create_Object(tbn.stdata, sizeof(MetaClass), 0)) == NULL) \n{\n");
	fprintf(f,"    rtl_printf(\"(rtl_metaclass) Error in creating the object Meta-KDM\n\");\n    return -1;\n  }\n#else\n  pMetaClass = Storage_Get_Object( tbn.stdata, 0);\n#endif\n\n");

	fprintf(f,"  pMetaClass = Storage_Get_Address( tbn.stdata, pMetaClass );\n\n");

	fprintf(f,"#ifdef MODULE  \n");
	
	fprintf(f,"  if (pMetaClass == NULL) {\n    rtl_printf(\"(rtl_metaclass) ERROR\n\");\n    return -1;\n  }\n\n");

	fprintf(f,"  pMetaClass->systemQuality.type = KDM_INT;\n  pMetaClass->systemQuality.value.i = 0;\n  pMetaClass->systemQuality.event = NO_EVENT;\n\n");
	fprintf(f,"  pMetaClass->maxQuality.type = KDM_INT;\n  pMetaClass->maxQuality.value.i = %d;\n  pMetaClass->maxQuality.event = NO_EVENT;\n\n",userdata->maxQuality);
	fprintf(f,"  pMetaClass->slsPolicy.type = KDM_INT;\n  pMetaClass->slsPolicy.value.i = %d;\n  pMetaClass->slsPolicy.event = NO_EVENT;\n\n",userdata->slsPolicy);
	fprintf(f,"  pMetaClass->reactivityDegree.type = KDM_INT;\n  pMetaClass->reactivityDegree.value.i = %d;\n  pMetaClass->reactivityDegree.event = NO_EVENT;\n\n",userdata->reactivityDegree);
	fprintf(f,"  pMetaClass->nExecDS.type = KDM_INT;\n  pMetaClass->nExecDS.value.i = 0;\n  pMetaClass->nExecDS.event = NO_EVENT;\n\n");
	fprintf(f,"  pMetaClass->eventRate.type = KDM_INT;\n  pMetaClass->eventRate.value.i = %d;\n  pMetaClass->eventRate.event = NO_EVENT;\n\n",userdata->eventRate);
	fprintf(f,"  pMetaClass->levelsCompleted.type = KDM_INT;\n  pMetaClass->levelsCompleted.value.i = 0;\n  pMetaClass->levelsCompleted.event = NO_EVENT;\n\n");
	fprintf(f,"  pMetaClass->levelsInterrupted.type = KDM_INT;\n  pMetaClass->levelsInterrupted.value.i = 0;\n  pMetaClass->levelsInterrupted.event = NO_EVENT;\n\n");
	fprintf(f,"  pMetaClass->nCurrentBehavior.type = KDM_INT;\n  pMetaClass->nCurrentBehavior.value.i = %d;\n  pMetaClass->nCurrentBehavior.event = NO_EVENT;\n\n",userdata->nCurrentBehavior);
	fprintf(f,"  pMetaClass->nLastBehavior.type = KDM_INT;\n  pMetaClass->nLastBehavior.value.i = -1;\n  pMetaClass->nLastBehavior.event = NO_EVENT;\n\n");
	fprintf(f,"#endif\n\n  return 0;\n\n} //meta_kdm_load()\n\n\n");

	fprintf(f,"int meta_kdm_unload(void)\n{\n  // The space reserved in the Storage is freed automatically when destroyed\n  return 0;\n  //Frees the used memory\n} //meta_kdm_unload()\n\n");
	fprintf(f,"#ifdef MODULE\nint init_module( void ) {\n  rtl_printf(\"(rtl_metaclass) Initializing meta-data\n\");\n  meta_kdm_load();\n  return 0;\n}\n\n");
	
	fprintf(f,"void cleanup_module (void) {\n}\n#endif\n");
 }
else printf("File metaclass.c could not be opened\n");
fclose(f);
}

int generate_metavalue(XML_State *userdata)
{
FILE *f;

printf("Generate metavalue.h\n");
if(f=fopen("./metavalue.h","w+"))
{
fprintf(f,"extern int meta_kdm_load (void);\n");
}
else printf("File metavalue.h could not be opened\n");
fclose(f);


printf("Generate metavalue.c\n");

if(f=fopen("./metavalue.c","w+"))
 {
	fprintf(f,"/* file: metavalue.c\n generated from metaclass.xml\n\n PLEASE, DO NOT MODIFIY THIS FILE\n*/\n\n\n");
	fprintf(f,"#ifdef MODULE\n#include <linux/module.h>\n#include <linux/kernel.h>\n#include <linux/version.h>\n");
	fprintf(f,"#include <rtl.h>\n#else\n#include <stdio.h>\n#include <stdlib.h>\n#endif\n\n");

	fprintf(f,"#include <artis/include/metaclass.h>\n\nMetaClass *pMetaClass;\n\n");

	fprintf(f,"int meta_kdm_load(void)\n{\n\n  int retvalue;\n\n");

	fprintf(f,"  /// Sets the default values for each instance of Meta-Class\n\n");

	fprintf(f,"#ifdef MODULE\n  if ((retvalue = metaclass_init_module()) < 0) {\n");
	fprintf(f,"    rtl_printf(\"(rtl-init) Init. of module 'Meta-KDM' failed, error %%d\",retvalue);\n");
	fprintf(f,"#else\n  if ((retvalue = meta_kdm_create()) < 0) {\n");
	fprintf(f,"    fprintf(stderr, \"(rtl-init) Init. of module 'Meta-KDM' failed, error %%d\",retvalue);\n#endif\n");
	fprintf(f,"    return retvalue;\n  }\n#ifdef ARTIS_INIT_DEBUG\n  rtl_printf(\"(rtl-init) Init. of module 'Meta-KDM' OK\");\n#endif\n\n");


	fprintf(f,"  if (pMetaClass == NULL) return -1;\n\n");

	fprintf(f,"  pMetaClass->systemQuality.type = KDM_INT;\n  pMetaClass->systemQuality.value.i = 0;\n  pMetaClass->systemQuality.event = NO_EVENT;\n\n");
	fprintf(f,"  pMetaClass->maxQuality.type = KDM_INT;\n  pMetaClass->maxQuality.value.i = %d;\n  pMetaClass->maxQuality.event = NO_EVENT;\n\n",userdata->maxQuality);
	fprintf(f,"  pMetaClass->slsPolicy.type = KDM_INT;\n  pMetaClass->slsPolicy.value.i = %d;\n  pMetaClass->slsPolicy.event = NO_EVENT;\n\n",userdata->slsPolicy);
	fprintf(f,"  pMetaClass->reactivityDegree.type = KDM_INT;\n  pMetaClass->reactivityDegree.value.i = %d;\n  pMetaClass->reactivityDegree.event = NO_EVENT;\n\n",userdata->reactivityDegree);
	fprintf(f,"  pMetaClass->nExecDS.type = KDM_INT;\n  pMetaClass->nExecDS.value.i = 0;\n  pMetaClass->nExecDS.event = NO_EVENT;\n\n");
	fprintf(f,"  pMetaClass->eventRate.type = KDM_INT;\n  pMetaClass->eventRate.value.i = %d;\n  pMetaClass->eventRate.event = NO_EVENT;\n\n",userdata->eventRate);
	fprintf(f,"  pMetaClass->levelsCompleted.type = KDM_INT;\n  pMetaClass->levelsCompleted.value.i = 0;\n  pMetaClass->levelsCompleted.event = NO_EVENT;\n\n");
	fprintf(f,"  pMetaClass->levelsInterrupted.type = KDM_INT;\n  pMetaClass->levelsInterrupted.value.i = 0;\n  pMetaClass->levelsInterrupted.event = NO_EVENT;\n\n");
	fprintf(f,"  pMetaClass->nCurrentBehavior.type = KDM_INT;\n  pMetaClass->nCurrentBehavior.value.i = %d;\n  pMetaClass->nCurrentBehavior.event = NO_EVENT;\n\n",userdata->nCurrentBehavior);
	fprintf(f,"  pMetaClass->nLastBehavior.type = KDM_INT;\n  pMetaClass->nLastBehavior.value.i = -1;\n  pMetaClass->nLastBehavior.event = NO_EVENT;\n\n");
	fprintf(f,"  return 0;\n\n} //meta_kdm_load()\n\n\n");

	fprintf(f,"int meta_kdm_unload(void)\n{\n  // The space reserved in the Storage is freed automatically when destroyed\n  return 0;\n  //Frees the used memory\n} //meta_kdm_unload()\n\n");
	fprintf(f,"#ifdef MODULE\nint init_module( void ) {\n  rtl_printf(\"(rtl_metaclass) Initializing meta-data\");\n  meta_kdm_load();\n  return 0;\n}\n\n");
	
	fprintf(f,"void cleanup_module (void) {\n}\n#endif\n");
 }
else printf("File metavalue.c could not be opened\n");
fclose(f);
}



void 
//-----------------------------------------------------------
analyse( const char *filename )
//-----------------------------------------------------------
{
  FILE *fin;               //xml source file
  XML_Parser parser;       //parser for the elements
  XML_State userdata;      //global user data
  char buff[MAXBUFF];      //string with element to parse
  int len, done, line;

  // open xml file 
  if( (fin = fopen( filename, "r" )) == NULL ){
    fprintf( stderr, "Error opening file %s\n", filename );
    exit( -1 );
  } // if error opening file

  // create parser and analyse document
  parser = XML_ParserCreate( NULL );
  if( !parser ){
    fprintf(stderr, "Couldn't allocate memory for parser\n");
    exit( -1 );
  } // if not enough memory
  
  // define handlers for analysing xml code
  XML_SetUserData( parser, &userdata );
  XML_SetXmlDeclHandler( parser, xmldecl_handler );
  if( strstr( filename, ".xml" ) != NULL )
    XML_SetElementHandler( parser, xmlstartelem_handler, xmlendelem_handler );
  else{
    fprintf( stderr, "Error: invalid filetype\n" );
    XML_ParserFree( parser );
    exit( -1 );
  } //else (not an xml or xsd file)
  XML_SetCharacterDataHandler( parser, xmltext_handler );

  // analising xml file
  userdata.parser = parser;
  userdata.wrong_version = 0;
  userdata.depth = 0;
  //userdata.is_a_class = 0;
  for( line = 1, done = 0; fgets( buff, MAXBUFF, fin ) != NULL; line++ ){
#ifdef DEBUG    
    printf( "line %d(%d): %s\n", line, strlen(buff), buff );
#endif    
    if( XML_Parse( parser, buff, strlen( buff ), done) == XML_STATUS_ERROR) {
	fprintf( stderr, "Parse error at line %d:\n%s\n",
              XML_GetCurrentLineNumber( parser ),
              XML_ErrorString( XML_GetErrorCode( parser )));
	exit( -1 );
    } //if error in xml file 
  
  } //for !EOF


/* Aqui las invocaciones a funciones de generación de código */

  //print_mclass(&userdata);
  generate_metaclass(&userdata);
  generate_metavalue(&userdata);

  // free memory and close files
  XML_ParserFree( parser );
  fclose( fin );

} //analyse()


int 
//------------------------------------------------------------
main( int  argc, 
      char *argv[] )
//------------------------------------------------------------
{
  // check sintax
  if( argc < 2 ){
    fprintf( stderr, "Usage: parser <file.xml>\n" );
    return -1;
  } //if bad sintax

  // analyse xml file
//  kdm_beginparsing( argv[1] );
  analyse( argv[1] );
//  kdm_endparsing(argv[1] );
  
  printf( "** XML file processed **\n" );
  return 0;

} //main()

