/*
   FILE:								       
	mem2trace.c						       
									
   DESCRIPTION:								
	Program that examines the trace written by the rt scheduler in
	shared (phisical) memory, and writes some files with that information.


   Copyright (C) 1998
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet



  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <mbuff.h>

#include "../../../trace/include/trace.h"


static volatile event_t *memory_events;

static FILE *interm_file;





static void
do_exit (int error)
{
  mbuff_detach (MBUFF_TRACE_EVENTS_NAME, (void *) memory_events);
  exit (error);
}





int 
main (int argc, char *argv[])
{
  char interm_filename[32];

  int i, max_events;


  /* Test correct number of arguments */
  if (argc < 2) {
    fprintf (stderr, "Use: mem2trace outputfile\n");
    exit (1);
  }


  /* Create the output intermediate trace file */
  strcpy (interm_filename, argv[1]);
  strcat (interm_filename, ".tri");
  if ((interm_file = fopen (interm_filename, "w")) == NULL) {
    fprintf (stderr, "mem2trace: Error in creating intermediate file %s.msg\n",
	     argv[1]);
    exit (1);
  }


  /* Open the memory with mbuff */
  memory_events = (volatile event_t *) mbuff_attach (MBUFF_TRACE_EVENTS_NAME, 
						     MBUFF_TRACE_EVENTS_SIZE);

  if (memory_events == NULL) {
    fprintf (stderr, "mem2trace: Error opening mbuff %s\n",
	     MBUFF_TRACE_EVENTS_NAME);
    exit (1);
  }


  /* The first event should be the amount of application rt-tasks */
  if (memory_events[0].type != NTASKS) {
    fprintf (stderr, "mem2trace: Don't know how many application tasks.\n");
    do_exit (1);
  }


  // Write the first event:
  fprintf (interm_file, "%lld %d %d\n", 
	   memory_events[0].timestamp,
	   memory_events[0].type, 
	   memory_events[0].id);


  // Write the rest of events:
  max_events = (int) (MBUFF_TRACE_EVENTS_SIZE / sizeof (event_t)) - 1;

  i = 1;
  while(i < max_events && memory_events[i].type != VOID_EVENT_TYPE) {
      fprintf (interm_file, "%lld %d %d %lld %lld\n", 
	       memory_events[i].timestamp, 
	       memory_events[i].type,
	       memory_events[i].id,
	       memory_events[i].data1,
	       memory_events[i].data2);
      i += 1;
  }
  
  do_exit (0);
  return 0;
}

